//Use this method to analyze the regulatory potential of the conserved
//islands. This program can be used to launch Cluster Buster jobs as
//well as analyzing the output. For the post-processing, we want
//remove some useless junk from the files and then find the category
//for each island. For each category, we calculate a length-normalized
//distribution of scores.
//
//
//java -Xmx1024m AnalyzeClusterBusterResults -o -c 19
//

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class AnalyzeClusterBusterResults {

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option clusterBusterResultsFileNameFlag = argParser.addStringOption('f', "clusterBusterResultsFileName");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "hs";
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : "Hs/v36") + "/RegulatoryPotential/conservation_islands_regulatory_scores_" + (species.equals("mm") ? "" : "hs_") + "w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".txt";
	    tmp = argParser.getOptionValue(clusterBusterResultsFileNameFlag);
	    String clusterBusterResultsFileName = tmp!=null ? (String)tmp : "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : "Hs/v36") + "/RegulatoryPotential/cbust_conservation_islands_" + (species.equals("mm") ? "" : "hs_") + "w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".txt";
	    //////////////////////////////////////////////////////////////////
	    int nChr = 24;

	    for (int i=0; i<nChr; i++) {
		if (i==chr || chr==-1) {
		    //Parse the conservation islands and put them into categories
		    ConservationIslands conservationIslands = new ConservationIslands(i, orchestra, species);
		    //Open the Cluster Buster scores and assign the score
		    parseClusterBusterFile(conservationIslands, clusterBusterResultsFileName);
		    //Loop through the catgories and extract the scores, save to file
		    //for (int j=0; j<FindConservationIslands.NUMBER_OF_CI_TYPES; j++) {
		    //TreeSet<ConservationIsland> cisType = conservationIslands.getType(j);
			BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
			for (ConservationIsland ci : conservationIslands.getConservationIslands()) 
			    bw.write(ci.getChromosome() + "\t" + ci.getStart() + "\t" + ci.getEnd() + "\t" + ci.getType() + "\t" + format.format(ci.getRegulatoryScore()) + "\n");
			bw.close();
			//}
			
		}
	    }
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    public static void parseClusterBusterFile(ConservationIslands consIslands, String clusterBusterResultsFile) {
	try {
	    //The conserved islands and the results in the clusterBusterResultsFile should be in the same order. We use that to speed up the parsing
	    BufferedReader br = new BufferedReader(new FileReader(clusterBusterResultsFile));
	    String line = br.readLine();
	    Iterator<ConservationIsland> it = consIslands.getConservationIslands().iterator();
	    Pattern scorePattern = Pattern.compile("^([\\d|\\.|e|-]+)\\s+(\\d+)\\s+(\\d+)");
	    Pattern islandPatter = Pattern.compile(">chr(\\d+):(\\d+)-(\\d+)");
	    ConservationIsland ci = it.next();
	    while (line!=null) {
		if (!line.startsWith("#") && line.length()>1) { //Ignore these lines
		    if (line.startsWith(">")) { //Find the CI
			Matcher m = islandPatter.matcher(line);
			ConservationIsland ciFound = null;
			if (m.find()) 
			    ciFound = new ConservationIsland(Integer.parseInt(m.group(1)), Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)));
			else
			    System.err.println("Could not find island for " + line);
			while ((ci.compareTo(ciFound)<0) && it.hasNext())
			    ci = it.next();
			//System.out.println(ci.toString());
		    }
		    else if (line.substring(0, 1).matches("\\d")) { //Parse the score and assign to CI
			Matcher m = scorePattern.matcher(line);
			String tokens[] = line.split("\\t");
			if (m.find()) {
			//if (tokens.length==128) {
			    double s = Double.parseDouble(m.group(1))/ci.getLength(); //*(Integer.parseInt(m.group(3)) - Integer.parseInt(m.group(2)))/ci.getLength();
			    //for (int i=3; i<tokens.length; i++)
			    //	s += tokens[i].equals("0") ? -36 : Math.log(Double.parseDouble(tokens[i]));
			    ci.setRegulatoryScore(s);
			}
			    //else
			    //System.err.println("Could not find score for " + line);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing clusterbuster results " + e.toString());
	}
    }


}