//Use this to study the GC-bias of the ChIP-Seq experiments. The idea
//is to check the input and build a histogram comparing the actual
//number of sequenced tags with a given GC-contents with the number
//based what we would excpect from the null model.

import jargs.gnu.CmdLineParser;
import java.io.*;
import java.text.DecimalFormat;

public class AnalyzeGCBias {

    public static DecimalFormat format = new DecimalFormat("###.####");

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .fa-files
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFile");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag if we want to extend the reads to length fragmentLength
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    //For analyzing RNA-Seq
	    CmdLineParser.Option rnaSeqFlag = argParser.addBooleanOption('r', "RNASeq");
	    //Use this flag to decide which species to
	    //investigate. Available species are: 
	    //Mouse (Mm) Default
	    //Yeast (Sc)
	    CmdLineParser.Option speciesFlag = argParser.addStringOption('s', "species");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    String GENOME_LOC = "./";
	    int nChr = 0;
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    boolean unique = argParser.getOptionValue(uniqueFlag)!=null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean extend = argParser.getOptionValue(extendFlag)!=null;
	    boolean rnaSeq = argParser.getOptionValue(rnaSeqFlag)!=null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? tmp.toString() : "Mm";
	    if (species.equals("Mm")) {
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
		nChr = 21;
	    }
	    else if (species.equals("Sc")) {
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Sc/sacCer1/" : "/media/exthd/Genomes/Sc/sacCer1/";
		nChr = 17;
	    }
	    //////////////////////////////////////////////////////////////////
	    int histRead[] = new int[AnalyzeControls.READ_LENGTH + 1], histFrag[] = new int[fragmentLength + 1];
	    int histReadChr[] = new int[AnalyzeControls.READ_LENGTH + 1], histFragChr[] = new int[fragmentLength + 1];
	    int hist[], histC[] = null;
	    int gcMismatches[] = new int[AnalyzeControls.READ_LENGTH + 1];
	    double gcQualityScores[] = new double[AnalyzeControls.READ_LENGTH + 1];
	    if (rnaSeq)
		histReadChr = chrGCContent(GENOME_LOC + species + "_rna.fa", AnalyzeControls.READ_LENGTH);
	    for (int i=1; i<=nChr; i++) {
		//Go through all the reads and put them into bins
		//depending on the GC-content of either the actual read
		//(35 bps) or the fragment (120 bps).
		ABIRead reads[] = PeakFinder2.parseFiles(i, inputFileName, fragmentLength, extend, 0, 0, null);
		FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(i, species));
		//Extract the sequences for all the reads
		String seqs[] = fastaSeq.getSequences(reads);
		//System.out.println("Found " + reads.length + " and " + seqs.length + " sequences." + reads[0].getQualityScore() + "\t" + reads[0].numberOfMismatches() + );
		hist = analyzeGCBias(seqs, reads, AnalyzeControls.READ_LENGTH);
		//Find the average quality score and no mismatches
		for (int j=0; j<reads.length; j++) {
		    if (reads[j].getSequence()!=null) {
			gcQualityScores[gcContent(reads[j].getSequence())] += reads[j].getQualityScore();
			gcMismatches[gcContent(reads[j].getSequence())] += reads[j].numberOfMismatches();
		    }
		}
		//if (!rnaSeq)
		    //histC = chrGCContent(GENOME_LOC + AnalyzeControls.getChromosomeString(i, species), AnalyzeControls.READ_LENGTH);
		for (int j=0; j<hist.length; j++) 
		    histRead[j] += hist[j];
		/*if (!rnaSeq)
			histReadChr[j] += histC[j];
			}*/
		if (!rnaSeq) {
		    //hist = analyzeGCBias(seqs, reads, fragmentLength);
		    //histC = chrGCContent(GENOME_LOC + AnalyzeControls.getChromosomeString(i, species), fragmentLength);
		    /*for (int j=0; j<hist.length; j++) {
			histFrag[j] += hist[j];		
			histFragChr[j] += histC[j];
			}*/
		}
		//Save the results
		if (!rnaSeq) {
		    BufferedWriter bw = new BufferedWriter(new FileWriter("input_GC_bias_" + AnalyzeControls.READ_LENGTH + "_" + species + "_chr" + i + ".txt"));
		    bw.write("#GC\tReads\tQS\tMismatches\n");
		    for (int j=0; j<histRead.length; j++) {
			bw.write(j + "\t" + histRead[j] + "\t");
			if (histRead[j]>0)
			    bw.write(format.format(gcQualityScores[j]/histRead[j]) + "\t" + format.format((double)gcMismatches[j]/histRead[j]) + "\n");
			else
			    bw.write("0\t0\n");
		    }
		    bw.close();
		    break;
		    /*bw = new BufferedWriter(new FileWriter("input_GC_bias_" + fragmentLength + "_" + species + "_chr" + i + ".txt"));
		    bw.write("#GC\tReads\tChr\n");
		    for (int j=0; j<histFrag.length; j++)
			bw.write(j + "\t" + histFrag[j] + "\t" + histFragChr[j] + "\n");
			bw.close();*/
		}
	    }
	    if (rnaSeq) {
		/*BufferedWriter bw = new BufferedWriter(new FileWriter("input_GC_bias_" + AnalyzeControls.READ_LENGTH + "_" + species + "_RNASeq.txt"));
		bw.write("#GC\tReads\tChr\n");
		for (int j=0; j<histRead.length; j++)
		    bw.write(j + "\t" + histRead[j] + "\t" + histReadChr[j] + "\n");
		    bw.close();*/
	    }
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    public static int[] analyzeGCBias(String seqs[], ABIRead reads[], int len) {
	int histogram[] = new int[len + 1];
	for (int j=0; j<seqs.length; j++) {
	    if (reads[j].getStrand())
		reads[j].setSequence(seqs[j].substring(0, AnalyzeControls.READ_LENGTH));
	    else
		reads[j].setSequence(MerTree.reverseComplement(seqs[j].substring(0, AnalyzeControls.READ_LENGTH)));
	    if (seqs[j]==null)
		System.out.println("Found null for read " + reads[j].toString());
	    else 
		histogram[gcContent(seqs[j], len, reads[j].getStrand())]++;
	}
	return histogram;
    }

    public static int gcContent(String r) { return gcContent(r.toCharArray()); }

    public static int gcContent(String r, int len, boolean s) { return gcContent(r.toCharArray(), len, s); }

    public static int gcContent(char[] ra) {
	//Count the number of Gs or Cs in a read r
	int gc = 0;
	for (int i=0; i<ra.length; i++) {
	    if (ra[i]=='c' || ra[i]=='C' || ra[i]=='g' || ra[i]=='G')
		gc++;
	}
	return gc;
    }

    public static int gcContent(char[] ra, int len, boolean s) {
	//Count the number of Gs or Cs in a read r
	int gc = 0;
	if (s) {
	    for (int i=0; i<len; i++) {
		if (ra[i]=='c' || ra[i]=='C' || ra[i]=='g' || ra[i]=='G')
		    gc++;
	    }
	}
	else {
	    for (int i=0; i<len; i++) {
		if (ra[len-1-i]=='c' || ra[len-1-i]=='C' || ra[len-1-i]=='g' || ra[len-1-i]=='G')
		    gc++;
	    }
	}
	return gc;
    }

    public static int[] chrGCContent(String fileName, int readLength) {
	//Use this to calculate the number of reads with a given
	//GC-content from the entire genome.
	int hist[] = new int[readLength + 1], nwc = 0, nrep = 0;
	char[] mer = new char[readLength];
	boolean initialize = true;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    while (br.ready()) {
		if (initialize) {
		    MerTree.initialize(mer, br);
		    for (int i=0; i<readLength-1; i++) {
			if (mer[i]=='n' || mer[i]=='N') { nwc++; }
			if (mer[i]=='a' || mer[i]=='c' || mer[i]=='g' || mer[i]=='t') { nrep++; }
		    }
		    initialize = false;
		}
		char b = MerTree.removeComments(br);
		if (b!='\n') {
		    if (b=='n' || b=='N') { nwc++; }
		    if (b=='a' || b=='c' || b=='g' || b=='t') { nrep++; }
		    mer[readLength-1] = b; //Read a k-mer
		    if (nwc==0 && nrep<readLength)
			hist[gcContent(mer)]++;
		    if (mer[0]=='n' || mer[0]=='N') { nwc--; } 
		    if (mer[0]=='a' || mer[0]=='c' || mer[0]=='g' || mer[0]=='t') { nrep--; }
		    for (int i=1; i<readLength; i++) { mer[i-1] = mer[i]; }
		}
		else {
		    //If we reach a comment we should empty the buffer
		    //since a new rna starts and we don't want to consider
		    //the boundary.
		    br.mark(1);
		    if ((char)br.read()=='>')
			initialize = true;
		    br.reset();
		}

	    }
	    br.close();
	}
	catch(IOException e) { }
	return hist;
    }

    public static int[] chrGCContent(String fileName, int readLength, int windowSize) {
	//Use this to calculate the number of reads with a given
	//GC-content from the entire genome. Calculate the average GC
	//proportion in a windowSize region and give it a weight equal
	//to the total number of reads which can be mapped into the
	//window. By comparing with the actual results from the reads,
	//we can figure out if there are any regions which are
	//underrepresented.
	int hist[] = new int[readLength + 1], nwc = 0, nrep = 0;
	char[] mer = new char[readLength];
	boolean initialize = true;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    while (br.ready()) {
		if (initialize) {
		    MerTree.initialize(mer, br);
		    for (int i=0; i<readLength-1; i++) {
			if (mer[i]=='n' || mer[i]=='N') { nwc++; }
			if (mer[i]=='a' || mer[i]=='c' || mer[i]=='g' || mer[i]=='t') { nrep++; }
		    }
		    initialize = false;
		}
		char b = MerTree.removeComments(br);
		if (b!='\n') {
		    if (b=='n' || b=='N') { nwc++; }
		    if (b=='a' || b=='c' || b=='g' || b=='t') { nrep++; }
		    mer[readLength-1] = b; //Read a k-mer
		    if (nwc==0 && nrep<readLength)
			hist[gcContent(mer)]++;
		    if (mer[0]=='n' || mer[0]=='N') { nwc--; } 
		    if (mer[0]=='a' || mer[0]=='c' || mer[0]=='g' || mer[0]=='t') { nrep--; }
		    for (int i=1; i<readLength; i++) { mer[i-1] = mer[i]; }
		}
		else {
		    //If we reach a comment we should empty the buffer
		    //since a new rna starts and we don't want to consider
		    //the boundary.
		    br.mark(1);
		    if ((char)br.read()=='>')
			initialize = true;
		    br.reset();
		}

	    }
	    br.close();
	}
	catch(IOException e) { }
	return hist;
    }

}
