//Use this to analyze the distances between 

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class AnalyzeMerDistance{

    public static String GENOME_LOC;


    protected static int nMismatches, merLength, blockSize, chrRef, maxBases;
    protected static boolean verbose, unique, orchestra, pseudo;
    protected static String species = "Mm", saveDir, fileName;

    public static DecimalFormat format = new DecimalFormat("###.########");

    //////////////////////////////////////////////////////////////////

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    //Use this flag for the chromosome which we are checking against
	    CmdLineParser.Option chromosomeReferenceFlag = argParser.addIntegerOption('C', "chromosomeReference");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    //Use this flag to determine the the maximum number of
	    //bases to analyze (excluding wildcards)
	    CmdLineParser.Option maxBasesFlag = argParser.addIntegerOption('M', "maxBases");
	    //Length of the mers
	    CmdLineParser.Option merLengthFlag = argParser.addIntegerOption('k', "length");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag to decide which species to
	    //investigate. Available species are: 
	    //Mouse (Mm) Default
	    //Yeast (Sc)
	    CmdLineParser.Option speciesFlag = argParser.addStringOption('s', "species");
	    //Use this flag to determine where we should store the
	    //resulting .bed-files
	    CmdLineParser.Option saveDirectoryFlag = argParser.addStringOption('d', "saveDirectory");
	    //Use this flag to decide how many bases to read
	    CmdLineParser.Option blockSizeFlag = argParser.addIntegerOption('b', "blockSize");
	    //Use pseudo color space instead of real code.
	    CmdLineParser.Option pseudoColorSpaceFlag = argParser.addBooleanOption('p', "pseudoColorSpace");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(chromosomeReferenceFlag);
	    chrRef = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(maxBasesFlag);
	    maxBases = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(merLengthFlag);
	    merLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    verbose = argParser.getOptionValue(verboseFlag)!=null;
	    unique = argParser.getOptionValue(uniqueFlag)!=null;
	    orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    pseudo = argParser.getOptionValue(pseudoColorSpaceFlag)!=null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    species = tmp!=null ? tmp.toString() : "Mm";
	    tmp = argParser.getOptionValue(saveDirectoryFlag);
	    saveDir = tmp!=null ? tmp.toString() : "./";
	    //make sure that it ends with '/'
	    if (!saveDir.endsWith("/")) { saveDir += "/"; }
	    tmp = argParser.getOptionValue(blockSizeFlag);
	    blockSize = tmp!=null ? ((Integer)tmp).intValue() : 1000000;
	    //////////////////////////////////////////////////////////////////
	    if (species.equals("Mm"))
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	    else if (species.equals("Sc"))
		GENOME_LOC = orchestra ? "/home/mh168/Genomes/Sc/sacCer1/" : "/media/exthd/Genomes/Sc/sacCer1/";
	    if (pseudo) { GENOME_LOC += "PseudoColorCode/"; }
	    if (chr==-1) { //Look at all chromosomes
		for (int i=1; i<=21; i++)
		    analyze(i);
	    }
	    else 
		analyze(chr);
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////
    
    public static void analyze(int c) {
	String fName = null;
	if (fileName==null)
	    fName = GENOME_LOC + AnalyzeControls.getChromosomeString(c, species);
	else
	    fName = fileName;
	//Split into multiple jobs of size blockSize
	int start = 0, end = (int)findFileLength(fName), max = maxBases<0 ? end : maxBases;
	System.out.println(fName + " has " + end + " bases.");
	while (start<end && start<max) {
	    HashMap minDists = new HashMap(blockSize/10);
	    minDists = findMinDistance(fName, start, minDists);
	    end = getLastKey(minDists);
	    int prevDist = saveMinDists(c, start, end, minDists); //Save the results
	    start = end + 1;
	}
    }

    protected static HashMap findMinDistance(String fName, int start, HashMap minDists) {
	MerTree merTree = new MerTree(orchestra);
	String refFileName = GENOME_LOC + AnalyzeControls.getChromosomeString(chrRef, species);
	int s = 0, basesRead = 0, e = maxBases>0 ? maxBases : (int)findFileLength(refFileName);
	//Read the chunk which we will be comparing and keep it in
	//memory to avoid having to re-read every iteration.
	char[] chrPart = MerTree.readGenome(fName, start, blockSize);
	//System.out.println("Searching for " + chrPart.length + " bases from " + start);
	//System.out.println(refFileName + " has " + e + " bases.");
	while (s<e) { //Go through the chromosome that we are comparing with
	    //Create an index tree for a chunk of nucleotides
	    basesRead = merTree.populateTree(merLength, refFileName, s, blockSize) - s;
	    s += basesRead + 1;
	    //System.out.println("Found " + merTree.numberOfNodes() + " nodes and " + merTree.numberOfLeaves() + " leaves.\nGoing through " + fName + ", starting at " + start);
	    //determine the minimum distances for the chunk which we are interested in
	    //minDists = merTree.findMinimumDistances(merLength, nMismatches, fName, start, blockSize, minDists);
	    minDists = merTree.findMinimumDistances(merLength, nMismatches, start, chrPart, minDists);
	}
	return minDists;
    }

    protected static int saveMinDists(int c, int start, int end, HashMap minDists) {
	String saveFileName = saveDir + "chr" + AnalyzeControls.getChromosomeNumber(c) + "_" + AnalyzeControls.getChromosomeNumber(chrRef) + "_" + merLength + "_" + nMismatches + "_" + start + "_" + end + ".bed";
	Integer key = new Integer(start);
	int dist = 0, s = start, prevDist = minDists.containsKey(key) ? ((MerInfo)minDists.get(key)).dist : 0;
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    for (int i=start + 1; i<=end; i++) {
		key = new Integer(i);
		dist = minDists.containsKey(key) ? ((MerInfo)minDists.get(key)).dist : 0;
		if (dist!=prevDist) {
		    bw.write("chr" + AnalyzeControls.getChromosomeNumber(c) + "\t" + s + "\t" + i + "\t" + prevDist + "\n");
		    s = i;
		    prevDist = dist;
		}
	    }
	    //save the last chunk
	    bw.write("chr" + AnalyzeControls.getChromosomeNumber(c) + "\t" + s + "\t" + end + "\t" + dist + "\n");
	    prevDist = dist;
	    bw.close();
	}
	catch(IOException ex) { }
	return prevDist;
    }    

    protected static int getLastKey(HashMap minDists) {
	Vector keys = new Vector(minDists.keySet());
	Collections.sort(keys);
	return ((Integer)keys.lastElement()).intValue();
    }

    public static long findFileLength(String fileName) {
	//Use this to find the number of bases in a Fasta-file
	long nBases = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    while (br.ready()) {
		char b = MerTree.removeComments(br);
		nBases += b=='\n' ? 0 : 1;
	    }
	    br.close();
	}
	catch (IOException e) { }
	return nBases;
    }

    public static long findFileLength(String fileName, boolean nonRep) {
	//Use this to find the number of bases in a Fasta-file
	long nBases = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    while (br.ready()) {
		char b = MerTree.removeComments(br);
		if (!nonRep || (b!='\n' && b!='n' && b!='N' && b!='a' && b!='c' && b!='g' && b!='t'))
		    nBases++;
	    }
	    br.close();
	}
	catch (IOException e) { }
	return nBases;
    }

}
