//Use this to pull out a list of all peaks which overlap and from two
//or more TFs.

import jargs.gnu.CmdLineParser;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;
import java.util.regex.*;

public class AnalyzeTFColocalization {

    public static DecimalFormat format = new DecimalFormat("###.########");

    public static final String MEF2_STR = "MEF2";
    public static final String NPAS4_STR = "NPAS4";
    public static final String CBP_STR = "CBP";
    public static final String CREB_STR = "CREB";
    public static final String CTCF_KCL_STR = "CTCF_KCL";
    public static final String CTCF_UN_STR = "CTCF_UN";
    //Variable names cannot start with a number...
    public static final String H8_KCL_STR = "4H8_KCL";
    public static final String H8_UN_STR = "4H8_UN";
    public static final String WG16_KCL_STR = "8WG16_KCL";
    public static final String WG16_UN_STR = "8WG16_UN";

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .fa-files
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFile");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag if we want to extend the reads to length fragmentLength
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean extend = argParser.getOptionValue(extendFlag)!=null;
	    Object tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    //////////////////////////////////////////////////////////////////
	    String intergenicFileNames[] = new String[4], tfNames[] = new String[]{MEF2_STR, NPAS4_STR, CBP_STR, CREB_STR}, fileStart = orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/";
	    //String fileNames[] = new String[10], tfNames[] = new String[]{MEF2_STR, NPAS4_STR, CBP_STR, CREB_STR, CTCF_KCL_STR, CTCF_UN_STR, H8_KCL_STR, H8_UN_STR, WG16_KCL_STR, WG16_UN_STR}, fileStart = orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/";
	    /*fileNames[0] = fileStart + "Mef2/Mef2_KCl/Mef2_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[1] = fileStart + "Npas4/Npas4_KCl/Npas4_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[2] = fileStart + "CBP/KCl/CBP_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[3] = fileStart + "CREB/KCl/CREB_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[4] = fileStart + "CTCF/KCl/CTCF_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[5] = fileStart + "CTCF/un/CTCF_un_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[6] = fileStart + "Pol2/4H8_KCl/4H8_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[7] = fileStart + "Pol2/4H8_un/4H8_un_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[8] = fileStart + "Pol2/8WG16_KCl/8WG16_KCl_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    fileNames[9] = fileStart + "Pol2/8WG16_un/8WG16_un_genes_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    findTFColocalization(0, fileNames, tfNames, "tf_colocalization.txt");*/
	    //For comparing intergenic sites only
	    intergenicFileNames[0] = fileStart + "Mef2/Mef2_KCl/Mef2_KCl_intergenic_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    intergenicFileNames[1] = fileStart + "Npas4/Npas4_KCl/Npas4_KCl_intergenic_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    intergenicFileNames[2] = fileStart + "CBP/KCl/CBP_KCl_intergenic_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    intergenicFileNames[3] = fileStart + "CREB/KCl/CREB_KCl_intergenic_" + (extend ? "E" : "") + fragmentLength + ".txt";
	    findTFColocalization(0, intergenicFileNames, tfNames, "tf_intergenic_colocalization.txt");
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    public static void findTFColocalization(int minDist, String fileNames[], String tfNames[], String saveFileName) {
	int nTFs = fileNames.length, nColocalizedPeaks[] = new int[nTFs], nTotPeaks = 0;
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //Write the header
	    bw.write("Chr\t");
	    for (int i=0; i<nTFs; i++) 
		bw.write(tfNames[i] + ",start\t" + tfNames[i] + ",end\t" + tfNames[i] + ",nReads\t");
	    bw.write("\n");
	    TreeSet colocalizedPeaks = new TreeSet();
	    for (int i=0; i<nTFs; i++) 
		colocalizedPeaks.addAll(parsePeaks(tfNames[i], fileNames[i]));
	    //Go through the list of peaks and find the groups which
	    //are overlapping
	    Iterator it = colocalizedPeaks.iterator();
	    TreeSet peaksInGroups = new TreeSet();
	    while (it.hasNext()) {
		TFPeak p = (TFPeak)it.next();
		if (!peaksInGroups.contains(p)) {
		    TreeSet peakGroup = new TreeSet(), tfsInGroup = new TreeSet();
		    peakGroup.add(p);
		    tfsInGroup.add(p.getTFName());
		    peaksInGroups.add(p);
		    int chr = p.getChromosome();
		    //Find the peaks which overlap with this
		    Iterator it2 = colocalizedPeaks.iterator();
		    while (it2.hasNext()) {
			p = (TFPeak)it2.next();
			if (p.overlaps(peakGroup) && !tfsInGroup.contains(p.getTFName()) && !peaksInGroups.contains(p)) {
			    peakGroup.add(p);
			    tfsInGroup.add(p.getTFName());
			    peaksInGroups.add(p);
			}
		    }
		    if (peakGroup.size()>nTFs)
			System.out.println("Found group with " + peakGroup.size());
		    if (peakGroup.size()>1) { //Save the group
			bw.write("chr" + chr + "\t");
			//Make sure to write them in the same order 
			nColocalizedPeaks[0] += savePeakInfo(getPeak(MEF2_STR, peakGroup), bw) ? 1 : 0;
			nColocalizedPeaks[1] += savePeakInfo(getPeak(NPAS4_STR, peakGroup), bw) ? 1 : 0;
			nColocalizedPeaks[2] += savePeakInfo(getPeak(CBP_STR, peakGroup), bw) ? 1 : 0;
			nColocalizedPeaks[3] += savePeakInfo(getPeak(CREB_STR, peakGroup), bw) ? 1 : 0;
			/*nColocalizedPeaks[4] += savePeakInfo(getPeak(CTCF_KCL_STR, peakGroup), bw) ? 1 : 0;
			  nColocalizedPeaks[5] += savePeakInfo(getPeak(CTCF_UN_STR, peakGroup), bw) ? 1 : 0;
			  nColocalizedPeaks[6] += savePeakInfo(getPeak(H8_KCL_STR, peakGroup), bw) ? 1 : 0;
			  nColocalizedPeaks[7] += savePeakInfo(getPeak(H8_UN_STR, peakGroup), bw) ? 1 : 0;
			  nColocalizedPeaks[8] += savePeakInfo(getPeak(WG16_KCL_STR, peakGroup), bw) ? 1 : 0;
			  nColocalizedPeaks[9] += savePeakInfo(getPeak(WG16_UN_STR, peakGroup), bw) ? 1 : 0;*/
			nTotPeaks++;
			bw.write("\n");
		    }
		}
	    }
	    bw.close();
	}
	catch (IOException e) { }
	System.out.println("Found " + nTotPeaks);
	for (int i=0; i<nTFs; i++) 	
	    System.out.println("Found " + nColocalizedPeaks[i] + " for " + tfNames[i] + " out of a total " + parsePeaks(tfNames[i], fileNames[i]).size());
    }

    protected static TFPeak getPeak(String tfName, TreeSet peaks) {
	//Find the first peak with name tfName
	Iterator it = peaks.iterator();
	while (it.hasNext()) {
	    TFPeak p = (TFPeak)it.next();
	    if (p.getTFName().equals(tfName))
		return p;
	}
	return null;
    }

    protected static boolean savePeakInfo(TFPeak p, BufferedWriter bw) throws IOException {
	if (p!=null) {
	    bw.write(p.toString() + "\t");
	    return true;
	}
	bw.write("-\t-\t-\t");
	return false;
    }

	    
    
    public static Vector parsePeaks(String tf, String fName) {
	Vector peaks = new Vector();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fName));
	    String line = br.readLine();
	    while (line!=null) {
		peaks.addElement(new TFPeak(tf, line));
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error opening " + fName + " : " + e.toString());
	}
	return peaks;
    }

}


