//
//
//java -Xmx512m CalculateSequenceDistances -m 3 -M 8
//
//For the DHS categories
//
//java -Xmx512m CalculateSequenceDistances -m 3 -M 8 -f /home/mh168/Projects/ChIP-Seq/TKData/Wui/category_seq_names.txt
//
//bsub -q shared_1d java -Xmx6512m CalculateSequenceDistances -m 6 -M 6 -f /home/mh168/Projects/ChIP-Seq/TKData/Wui/category_seq_all_names.txt
//
//bsub -q shared_1d java -Xmx6512m CalculateSequenceDistances -m 2 -M 6 -f /home/mh168/Projects/ChIP-Seq/TKData/Wui/KLdists/dnase1_stam1_distal_1000_real_control_names.txt
//
//When we do not group things:
//
//bsub -q shared_2h -o /dev/null 'for ((c=1; c<=72; c++)) ; do  for ((C=1; C<=72; C++)) ; do bsub -o /dev/null -q all_15m java -Xmx6512m CalculateSequenceDistances -m 6 -M 6 -s /home/mh168/Projects/ChIP-Seq/TKData/Wui/KLdists/Words6/dnase1_stam1_distal_5000_group_$c\_cell_$C\_kl_dists_6.txt -c $c -C $C ; done ; done'
//
//bsub -q shared_2h -o /dev/null 'for ((c=1; c<=72; c++)) ; do  for ((C=1; C<=72; C++)) ; do bsub -o /dev/null -q all_15m java -Xmx6512m CalculateSequenceDistances -m 6 -M 6 -s /home/mh168/Projects/ChIP-Seq/TKData/Wui/KLdists/WordsControl6/dnase1_stam1_distal_5000_group_$c\_cell_$C\_kl_dists_6.txt -c $c -C $C --control ; done ; done'


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class CalculateSequenceDistances {



    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option motifFlag = argParser.addBooleanOption("motifs");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option seqFileNameFlag = argParser.addStringOption('f', "seqFileName");
	    CmdLineParser.Option pValFlag = argParser.addDoubleOption('p', "pVal");
	    CmdLineParser.Option minLenFlag = argParser.addIntegerOption('m', "minLen");
	    CmdLineParser.Option maxLenFlag = argParser.addIntegerOption('M', "maxLen");
	    CmdLineParser.Option subsampleFlag = argParser.addIntegerOption('S', "subsample");
	    CmdLineParser.Option celltypeFlag = argParser.addIntegerOption('c', "celltype");
	    CmdLineParser.Option commonalityFlag = argParser.addIntegerOption('C', "commonality");
	    CmdLineParser.Option tssDistFlag = argParser.addIntegerOption('t', "tssDist");
	    CmdLineParser.Option randpermFlag = argParser.addBooleanOption("randperm");
	    CmdLineParser.Option controlFlag = argParser.addBooleanOption("control");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean motifs = argParser.getOptionValue(motifFlag)!=null;
	    boolean randperm = argParser.getOptionValue(randpermFlag)!=null;
	    boolean control = argParser.getOptionValue(controlFlag)!=null;
	    tmp = argParser.getOptionValue(pValFlag);
	    double pVal = tmp!=null ? ((Double)tmp).doubleValue() : .00001;
	    tmp = argParser.getOptionValue(minLenFlag);
	    int minLen = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(maxLenFlag);
	    int maxLen = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(subsampleFlag);
	    int subsample = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(celltypeFlag);
	    int celltype = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(commonalityFlag);
	    int commonality = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(tssDistFlag);
	    int tssDist = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(seqFileNameFlag);
	    String seqFileName = tmp!=null ? (String)tmp : "seq_names.txt";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "tmp.txt";
	    //////////////////////////////////////////////////////////////////
	    int nChr = 24;
	    //Find the list of sequences to compare
	    Vector<String> fileNames = new Vector<String>();
	    //When we want to compare everything to one cell-type/specificity
	    String fileName = "/groups/kreiman/hemberg/SeqData/Wui/PeakLocs/Sequences/Tier3/Many/dnase_stam1_manygroups_distal_" + tssDist + "_names_" + (control ? "control_" : randperm ? "dhs_rand_perm_" : "") + "group_" + celltype + "_cell_" + commonality + "_locs_w_50.fa";
	    boolean oneGroupOneType = false;
	    //Keep track of what cell-type and commonality each file corresponds to. Not an elegant solution....
	    HashMap<String, int[]> fileNamesKey = new HashMap<String, int[]>();
	    HashMap<Integer, HashMap<Integer, String>> keyFileNames = new HashMap<Integer, HashMap<Integer, String>>();
	    Vector<String> seq = new Vector<String>();
	    int nGroups = 72, nCellTypes = 72;
	    String dirStart = "/groups/kreiman/hemberg/SeqData/Wui/PeakLocs/Sequences/Tier3/Many/";
	    if (celltype>0 && commonality>0) {
		//Use when we have one group per cell-type
		EnrichedWordFinder.parseSequenceFile(fileName, seq, nChr);
		//Load all 
		String fileStart = dirStart + "dnase_stam1_manygroups_distal_" + tssDist + "_names_" + (control ? "control_" : randperm ? "dhs_rand_perm_" : "") + "group_" + celltype + "_cell_" + commonality + "_locs_w_50.fa";
		for (int i=1; i<=nGroups; i++) {
		    keyFileNames.put(i, new HashMap<Integer, String>());
		    for (int j=1; j<=nCellTypes; j++) {
			if (!(j==celltype && i==commonality)) {
			    String fName = "/groups/kreiman/hemberg/SeqData/Wui/PeakLocs/Sequences/Tier3/Many/dnase_stam1_manygroups_distal_" + tssDist + "_names_" + (control ? "control_" : randperm ? "dhs_rand_perm_" : "") + "group_" + j + "_cell_" + i + "_locs_w_50.fa";
			    fileNames.addElement(fName);
			    fileNamesKey.put(fName, new int[]{j, i});
			    keyFileNames.get(i).put(j, fName);
			}
		    }
		}
		oneGroupOneType = true;
	    }
	    else {
		try {
		    BufferedReader br = new BufferedReader(new FileReader(seqFileName));
		    String line = br.readLine();
		    while (line!=null) {
			if (line.length()>0 && !line.startsWith("#"))
			    fileNames.addElement(line.trim());
			line = br.readLine();
		    }
		    br.close();
		}
		catch (IOException e) {
		    System.err.println("Error reading seq files " + e.toString());
		}
	    }
	    Vector<Vector<String>> seqs = new Vector<Vector<String>>();
	    for (String fName : fileNames) {
		//Vector<String> s = new Vector<String>();
		//EnrichedWordFinder.parseSequenceFile(fName, s, nChr);
		//seqs.addElement(s);
		//System.out.println("Found " + s.size() + " for " + fName);
	    }
	    ///////////////////////////////////////////////////////
	    //Calculate the distribution of single and di-nucleotides across sequences
	    double gc, dinucleotides[] = new double[12];
	    for (int k=0; k<seqs.size(); k++) {
		//char[] seq = seqs.elementAt(k).toCharArray();
		//for () {

		//}
	    }
	    ///////////////////////////////////////////////////////
	    //Calculate the frequency of the words
	    double pWords[][][] = new double[maxLen - minLen + 1][0][0], klDists[][][] = oneGroupOneType ? new double[maxLen - minLen + 1][1][fileNames.size()] : new double[maxLen - minLen + 1][fileNames.size()][fileNames.size()], entropy[][] = new double[maxLen - minLen + 1][fileNames.size()];
	    int nWords[][] = new int[maxLen - minLen + 1][1+fileNames.size()];
	    for (int i=minLen; i<=maxLen; i++) {
		Vector<String> words = new Vector<String>(EnrichedWordFinder.getAllWords(i));
		pWords[i-minLen] = new double[fileNames.size()+1][words.size()];
		for (int k=0; k<fileNames.size(); k++) {
		    String wordDirName = dirStart + "Words" + i + (control ? "_control" : randperm ? "_dhs_rand_perm" : "") + "/";
		    File dir = new File(wordDirName);
		    dir.mkdirs();
		    String wordFileName = fileNames.elementAt(k).replace("locs", "words_" + i).replace("Many/", "Many/Words" + i + (control ? "_control" : randperm ? "_dhs_rand_perm" : "") + "/").replace(".fa", ".txt");
		    try {
			BufferedReader br = new BufferedReader(new FileReader(wordFileName));
			for (int j=0; j<words.size(); j++) {
			    pWords[i-minLen][k][j] = (double)Integer.parseInt(br.readLine());
			    nWords[i-minLen][k] += (int)pWords[i-minLen][k][j];
			}
			br.close();
		    }
		    catch (Exception e) {
			//System.out.println("Failed to read words for " + wordFileName);
			Vector<String> s = new Vector<String>();
			EnrichedWordFinder.parseSequenceFile(fileNames.elementAt(k), s, nChr);
			BufferedWriter bw = new BufferedWriter(new FileWriter(wordFileName));
			for (int j=0; j<words.size(); j++) {
			    pWords[i-minLen][k][j] = (double)EnrichedWordFinder.numberOfWords(words.elementAt(j), s);
			    nWords[i-minLen][k] += (int)pWords[i-minLen][k][j];
			    bw.write((int)pWords[i-minLen][k][j] + "\n");
			}
			bw.close();
			//System.out.println("Calculated probabilities for word " + j + " for " + seqs.size());
		    }
		}
		//Calculate the slope for the distribution of word frequencies
		for (int j=0; j<words.size(); j++) {
		    for (int k=0; k<fileNames.size(); k++) 
			pWords[i-minLen][k][j] /= nWords[i-minLen][k];
		}
		if (oneGroupOneType) {
		    //Store the information for the seqs that we are comparing to last
		    for (int j=0; j<words.size(); j++) {
			pWords[i-minLen][fileNames.size()][j] = (double)EnrichedWordFinder.numberOfWords(words.elementAt(j), seq);
			nWords[i-minLen][fileNames.size()] += (int)pWords[i-minLen][fileNames.size()][j];
		    }
		    for (int j=0; j<words.size(); j++) 
			pWords[i-minLen][fileNames.size()][j] /= nWords[i-minLen][fileNames.size()];
		    System.out.println("Calculated probabilities for " + nWords[i-minLen].length);
		    /////////////////////////////////
		    for (int j=0; j<fileNames.size(); j++) {
			klDists[i-minLen][0][j] = calculateKLDistance(pWords[i-minLen][j], pWords[i-minLen][fileNames.size()]);
			//System.out.println("KL distance between " + fileNames.elementAt(j).substring(fileNames.elementAt(j).lastIndexOf("/") + 1) + " and " + fileName.substring(fileName.lastIndexOf("/") + 1) + " for words of length " + i + " is " + format.format(klDists[i-minLen][0][j]));
		    }
		}
		else {
		    
		    //Calculate the pairwise KL-distances
		    for (int j=0; j<fileNames.size(); j++) {
			entropy[i-minLen][j] = calculateSequenceEntropy(pWords[i-minLen][j]);
			System.out.println("Entropy is " + format.format(entropy[i-minLen][j]) + " for " + fileNames.elementAt(j).substring(fileNames.elementAt(j).lastIndexOf("/") + 1));
			for (int k=0; k<fileNames.size(); k++) {
			    if (j!=k) {
				klDists[i-minLen][j][k] = calculateKLDistance(pWords[i-minLen][j], pWords[i-minLen][k]);
				//System.out.println("KL distance between " + fileNames.elementAt(j).substring(fileNames.elementAt(j).lastIndexOf("/") + 1) + " and " + fileNames.elementAt(k).substring(fileNames.elementAt(k).lastIndexOf("/") + 1) + " for words of length " + i + " is " + format.format(klDists[i-minLen][j][k]));
			    }
			}
		    }
		}
		if (oneGroupOneType) {
		    //Use a different format for these.
		    //In the file-name, we list the cell-type and the specificity, then the three columns show other cell-type, other specificity and KL-distance
		    System.out.println("Saving to " + saveFileName);
		    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		    
		    for (int j=1; j<=nCellTypes; j++) {
			for (int k=1; k<=nGroups; k++) {
			    bw.write(j + "\t" + k + "\t");
			    if (keyFileNames.containsKey(j) && keyFileNames.get(j).containsKey(k))
				bw.write(format.format(klDists[i-minLen][0][fileNames.indexOf(keyFileNames.get(j).get(k))]) + "\n");
			    else
				bw.write("0\n");
			}
		    }
		    bw.close();
		}
		else {
		    BufferedWriter bw = new BufferedWriter(new FileWriter(seqFileName.replace(".txt", "_" + i + "_kl_dists.txt")));
		    for (int j=0; j<seqs.size(); j++) {
			bw.write(format.format(-entropy[i-minLen][j]) + "\t");
			for (int k=0; k<seqs.size(); k++) 
			    bw.write(format.format(klDists[i-minLen][j][k]) + "\t");
			bw.write("\n");
		    }
		    bw.close();
		}
	    }
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static double calculateKLDistance(double a[], double b[]) {
	double s = 0;
	if (a.length==b.length) {
	    for (int i=0; i<a.length; i++) {
		if (b[i]>0 && a[i]>0)
		    s += a[i]*Math.log(a[i]/b[i]);
	    }
	}
	return s;
    }

    public static double calculateSequenceEntropy(double a[]) {
	double s = 0;
	for (int i=0; i<a.length; i++) {
	    if (a[i]>0)
		s += a[i]*Math.log(a[i]);
	}
	return s;
    }

}





