//Use this method to analyze the presence of enhancers, promoters and
//TFBSs at conservation islands.
//
//bsub -q shared_2h java -Xmx4096m -server ConservationIslandOverlap -o -c 19


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class ConservationIslandOverlap {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option excludeExonsFlag = argParser.addBooleanOption('e', "excludeExons");
	    CmdLineParser.Option excludePromotersFlag = argParser.addIntegerOption('p', "excludePromoters");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option minDistFlag = argParser.addIntegerOption('d', "minDist");	    
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean excludeExons = argParser.getOptionValue(excludeExonsFlag)!=null;
	    tmp = argParser.getOptionValue(excludePromotersFlag);
	    int promoterDist = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(minDistFlag);
	    int minIslandDist = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "/groups/neuroduo/martin/Genomes/Mm/v9/conservation_islands_" + (excludeExons ? "exons_" : "") + (promoterDist>0 ? "promoters_" : "") + "w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".bed";
	    ///////////////////////////////////////////////////////
	    int nChr = 21, nWindows = 0;;
	    String exonFileName = "/groups/neuroduo/martin/Genomes/Mm/v9/all_EXNs.tsv";
	    RefSeqUCSCEnsemblGenes geneInfo[] = new RefSeqUCSCEnsemblGenes[nChr];
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));

	    System.out.println("Saving results to " + saveFileName);
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    //Load the annotation
		    geneInfo[i-1] = new RefSeqUCSCEnsemblGenes(i, orchestra, species);
		    //Load the transcripts
		    //TreeSet<TranscribedRegion2> novelTranscripts = ;
		    //Load the enhancers

		    //Load the peaks
		    TreeMap<String, Vector<Peak>> category2peaks = new TreeMap<String, Vector<Peak>>();
		    for (int i=0; i<PeaksNearTranscripts.PEAK_CATEGORIES.length; i++)
			category2peaks.put(PeaksNearTranscripts.PEAK_CATEGORIES[i], loadPeaks(chr));
		    //Load the conservation islands
		    ConservationIslands consIslands = new ConservationIslands(i);
		    //Find out how many of the islands are near a promoter, enhancer or peak
		    for (ConservationIsland consIsland : consIslands) {
			if (consIsland.overlaps(geneInfo, promoterDist))
			    consIslandsCategoriesMap.put(PROMOTER_CONS_ISLAND_STR, consIslandsCategoriesMap.get(PROMOTER_CONS_ISLAND_STR) + 1);
			else if (consIsland.overlaps(enhancers))
			    consIslandsCategoriesMap.put(ENHANCER_CONS_ISLAND_STR, consIslandsCategoriesMap.get(ENHANCER_CONS_ISLAND_STR) + 1);
		    }

		    //Get all the exons as well
		    TreeSet<Exon> exons = excludeExons ? parseExons(exonFileName, i) : new TreeSet<Exon>();
		    PhastConsParser2 phastCons = new PhastConsParser2(i, true, orchestra);
		    //Go through the genome one bin at a time, exclude the ones that are near promoters or overlapping exons
		    int windowStart = 0, islandStart = -1;
		    double meanConsIsland = 0;
		    Iterator<Exon> exonIt = exons.iterator();
		    Exon exon = exonIt.hasNext() ? exonIt.next() : null;
		    Iterator<RefSeqUCSCEnsemblGene> geneIt = geneInfo[i-1]==null ? (new TreeSet<RefSeqUCSCEnsemblGene>()).iterator() : geneInfo[i-1].getGenes().iterator();
		    RefSeqUCSCEnsemblGene gene = geneIt.hasNext() ? geneIt.next() : null;
		    TreeMap<Integer, Integer> islandStartEnd = new TreeMap<Integer, Integer>();
		    while (windowStart + windowSize<RefSeqUCSCEnsemblGenes.CHROMOSOME_LENGTH[i-1] && phastCons.hasMoreIslands()) {
			//if (!overlapsExon(windowStart, windowStart + windowSize, exons) && !overlapsPromoter(windowStart, windowStart + windowSize, geneInfo[i-1], promoterDist)) {
			if ((exon==null || windowStart + windowSize<exon.getStart()) && (gene==null || windowStart + windowSize<gene.getStart() - promoterDist)) {
			    double cons[] = phastCons.getNextIsland(windowSize), totCons = 0;
			    for (int j=0; j<cons.length; j++) { totCons += cons[j]; }
			    if (totCons>cons.length/2 /*isConservationIsland()*/) { //Store in .bed format
				if (islandStart==-1)
				    islandStart = windowStart;
				nWindows++;
				meanConsIsland += totCons/windowSize;
			    }
			    else if (islandStart!=-1) {
				islandStartEnd.put(islandStart, windowStart + windowSize);
				islandStart = -1;
				meanConsIsland = 0;
			    }
			}
			else 
			    phastCons.getNextIsland(windowSize);
			windowStart = phastCons.getNextIslandStart();
			if (exonIt.hasNext() && windowStart>exon.getEnd()) 
			    exon = exonIt.next();
			if (geneIt.hasNext() && windowStart>gene.getStart() + promoterDist) 
			   gene = geneIt.next();
		    }
		    if (islandStart!=-1)  //Write the last window
			islandStartEnd.put(islandStart, windowStart);
		    //Merge islands that are within minIslandDist
		    int islandEnd = islandStartEnd.get(islandStartEnd.firstKey());
		    islandStart = islandStartEnd.firstKey();
		    for (Integer s : islandStartEnd.keySet()) {
			if (s - islandEnd>minIslandDist) {
			    String chrStr = Integer.toString(i);
			    if (i==20)
				chrStr = "X";
			    else if(i==21)
				chrStr = "Y";
			    bw.write("chr" + chrStr + "\t" + islandStart + "\t" + islandEnd + "\t1\n");
			    islandStart = s;
			    islandEnd = islandStartEnd.get(s);
			}
			else
			    islandEnd = islandStartEnd.get(s);
		    }
		    //Write the last window
		    String chrStr = Integer.toString(i);
		    if (i==20)
			chrStr = "X";
		    else if(i==21)
			chrStr = "Y";
		    bw.write("chr" + chrStr + "\t" + islandStart + "\t" + islandStartEnd.get(islandStartEnd.lastKey()) + "\t1\n");
		}
	    }
	    bw.close();
	    //Convert to .bw format and create a load-line
	    String trackInfoStr = "track type=bigWig name=\"conservation_islands" + (excludeExons ? "_exons" : "") + (promoterDist>0 ? "_promoters" : "") + "\" description=\"conservation_islands" + (excludeExons ? "_exons" : "") + (promoterDist>0 ? "_promoters" : "") + "\" priority=" + (1567 + (excludeExons ? 1 : 0) + (promoterDist>0 ? 2 : 0)) + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:1 color=\"255,127,0\" smoothingWindow=off bigDataUrl=http://neuroduo.hms.harvard.edu/bigWig/mh/" + (saveFileName.substring(saveFileName.lastIndexOf("/") + 1)).replace("bed", "bw");
	    CreateWigFiles.addTrackInfo(trackInfoStr, CreateWigFiles.TRACK_INFO_FILE_NAME);
	    CreateWigFiles.createWWWLink(saveFileName, "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/");
	    System.out.println("Found " + nWindows + " that were conserved.");
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static boolean overlapsExon(int start, int end, Vector<Exon> exons) {
	for (Exon e : exons) {
	    if (e.overlaps(start, end))
		return true;
	}
	return false;
    }

    public static boolean overlapsPromoter(int start, int end, RefSeqUCSCEnsemblGenes geneInfo, int promoterDist) {
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes()) {
	    if ((g.getStart() - promoterDist<=start && g.getStart() + promoterDist>=start) || (g.getStart() - promoterDist<=end && g.getStart() + promoterDist>=end) || (g.getStart() - promoterDist>start && g.getStart() + promoterDist<end))
		return true;
	}
	return false;
    }

    public static boolean isConservationIsland(double consScores[]) {
	double consSum = 0, maxCons = 0;
	for (int i=0; i<consScores.length; i++) {
	    consSum += consScores[i];
	    maxCons = consScores[i]>maxCons ? consScores[i] : maxCons;
	}
	return /*maxCons>.9 || */consSum>consScores.length/2;
    }

    public static TreeSet<Exon> parseExons(String exonFileName, int chr) {
	TreeSet<Exon> exons = new TreeSet<Exon>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(exonFileName));
	    String line = br.readLine();
	    Pattern exonPattern = Pattern.compile("([\\w|-|_]+)\\s+chr(\\w+)\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = exonPattern.matcher(line);
		if (m.find()) {
		    int ch;
		    if (m.group(2).equals("X"))
			ch = 20;
		    else if (m.group(2).equals("Y"))
			ch = 21;
		    else if (m.group(2).equals("M") || m.group(2).contains("random")) {
			line = br.readLine();
			continue;
		    }
		    else
			ch = Integer.parseInt(m.group(2));
		    if (ch==chr) {
			Exon ex = new Exon(chr, Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4)), m.group(1));
			if (!exons.contains(ex))
			    exons.add(ex);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing exon file " + e.toString());
	}
	System.out.println("Found " + exons.size() + " exons for chr " + chr);
	return exons;
    }

}


