import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class ConservationIslands {

    protected int chr = -1;
    protected String species;
    protected TreeSet<ConservationIsland> conservationIslands = new TreeSet<ConservationIsland>();

    public ConservationIslands(boolean orchestra, String sp) {
	species = sp;
	parseFile(orchestra, false);
    }

    public ConservationIslands(int ch, boolean orchestra, String sp) {
	chr = ch;
	species = sp;
	parseFile(orchestra, false);
    }

    public ConservationIslands(int ch, boolean orchestra, String sp, boolean hg19) {
	chr = ch;
	species = sp;
	parseFile(orchestra, hg19);
    }

    public ConservationIslands(int ch, String sp) {
	//Use when not parsing from file
	chr = ch;
	species = sp;
    }

    protected void parseFile(boolean orchestra, boolean hg19) {
	String fileName = (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/") + "Genomes/" + (species.equals("mm") ? "Mm/v9/conservation_islands_w_10" + (chr==-1 ? "" : "_" + chr) + ".bed" : "Hs/v3" + (hg19 ? "7" : "6") + "/conservation_islands_hs_w_10" + (chr==-1 ? "" : "_" + chr) + ".bed");
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    System.out.println(line);
	    //chr start end type
	    Pattern bedPattern = Pattern.compile("chr(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = bedPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(1).equals("X"))
			c = species.equals("mm") ? 20 : 23;
		    else if (m.group(1).equals("Y"))
			c = species.equals("mm") ? 21 : 24;
		    else
			c = Integer.parseInt(m.group(1));
		    if (chr==-1 || c==chr)
			conservationIslands.add(new ConservationIsland(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4))));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing conservationIslands: " + e.toString());
	}
	System.out.println("Found " + conservationIslands.size() + " conservation islands from " + fileName + " with first " + conservationIslands.iterator().next().toString());
    }

    public boolean overlaps(Peak p) { return overlaps(p.getChromosome(), p.getCenter()); }

    public boolean overlapsWithin(int chr, int p, int dist) {
	//Returns true if there is a cons island within dist in either direction
	return overlaps(chr, p-dist, p+dist);
    }

    public ConservationIsland getOverlappingConservationIsland(int chr, int p) {
	for (ConservationIsland c : conservationIslands) {
	    if (c.getChromosome()==chr && c.overlaps(p)) {
		//System.out.println(p + " overlaps " + c.toString());
		return c;
	    }
	}
	return null;
    }

    public boolean overlaps(int chr, int p) { return getOverlappingConservationIsland(chr, p)!=null; }

    public boolean overlaps(Exon e) { return overlaps(e.getChromosome(), e.getStart(), e.getEnd()); }

    public boolean overlaps(int chr, int s, int e) { return getOverlappingConservationIslands(chr, s, e).size()>0; }

    public TreeSet<ConservationIsland> getOverlappingConservationIslands(TranscribedRegion2 t) { return getOverlappingConservationIslands(t.getChromosome(), t.getStart(), t.getEnd()); }

    public TreeSet<ConservationIsland> getOverlappingConservationIslands(Exon e) { return getOverlappingConservationIslands(e.getChromosome(), e.getStart(), e.getEnd()); }

    public TreeSet<ConservationIsland> getOverlappingConservationIslands(int chr, int s, int e) { return getOverlappingConservationIslands(chr, s, e, 0); }

    public TreeSet<ConservationIsland> getOverlappingConservationIslands(int chr, int s, int e, int includeStartOverlap) {
	//The last argument is used to exclude the islands that overlap the start of the transcript. If the values is zero, then we don't bother. If the value is >0, then we exclude CIs that overlap s and for <0 CIs that overlap e.
	TreeSet<ConservationIsland> overlapping = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : conservationIslands) {
	    if (c.getChromosome()==chr && c.overlaps(s, e)) {
		if (includeStartOverlap==0 || (includeStartOverlap>0 && !c.overlaps(s)) || (includeStartOverlap<0 && !c.overlaps(e))) 
		    overlapping.add(c);
	    }
	}
	return overlapping;
    }

    public TreeMap<ConservationIsland, Exon> conservedExons(TreeSet<Exon> exons) {
	TreeMap<ConservationIsland, Exon> conservedExons = new TreeMap<ConservationIsland, Exon>();
	for (Exon e : exons) {
	    for (ConservationIsland c : getOverlappingConservationIslands(e))
		conservedExons.put(c, e);
	}
	//System.out.println("Found " + nOverlap + " overlapping islands.");
	return conservedExons;
    }

    public void add(ConservationIsland c) {
	if (c.getChromosome()==chr)
	    conservationIslands.add(c);
    }

    public TreeSet<ConservationIsland> getConservationIslands() { return conservationIslands; }

    public TreeSet<ConservationIsland> getType(int t) { 
	TreeSet<ConservationIsland> cis = new TreeSet<ConservationIsland>();
	for (ConservationIsland ci : conservationIslands) {
	    if (ci.getType()==t)
		cis.add(ci);
	}
	return cis; 
    }

}

///////////////////////////////////////////////////////

