//Use this method to find clusters of ESTs and to compare against various features of the annotation
//
//bsub -q shared_2h java -Xmx4096m -server ESTClusterAnalysis -o -c 22 --species hs --nMC 100 --spliced 1 --hela
//

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.Tally;


public class ESTClusterAnalysis {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option reverseFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option helaFlag = argParser.addBooleanOption("hela");
	    CmdLineParser.Option k562Flag = argParser.addBooleanOption("k562");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option maxDistanceFlag = argParser.addIntegerOption('d', "maxDistance");	    
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption("nMC");	    
	    CmdLineParser.Option spliceFlag = argParser.addIntegerOption("spliced");	    
	    //Get the lncRNA for the 
	    CmdLineParser.Option grimmondLncRNAFlag = argParser.addStringOption("grimmondLncRNA");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean forward = argParser.getOptionValue(reverseFlag)==null;
	    boolean hela = argParser.getOptionValue(helaFlag)!=null;
	    boolean k562 = argParser.getOptionValue(k562Flag)!=null;
	    //tmp = argParser.getOptionValue(peakFileNameFlag);
	    //String peakFileName = tmp!=null ? (String)tmp : (!orchestra ? "/home/mhemberg" : "/home/mh168") + "/Projects/ChIP-Seq/Encode/PeakFiles/" + (hela ? "wgEncodeUwDnaseSeqPeaksRep1Helas3.narrowPeak" : (k562 ? "wgEncodeUwDnaseSeqPeaksRep1K562.narrowPeak" : "dnase_names.txt"));
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(spliceFlag);
	    int spliced = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "./EST/" + species + "_EST_distribution_" + (spliced==0 ? "" : (spliced<0 ? "un" : "") + "spliced_" )+ (hela ? "hela_" : (k562 ? "k562_" : "")) + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    tmp = argParser.getOptionValue(maxDistanceFlag);
	    int maxDist = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(grimmondLncRNAFlag);
	    String grimmondLncRNAFileName = tmp!=null ? (String)tmp : species.equals("hs") ? "/home/mh168/Projects/ChIP-Seq/grimmond_lncrna_names.txt" : null;
	    ///////////////////////////////////////////////////////
	    int nChr = FindTranscripts.numberOfChromosomes(species) + 1, maxClusterDist = 100, maxPromoterDist = 100, maxEnhancerDist = 100;
	    String estDirName = !species.equals("mm") ? "/groups/neuroduo/jesse/human_annot_for_mh/human_ESTs/" : "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/";
	    String lncRNAFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    String macroRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    String extragenicEnhancerFileName = !species.equals("mm") ? (hela ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed" : "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_k562_enhancer_list_hg18.tsv") : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lRNA_lMe3_real.tsv";
	    String peakFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? (!orchestra ? "/home/mhemberg" : "/home/mh168") + "/Projects/ChIP-Seq/Encode/PeakFiles/" + (hela ? "hela_tf_names.txt" : (k562 ? "wgEncodeUwDnaseSeqPeaksRep1K562.narrowPeak" : "dnase_names.txt")) : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/tf_kcl_names";
	    String k4me3PeakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/H3K4Me3_KCl_peaks_names";

	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //Save the number of EST clusters as a function of the number of cell types
	    BufferedWriter bwCellTypes = new BufferedWriter(new FileWriter(saveFileName.replace("distribution", "celltype_distribution")));
	    String categories[] = new String[]{"Promoters", "ncPromoters", "Intronic", /*"LncRNAs", "LncRNAs_cons",*/ "Enhancers", "Enhancers_cons", "Peaks", "Peaks_cons", "Ctcf", "Ctcf_cons", "K4me3", "K4me3_cons", "Remaining", "Remaining_cons"};
	    if (chr==-1 || chr==1) {
		bw.write("#chr");
		bwCellTypes.write("#CellTypes");
		for (int i=0; i<categories.length; i++) {
		    bw.write("\t" + categories[i] + "\t" + categories[i] + "_clust");
		    bwCellTypes.write("\t" + categories[i] + "\t" + categories[i] + "_clust");
		}
		bw.write("\n");
		bwCellTypes.write("\n");
	    }
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    //Load the ESTs
		    TreeSet<EST> ests = new TreeSet<EST>(CompareCageEstRNASeqStarts.parseEstTags(estDirName, chr, forward, spliced, true, species).values());
		    //Find the total number of tissue types
		    TreeMap<Integer, Integer> tissueTypes = new TreeMap<Integer, Integer>();
		    for (EST est : ests) {
			if (!tissueTypes.containsKey(est.getTissueType()))
			    tissueTypes.put(est.getTissueType(), 0);
			tissueTypes.put(est.getTissueType(), tissueTypes.get(est.getTissueType()) + 1);
		    }
		    //Exclude 0 and -1
		    //System.out.println("Found " + tissueTypes.remove(0).toString() + " ESTs with tissue type 0");
		    //System.out.println("Found " + tissueTypes.remove(-1).toString() + " ESTs with tissue type -1");
		    //for (Integer k : tissueTypes.keySet())
		    //System.out.println(k + "\t" + tissueTypes.get(k));

		    TreeMap<EST, RefSeqUCSCEnsemblGene> ests2Promoters = new TreeMap<EST, RefSeqUCSCEnsemblGene>(), ests2NcPromoters = new TreeMap<EST, RefSeqUCSCEnsemblGene>(), ests2Intronic = new TreeMap<EST, RefSeqUCSCEnsemblGene>();

		    TreeMap<EST, TranscribedRegion2> ests2LncRNAs = new TreeMap<EST, TranscribedRegion2>(), ests2LncRNAsCons = new TreeMap<EST, TranscribedRegion2>();

		    TreeMap<EST, Enhancer2> ests2Enhancers = new TreeMap<EST, Enhancer2>(), ests2EnhancersCons = new TreeMap<EST, Enhancer2>();
		    TreeMap<EST, Peak> ests2Peak = new TreeMap<EST, Peak>(), ests2PeakCons = new TreeMap<EST, Peak>();
		    TreeMap<EST, Peak> ests2K4me3Peak = new TreeMap<EST, Peak>(), ests2K4me3PeakCons = new TreeMap<EST, Peak>();
		    TreeMap<EST, Peak> ests2Ctcf = new TreeMap<EST, Peak>(), ests2CtcfCons = new TreeMap<EST, Peak>();

		    //Get the conservation info
		    ConservationIslands conservationIslands = new ConservationIslands(i, orchestra, species);
		    //Load the annotations
		    RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(chr, orchestra, species);
		    //Load macro and lincRNAs
		    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species, i);
		    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species, i);
		    //Parse the LncRNA obtained from the Grimmond data
		    //Vector<ReducedModel2> lncRNAsForward = parseGrimmondLncRNA(grimmondLncRNAFileName, i, true, true), lncRNAsReverse = parseGrimmondLncRNA(grimmondLncRNAFileName, i, false, true);
		    //Enhancers
		    TreeSet<Enhancer2> extragenicEnhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName, i);
		    Vector<Peak> peaks = new Vector<Peak>();
		    String tfNamesInStudy[] = null;
		    String cellType = hela ? "Helas3" : "K562";
		    if (species.equals("mm")) {
			if (peakFileName.contains("tf_kcl_names")) {
			    peaks = new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values()); 
			    tfNamesInStudy = new String[]{"CBP_Millipore_KCl_B1","CREB_SC_KCl_B1","Npas4_Npas4_KCl_B1","SRF_H300_KCl_B1"};
			}
		    }
		    else 
			peaks = new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(peakFileName, cellType, chr));
			//peaks = new Vector<Peak>(EncodePeaks.parseEncodePeakFileTF(peakFileName, "Dnase", chr));
		    System.out.println("Found " + peaks.size() + " peaks");

		    Vector<Peak> k4me3Peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(k4me3PeakFileName, cellType, chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(k4me3PeakFileName, chr).values());


		    //Get the CTCF peaks
		    String ctcfPeakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1" + cellType + "Ctcf.narrowPeak" : null;
		    String ctcfNames[] = new String[]{"Ctcf"};
		    Vector<Peak> ctcfPeaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(ctcfPeakFileName, cellType, chr)) : new Vector<Peak>();
		    System.out.println("Found " + ctcfPeaks.size() + " CTCF peaks");

		    //Remove the ESTs that are close to a known promoter
		    TreeSet<RefSeqUCSCEnsemblGene> genesWithESTs = new TreeSet<RefSeqUCSCEnsemblGene>(), ncGenesWithESTs = new TreeSet<RefSeqUCSCEnsemblGene>(), overlappingGeneESTs = new TreeSet<RefSeqUCSCEnsemblGene>();
		    for (EST est : ests) {
			RefSeqUCSCEnsemblGene nearestGene = geneInfo.getNearestTSSGene(est.getStart(), forward);
			if (!nearestGene.isNonCoding() && Math.abs(est.getStart() - nearestGene.getStart())<maxPromoterDist) {
			    ests2Promoters.put(est, nearestGene);
			    genesWithESTs.add(nearestGene);
			    //Also consider the conserved genes

			}
		    }
		    for (EST est : ests2Promoters.keySet())
			ests.remove(est);
		    ////////////////////////////////////////////
		    //ncRNAs
		    for (EST est : ests) {
			RefSeqUCSCEnsemblGene nearestGene = geneInfo.getNearestTSSGene(est.getStart(), forward);
			if (nearestGene.isNonCoding() && Math.abs(est.getStart() - nearestGene.getStart())<maxPromoterDist) {
			    ests2NcPromoters.put(est, nearestGene);
			    ncGenesWithESTs.add(nearestGene);
			}
			else if (species.equals("mm") && FindConservationIslands.overlapsMacroRNAPromoter(i, est.getStart() - maxPromoterDist, est.getStart() + maxPromoterDist, macroRNAs)) {
			    ests2NcPromoters.put(est, nearestGene);
			    ncGenesWithESTs.add(nearestGene);
			}
			else if (species.equals("mm") && FindConservationIslands.overlapsLncRNAPromoter(i, est.getStart() - maxPromoterDist, est.getStart() + maxPromoterDist, lncRNAs)) {
			    ests2NcPromoters.put(est, nearestGene);
			    ncGenesWithESTs.add(nearestGene);
			}
		    }
		    for (EST est : ests2NcPromoters.keySet())
			ests.remove(est);
		    ////////////////////////////////////////////
		    //Remove the ones that are inronic to known genes
		    for (EST est : ests) {
			RefSeqUCSCEnsemblGene overlappingGene = geneInfo.getOverlappingGene(est.getStart(), est.getStart(), forward);
			if (overlappingGene!=null) {
			    ests2Intronic.put(est, overlappingGene);
			    overlappingGeneESTs.add(overlappingGene);
			}
		    }
		    for (EST est : ests2Intronic.keySet())
			ests.remove(est);
		    //LncRNAs
		    /*if (species.equals("hs")) {
			TreeSet<TranscribedRegion2> overlappingLncRNAs = new TreeSet<TranscribedRegion2>(), overlappingLncRNAsCons = new TreeSet<TranscribedRegion2>();
			for (EST est : ests) {
			    for (int j=0; j<lncRNAsForward.size(); j++) {
				TranscribedRegion2 overlapsLncRNA = getOverlappingTranscript(lncRNAsForward.elementAt(j).getSegments(), i, true, est.getStart(), est.getStart(), maxDist);
				if (overlapsLncRNA==null) 
				    overlapsLncRNA = getOverlappingTranscript(lncRNAsReverse.elementAt(j).getSegments(), i, false, est.getStart(), est.getStart(), maxDist);
				if (overlapsLncRNA!=null) {
				    System.out.println("Found EST " + est.toString() + " overlapping " + overlapsLncRNA.toString());
				    if (conservationIslands.overlaps(i, est.getStart(), est.getStart())) {
					overlappingLncRNAsCons.add(overlapsLncRNA);
					ests2LncRNAsCons.put(est, overlapsLncRNA);
				    }
				    else {
					overlappingLncRNAs.add(overlapsLncRNA);
					ests2LncRNAs.put(est, overlapsLncRNA);
				    }
				    break;
				}
			    }
			}
			for (EST est : ests2LncRNAs.keySet())
			    ests.remove(est);
			for (EST est : ests2LncRNAsCons.keySet())
			    ests.remove(est);
			    }*/
		    //Enhancers
		    TreeSet<Enhancer2> overlappingEnhancers = new TreeSet<Enhancer2>(), overlappingEnhancersCons = new TreeSet<Enhancer2>();
		    for (EST est : ests) {
			Enhancer2 overlapsEnhancer = FindConservationIslands.getOverlappingEnhancer(extragenicEnhancers, i, est.getStart(), est.getStart(), maxEnhancerDist);
			if (overlapsEnhancer!=null) {
			    if (conservationIslands.overlaps(i, est.getStart(), est.getStart())) {
				overlappingEnhancersCons.add(overlapsEnhancer);
				ests2EnhancersCons.put(est, overlapsEnhancer);
				System.out.println("Found " + est.isSpliced() + " EST " + est.toString() + " overlapping conserved enhancer " + overlapsEnhancer.toString());
			    }
			    else {
				overlappingEnhancers.add(overlapsEnhancer);
				ests2Enhancers.put(est, overlapsEnhancer);
			    }
			}
		    }
		    for (EST est : ests2Enhancers.keySet())
			ests.remove(est);
		    for (EST est : ests2EnhancersCons.keySet())
			ests.remove(est);
		    //Peaks
		    TreeSet<Peak> peakClusters = new TreeSet<Peak>(), peakClustersCons = new TreeSet<Peak>(), k4me3PeakClusters = new TreeSet<Peak>(), k4me3PeakClustersCons = new TreeSet<Peak>();
		    for (EST est : ests) {
			Peak p = FindConservationIslands.getOverlappingPeak(k4me3Peaks, i, est.getStart(), est.getStart(), maxDist, null);
			if (p!=null) {
			    if (conservationIslands.overlaps(i, est.getStart(), est.getStart())) {
				ests2K4me3PeakCons.put(est, p);
				k4me3PeakClustersCons.add(p);
				//System.out.println("Found " + est.isSpliced() + " EST" + est.toString() + " overlapping conserved DHS " + p.toString());
			    }
			    else {
				ests2K4me3Peak.put(est, p);
				k4me3PeakClusters.add(p);
			    }
			}
			else {
			    p = FindConservationIslands.getOverlappingPeak(peaks, i, est.getStart(), est.getStart(), maxDist, null);
			    if (p!=null) {
				if (conservationIslands.overlaps(i, est.getStart(), est.getStart())) {
				    ests2PeakCons.put(est, p);
				    peakClustersCons.add(p);
				    //System.out.println("Found " + est.isSpliced() + " EST" + est.toString() + " overlapping conserved DHS " + p.toString());
				}
				else {
				    ests2Peak.put(est, p);
				    peakClusters.add(p);
				}
			    }
			}
		    }
		    for (EST est : ests2Peak.keySet())
			ests.remove(est);
		    for (EST est : ests2PeakCons.keySet())
			ests.remove(est);
		    //Insulators
		    TreeSet<Peak> ctcfClusters = new TreeSet<Peak>(), ctcfClustersCons = new TreeSet<Peak>();
		    for (EST est : ests) {
			Peak p = FindConservationIslands.getOverlappingPeak(ctcfPeaks, i, est.getStart(), est.getStart(), maxDist, null);
			if (p!=null) {
			    if (conservationIslands.overlaps(i, est.getStart(), est.getStart())) {
				ests2CtcfCons.put(est, p);
				ctcfClustersCons.add(p);
			    }
			    else {
				ests2Ctcf.put(est, p);
				ctcfClusters.add(p);
			    }
			}
		    }
		    for (EST est : ests2Ctcf.keySet())
			ests.remove(est);
		    for (EST est : ests2CtcfCons.keySet())
			ests.remove(est);
		    //Group the remaining ESTs into clusters
		    Vector<TreeSet<EST>> estClusters = new Vector<TreeSet<EST>>(), estClustersCons = new Vector<TreeSet<EST>>();
		    int nNonpromoterESTs = 0, nNonpromoterESTsCons = 0;
		    TreeSet<EST> cluster = new TreeSet<EST>();
		    int clusterStart = Integer.MAX_VALUE, clusterEnd = 0;
		    for (EST est : ests) {
			if (cluster.size()==0 || Math.abs(est.getStart() - cluster.last().getStart())<maxClusterDist) {
			    cluster.add(est);
			    clusterStart = Math.min(clusterStart, est.getStart());
			    clusterEnd = Math.max(clusterEnd, est.getStart());
			}
			else {
			    if (conservationIslands.overlaps(i, clusterStart, clusterEnd)) {
				estClustersCons.addElement(cluster);
				nNonpromoterESTsCons += cluster.size();
			    }
			    else {
				estClusters.addElement(cluster);
				nNonpromoterESTs += cluster.size();
			    }
			    cluster = new TreeSet<EST>();
			    clusterStart = Integer.MAX_VALUE;
			    clusterEnd = 0;
			}
		    }
		    bw.write(i + "\t" + ests2Promoters.size() + "\t" + genesWithESTs.size() + "\t" + ests2NcPromoters.size() + "\t" + ncGenesWithESTs.size() + "\t" + ests2Intronic.size() + "\t" + overlappingGeneESTs.size() +"\t" + ests2Enhancers.size() + "\t" + overlappingEnhancers.size() + "\t" + ests2EnhancersCons.size() + "\t" + overlappingEnhancersCons.size() + "\t" + ests2Peak.size() + "\t" + peakClusters.size() + "\t" + ests2PeakCons.size() + "\t" + peakClustersCons.size() + "\t" + ests2Ctcf.size() + "\t" + ctcfClusters.size() + "\t" + ests2CtcfCons.size() + "\t" + ctcfClustersCons.size() + "\t" + ests2K4me3Peak.size() + "\t" + k4me3PeakClusters.size() + "\t" + ests2K4me3PeakCons.size() + "\t" + k4me3PeakClustersCons.size() + "\t" + nNonpromoterESTs + "\t" + estClusters.size() + "\t" + nNonpromoterESTsCons + "\t" + estClustersCons.size() + "\n");
		    //Now we can find out how the number of clusters change as a function of the number of cell types sampled
		    int nCellTypes = 0;
		    while (++nCellTypes<tissueTypes.size()) {
			bwCellTypes.write(Integer.toString(nCellTypes));
			int nPromoterESTs[] = numberOfUniqueESTsGene(ests2Promoters, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nPromoterESTs[0] + "\t" + nPromoterESTs[1]);
			int nNcPromoterESTs[] = numberOfUniqueESTsGene(ests2NcPromoters, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nNcPromoterESTs[0] + "\t" + nNcPromoterESTs[1]);
			int nIntronicESTs[] = numberOfUniqueESTsGene(ests2Intronic, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nIntronicESTs[0] + "\t" + nIntronicESTs[1]);
			int nLncRNAESTs[] = numberOfUniqueESTsTranscript(ests2LncRNAs, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nLncRNAESTs[0] + "\t" + nLncRNAESTs[1]);
			int nLncRNAConsESTs[] = numberOfUniqueESTsTranscript(ests2LncRNAsCons, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nLncRNAConsESTs[0] + "\t" + nLncRNAConsESTs[1]);
			int nEnhancerESTs[] = numberOfUniqueESTsEnhancer(ests2Enhancers, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nEnhancerESTs[0] + "\t" + nEnhancerESTs[1]);
			int nEnhancerConsESTs[] = numberOfUniqueESTsEnhancer(ests2EnhancersCons, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nEnhancerConsESTs[0] + "\t" + nEnhancerConsESTs[1]);
			int nPeakEST[] = numberOfUniqueESTsPeak(ests2Peak, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nPeakEST[0] + "\t" + nPeakEST[1]);
			int nPeakConsESTs[] = numberOfUniqueESTsPeak(ests2PeakCons, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nPeakConsESTs[0] + "\t" + nPeakConsESTs[1]);
			int nCtcfEST[] = numberOfUniqueESTsPeak(ests2Ctcf, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nCtcfEST[0] + "\t" + nCtcfEST[1]);
			int nCtcfConsESTs[] = numberOfUniqueESTsPeak(ests2CtcfCons, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nCtcfConsESTs[0] + "\t" + nCtcfConsESTs[1]);
			int nClusterESTs[] = numberOfUniqueESTsCluster(estClusters, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nClusterESTs[0] + "\t" + nClusterESTs[1]);
			int nClusterConsESTs[] = numberOfUniqueESTsCluster(estClustersCons, new TreeSet<Integer>(tissueTypes.keySet()), nCellTypes, nMC);
			bwCellTypes.write("\t" + nClusterConsESTs[0] + "\t" + nClusterConsESTs[1]);
			bwCellTypes.write("\n");
		    }
		}
	    }
	    bw.close();
	    bwCellTypes.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static Vector<ReducedModel2> parseGrimmondLncRNA(String grimmondLncRNAFileName, int chr, boolean forward, boolean convert2hg18) {
	Vector<ReducedModel2> lncRNAs = new Vector<ReducedModel2>();
	Vector<String> bedFileNames = new Vector<String>();
	try {
	    if (grimmondLncRNAFileName!=null) {
		if (grimmondLncRNAFileName.endsWith(".bed"))
		    bedFileNames.addElement(grimmondLncRNAFileName);
		else {
		    BufferedReader br = new BufferedReader(new FileReader(grimmondLncRNAFileName));
		    String line = br.readLine();
		    while (line!=null) {
			if (!line.startsWith("#"))
			    bedFileNames.addElement(line);
			line = br.readLine();
		    }
		    br.close();
		}
		for (String fName : bedFileNames) {
		    if (fName.contains("Grimmond")) {
			Runtime runtime = Runtime.getRuntime();
			String newFName = fName.replace(".bed", "_hg18.bed");
			String liftOverCmd = "liftOver " + fName + " /groups/neuroduo/martin/Genomes/Hs/hg19ToHg18.over.chain " + newFName + " /dev/null";
			System.out.println(liftOverCmd);
			runtime.exec(liftOverCmd);
			fName = newFName;
		    }
		    lncRNAs.addElement(new ReducedModel2(fName, chr, forward, "hs"));
		    System.out.println("Found " + lncRNAs.lastElement().getSegments().size() + " transcripts for " + fName.substring(fName.lastIndexOf("/") + 1));
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Error parsing lncRNA file " + e.toString());
	}
	return lncRNAs;
    }

    public static TranscribedRegion2 getOverlappingTranscript(TreeSet<TranscribedRegion2> segments, int c, boolean forward, int start, int end, int maxDist) {
	for (TranscribedRegion2 t : segments) {
	    if (forward && t.getChromosome()==c && t.getStart()>start - maxDist && t.getStart()<end + maxDist)
		return t;
	    else if (!forward && t.getChromosome()==c && t.getEnd()>end - maxDist && t.getEnd()<end + maxDist)
		return t;
	}
	return null;
    }

    public static int[] numberOfUniqueESTsGene(TreeMap<EST, RefSeqUCSCEnsemblGene> ests2Promoters, TreeSet<Integer> tissueTypes, int nCellTypes, int nMC) {
	//Use this method to calculate the number of ESTs and EST
	//clusters that we will observe when considering only
	//nCellTypes. Since the choice of cell types matters, we take
	//the average of nMC trials. The first element of the vector
	//contains the number of ESTs and the second the number of
	//clusters. 
	Tally estsFoundTally = new Tally(), estClustersFoundTally = new Tally();
	for (int i=0; i<nMC; i++) {
	    //Select a random subset of tissue types
	    HashSet<Integer> tissues = randomSubset(tissueTypes, nCellTypes);
	    //if (nCellTypes<10)
	    //	System.out.println("Found tissues " + tissues.toString());
	    //Keep only the ESTs that come from this subset
	    HashSet<EST> ests = new HashSet<EST>();
	    TreeSet<RefSeqUCSCEnsemblGene> genes = new TreeSet<RefSeqUCSCEnsemblGene>();
	    for (EST est : ests2Promoters.keySet()) {
		if (tissues.contains(est.getTissueType())) {
		    ests.add(est);
		    genes.add(ests2Promoters.get(est));
		}
	    }
	    //System.out.println("Found " + ests.size() + " ESTs for " + nCellTypes + " cell types.");
	    estsFoundTally.add(ests.size());
	    estClustersFoundTally.add(genes.size());
	}
	return new int[]{(int)Math.round(estsFoundTally.average()), (int)Math.round(estClustersFoundTally.average())};
    }

    public static int[] numberOfUniqueESTsTranscript(TreeMap<EST, TranscribedRegion2> ests2LncRNAs, TreeSet<Integer> tissueTypes, int nCellTypes, int nMC) {
	Tally estsFoundTally = new Tally(), estClustersFoundTally = new Tally();
	for (int i=0; i<nMC; i++) {
	    //Select a random subset of tissue types
	    HashSet<Integer> tissues = randomSubset(tissueTypes, nCellTypes);
	    HashSet<EST> ests = new HashSet<EST>();
	    TreeSet<TranscribedRegion2> lncRNAs = new TreeSet<TranscribedRegion2>();
	    for (EST est : ests2LncRNAs.keySet()) {
		if (tissues.contains(est.getTissueType())) {
		    ests.add(est);
		    lncRNAs.add(ests2LncRNAs.get(est));
		}
	    }
	    //System.out.println("Found " + ests.size() + " ESTs for " + nCellTypes + " cell types.");
	    estsFoundTally.add(ests.size());
	    estClustersFoundTally.add(lncRNAs.size());
	}
	return new int[]{(int)Math.round(estsFoundTally.average()), (int)Math.round(estClustersFoundTally.average())};
    }

    public static int[] numberOfUniqueESTsEnhancer(TreeMap<EST, Enhancer2> ests2Enhancers, TreeSet<Integer> tissueTypes, int nCellTypes, int nMC) {
	Tally estsFoundTally = new Tally(), estClustersFoundTally = new Tally();
	for (int i=0; i<nMC; i++) {
	    //Select a random subset of tissue types
	    HashSet<Integer> tissues = randomSubset(tissueTypes, nCellTypes);
	    //Keep only the ESTs that come from this subset
	    HashSet<EST> ests = new HashSet<EST>();
	    TreeSet<Enhancer2> enhancers = new TreeSet<Enhancer2>();
	    for (EST est : ests2Enhancers.keySet()) {
		if (tissues.contains(est.getTissueType())) {
		    ests.add(est);
		    enhancers.add(ests2Enhancers.get(est));
		}
	    }
	    estsFoundTally.add(ests.size());
	    estClustersFoundTally.add(enhancers.size());
	}
	return new int[]{(int)Math.round(estsFoundTally.average()), (int)Math.round(estClustersFoundTally.average())};
    }

    public static int[] numberOfUniqueESTsPeak(TreeMap<EST, Peak> ests2Peak, TreeSet<Integer> tissueTypes, int nCellTypes, int nMC) {
	Tally estsFoundTally = new Tally(), estClustersFoundTally = new Tally();
	for (int i=0; i<nMC; i++) {
	    //Select a random subset of tissue types
	    HashSet<Integer> tissues = randomSubset(tissueTypes, nCellTypes);
	    //Keep only the ESTs that come from this subset
	    HashSet<EST> ests = new HashSet<EST>();
	    TreeSet<Peak> peaks = new TreeSet<Peak>();
	    for (EST est : ests2Peak.keySet()) {
		if (tissues.contains(est.getTissueType())) {
		    ests.add(est);
		    peaks.add(ests2Peak.get(est));
		}
	    }
	    estsFoundTally.add(ests.size());
	    estClustersFoundTally.add(peaks.size());
	}
	return new int[]{(int)Math.round(estsFoundTally.average()), (int)Math.round(estClustersFoundTally.average())};
    }

    public static int[] numberOfUniqueESTsCluster(Vector<TreeSet<EST>> estsCluster, TreeSet<Integer> tissueTypes, int nCellTypes, int nMC) {
	Tally estsFoundTally = new Tally(), estClustersFoundTally = new Tally();
	for (int i=0; i<nMC; i++) {
	    //Select a random subset of tissue types
	    HashSet<Integer> tissues = randomSubset(tissueTypes, nCellTypes);
	    //Keep only the ESTs that come from this subset
	    HashSet<EST> ests = new HashSet<EST>();
	    int nClusters = 0;
	    for (TreeSet<EST> estCluster : estsCluster) {
		boolean estFoundInCluster = false;
		for (EST est : estCluster) {
		    if (tissues.contains(est.getTissueType())) {
			ests.add(est);
			if (!estFoundInCluster) {
			    nClusters++;
			    estFoundInCluster = true;
			}
		    }
		}
	    }
	    estsFoundTally.add(ests.size());
	    estClustersFoundTally.add(nClusters);
	}
	return new int[]{(int)Math.round(estsFoundTally.average()), (int)Math.round(estClustersFoundTally.average())};
    }

    public static HashSet<Integer> randomSubset(TreeSet<Integer> tissueTypes, int nCellTypes) {
	//Get a random subset of size nCellTypes from tissueTypes. If
	//nCellTypes is larger than the size of tissueTypes return
	//null.
	if (nCellTypes>tissueTypes.size())
	    return null;
	else {
	    Random random = new Random();
	    int array[] = new int[tissueTypes.size()];
	    for (int i=0; i<array.length; i++) 
		array[i] = i;
	    for (int i = array.length; i > 1; i--) {
		// Pick a random element to swap with the i-th element.
		int j = random.nextInt(i);  // 0 <= j <= i-1 (0-based array)
		// Swap array elements.
		int tmp = array[j];
		array[j] = array[i-1];
		array[i-1] = tmp;
	    }
	    HashSet<Integer> tissues = new HashSet<Integer>();
	    for (int i=0; i<nCellTypes; i++) {
		Iterator it = tissueTypes.iterator();
		for (int j=0; j<array[i]-1; j++)
		    it.next();
		tissues.add((Integer)it.next());
	    }
	    return tissues;
	}
    }

}


