import edu.uci.ics.jung.graph.impl.*;
import java.util.*;

public class Enhancer extends UndirectedSparseVertex implements Comparable {

    protected Peak peak;
    protected RefSeqGene nearestGene, overlappingGene;
    protected int nReadsForward, nReadsReverse, nearestDistance, timePoint;
    protected int chr, start, end;
    protected boolean nonGenic;
    protected double h3K4Me3KCl, h3K4Me1KCl, h3K4Me3Un, h3K4Me1Un;
    /////////////////////////////////
    protected ABIRead[] forwardReads = null, reverseReads = null, forwardReadsSmall = null, reverseReadsSmall = null;
    protected int reverseStart = Integer.MIN_VALUE, reverseEnd = Integer.MIN_VALUE, nReadsReverseTranscribedRegion = 0;
    protected int forwardStart = Integer.MIN_VALUE, forwardEnd = Integer.MIN_VALUE, nReadsForwardTranscribedRegion = 0;			    			    
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();


    //////////////////////////////////////////////////////////////////

    public Enhancer(Peak p, int width, int nearestDist, RefSeqGene gene, boolean non, int nForward, int nReverse, int t) {
	peak = p;
	chr = p.getChromosome();
	start = p.getCenter() - width;
	end = p.getCenter() + width;
	nearestDistance = nearestDist;
	nearestGene = gene;
	nonGenic = non;
	nReadsForward = nForward;
	nReadsReverse = nReverse;
	timePoint = t;
    }

    public Enhancer(Peak p, int s, int e, int nForward, int nReverse) {
	peak = p;
	chr = p.getChromosome();
	start = s;
	end = e;
	nReadsForward = nForward;
	nReadsReverse = nReverse;
    }

    public Enhancer(Peak p, int s, int e, int nForward, int nReverse, double k4Me1K, double k4Me1U) {
	peak = p;
	chr = p.getChromosome();
	start = s;
	end = e;
	nReadsForward = nForward;
	nReadsReverse = nReverse;
	h3K4Me1KCl = k4Me1K;
	h3K4Me1Un = k4Me1U;
    }

    public Enhancer(int c, int s, int e, int nForward, int nReverse) {
	//Use this constructor when parsing from file
	chr = c;
	start = s;
	end = e;
	nReadsForward = nForward;
	nReadsReverse = nReverse;
    }

    public Enhancer(int c, int s, int e, int nForward, int nReverse, boolean n) {
	//Use this constructor when parsing from file
	chr = c;
	start = s;
	end = e;
	nReadsForward = nForward;
	nReadsReverse = nReverse;
	nonGenic = n;
    }

    public Enhancer(int c, int s) {
	//Use this constructor for random control regions
	chr = c;
	start = s;
	end = s;
    }

    public Enhancer(int c, int s, int e) {
	chr = c;
	start = s;
	end = e;
    }

    ///////////////////////////////////////////////////////

    public boolean isNonGenic() { return nonGenic; }

    public void setNonGenic(boolean n) { nonGenic = n; }

    public void setOverlappingGene(RefSeqGene g) { overlappingGene = g; nonGenic = g==null; }

    public int getChromosome() { return chr; }

    public int getCenter() { return (start + end)/2; }

    public int getTimePoint() { return timePoint; }

    public int getNearestDistance() { return nearestDistance; }

    public RefSeqGene getNearestGene() { return nearestGene; }

    public int numberOfReadsForward() { return nReadsForward; }

    public int numberOfReadsReverse() { return nReadsReverse; }

    public boolean highQuality(int minReads) {
	return nReadsForward>minReads && nReadsReverse>minReads && (peak.hasFactor("8WG16") || peak.hasFactor("4H8"));
    }

    public Peak getPeak() { return peak; }

    public double getH3K4Me3(boolean kcl) { return kcl ? h3K4Me3KCl : h3K4Me3Un; }

    public double getH3K4Me3() { return Math.max(getH3K4Me3(true), getH3K4Me3(false)); }
    public double getH3K4Me1(boolean kcl) { return kcl ? h3K4Me1KCl : h3K4Me1Un; }

    ////////////////////////////////////////////

    /*public void setReads(ABIRead[] reads, boolean forward, int minDist) {
	if (forward) {
	    forwardReads = reads;
	    if (forwardReads.length>1) {
		int tmp2[] = IsolateBidirectionalTranscription.findTranscribedRegionBoundaries(forwardReads, true, minDist);
		forwardStart = tmp2[0];
		forwardEnd = tmp2[1];
		nReadsForwardTranscribedRegion = tmp2[2];
	    }
	}
	else {
	    reverseReads = reads;
	    if (reverseReads.length>1) {
		int tmp2[] = IsolateBidirectionalTranscription.findTranscribedRegionBoundaries(reverseReads, false, minDist);
		reverseStart = tmp2[0];
		reverseEnd = tmp2[1];
		nReadsReverseTranscribedRegion = tmp2[2];
	    }
	}
    }

    public void setSmallReads(ABIRead[] reads, boolean forward) {
	if (forward)
	    forwardReadsSmall = reads;
	else
	    reverseReadsSmall = reads;
	    }*/

    public int getTranscribedRegionStart(String fName, boolean strand, int minDist) { return findTranscribedRegionBoundaries(fName, strand, minDist)[0]; }

    public int getTranscribedRegionEnd(String fName, boolean strand, int minDist) { return findTranscribedRegionBoundaries(fName, strand, minDist)[1]; }

    public int getTranscribedRegionLength(String fName, boolean strand, int minDist) { 
	int tmp[] = findTranscribedRegionBoundaries(fName, strand, minDist);
	return Math.abs(tmp[1] - tmp[0]);
    }

    public int numberOfReadsInTranscribedRegion(String fName, boolean strand, int minDist) { return findTranscribedRegionBoundaries(fName, strand, minDist)[2]; }

    private int[] findTranscribedRegionBoundaries(String fName, boolean forward, int minDist) {
	//Extract the start and end of the longest transcribed region amongst the reads.
	if (!factorReads.containsKey(fName)) { return new int[]{0, 0, 0}; }
	ABIRead[] rs = factorReads.get(fName);
	//Extract all the reads in the direction that we are interested in
	int n = 0, j = 0;
	for (int i=0; i<rs.length; i++) { n += rs[i].getStrand()==forward ? 1 : 0; }
	ABIRead reads[] = new ABIRead[n];
	for (int i=0; i<rs.length; i++) { 
	    if (rs[i].getStrand()==forward)
		reads[j++] = rs[i];
	}
	if (reads==null || reads.length==0)
	    return new int[]{0, 0, 0};
	Arrays.sort(reads);
	int longestStart = Integer.MAX_VALUE, longestEnd = Integer.MAX_VALUE;
	if (forward && Math.abs(reads[0].getPosition() - getCenter())<minDist)
	    longestStart = reads[0].getPosition();
	else if (!forward && Math.abs(reads[reads.length-1].getPosition() - getCenter())<minDist)
	    longestStart = reads[reads.length - 1].getPosition();
	int nReadsLongest = 0, nReads = 0;
	boolean validTranscript = longestStart!=Integer.MAX_VALUE;
	if (forward) {
	    int i = 1;
	    while (i<reads.length) {
		if (validTranscript && Math.abs(reads[i].getPosition() - reads[i-1].getPosition())<=minDist)
		    nReads++;
		else if (Math.abs(reads[i].getPosition() - getCenter())<minDist) {
		    nReads = 1;
		    longestStart = reads[i].getPosition();
		    validTranscript = true;
		}
		else if (Math.abs(reads[i].getPosition() - reads[i-1].getPosition())>minDist)
		    validTranscript = false;
		if (nReads>nReadsLongest) {
		    nReadsLongest = nReads;
		    longestEnd = reads[i].getPosition() + 35;
		}
		i++;
	    }
	}
	else {
	    int i = reads.length - 1;
	    while (i>0) {
		if (validTranscript && Math.abs(reads[i].getPosition() - reads[i-1].getPosition())<minDist)
		    nReads++;
		else if (Math.abs(reads[i-1].getPosition() - getCenter())<minDist) {
		    nReads = 1;
		    longestStart = reads[i-1].getPosition();
		    validTranscript = true;
		}
		else if (Math.abs(reads[i].getPosition() - reads[i-1].getPosition())>minDist)
		    validTranscript = false;
		if (nReads>nReadsLongest) {
		    nReadsLongest = nReads;
		    longestEnd = reads[i-1].getPosition() - 35;
		}
		i--;
	    }
	}
	if (longestStart==Integer.MAX_VALUE || longestEnd==Integer.MAX_VALUE)
	    return new int[]{0, 0, 0};
	System.out.println("Found length " + Math.abs(longestEnd - longestStart) + " with " + nReadsLongest + " from " + reads.length + " starting from " + (longestStart - getCenter()) + " at enhancer " + getCenter() + " strand " + forward);
	return new int[]{longestStart, longestEnd, nReadsLongest};
    }

    /*public boolean hasBidirectionalTranscription() {
	int overlap = reverseStart - forwardStart;
	return forwardStart!=Integer.MIN_VALUE && reverseStart!=Integer.MIN_VALUE && nReadsForwardTranscribedRegion>1 && nReadsReverseTranscribedRegion>1 && Math.abs(overlap)<1000 && ((forwardStart-getCenter()>0 && reverseStart-getCenter()<0) || (forwardStart-getCenter()<0 && reverseStart-getCenter()>0));
	}*/

    public boolean hasForwardTranscription() {
	return forwardStart!=Integer.MIN_VALUE && nReadsForwardTranscribedRegion>1 && Math.abs(forwardStart-getCenter())<1000;
    }

    public boolean hasReverseTranscription() {
	return reverseStart!=Integer.MIN_VALUE && nReadsReverseTranscribedRegion>1 && Math.abs(reverseStart-getCenter())<1000;
    }

    public int numberOfReadsWithin(String fName, int dist) {
	int n = 0;
	for (int i=0; i<factorReads.get(fName).length; i++) 
	    n += Math.abs(factorReads.get(fName)[i].getPosition() - getCenter())<dist ? 1 : 0;
	return n;
    }

    public ABIRead[] getForwardReads() { return forwardReads; }

    public ABIRead[] getReverseReads() { return reverseReads; }

    public ABIRead[] getForwardSmallReads() { return forwardReadsSmall; }

    public ABIRead[] getReverseSmallReads() { return reverseReadsSmall; }

    public void setFactorReads(String f, ABIRead[] reads) { factorReads.put(f, reads); }

    public void addFactorReads(String f, ABIRead[] reads) { 
	ABIRead rs[] = getFactorReads(f);
	if (rs!=null) {
	    ABIRead[] newRs = new ABIRead[rs.length + reads.length];
	    for (int i=0; i<rs.length; i++) { newRs[i] = rs[i]; }
	    for (int i=0; i<reads.length; i++) { newRs[i + rs.length] = reads[i]; }
	    factorReads.put(f, newRs); 
	}
	else
	    factorReads.put(f, reads);
    }

    public ABIRead[] getFactorReads(String f) { return factorReads.containsKey(f) ? factorReads.get(f) : null; }

    //////////////////////////////////////////////////////////////////
    //JUNG stuff

    public RefSeqGene getNearestTSS() {
	RefSeqGene nearest = null;
	if (!nonGenic) {
	    //Always choose the one that we are overlapping
	    return overlappingGene;
	}
	int d = Integer.MAX_VALUE;
	for (RefSeqGene g : new Vector<RefSeqGene>(getNeighbors())) {
	    if (Math.abs(g.getStart()-getCenter())<d) {
		d = Math.abs(g.getStart()-getCenter());
		nearest = g;
	    }
	}
	return nearest;
    }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = (peak!=null ? peak.toString() : "chr" + chr + ":" + start + "-" + end) + "\t" + nReadsForward + ", " + nReadsReverse;
	return str;
    }

    public int compareTo(Object o) {
	if (!(o instanceof Enhancer))
	    System.err.println("ERROR comparing enhancers!");
	Enhancer e = (Enhancer)o;
	if (this.chr<e.chr)
	    return -1;
	else if (this.chr>e.chr)
	    return 1;
	else {
	    if (getCenter()<e.getCenter())
		return -1;
	    else if (getCenter()>e.getCenter())
		return 1;
	    else 
		return 0;
	}
    }

    public boolean equals(Object o) {
	if (!(o instanceof Enhancer))
	    return false;
	Enhancer e = (Enhancer)o;
	return chr==e.chr && e.start==start && e.end==end;
    }

    public int hashCode() { return 5*chr + 13*start + 29*end; }

}
