//Use this mehtod to try to account for all the reads in the
//experiment and see how many fall into each category. Together with
//the matlab sript ismb_poster_plots.m, this is the file mainly used
//for creating table 1 in the paper, the one accounting for the
//distribution of reads and transcripts.
//
//This is an off-shoot of the EstimateGenicTranscription3.java file. The purpose of this is to collect the relevant data for Fig 2, whereas EstimateGenicTranscription3 is ued to analyze the entire transcriptome for tab 1.
//
//bsub -q shared_2h java -Xmx4096m -server EstimateConservedGenicTranscription -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/merged_total/merged_total -o -c 19
//
//bsub -q shared_2h java -Xmx4096m -server EstimateConservedGenicTranscription -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac -o --species hs -c 22
//
//bsub -q shared_2h java -Xmx4096m -server EstimateGenicTranscription3 -n /groups/neuroduo/martin/ChIP-Seq/GrimmondData/Hs/mRNA/brain/brain -o --species hs -c 22

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class EstimateConservedGenicTranscription {

    public static final int MIN_READS_FOR_ARTEFACT = 10;
    public static final double MIN_READS_FROM_ONE_BASE_PROPORTION_FOR_ARTEFACT = .90;
    public static final double MAX_VALUE_HISTOGRAM = 10;

    public static final String SAVE_DIR = "/home/mh168/Projects/ChIP-Seq/TKData/Transcripts/ProfilesScatter/";
    public static final String EXCLUDE_FILE_START = "/home/mh168/Projects/ChIP-Seq/TKData/Transcripts/excluded_loci";
    public static final String REPEAT_RNA_FILE_MM = "/groups/neuroduo/martin/Genomes/Mm/v9/mmRNARepeats.txt";
    public static final String REPEAT_RNA_FILE_HS_36 = "/groups/neuroduo/martin/Genomes/Hs/v36/hsRNARepeats.txt";
    public static final String REPEAT_RNA_FILE_HS_37 = "/groups/neuroduo/martin/Genomes/Hs/v37/hsRNARepeats.txt";
    public static final String HATRIC_REGIONS_SUFFIX = "_hatric_regions_";
	
    public static final String SEQUENCE_DIR = "/Sequences/";

    public static DecimalFormat format = new DecimalFormat("###.#######");
    public static Random random = new Random();

    ////////////////////////////////////////////
    public static LinkedHashMap<String, FeatureInfo> featuresCons = new LinkedHashMap<String, FeatureInfo>();
    //Keep track of the ESTs found at different features
    public static TreeSet<EST> ests;
    public static LinkedHashMap<String, ESTCIData> estFeatures = new LinkedHashMap<String, ESTCIData>();
    //Keep track of the ones that have already been analyzed
    public static HashSet<ConservationIsland> consIslandsOverlapped = new HashSet<ConservationIsland>();
    public static ConservationIslands consIslands;

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option intronicFlag = argParser.addBooleanOption('I', "intronic");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option reverseFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option bedFileNameFlag = argParser.addStringOption('b', "bedFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option maxDistanceFlag = argParser.addIntegerOption('d', "maxDistance");	    
	    CmdLineParser.Option maxEnhancerDistFlag = argParser.addIntegerOption('e', "maxEnhancerDist");	    
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option minReadsFlag = argParser.addIntegerOption('m', "minReads");	    
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    CmdLineParser.Option upstreamFlag = argParser.addIntegerOption('U', "upstream");
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('D', "downstream");
	    CmdLineParser.Option sequenceFlag = argParser.addIntegerOption("sequence");
	    CmdLineParser.Option spliceFlag = argParser.addIntegerOption("spliced");	    
	    //Indicate the use of hg19 (v37) instead of the default hg18
	    CmdLineParser.Option hg19Flag = argParser.addBooleanOption("hg19");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean intronic = argParser.getOptionValue(intronicFlag)!=null;
	    boolean forward = argParser.getOptionValue(reverseFlag)==null;
	    boolean hg19 = argParser.getOptionValue(hg19Flag)!=null; 
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(bedFileNameFlag);
	    String bedFileName = tmp!=null ? (String)tmp : fileName;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : fileName + "_" + bedFileName.substring(bedFileName.lastIndexOf("/") + 1) + "_reads_categories" + "_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    tmp = argParser.getOptionValue(maxDistanceFlag);
	    int maxDistance = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(maxEnhancerDistFlag);
	    int maxEnhancerDist = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(minReadsFlag);
	    int minReads = tmp!=null ? ((Integer)tmp).intValue() : species.equals("mm") ? 9 : 5;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : .001;
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(spliceFlag);
	    int spliced = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(sequenceFlag);
	    int minAAs = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    ///////////////////////////////////////////////////////
	    String cellType = "Helas3";
	    String exonFileName = "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : "Hs/v36") + "/all_EXNs.tsv";
	    boolean grimmondData = fileName.contains("Grimmond");

	    String extragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lRNA_lMe3_real.tsv";
	    extragenicEnhancerFileName = grimmondData ? null : extragenicEnhancerFileName;
	    TreeSet<Enhancer2> extragenicEnhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
	    String intragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_intragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_in_lMe3_real.tsv";
	    intragenicEnhancerFileName = grimmondData ? null : intragenicEnhancerFileName;
	    TreeSet<Enhancer2> intragenicEnhancers = FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName);
	    TreeSet<Enhancer2> allEnhancers = new TreeSet<Enhancer2>();
	    allEnhancers.addAll(extragenicEnhancers);
	    allEnhancers.addAll(intragenicEnhancers);
	    String lncRNAFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    String lncRNAExonFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/exonsNPC.gff";
	    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species);
	    String macroRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species);
	    String miRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
	    TreeSet<MiRNA> miRNAs = AnalyzeTranscriptReadProfiles.loadMiRNA(miRNAFileName, species);
	    String peakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/hela_tf_names.txt" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/tf_kcl_names";
	    String tfNamesInStudy[] = species.equals("mm") ? new String[]{"CBP_Millipore_KCl_B1","CREB_SC_KCl_B1","Npas4_Npas4_KCl_B1","SRF_H300_KCl_B1"} : EncodePeaks.getEncodeTFNames(peakFileName, cellType);
	    peakFileName = grimmondData ? null : peakFileName;
	    Vector<Peak> peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(peakFileName, cellType, chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values());
	    System.out.println("Found " + peaks.size() + " peaks");
	    //Get CTCF peaks if available
	    String ctcfPeakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3Ctcf.narrowPeak" : null;
	    String ctcfNames[] = new String[]{"Ctcf"};
	    ctcfPeakFileName = grimmondData ? null : ctcfPeakFileName;
	    Vector<Peak> ctcfPeaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(ctcfPeakFileName, cellType, chr)) : new Vector<Peak>();
	    System.out.println("Found " + ctcfPeaks.size() + " CTCF peaks");
	    //Get the H3K4me3 peaks if available
	    String k4me3PeakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/H3K4Me3_KCl_peaks_names";
	    Vector<Peak> k4me3Peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(k4me3PeakFileName, cellType, chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(k4me3PeakFileName, chr).values());
	    System.out.println("Found " + k4me3Peaks.size() + " H3K4me3 peaks");
	    String estDirName = !species.equals("mm") ? "/groups/neuroduo/jesse/human_annot_for_mh/human_ESTs/" : "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/";
	    //Remove repeat RNAs
	    String excludeLociFileName = EXCLUDE_FILE_START + "_" + species + "RNARepeats.txt";
	    String repeatFileName = species.toLowerCase().equals("mm") ? REPEAT_RNA_FILE_MM : (hg19 ? EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_37 : REPEAT_RNA_FILE_HS_36);
	    TreeSet<TranscribedRegion2> repeatRNAs = parseRepeatFile(repeatFileName, species, chr, forward);


	    int nChr = FindTranscripts.numberOfChromosomes(species), readLength = 35, maxESTDist = 100, ciSlackDist = 100;
	    String rnaFileName = fileName + ".tsv";
	    //RefSeqUCSCEnsemblGenes geneInfo[] = new RefSeqUCSCEnsemblGenes[nChr];

	    TreeMap<Integer, Integer> remainingReadsHist = new TreeMap<Integer, Integer>();
	    //use the normalization factor to calculate the densities of the regions that we pick up without using the transcript caller
	    double normFactor = 1e6/FindReadsAcrossFeature2.countRNAReads(fileName, nChr);
	    //Use this structure to keep track of clusters with
	    //significant transcription at divergent sites, near
	    //peaks, enhancers and other extragenic sites.
	    TreeMap<String, TreeMap<Integer, Integer>> unannotatedReadsHist = new TreeMap<String, TreeMap<Integer, Integer>>();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(chr==-1 ? "/dev/null" : saveFileName));
	    System.out.println("Saving results to " + saveFileName);
	    String headers[] = new String[]{"Genic", "ncRNA", "ncRNA_repeats", "macroRNA", "lncRNA", "Divergent", "Near_genes", "Near_ncRNA", "eRNA_extra", "eRNA_intra", "Other_AS", "Unannotated_distal", "Near_other_peaks", "Near_ctcf_peaks", "Near_k4me3_peaks", "Remaining"};
	    if (chr==-1 || chr==1) {
		bw.write("#chr");
		for (int i=0; i<headers.length; i++)
		    bw.write("\t" + headers[i] + "\t" + headers[i] + "_regs\t" + headers[i] + "_bp");
		bw.write("\n");
	    }

	    //String excludeLociFileName = EXCLUDE_FILE_START + "_" + species + ".txt";
	    //Vector<int[]> excludeLoci = parseExcludeLociFile(excludeLociFileName, nChr);
	    //For saving the profiles to use for aggregate plots
	    int nBins = FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1);
	    LinkedHashMap<String, FeatureInfo> features = new LinkedHashMap<String, FeatureInfo>();
	    features.put("TSS", new FeatureInfo("TSS", nChr, nBins));
	    features.put("ncTSS", new FeatureInfo("ncTSS", nChr, nBins));
	    features.put("novel", new FeatureInfo("novel", nChr, nBins));
	    features.put("extra_enhancer", new FeatureInfo("extra_enhancer", nChr, nBins));
	    features.put("intra_enhancer", new FeatureInfo("intra_enhancer", nChr, nBins));
	    features.put("RFBS", new FeatureInfo("RFBS", nChr, nBins));
	    features.put("insulator", new FeatureInfo("insulator", nChr, nBins));
	    features.put("k4me3", new FeatureInfo("k4me3", nChr, nBins));
	    
	    featuresCons.put("TSS", new FeatureInfo("TSS", nChr, nBins));
	    featuresCons.put("ncTSS", new FeatureInfo("ncTSS", nChr, nBins));
	    featuresCons.put("novel", new FeatureInfo("novel", nChr, nBins));
	    featuresCons.put("extra_enhancer", new FeatureInfo("extra_enhancer", nChr, nBins));
	    featuresCons.put("intra_enhancer", new FeatureInfo("intra_enhancer", nChr, nBins));
	    featuresCons.put("RFBS", new FeatureInfo("RFBS", nChr, nBins));
	    featuresCons.put("insulator", new FeatureInfo("insulator", nChr, nBins));
	    featuresCons.put("k4me3", new FeatureInfo("k4me3", nChr, nBins));
	    
	    //Load the ESTs
	    ests = new TreeSet<EST>(CompareCageEstRNASeqStarts.parseEstTags(estDirName, chr, forward, 0, true, species).values());
	    //Save a list of the transcribed regions identified by
	    //FindTranscripts3. The initial categorization is too
	    //generous in favor of novel transcribed regions and hence
	    //the below list will be more restrictive. First we need
	    //to empty the old file (if it exists) since the entries
	    //are added one at a time.
	    String saveFileNameHaTriC = rnaFileName.substring(0, rnaFileName.lastIndexOf(".")) + HATRIC_REGIONS_SUFFIX + (forward ? "f" : "r") + "_" + chr + ".txt";
	    BufferedWriter bwHaTriC = new BufferedWriter(new FileWriter(saveFileNameHaTriC));
	    bwHaTriC.close();

	    //Thresholds for determining the required number of reads for transcribed regions
	    int minReadsDivergent = minReads, minReadsEnhancers = minReads;
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    //Load the annotation
		    RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(i, orchestra, species,hg19);

		    bw.write(i + "\t");
		    //Load the conservation islands
		    consIslands = new ConservationIslands(i, orchestra, species);
		    //Load reads
		    ABIRead[] rnaReadsForward = ABIReads.loadMRNAReadsChr(rnaFileName, true, i, readLength, false, repeatFileName);
		    ABIRead[] rnaReadsReverse = ABIReads.loadMRNAReadsChr(rnaFileName, false, i, readLength, false, repeatFileName);
		    System.out.println("Found " + rnaReadsForward.length + " reads forward and " + rnaReadsReverse.length + " reverse reads for chr " + i);
		    //Count the repeat reads
		    TreeSet<TranscribedRegion2> rnaRepeats = ABIReads.loadRepeatReadsChr(rnaFileName, forward, i, readLength, false, repeatFileName);
		    //Set reads to genes and transcripts
		    geneInfo.setFactorReads(rnaFileName, rnaReadsReverse, -1);
		    geneInfo.setFactorReads(rnaFileName, rnaReadsForward, 1);
		    geneInfo.setGeneIDsKeep(true, rnaFileName, minDens);
		    geneInfo.setGeneIDsKeep(false, rnaFileName, minDens);
		    //Set reads to the ncRNAs
		    AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, rnaFileName, forward ? rnaReadsForward : rnaReadsReverse, i, forward ? 1 : -1);
		    AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, rnaFileName, forward ? rnaReadsForward : rnaReadsReverse, i);
		    AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, forward ? rnaReadsForward : rnaReadsReverse, i, forward ? 1 : -1);
		    //Load the called transcripts
		    ReducedModel2 reducedModel = new ReducedModel2(bedFileName  + FindTranscripts.BED_FILE_SUFFIX + (forward ? "_f.bed" : "_r.bed"), forward ? rnaReadsForward : rnaReadsReverse, rnaFileName, i, forward, species);
		    //Assign categories
		    reducedModel.compare2Annotation(maxDistance, geneInfo, lncRNAs, miRNAs, macroRNAs, allEnhancers);
		    reducedModel.categorizeUnannotatedTranscribedRegions(maxEnhancerDist, species.equals("mm"));
		    //reducedModel.categorizeTranscribedRegions();
		    /////////////////////////////////
		    //Keep track of which bps are covered by transcripts
		    String readDensSaveFile = fileName + "_transcripts_dens_" + (forward ? "f" : "r") + "_" + i + ".txt";
		    ////////////////////////////////////////////
		    int startReads = forward ? rnaReadsForward.length : rnaReadsReverse.length;
		    //First go through all of the conserved TSSs and get read densities
		    ABIRead readsRemaining[] = analyzeAnnotatedConservedGenes(forward ? rnaReadsForward : rnaReadsReverse, forward, true, geneInfo, reducedModel, bw, normFactor, upstream, downstream, maxESTDist, ciSlackDist);
		    //Subtract the annotated reads from the total
		    //ABIRead readsRemaining[] = analyzeAnnotatedGenes(forward ? rnaReadsForward : rnaReadsReverse, forward, true, geneInfo, rnaFileName, minDens, minReads, readDensSaveFile.replace("dens", "coding_dens"), normFactor, reducedModel, bw, features.get("TSS"), upstream, downstream, repeatRNAs, minAAs, consIslands);
		    //Do the non-coding annotated genes
		    estFeatures.put("ncTSS", analyzeAnnotatedConservedGenes(readsRemaining, forward, false, geneInfo, normFactor, featuresCons.get("ncTSS"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzeAnnotatedGenes(readsRemaining, forward, false, geneInfo, rnaFileName, minDens, minReads, readDensSaveFile.replace("dens", "ncRNA_dens"), normFactor, reducedModel, bw, features.get("ncTSS"), upstream, downstream, repeatRNAs, minAAs, consIslands);
		    //Save the repeat RNA transcription
		    saveTranscriptInfo(readDensSaveFile.replace("dens", "RepncRNA_dens"), rnaRepeats, rnaFileName, normFactor, bw);
		    saveHaTriCRegions(i, forward, rnaFileName, rnaRepeats, false);
		    //macroRNAs and lincRNAs
		    estFeatures.get("ncTSS").putAll(analyzeConservedMacroRNAs(readsRemaining, i, forward, macroRNAs, normFactor, featuresCons.get("ncTSS"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzeMacroRNAs(readsRemaining, i, forward, macroRNAs, rnaFileName, minDens, minReads, readDensSaveFile.replace("dens", "ncRNA_dens"), normFactor, reducedModel, bw, features.get("ncTSS"), upstream, downstream, repeatRNAs, minAAs, consIslands);
		    estFeatures.get("ncTSS").putAll(analyzeConservedLincRNAs(readsRemaining, i, forward, lncRNAs, normFactor, featuresCons.get("ncTSS"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzeLincRNAs(readsRemaining, i, forward, lncRNAs, rnaFileName, minDens, minReads, readDensSaveFile.replace("dens", "ncRNA_dens"), normFactor, reducedModel, bw, features.get("ncTSS"), upstream, downstream, repeatRNAs, minAAs, consIslands);
		    //Divergent
		    readsRemaining = analyzeDivergent(readsRemaining, windowSize, forward, geneInfo, rnaFileName, minReadsDivergent, readDensSaveFile.replace("dens", "divergent_dens"), normFactor, reducedModel, bw, repeatRNAs, minAAs);
		    //Remove stuff near genes, these are not assigned to transcripts, but merely "cleaned up" to make the analysis of the remaining categories cleaner.
		    readsRemaining = removeNearby(readsRemaining, maxDistance, windowSize, forward, geneInfo, true, rnaFileName, minDens, reducedModel, bw);
		    readsRemaining = removeNearby(readsRemaining, maxDistance, windowSize, forward, geneInfo, false, rnaFileName, minDens, reducedModel, bw);
		    //eRNAs
		    estFeatures.put("extra_enhancer", analyzeConservedExtragenicERNA(readsRemaining, i, forward, extragenicEnhancers, normFactor, featuresCons.get("extra_enhancer"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzeExtragenicERNA(readsRemaining, i, forward, windowSize, extragenicEnhancers, rnaFileName, minReadsEnhancers, readDensSaveFile.replace("dens", "ExtraeRNA_dens"), normFactor, reducedModel, bw, features.get("extra_enhancer"), upstream, downstream, repeatRNAs, minAAs, consIslands);
		    readsRemaining = analyzeIntragenicERNA(readsRemaining, i, forward, windowSize, geneInfo, intragenicEnhancers, rnaFileName, minReadsEnhancers, readDensSaveFile.replace("dens", "IntraeRNA_dens"), normFactor, reducedModel, bw, features.get("intra_enhancer"), upstream, downstream, repeatRNAs, minAAs, consIslands);
		    //Other AS
		    readsRemaining = analyzeAS(readsRemaining, forward, windowSize, geneInfo, rnaFileName, minReadsEnhancers, readDensSaveFile.replace("dens", "otherAS_dens"), normFactor, reducedModel, peaks, ctcfPeaks, bw, repeatRNAs, minAAs);
		    //Novel transcripts
		    estFeatures.put("novel", analyzeConservedNovel(readsRemaining, i, forward, minDens, minReads, normFactor, reducedModel, featuresCons.get("novel"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzeNovel(readsRemaining, i, forward, rnaFileName, minDens, minReads, readDensSaveFile.replace("dens", "novel_dens"), normFactor, reducedModel, bw, features.get("novel"), upstream, downstream, repeatRNAs, minAAs, consIslands, consIslandsOverlapped);
		    //Remove stuff near novel transcripts as well. Seems as if there is divergent transcription from some of these guys that could end up being classified as RFBS transcription.
		    readsRemaining = removeNearbyNovel(readsRemaining, maxDistance, windowSize, forward, rnaFileName, minDens, reducedModel);
		    readsRemaining = removeNearbyNovel(readsRemaining, maxDistance, windowSize, forward, rnaFileName, minDens, reducedModel);
		    //RFBSs
		    estFeatures.put("RFBS", analyzeConservedPeaks(readsRemaining, i, forward, peaks, k4me3Peaks, geneInfo, normFactor, featuresCons.get("RFBS"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzePeaks(readsRemaining, i, forward, windowSize, peaks, k4me3Peaks, minReadsEnhancers, readDensSaveFile.replace("dens", "peaks_dens"), normFactor, reducedModel, geneInfo, bw, features.get("RFBS"), upstream, downstream, repeatRNAs, consIslands);
		    //Insulators
		    readsRemaining = analyzePeaks(readsRemaining, i, forward, windowSize, ctcfPeaks, k4me3Peaks, minReadsEnhancers, readDensSaveFile.replace("dens", "ctcf_peaks_dens"), normFactor, reducedModel, geneInfo, bw, features.get("insulator"), upstream, downstream, repeatRNAs, consIslands);
		    //Other H3K4me3
		    estFeatures.put("k4me3", analyzeConservedPeaks(readsRemaining, i, forward, k4me3Peaks, null, geneInfo, normFactor, featuresCons.get("k4me3"), upstream, downstream, consIslands, consIslandsOverlapped, ests, maxESTDist, ciSlackDist));
		    readsRemaining = analyzePeaks(readsRemaining, i, forward, windowSize, k4me3Peaks, null, minReadsEnhancers, readDensSaveFile.replace("dens", "k4me3_peaks_dens"), normFactor, reducedModel, geneInfo, bw, features.get("k4me3"), upstream, downstream, repeatRNAs, consIslands);
		    //readsRemaining = analyzeH3K4me3Peaks(readsRemaining, i, forward, windowSize, k4me3Peaks, minReadsEnhancers, readDensSaveFile.replace("dens", "k4me3_peaks_dens"), normFactor, bw, upstream, downstream, consIslands);
		    //Remaining
		    System.out.println("There are " + readsRemaining.length + " reads remaining from " + startReads);
		    estFeatures.put("remaining", analyzeConservedRemaining(readsRemaining, i, forward, geneInfo, normFactor, consIslands, consIslandsOverlapped, ests, maxESTDist));
		    //Find the histogram for the remaining reads based on windowSize non-overlapping bins
		    remainingReadsHist.putAll(analyzeRemainingReads(readsRemaining, windowSize, i, forward, minReadsEnhancers, readDensSaveFile.replace("dens", "remaining_dens"), species, normFactor, bw, repeatRNAs));
		    bw.write("\n");
		    ////////////////////////////////////////////
		    //Save the EST information
		    String estSaveFileName = "./EST/" + species + "_cons_EST_distribution_spliced_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
		    saveESTData(estSaveFileName, chr, estFeatures, true, geneInfo, forward);
		    saveESTDistribution(estSaveFileName, chr, estFeatures, true);
		    saveESTData(estSaveFileName.replace("spliced", "unspliced"), chr, estFeatures, false, geneInfo, forward);
		    saveESTDistribution(estSaveFileName.replace("spliced", "unspliced"), chr, estFeatures, false);
		}
	    }
	    bw.close();
	    ////////////////////////////////////////////
	    //Save the aggregate profiles and the scatter plots
	    String scatterSaveFileStart = SAVE_DIR + "scatter_cons_" + fileName.substring(fileName.lastIndexOf("/") + 1) + "_" + (fileName.toLowerCase().contains("rna") ? (forward ? "f_" : "r_") : "") + "w_" + windowSize + (upstream==0 ? "d" : (downstream==0 ? "u" : ""));
	    String profileSaveFile = SAVE_DIR + "profile_cons_" + fileName.substring(fileName.lastIndexOf("/") + 1) + (fileName.toLowerCase().contains("rna") ? (forward ? "_f" : "_r") : "") + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt";
	    saveScatterAndProfile(features, chr, nChr, upstream, scatterSaveFileStart, profileSaveFile);
	    saveScatterAndProfile(featuresCons, chr, nChr, upstream, scatterSaveFileStart.replace("cons", "cons_islands"), profileSaveFile.replace("cons", "cons_islands"));

	    String saveName = chr==-1 ? "/dev/null" : "RemainingReadsHistograms/" + fileName.substring(fileName.lastIndexOf("/") + 1) + "_remaining_reads_histogram_" + (forward ? "f" : "r") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    System.out.println("Saving histogram to " + saveName);
	    bw = new BufferedWriter(new FileWriter(saveName));
	    for (Integer k : remainingReadsHist.keySet())
		bw.write(k + "\t" + remainingReadsHist.get(k) + "\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static Vector<int[]> parseExcludeLociFile(String excludeLociFileName, int nChr) {
        Vector<int[]> excludedLoci = new Vector<int[]>();
        excludedLoci.setSize(nChr);
        TreeMap<Integer, TreeSet<Integer>> chrLoci = new TreeMap<Integer, TreeSet<Integer>>();
        try {
            BufferedReader br = new BufferedReader(new FileReader(excludeLociFileName));
            String line = br.readLine();
            while (line!=null) {
                String tokens[] = line.split("\\t");
                int c = Integer.parseInt(tokens[0]);
                if (!chrLoci.containsKey(c)) 
                    chrLoci.put(c, new TreeSet<Integer>());
                chrLoci.get(c).add(Integer.parseInt(tokens[1]));
                line = br.readLine();
            }
            br.close();
        }
        catch (IOException e) {
            System.err.println("Error opening file " + excludeLociFileName);
        }
        for (int c=1; c<=nChr; c++) {
            if (chrLoci.containsKey(c)) {
                int loci[] = new int[chrLoci.get(c).size()], i = 0;
                for (Integer l : chrLoci.get(c))
                    loci[i++] = l;
                excludedLoci.setElementAt(loci, c-1);
            }
            else
                excludedLoci.setElementAt(new int[]{}, c-1);
        }
        return excludedLoci;
    }

    public static TreeSet<TranscribedRegion2> parseRepeatFile(String repeatFileName, String species, int chr, boolean forward) { return parseRepeatFile(repeatFileName, species, chr, forward, true); }

    public static TreeSet<TranscribedRegion2> parseRepeatFile(String repeatFileName, String species, int chr, boolean forward, boolean includeAddedByMH) {
	//Header is:
	//#bin    swScore milliDiv        milliDel        milliIns        genoName        genoStart       genoEnd genoLeft        strand  repName repClass        repFamily       repStart        repEnd  repLeft id

        TreeSet<TranscribedRegion2> repeatRNAs = new TreeSet<TranscribedRegion2>();
        try {
            BufferedReader br = new BufferedReader(new FileReader(repeatFileName));
            String line = br.readLine();
            while (line!=null) {
		//if (line.contains("rRNA") || line.contains("tRNA")) {
		    String tokens[] = line.split("\\t");
		    String chrStr = tokens[5].substring(3);
		    int c = -1, start = Integer.parseInt(tokens[6]), end = Integer.parseInt(tokens[7]);
		    if (chrStr.equals("X"))
			c = species.toLowerCase().equals("hs") ? 23 : 20;
		    else if (chrStr.equals("Y"))
			c = species.toLowerCase().equals("hs") ? 24 : 21;
		    else if (!chrStr.equals("M") && chrStr.length()<=2)
			c = Integer.parseInt(chrStr);
		    boolean strand = tokens[9].equals("+");
		    boolean addedByMH = line.contains("MH");
		    if ((c==chr || chr==-1) && (includeAddedByMH || !addedByMH)/* && forward==strand*/) {
			TranscribedRegion2 t = new TranscribedRegion2(start, end, strand, false, c, 1);
			t.setType(TranscribedRegion2.NON_CODING_REPEAT_STR);
			repeatRNAs.add(t);
		    }
		line = br.readLine();
            }
            br.close();
        }
        catch (IOException e) {
            System.err.println("Error opening file " + repeatFileName);
        }
	System.out.println("Found " + repeatRNAs.size() + " repeat RNAs.");
        return repeatRNAs;
    }
    
    public static void saveESTDistribution(String estSaveFileName, int chr, LinkedHashMap<String, ESTCIData> estFeatures, boolean spliced) {
	/*for (String h : estFeatures.keySet()) {
	    System.out.println("Found " + estFeatures.get(h).size() + " for " + h);
	    int n = 0;
	    HashMap features = new HashMap();
	    Iterator it = estFeatures.get(h).keySet().iterator();
	    //Find how many ESTs correspond to each feature
	    while(it.hasNext()) {
		EST est = (EST)it.next();
		if (est.isSpliced()==spliced) {
		    n++;
		    if (!features.containsKey(estFeatures.get(h).get(est)))
			features.put(estFeatures.get(h).get(est), new Integer(0));
		    features.put(estFeatures.get(h).get(est), 1 + (Integer)features.get(estFeatures.get(h).get(est)));
		}
	    }
	    System.out.println("Found " + n + " ESTs for " + h);
	    //Count so that we have number of features with a given number of ESTs
	    TreeMap<Integer, Integer> histogram = new TreeMap<Integer, Integer>();
	    it = features.keySet().iterator();
	    while(it.hasNext()) {
		Object k = it.next();
		if (!histogram.containsKey((Integer)features.get(k)))
		    histogram.put((Integer)features.get(k), new Integer(0));
		histogram.put((Integer)features.get(k), 1 + (Integer)histogram.get((Integer)features.get(k)));
		if (!h.contains("TSS"))
		    System.out.println("Found " + ((Integer)features.get(k)) + " features for " + k.toString());
	    }
	    for (Integer k : histogram.keySet())
		System.out.println(k + "\t" + histogram.get(k));
		}*/
    }
    
    public static void saveESTData(String estSaveFileName, int chr, LinkedHashMap<String, ESTCIData> estFeatures, boolean spliced, RefSeqUCSCEnsemblGenes geneInfo, boolean forward) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(estSaveFileName));
	    if (chr==1 || chr==-1) {
		bw.write("#chr");
		for (String h : estFeatures.keySet())
		    bw.write("\t" + h + "_noESTs\t" + h + "_total\t" + h + "_cross");
		bw.write("\n");
	    }
	    bw.write((new Integer(chr)).toString());
	    for (String h : estFeatures.keySet()) {
		System.out.println("Found " + estFeatures.get(h).numberOfFeaturesNoESTs(spliced) + " with zero ESTs " + estFeatures.get(h).numberOfFeatures() + " and " + estFeatures.get(h).numberOfFeaturesESTCrossTSS(geneInfo, forward) + " for " + h + " with " + format.format(estFeatures.get(h).ciHasESTProb(spliced)));
		/*int n = 0;
		HashMap features = new HashMap();
		Iterator it = estFeatures.get(h).keySet().iterator();
		//for (EST est : estFeatures.get(h))
		//n += est.isSpliced()==spliced ? 1 : 0;
		//bw.write("\t" + n);
		while(it.hasNext()) {
		    EST est = (EST)it.next();
		    if (est.isSpliced()==spliced) {
			if (!features.containsKey(estFeatures.get(h).get(est)))
			    features.put(estFeatures.get(h).get(est), new Integer(0));
			features.put(estFeatures.get(h).get(est), 1 + (Integer)features.get(estFeatures.get(h).get(est)));
		    }
		    }*/
		bw.write("\t" + estFeatures.get(h).numberOfFeaturesNoESTs(spliced) + "\t" + estFeatures.get(h).numberOfFeatures() + "\t" + estFeatures.get(h).numberOfFeaturesESTCrossTSS(geneInfo, forward));
	    }
	    bw.write("\n");
	    bw.close();
	}
	catch (IOException e) {
	    System.out.println("Error saving EST data " + e.toString());
	}
    }
    


    public static void saveScatterAndProfile(LinkedHashMap<String, FeatureInfo> features, int chr, int nChr, int upstream, String scatterFileStart, String saveProfileName) {
	//
	try {
	    //Save the scatter-stuff
	    String scatterFileName = scatterFileStart + (chr!=-1 ? "_" + chr : "") + ".txt";
	    System.out.println("Saving scatter info to " + scatterFileName);
	    BufferedWriter scatterBw = new BufferedWriter(new FileWriter(scatterFileName));
	    if (chr==-1 || 1==chr) 
		scatterBw.write("#chr\tpos\tstrand\treads\tfeature\n");
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) { 
		    for (String feature : features.keySet()) {
			for (int j=0; j<features.get(feature).getLocVec(i).size(); j++) 
			    scatterBw.write(i + "\t" + features.get(feature).getLocVec(i).elementAt(j) + "\t" + (features.get(feature).getStrandVec(i).elementAt(j) ? "1" : "0") + "\t" + format.format(features.get(feature).getReadsInWindow(i, true).elementAt(j) + features.get(feature).getReadsInWindow(i, false).elementAt(j)) + "\t" + featureNumber(feature) + "\n");
		    }
		}
	    }
	    scatterBw.close();
	    //Add them all up and save to file
	    for (String feature : features.keySet()) 
		features.get(feature).calculateMeanHist(); //MAX_VALUE_HISTOGRAM, !feature.equals("TSS") ? MAX_VALUE_HISTOGRAM : Double.MAX_VALUE);
	    if (saveProfileName!=null) {
		System.out.println("Saving profile to " + saveProfileName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveProfileName));
		bw.write("#pos");
		for (String feature : features.keySet())
		    bw.write("\t" + feature);
		bw.write("\n");
		int n = features.values().iterator().next().getSumHistMean().length;
		for (int i=0; i<n; i++) {
		    bw.write("" + (-upstream + i));
		    for (String feature : features.keySet())
			bw.write("\t" + format.format(features.get(feature).getSumHistMean()[i]));
		    bw.write("\n");
		}
		bw.close();
	    }
	}
	catch (IOException e) {
	    System.out.println("Error saving scatter and profile data " + e.toString());
	}
    }

    public static int featureNumber(String feature) {
        if (feature.equals("TSS"))
            return 0;
	else if (feature.equals("ncTSS"))
            return 1;
	else if (feature.equals("novel"))
            return 2;
        else if (feature.equals("extra_enhancer"))
            return 3;
	else if (feature.equals("intra_enhancer"))
            return 4;
        else if (feature.equals("RFBS"))
            return 5;
        else if (feature.equals("insulator"))
            return 6;
	else if (feature.equals("k4me3"))
            return 7;
        return -1;
    }

	
    public static ABIRead[] getReadSubset(ABIRead reads[], int chr, boolean forward, int start, int end) {
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, start));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, end));
	endInd = endInd<0 ? -endInd-1 : endInd;
	ABIRead rs[] = new ABIRead[endInd - startInd];
	for (int i=startInd; i<endInd; i++) 
	    rs[i-startInd] = reads[i]; 
	Arrays.sort(rs);
	return rs;
    }

    public static boolean isArtefact(ABIRead rs[]) {
	//Use this method to look for spikes with lots of reads in the same location and no others nearby.
	int nReads = rs.length, nReadsFromOneBase = 0, maxReadsFromOneBase = 0;
	if (nReads>=MIN_READS_FOR_ARTEFACT) {
	    if (readsOverlapping(rs))
		return true;
	    for (int i=1; i<rs.length; i++) {
		if (rs[i-1].equals(rs[i])) {
		    nReadsFromOneBase++;
		    maxReadsFromOneBase = nReadsFromOneBase>maxReadsFromOneBase ? nReadsFromOneBase : maxReadsFromOneBase;
		}
		else 
		    nReadsFromOneBase = 0;
	    }
	    return (double)maxReadsFromOneBase/nReads>=MIN_READS_FROM_ONE_BASE_PROPORTION_FOR_ARTEFACT || maxReadsFromOneBase>MIN_READS_FROM_ONE_BASE_PROPORTION_FOR_ARTEFACT*10;
	}
	return false;
    }

    public static boolean readsOverlapping(ABIRead reads[]) {
	//Returns true if all of the reads are part of an overlapping cluster
	for (int i=1; i<reads.length; i++) {
	    if (!reads[i-1].overlaps(reads[i]))
		return false;
	}
	return true;
    }

    public static ABIRead[] subtractReads(ABIRead[] reads, ABIRead[] subtractReads) {
	Vector<ABIRead> readsKeep = new Vector<ABIRead>();
	for (int i=0; i<reads.length; i++) {
	    if (Arrays.binarySearch(subtractReads, reads[i])<0)
		readsKeep.addElement(reads[i]);
	}
	ABIRead[] rs = new ABIRead[readsKeep.size()];
	for (int i=0; i<rs.length; i++)
	    rs[i] = readsKeep.elementAt(i);
	return rs;
    }

	

    public static ABIRead[] analyzeAnnotatedGenes(ABIRead[] rnaReads, boolean forward, boolean coding, RefSeqUCSCEnsemblGenes geneInfo, String rnaFileName, double minDens, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs, ConservationIslands consIslands) {
	int startReads = rnaReads.length;
	int nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> codingRegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the genes and create a region for each gene that
	//is expressed above the min dens
	for (RefSeqUCSCEnsemblGene g : geneInfo.getNonOverlappingGenes(forward)) {
	    if (!g.isNonCoding()==coding) {
		TranscribedRegionMerge t = new TranscribedRegionMerge(g);
		t.assignReads(rnaReads);
		if (t.getReadDensity()>minDens && t.getReads().length>minReads) {
		    //Check for overlap with repeat RNAs
		    //Find the histogram to be used for the aggregate profile
		    if (coding) { 
			//Go through the transcribed regions and match
			//the ones that can be uniquely matched to
			//genes and in the process adding the reads
			//that were missed by the annotation
			TreeSet<TranscribedRegion2> overlappingCoding = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.CODING_STR));
			TreeSet<TranscribedRegion2> overlappingFragments = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.CODING_FRAGMENT_STR));
			TreeSet<TranscribedRegion2> overlappingMulti = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.MULTI_GENE_STR));
			//Save these regions to a file that will be used to assess the accuracy of transcript-calling
			saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingCoding);
			saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingFragments);
			saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingMulti);
			
			if (t.merge(overlappingCoding) || t.merge(overlappingFragments) || t.merge(overlappingMulti)) 
			    //Find the start profile based on actual start-pos
			    //featureInfo.addLocInfo(geneInfo.getChromosome(), g.getStart(), forward, FindReadsAcrossFeature2.getReadsHistogram(geneInfo.getChromosome(), g.getStart(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream);
			    nFoundByHaTriC++;
		    }
		    else {
			TreeSet<TranscribedRegion2> overlappingNC = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.NON_CODING_ANNOTATED_STR));
			saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingNC);
			if (t.merge(overlappingNC)) 
			    nFoundByHaTriC++;
			//System.out.println("Found transcript " + t.toString() + " with " + t.getReads().length + " reads for gene " + g.toString());
		    }
		    boolean conserved = consIslands !=null ? consIslands.overlaps(geneInfo.getChromosome(), forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_PROMOTER_DIST) : true;
		    t.setConservedStart(conserved);
		    featureInfo.addLocInfo(geneInfo.getChromosome(), forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(geneInfo.getChromosome(), forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
		    t.assignReads(rnaReads);
		    codingRegions.add(t);
		    rnaReads = subtractReads(rnaReads, t.getReads());
		    //System.out.println("Found transcript " + t.toString() + " with " + t.getReads().length + " reads for gene " + g.toString());
		}
	    }
	}
	//Clean up by removing any other reads overlapping genes
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes()) {
	    if (!g.isNonCoding()==coding && g.getStrand()==forward) {
		TranscribedRegionMerge t = new TranscribedRegionMerge(g);
		t.assignReads(rnaReads);
		rnaReads = subtractReads(rnaReads, t.getReads());
	    }
	}
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, false, codingRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = coding ? "coding" : "ncRNA";
	    saveSequence(geneInfo.getChromosome(), forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + geneInfo.getChromosome() + ".fa", false, minAAs, codingRegions, type);
	}
	return rnaReads;
    }

    public static ABIRead[] analyzeAnnotatedConservedGenes(ABIRead[] rnaReads, boolean forward, boolean coding, RefSeqUCSCEnsemblGenes geneInfo, ReducedModel2 reducedModel, BufferedWriter bw, double normFactor, FeatureInfo featureInfo, int upstream, int downstream, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped, TreeSet<EST> ests, int maxESTDist, int ciSlackDist) {
	ESTCIData estciData = new ESTCIData();
	int startReads = rnaReads.length;
	HashMap<EST, RefSeqUCSCEnsemblGene> ests2Promoters = new HashMap<EST, RefSeqUCSCEnsemblGene>();
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes(forward)) {
	    //Check if overlapping conserved island
	    if (!g.isNonCoding()==coding && g.getStrand()==forward) {
		TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(geneInfo.getChromosome(), g.getStart() - ciSlackDist, g.getStart() + ciSlackDist);
		if (cis!=null && !cis.isEmpty()) { 
		    boolean foundCI = false;
		    for (ConservationIsland ci : cis) {
			if (!consIslandsOverlapped.contains(ci)) {
			    consIslandsOverlapped.add(ci);
			    foundCI = true;
			}
		    }
		    if (foundCI) {
			estciData.addCI(cis);
			featureInfo.addLocInfo(geneInfo.getChromosome(), g.getStart(), forward, FindReadsAcrossFeature2.getReadsHistogram(geneInfo.getChromosome(), g.getStart(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			if (ests!=null) { //Check how many ESTs there are near this promoter
			    for (EST est : ests) {
				if (est.getStrand()==forward && Math.abs(est.getStart() - g.getStart())<maxESTDist) {
				    ests2Promoters.put(est, g);
				    estciData.addCI(cis, est);
				}
			    }
			for (EST est : ests2Promoters.keySet())
			    ests.remove(est);
			//System.out.println("Found " + ests2Promoters.size() + " ESTs for " + g.toString());
			}
		    }
		}
	    }
	}
	return estciData;
    }

    public static void saveHaTriCRegions(int chr, boolean strand, String rnaFileName, TreeSet<TranscribedRegion2> regions) { saveHaTriCRegions(chr, strand, rnaFileName, regions, true); }

    public static void saveHaTriCRegions(int chr, boolean strand, String rnaFileName, TranscribedRegion2 t) { 
	TreeSet<TranscribedRegion2> regs = new TreeSet<TranscribedRegion2>();
	regs.add(t);
	saveHaTriCRegions(chr, strand, rnaFileName, regs, true); 
    }

    public static void saveHaTriCRegions(int chr, boolean strand, String rnaFileName, TreeSet<TranscribedRegion2> regions, boolean append) {
	//chr start end nReads category
	if (regions.size()>0) {
	    try {
		String saveFileName = rnaFileName.substring(0, rnaFileName.lastIndexOf(".")) + HATRIC_REGIONS_SUFFIX + (strand ? "f" : "r") + "_" + chr + ".txt";
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName, append));
		if (!append)
		    bw.write("#chr\tstart\tend\tnReads\tcategory\n");
		for (TranscribedRegion2 r : regions) 
		    bw.write(chr + "\t" + r.getStart() + "\t" + r.getEnd() + "\t" + r.getFactorReads(rnaFileName).length + "\t" + getCategoryNumber(r.getType()) + "\n");
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error writing HaTriC regions summary file " + e.toString());
	    }
	}
    }

    public static int getCategoryNumber(String category) {
	if (category.equals(TranscribedRegion2.CODING_STR))
	    return 0;
	else if (category.equals(TranscribedRegion2.CODING_FRAGMENT_STR))
	    return 1;
	else if (category.equals(TranscribedRegion2.MULTI_GENE_STR))
	    return 2;
	else if (category.equals(TranscribedRegion2.NON_CODING_ANNOTATED_STR) || category.equals(TranscribedRegion2.MACRO_RNA_STR) || category.equals(TranscribedRegion2.LNC_RNA_STR) || category.equals(TranscribedRegion2.MI_RNA_STR))
	    return 3;
	else if (category.equals(TranscribedRegion2.DIVERGENT_STR))
	    return 4;
	else if (category.equals(TranscribedRegion2.NON_CODING_AS_STR))
	    return 5;
	else if (category.equals(TranscribedRegion2.ERNA_STR))
	    return 6;
	else if (category.equals(TranscribedRegion2.NON_CODING_OTHER_RNA_STR) || category.equals(TranscribedRegion2.PROMOTER_DISTAL_NCRNA_STR))
	    return 7;
	else if (category.equals(TranscribedRegion2.NON_CODING_REPEAT_STR))
	    return 8;
	else {
	    System.out.println("Could not find type for " + category);
	    return -1;
	}
    }

    public static void saveTranscriptInfo(TreeSet<TranscribedRegionMerge> regions, int readsRemaining, BufferedWriter bw) { 
	int nbpsCovered = 0;
	try {
	    for (TranscribedRegionMerge t : regions) 
		nbpsCovered += t.getLength();
	    bw.write(readsRemaining + "\t" + nbpsCovered + "\t" + regions.size() + "\t");
	}
	catch (IOException e) {
	    System.err.println("Error writing file " + e.toString());
	}
    }

    public static void saveTranscriptInfo(String readDensSaveFile, TreeSet<TranscribedRegion2> regions, String rnaFileName, double normFactor, BufferedWriter bw) { 
	//Use this method for the repeat regions
	int nbpsCovered = 0, nReads = 0, nRegions = 0;
	try {
	    BufferedWriter bwDens = new BufferedWriter(new FileWriter(readDensSaveFile));
	    for (TranscribedRegion2 t : regions) {
		ABIRead[] reads = t.getFactorReads(rnaFileName);
		if (reads.length>0) {
		    nbpsCovered += t.getLength();
		    nReads += reads.length;
		    nRegions++;
		    bwDens.write(format.format(normFactor*t.getReadDensity(rnaFileName)) + "\t" + t.getChromosome() + "\t" + t.getStart() + "\t" + t.getEnd() + "\n");	    
		}
	    }
	    bwDens.close();
	    bw.write(nReads + "\t" + nbpsCovered + "\t" + nRegions + "\t");
	    System.out.println("For repeat RNAs found " + nReads + "\t" + nbpsCovered + "\t" + nRegions + "\t");
	}
	catch (IOException e) {
	    System.err.println("Error writing file " + e.toString());
	}
    }

    public static void saveTranscriptInfo(String readDensSaveFile, boolean append, TreeSet<TranscribedRegionMerge> regions, int nFoundByHaTriC, double normFactor, int readsRemaining, BufferedWriter bw) {
	int nbpsCovered = 0, nReads = 0;
	try {
	    BufferedWriter bwDens = new BufferedWriter(new FileWriter(readDensSaveFile, append));
	    BufferedWriter bwDensCons = new BufferedWriter(new FileWriter(readDensSaveFile.replace("transcripts_", "transcripts_cons_"), append));
	    System.out.println("Saving " + regions.size() + " densities to " + readDensSaveFile);
	    for (TranscribedRegionMerge t : regions) {
		nbpsCovered += t.getLength();
		nReads += t.getReads().length;
		bwDens.write(format.format(normFactor*t.getReadDensity()) + "\t" + t.getChromosome() + "\t" + t.getStart() + "\t" + t.getEnd() + "\n");	    
		if (t.isConservedStart()) 
		    bwDensCons.write(format.format(normFactor*t.getReadDensity()) + "\t" + t.getChromosome() + "\t" + t.getStart() + "\t" + t.getEnd() + "\n");	    
		//if (t.getGene()!=null)
		//  System.out.println("Saving " + t.toString() + " overlapping " + t.getGene().toString());
	    }
	    bwDens.close();
	    bwDensCons.close();
	    System.out.println("Counted " + nReads + " from transcripts and saving " + readsRemaining);
	    bw.write(readsRemaining + "\t" + nbpsCovered + "\t" + regions.size() + "\t");
	}
	catch (IOException e) {
	    System.err.println("Error writing file " + e.toString());
	}
    }

    public static void saveSequence(int chr, boolean forward, String saveFileName, boolean append, int minAAs, TreeSet<TranscribedRegionMerge> regions, String type) {
	try {
	    FastaSequence2 fastaSeq = new FastaSequence2(chr, true);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName, append)), aabws[] = new BufferedWriter[3];
	    if (minAAs>=0) {
		File dirName = new File(saveFileName.substring(0, saveFileName.lastIndexOf("/")+1) + FindTranscripts2.AA_DIR_NAME);
		//System.out.println("Saving AAs to " + dirName + saveFileName.substring(saveFileName.lastIndexOf("/") + 1).replace("transcripts", "aas_0"));
		dirName.mkdirs();
		for (int i=0; i<3; i++)
		    aabws[i] = new BufferedWriter(new FileWriter(dirName + "/" + saveFileName.substring(saveFileName.lastIndexOf("/") + 1).replace("transcripts", "aas_" + i), append));
	    }
	    for (TranscribedRegionMerge t : regions) {
		String header = ">chr" + chr + ":" + t.getStart() + "-" + t.getEnd() + "(" + (forward ? "+" : "-") + ")" + "_" + type;
		String seq = fastaSeq.getSequence(t.getStart(), t.getLength(), !forward);
		bw.write(header + "\n" + seq + "\n");
		if (minAAs>0) {
		    //Get the AA sequence for all three different reading frames
		    for (int i=0; i<3; i++) {
			//We want to split the sequences when we find
			//a stop-codon. Information about this should
			//be stored in the header
			String aaSeqs[] = (FastaSequence2.convert2AAs(seq, i)).split("\\*");
			for (int j=0; j<aaSeqs.length; j++) {
			    if (aaSeqs[j].length()>=minAAs)
				aabws[i].write(header + "_frag_" + j + "\n" + aaSeqs[j] + "\n");
			}
		    }
		}
	    }
	    bw.close();
	    if (minAAs>0) {
		for (int i=0; i<3; i++)
		    aabws[i].close();
	    }
	
	}
	catch (IOException e) {
	    System.err.println("Error writing file " + e.toString());
	}
    }

    public static ABIRead[] analyzeMacroRNAs(ABIRead[] rnaReads, int chr, boolean forward, TreeSet<MacroRNA> macroRNAs, String rnaFileName, double minDens, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs, ConservationIslands consIslands) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> macroRNARegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the genes and create a region for each gene that
	//is expressed above the min dens
	for (MacroRNA m : macroRNAs) {
	    if (m.getStrand()==forward && m.getChromosome()==chr) {
		TranscribedRegionMerge t = new TranscribedRegionMerge(m);
		t.assignReads(rnaReads);
		if (t.getReadDensity()>minDens && t.getReads().length>minReads/* && (t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
		    macroRNARegions.add(t);
		    rnaReads = subtractReads(rnaReads, t.getReads());
		    //System.out.println("Found transcript " + t.toString() + " with " + t.getReads().length + " reads for macroRNA " + m.toString());
		}
	    }
	}
	for (TranscribedRegionMerge t : macroRNARegions) {
	    TreeSet<TranscribedRegion2> overlappingMacro = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.MACRO_RNA_STR));
	    saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingMacro);
	    t.merge(overlappingMacro);
	    boolean conserved = consIslands!=null ? consIslands.overlaps(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_PROMOTER_DIST) : true;
	    featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
	    t.setConservedStart(conserved);
	    t.assignReads(rnaReads);
	    rnaReads = subtractReads(rnaReads, t.getReads());
	    nFoundByHaTriC++;
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, true, macroRNARegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "ncRNA";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", true, minAAs, macroRNARegions, type);
	}
	return rnaReads;
    }

    public static ESTCIData analyzeConservedMacroRNAs(ABIRead[] rnaReads, int chr, boolean forward, TreeSet<MacroRNA> macroRNAs, double normFactor, FeatureInfo featureInfo, int upstream, int downstream, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped, TreeSet<EST> ests, int maxESTDist, int ciSlackDist) {
	//Go through the genes and create a region for each gene that
	//is expressed above the min dens
	ESTCIData estciData = new ESTCIData();
	HashMap<EST, MacroRNA> ests2Macros = new HashMap<EST, MacroRNA>();
	for (MacroRNA m : macroRNAs) {
	    if (m.getStrand()==forward && m.getChromosome()==chr) {
		TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(chr, (forward ? m.getStart() : m.getEnd()) - ciSlackDist, (forward ? m.getStart() : m.getEnd()) + ciSlackDist);
		if (cis!=null && !cis.isEmpty()) {
		    boolean foundCI = false;
		    for (ConservationIsland ci : cis) {
			if (!consIslandsOverlapped.contains(ci)) {
			    consIslandsOverlapped.add(ci);
			    foundCI = true;
			}
		    }
		    if (foundCI) {
			estciData.addCI(cis);
			featureInfo.addLocInfo(chr, forward ? m.getStart() : m.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? m.getStart() : m.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			if (ests!=null) { //Check how many ESTs there are near this promoter
			    for (EST est : ests) {
				if (est.getStrand()==forward && Math.abs(est.getStart() - m.getStart())<maxESTDist) 
				    //ests2Macros.put(est, m);
				    estciData.addCI(cis, est);
			    }
			    for (EST est : ests2Macros.keySet())
				ests.remove(est);
			}
		    }
		}
	    }
	}
	return estciData;
    }

    public static ABIRead[] analyzeLincRNAs(ABIRead[] rnaReads, int chr, boolean forward, TreeSet<LncRNA> lncRNAs, String rnaFileName, double minDens, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs, ConservationIslands consIslands) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> lncRNARegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the genes and create a region for each gene that
	//is expressed above the min dens
	for (LncRNA m : lncRNAs) {
	    if (m.getChromosome()==chr) {
		TranscribedRegionMerge t = new TranscribedRegionMerge(m, forward);
		t.assignReads(rnaReads);
		if (t.getReadDensity()>minDens && t.getReads().length>minReads/* && (t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
		    lncRNARegions.add(t);
		    rnaReads = subtractReads(rnaReads, t.getReads());
		    //System.out.println("Found transcript " + t.toString() + " with " + t.getReads().length + " reads for lincRNA " + m.toString());
		}
	    }
	}
	for (TranscribedRegionMerge t : lncRNARegions) {
	    TreeSet<TranscribedRegion2> overlappingLinc = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.LNC_RNA_STR));
	    saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingLinc);
	    t.merge(overlappingLinc);
	    t.assignReads(rnaReads);
	    boolean conserved = consIslands!=null ? consIslands.overlaps(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_PROMOTER_DIST) : true;
	    featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
	    t.setConservedStart(conserved);
	    rnaReads = subtractReads(rnaReads, t.getReads());
	    nFoundByHaTriC++;
	    //System.out.println("Found transcript " + t.toString() + " with " + t.getReads().length + " reads for lincRNA");
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, true, lncRNARegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "ncRNA";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", true, minAAs, lncRNARegions, type);
	}
	return rnaReads;
    }

    public static ESTCIData analyzeConservedLincRNAs(ABIRead[] rnaReads, int chr, boolean forward, TreeSet<LncRNA> lncRNAs, double normFactor, FeatureInfo featureInfo, int upstream, int downstream, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped, TreeSet<EST> ests, int maxESTDist, int ciSlackDist) {
	ESTCIData estciData = new ESTCIData();
	HashMap<EST, LncRNA> ests2Lncs = new HashMap<EST, LncRNA>();
	for (LncRNA m : lncRNAs) {
	    if (m.getChromosome()==chr && consIslands.overlaps(chr, forward ? m.getStart() : m.getEnd())) {
		TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(chr, (forward ? m.getStart() : m.getEnd()) - ciSlackDist, (forward ? m.getStart() : m.getEnd()) + ciSlackDist);
		if (cis!=null && !cis.isEmpty()) {
		    boolean foundCI = false;
		    for (ConservationIsland ci : cis) { 
			if (!consIslandsOverlapped.contains(ci)) {
			    consIslandsOverlapped.add(ci);
			    foundCI = true;
			}
		    }
		    if (foundCI) {
			estciData.addCI(cis);
			featureInfo.addLocInfo(chr, forward ? m.getStart() : m.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? m.getStart() : m.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			if (ests!=null) { //Check how many ESTs there are near this promoter
			    for (EST est : ests) {
				if (est.getStrand()==forward && Math.abs(est.getStart() - m.getStart())<maxESTDist) 
				    //ests2Lncs.put(est, m);
				    estciData.addCI(cis, est);
			    }
			    for (EST est : ests2Lncs.keySet())
				ests.remove(est);
			}
		    }
		}
	    }
	}
	return estciData;
    }

    public static ABIRead[] analyzeDivergent(ABIRead[] rnaReads, int windowSize, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, String rnaFileName, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> codingRegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the genes and create a region for each gene that
	//is expressed above the min dens
	for (RefSeqUCSCEnsemblGene g : geneInfo.getNonOverlappingGenes(!forward)) {
	    if (!g.isNonCoding() && g.getStrand()!=forward) {
		//Get the reads in the upstream region
		TranscribedRegionMerge t = findDivergentTranscript(g, rnaReads, windowSize, windowSize, windowSize, minReads);
		if (t!=null/* && (t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
		    t.assignReads(rnaReads);
		    //if (t.getReads().length/t.getLength()>minReads) {
			codingRegions.add(t);
			rnaReads = subtractReads(rnaReads, t.getReads());
			//System.out.println("Found divergent transcript " + t.toString() + " with " + t.getReads().length + " reads for gene " + g.toString());
			//}
		}
	    }
	}
	System.out.println("Found " + reducedModel.getSegments(TranscribedRegion2.DIVERGENT_STR).size() + " divergent transcripts from transcript caller.");
	for (TranscribedRegionMerge t : codingRegions) {
	    TreeSet<TranscribedRegion2> overlappingDivergent = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.DIVERGENT_STR));
	    saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingDivergent);
	    if (t.merge(overlappingDivergent)) {
		t.assignReads(rnaReads);
		rnaReads = subtractReads(rnaReads, t.getReads());
		nFoundByHaTriC++;
		//System.out.println("Found divergent transcript " + t.toString() + " with " + t.getReads().length);
	    }
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, false, codingRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "divergent";
	    saveSequence(geneInfo.getChromosome(), forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + geneInfo.getChromosome() + ".fa", false, minAAs, codingRegions, type);
	}
	return rnaReads;
    }

    public static TranscribedRegionMerge findDivergentTranscript(RefSeqUCSCEnsemblGene g, ABIRead[] reads, int upstream, int downstream, int stepSize, int minReads) {
	boolean forward = !g.getStrand();
	int start = g.getStart() + (forward ? 0 : -downstream);
	int end = g.getStart() + (forward ? downstream : 0);
	ABIRead[] rs = getReadSubset(reads, g.getChromosome(), !g.getStrand(), start, end);
	if (rs!=null && rs.length>minReads) {
	    //contiue stepping outwards
	    int nSteps = 0;
	    while (rs!=null && rs.length>minReads) {
		nSteps++;
		int s = (forward ? end + (nSteps - 1)*stepSize : start - nSteps*stepSize);
		int e = (forward ? end + nSteps*stepSize : start - (nSteps - 1)*stepSize);
		rs = getReadSubset(reads, g.getChromosome(), !g.getStrand(), s, e);
	    }
	    //Also look upstream in case the transcript overlaps the annotated TS
	    ABIRead[] upstreamReads = getReadSubset(reads, g.getChromosome(), !g.getStrand(), (forward ? start - upstream : end), forward ? start : end + upstream);
	    boolean upstreamStart = upstreamReads!=null && upstreamReads.length>minReads;
	    return new TranscribedRegionMerge(g.getChromosome(), !g.getStrand(), start + (forward ? (upstreamStart ? -upstream : 0) : -(nSteps - 1)*stepSize), end + (forward ? (nSteps - 1)*stepSize : (upstreamStart ? 0 : upstream)));
	}
	return null;
    }

    public static ABIRead[] removeNearby(ABIRead[] rnaReads, int windowSize, int stepSize, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, boolean coding, String rnaFileName, double minDens, ReducedModel2 reducedModel, BufferedWriter bw) {
	int startReads = rnaReads.length;
	//Go through the genes and create a region for each gene that
	//is expressed above the min dens
	for (RefSeqUCSCEnsemblGene g : geneInfo.getNonOverlappingGenes(forward)) {
	    if (!g.isNonCoding()==coding && g.getStrand()==forward && g.getReadDensity(rnaFileName)>minDens) {
		//Get the reads in the upstream region
		rnaReads = subtractNearbyReads(g, rnaReads, true, windowSize, 0, stepSize);
		rnaReads = subtractNearbyReads(g, rnaReads, false, 0, windowSize, stepSize);
	    }
	}
	//Remove reads near the called transcripts as well
	String categories[] = coding ? new String[]{TranscribedRegion2.CODING_STR, TranscribedRegion2.CODING_FRAGMENT_STR, TranscribedRegion2.MULTI_GENE_STR} : new String[]{TranscribedRegion2.NON_CODING_ANNOTATED_STR};
	for (int i=0; i<categories.length; i++) {
	    for (TranscribedRegion2 t2 : reducedModel.getSegments(categories[i])) {
		if (t2.getReadDensity(rnaFileName)>minDens) {
		    rnaReads = subtractNearbyReads(t2, rnaReads, true, windowSize, 0, stepSize);
		    rnaReads = subtractNearbyReads(t2, rnaReads, false, 0, windowSize, stepSize);
		}
	    }
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(new TreeSet<TranscribedRegionMerge>(), startReads - rnaReads.length, bw);
	return rnaReads;
    }

    public static ABIRead[] removeNearbyNovel(ABIRead[] rnaReads, int windowSize, int stepSize, boolean forward, String rnaFileName, double minDens, ReducedModel2 reducedModel) {
	System.out.println("Found " + rnaReads.length + " reads before cleaning.");
	for (TranscribedRegion2 t2 : reducedModel.getSegments(TranscribedRegion2.PROMOTER_DISTAL_NCRNA_STR)) {
	    if (t2.getReadDensity(rnaFileName)>minDens) {
		rnaReads = subtractNearbyReads(t2, rnaReads, true, windowSize, 0, stepSize);
		rnaReads = subtractNearbyReads(t2, rnaReads, false, 0, windowSize, stepSize);
	    }
	}
	for (TranscribedRegion2 t2 : reducedModel.getSegments(TranscribedRegion2.NON_CODING_OTHER_RNA_STR)) {
	    if (t2.getReadDensity(rnaFileName)>minDens) {
		rnaReads = subtractNearbyReads(t2, rnaReads, true, windowSize, 0, stepSize);
		rnaReads = subtractNearbyReads(t2, rnaReads, false, 0, windowSize, stepSize);
	    }
	}
	System.out.println("Found " + rnaReads.length + " reads after cleaning.");
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	//saveTranscriptInfo(new TreeSet<TranscribedRegionMerge>(), startReads - rnaReads.length, bw);
	return rnaReads;
    }
    
    public static ABIRead[] subtractNearbyReads(TranscribedRegion2 t2, ABIRead[] reads, boolean start, int upstream, int downstream, int stepSize) {
	boolean forward = t2.getStrand();
	int s = (start ? t2.getStart() : t2.getEnd()) + (forward ? -upstream : -downstream);
	int e = (start ? t2.getStart() : t2.getEnd()) + (forward ? downstream : upstream);
	ABIRead[] rs = getReadSubset(reads, t2.getChromosome(), t2.getStrand(), s, e);
	int nSteps = 0;
	while (rs!=null && rs.length>0) {
	    reads = subtractReads(reads, rs);
	    //System.out.println("Removed " + rs.length + " reads between " + s + "-" + e + " near " + t2.toString());
	    nSteps++;
	    s += (forward ? (start ? -stepSize : 0) : (start ? 0 : -stepSize));
	    e += (forward ? (start ? 0 : stepSize) : (start ? stepSize : 0));	
	    rs = getReadSubset(reads, t2.getChromosome(), !t2.getStrand(), s, e);
	}
	return reads;
    }

    public static ABIRead[] subtractNearbyReads(RefSeqUCSCEnsemblGene g, ABIRead[] reads, boolean start, int upstream, int downstream, int stepSize) {
	boolean forward = g.getStrand();
	int s = (start ? g.getStart() : g.getEnd()) + (forward ? -upstream : -downstream);
	int e = (start ? g.getStart() : g.getEnd()) + (forward ? downstream : upstream);
	ABIRead[] rs = getReadSubset(reads, g.getChromosome(), g.getStrand(), s, e);
	int nSteps = 0;
	while (rs!=null && rs.length>0) {
	    reads = subtractReads(reads, rs);
	    //System.out.println("Removed " + rs.length + " reads between " + s + "-" + e + " near " + g.toString());
	    nSteps++;
	    s += (forward ? (start ? -stepSize : 0) : (start ? 0 : -stepSize));
	    e += (forward ? (start ? 0 : stepSize) : (start ? stepSize : 0));
	    rs = getReadSubset(reads, g.getChromosome(), !g.getStrand(), s, e);
	}
	return reads;
    }

    public static ABIRead[] analyzeExtragenicERNA(ABIRead[] rnaReads, int chr, boolean forward, int windowSize, TreeSet<Enhancer2> extragenicEnhancers, String rnaFileName, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs, ConservationIslands consIslands) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> enhancerRegions = new TreeSet<TranscribedRegionMerge>();
	for (Enhancer2 e : extragenicEnhancers) {
	    if (e.getChromosome()==chr) {
		TranscribedRegionMerge t = findERNATranscript(rnaReads, minReads, chr, e.getCenter() + (forward ? 0 : -windowSize), e.getCenter() + (forward ? windowSize : 0), forward, windowSize);
		if (t!=null) {
		    if (true/*(t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
			boolean conserved = consIslands!=null ? consIslands.overlapsWithin(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_ENHANCER_DIST) : true;
			featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
			t.setConservedStart(conserved);
			t.assignReads(rnaReads);
			System.out.println("Found eRNA transcript " + t.toString() + " with " + t.getReads().length);
			//if (t.getReadDensity()>minDens) {
			enhancerRegions.add(t);
			rnaReads = subtractReads(rnaReads, t.getReads());
		    }
		}
	    }
	}
	//Consider the transcripts that were called
	System.out.println("Found " + reducedModel.getSegments(TranscribedRegion2.ERNA_STR).size() + " eRNA transcripts from transcript caller.");
	for (TranscribedRegionMerge t : enhancerRegions) {
	    TreeSet<TranscribedRegion2> overlappingERNA = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.ERNA_STR));
	    saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingERNA);
	    if (t.merge(overlappingERNA)) {
		boolean conserved = consIslands!=null ? consIslands.overlapsWithin(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_ENHANCER_DIST) : true;
		featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
		t.setConservedStart(conserved);
		t.assignReads(rnaReads);
		rnaReads = subtractReads(rnaReads, t.getReads());
		nFoundByHaTriC++;
	    }
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, false, enhancerRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "eRNA";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", false, minAAs, enhancerRegions, type);
	}
	return rnaReads;
    }

    public static ESTCIData analyzeConservedExtragenicERNA(ABIRead[] rnaReads, int chr, boolean forward, TreeSet<Enhancer2> extragenicEnhancers, double normFactor, FeatureInfo featureInfo, int upstream, int downstream, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped, TreeSet<EST> ests, int maxESTDist, int ciSlackDist) {
	ESTCIData estciData = new ESTCIData();
	HashMap<EST, Enhancer2> ests2Enhancers = new HashMap<EST, Enhancer2>();
	for (Enhancer2 e : extragenicEnhancers) {
	    if (e.getChromosome()==chr) {
		TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(chr, e.getCenter() - ciSlackDist, e.getCenter() + ciSlackDist);
		if (cis!=null && !cis.isEmpty()) {
		    boolean foundCI = false;
		    for (ConservationIsland ci : cis) {
			if (!consIslandsOverlapped.contains(ci)) {
			    consIslandsOverlapped.add(ci);
			    foundCI = true;
			}
		    }
		    if (foundCI) {
			estciData.addCI(cis);
			featureInfo.addLocInfo(chr, e.getCenter(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, e.getCenter(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			if (ests!=null) { //Check how many ESTs there are near this promoter
			    for (EST est : ests) {
				if (est.getStrand()==forward && Math.abs(est.getStart() - e.getCenter())<maxESTDist) {
				    ests2Enhancers.put(est, e);
				    estciData.addCI(cis, est);
				}
			    }
			    for (EST est : ests2Enhancers.keySet())
				ests.remove(est);
			}
		    }
		}
	    }
	}
	return estciData;
    }
    
    public static TranscribedRegionMerge findERNATranscript(ABIRead[] rnaReads, int minReads, int chr, int start, int end, boolean forward, int stepSize) {
	ABIRead[] rs = getReadSubset(rnaReads, chr, forward, start, end);
	if (rs!=null && rs.length>minReads) {
	    //Get a more fine-grained estimate of the length by measuring the distance between the first and last reads
	    int s = rs[0].getPosition(), e = rs[rs.length-1].getPosition();
	    int nSteps = 0;
	    while (rs!=null && rs.length>minReads) {
		nSteps++;
		rs = getReadSubset(rnaReads, chr, forward, (forward ? end + (nSteps - 1)*stepSize : start -nSteps*stepSize), (forward ? end + nSteps*stepSize : start -(nSteps - 1)*stepSize));
		if (rs!=null && rs.length>minReads) {
		    if (forward) 
			e = rs[rs.length-1].getPosition();
		    else
			s = rs[0].getPosition();
		}
	    }
	    return new TranscribedRegionMerge(chr, forward, s, e);
	}
	return null;
    }

    public static ABIRead[] analyzeIntragenicERNA(ABIRead[] rnaReads, int chr, boolean forward, int windowSize, RefSeqUCSCEnsemblGenes geneInfo, TreeSet<Enhancer2> intragenicEnhancers, String rnaFileName, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs, ConservationIslands consIslands) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> enhancerRegions = new TreeSet<TranscribedRegionMerge>();
	for (Enhancer2 e : intragenicEnhancers) {
	    if (e.getChromosome()==chr) {
		RefSeqUCSCEnsemblGene overlappingGene = geneInfo.getOverlappingGene(e.getCenter());
		if (overlappingGene!=null && overlappingGene.getStrand()!=forward) {
		    TranscribedRegionMerge t = findERNATranscript(rnaReads, minReads, chr, e.getCenter() + (forward ? 0 : -windowSize), e.getCenter() + (forward ? windowSize : 0), forward, windowSize);
		    if (t!=null/* && (t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
			boolean conserved = consIslands!=null ? consIslands.overlapsWithin(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_ENHANCER_DIST) : true;
			featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
			t.setConservedStart(conserved);
			t.assignReads(rnaReads);
			System.out.println("Found eRNA transcript " + t.toString() + " with " + t.getReads().length + " overlapping gene " + overlappingGene.toString());
			//if (t.getReadDensity()>minDens) {
			enhancerRegions.add(t);
			rnaReads = subtractReads(rnaReads, t.getReads());
			//}
		    }
		}
	    }
	}
	for (TranscribedRegionMerge t : enhancerRegions) {
	    TreeSet<TranscribedRegion2> overlappingERNA = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.ERNA_STR));
	    saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingERNA);
	    if (t.merge(overlappingERNA)) {
		boolean conserved = consIslands!=null ? consIslands.overlapsWithin(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_ENHANCER_DIST) : true;
		featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
		t.setConservedStart(conserved);
		t.assignReads(rnaReads);
		rnaReads = subtractReads(rnaReads, t.getReads());
		nFoundByHaTriC++;
	    }
	}
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, true, enhancerRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "eRNA";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", true, minAAs, enhancerRegions, type);
	}
	return rnaReads;
    }
	
    public static ABIRead[] analyzeAS(ABIRead[] rnaReads, boolean forward, int windowSize, RefSeqUCSCEnsemblGenes geneInfo, String rnaFileName, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, Vector<Peak> peaks, Vector<Peak> ctcfPeaks, BufferedWriter bw, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs) {
	int startReads = rnaReads.length, chr = geneInfo.getChromosome(), nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> asRegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the peaks and create a region for each peak that has transcripts and is intragenic
	for (Peak p : peaks) {
	    if (p.getChromosome()==chr) {
		RefSeqUCSCEnsemblGene overlappingGene = geneInfo.getOverlappingGene(p.getCenter());
		if (overlappingGene!=null && overlappingGene.getStrand()!=forward) {
		    TranscribedRegionMerge t = findERNATranscript(rnaReads, minReads, chr, p.getCenter() + (forward ? 0 : -windowSize), p.getCenter() + (forward ? windowSize : 0), forward, windowSize);
		    if (t!=null/* && (t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
			t.assignReads(rnaReads);
			asRegions.add(t);
			rnaReads = subtractReads(rnaReads, t.getReads());
			//System.out.println("Found AS transcript " + t.toString() + " with " + t.getReads().length + " reads for gene " + overlappingGene.toString());
		    }
		}
	    }
	}
	for (TranscribedRegionMerge t : asRegions) {
	    TreeSet<TranscribedRegion2> overlappingAS = t.findOverlappingTranscript(reducedModel.getSegments(TranscribedRegion2.NON_CODING_AS_STR));
	    saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, overlappingAS);
	    if (t.merge(overlappingAS)) {
		t.assignReads(rnaReads);
		rnaReads = subtractReads(rnaReads, t.getReads());
		nFoundByHaTriC++;
		//System.out.println("Found AS transcript " + t.toString() + " with " + t.getReads().length + " reads.");
	    }
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, false, asRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "AS";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", false, minAAs, asRegions, type);
	}
	return rnaReads;
    }

    public static ABIRead[] analyzeNovel(ABIRead[] rnaReads, int chr, boolean forward, String rnaFileName, double minDens, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, int minAAs, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> novelRegions = new TreeSet<TranscribedRegionMerge>();
	TreeSet<TranscribedRegion2> overlappingNovel = reducedModel.getSegments(TranscribedRegion2.NON_CODING_OTHER_RNA_STR);
	overlappingNovel.addAll(reducedModel.getSegments(TranscribedRegion2.PROMOTER_DISTAL_NCRNA_STR));
	for (TranscribedRegion2 t2 : overlappingNovel) {
	    if (t2.getChromosome()==chr && t2.getStrand()==forward) {
		TranscribedRegionMerge t = new TranscribedRegionMerge(t2);
		t.assignReads(rnaReads);
		if (true /*(t.findOverlappingTranscript(repeatRNAs)).isEmpty()*/) {
		    if (t.getReadDensity()>minDens && t.getReads().length>minReads) {
			ConservationIsland ci = consIslands.getOverlappingConservationIsland(chr, forward ? t.getStart() : t.getEnd());
			if (ci!=null && !consIslandsOverlapped.contains(ci)) {
			    featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			    t.setConservedStart(true);
			}
			novelRegions.add(t);
			nFoundByHaTriC++;
			//System.out.println("Found novel transcript " + t.toString() + " with " + t.getReads().length);
			
			//System.out.println("saveHaTriCRegions(" + reducedModel.getChromosome() + ", " + reducedModel.getStrand() + ", " + rnaFileName + ", " + t2.toString());
			saveHaTriCRegions(reducedModel.getChromosome(), reducedModel.getStrand(), rnaFileName, t2);
			rnaReads = subtractReads(rnaReads, t.getReads());
		    }
		    //else 
		    //System.out.println("Found novel transcript " + t.toString() + " has too few reads " + t.getReads().length);
		}
		else
		    System.out.println("Found novel transcript " + t.toString() + " with " + t.getReads().length + " reads overlapping repeat RNA.");
	    }
	}	
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, false, novelRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	if (minAAs>0) {
	    String type = "novel";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", false, minAAs, novelRegions, type);
	}
	return rnaReads;
    }

    public static ABIRead[] analyzeConservedNovel(ABIRead[] rnaReads, int chr, boolean forward, double minDens, int minReads, double normFactor, ReducedModel2 reducedModel, int upstream, int downstream, int maxESTDist, int ciSlackDist) {
	ESTCIData estciData = new ESTCIData();
	HashMap<EST, TranscribedRegionMerge> ests2Novel = new HashMap<EST, TranscribedRegionMerge>();
	TreeSet<TranscribedRegionMerge> novelRegions = new TreeSet<TranscribedRegionMerge>();
	TreeSet<TranscribedRegion2> overlappingNovel = reducedModel.getSegments(TranscribedRegion2.NON_CODING_OTHER_RNA_STR);
	overlappingNovel.addAll(reducedModel.getSegments(TranscribedRegion2.PROMOTER_DISTAL_NCRNA_STR));
	for (TranscribedRegion2 t2 : overlappingNovel) {
	    if (t2.getChromosome()==chr && t2.getStrand()==forward) {
		TranscribedRegionMerge t = new TranscribedRegionMerge(t2);
		t.assignReads(rnaReads);
		if (t.getReadDensity()>minDens && t.getReads().length>minReads) {
		    System.out.println("Found novel transcript " + t.toString() + " with " + t.getReads().length);
		    TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(chr, (forward ? t.getStart() : t.getEnd()) - ciSlackDist, (forward ? t.getStart() : t.getEnd()) + ciSlackDist);
		    if (cis!=null && !cis.isEmpty()) {
			boolean foundCI = false;
			for (ConservationIsland ci : cis) {
			    if (!consIslandsOverlapped.contains(ci)) {
				consIslandsOverlapped.add(ci);
				foundCI = true;
			    }
			}
			if (foundCI) {
			    estciData.addCI(cis);
			    System.out.println("Found conserved novel transcript " + t.toString() + " with " + t.getReads().length);
			    featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			    if (ests!=null) { //Check how many ESTs there are near this promoter
				for (EST est : ests) {
				    if (est.getStrand()==forward && Math.abs(est.getStart() - (forward ? t.getStart() : t.getEnd()))<maxESTDist) {
					ests2Novel.put(est, t);
					estciData.addCI(cis, est);
				    }
				}
				for (EST est : ests2Novel.keySet())
				    ests.remove(est);
			    }
			}
		    }
		}
	    }
	}
	saveTranscriptInfo(readDensSaveFile, false, novelRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	
	return estciData;
    }

    public static ABIRead[] analyzePeaks(ABIRead[] rnaReads, int chr, boolean forward, int windowSize, Vector<Peak> peaks, Vector<Peak> k4me3peaks, int minReads, String readDensSaveFile, double normFactor, ReducedModel2 reducedModel, RefSeqUCSCEnsemblGenes geneInfo, BufferedWriter bw, FeatureInfo featureInfo, int upstream, int downstream, TreeSet<TranscribedRegion2> repeatRNAs, ConservationIslands consIslands) {
	int startReads = rnaReads.length, nFoundByHaTriC = 0;
	TreeSet<TranscribedRegionMerge> peakRegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the peaks and create a region for each peak that has transcripts and is intragenic
	for (Peak p : peaks) {
	    //Skip peaks that overlap an me3 peak
	    RefSeqUCSCEnsemblGene nearestTSS = geneInfo.getNearestTSSGene(p.getCenter());
	    if (p.getChromosome()==chr && !geneInfo.overlapsGene(p.getStart(), p.getEnd()) && Math.abs(p.getCenter() - nearestTSS.getStart())>2000) {
		if (k4me3peaks==null || !p.overlaps(k4me3peaks)) {
		    TranscribedRegionMerge t = findERNATranscript(rnaReads, minReads, chr, p.getCenter() + (forward ? 0 : -windowSize), p.getCenter() + (forward ? windowSize : 0), forward, windowSize);
		    if (t!=null) {
			t.assignReads(rnaReads);
			boolean conserved = consIslands!=null ? consIslands.overlapsWithin(geneInfo.getChromosome(), forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_ENHANCER_DIST) : null;
			featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
			t.setConservedStart(conserved);
			peakRegions.add(t);
			rnaReads = subtractReads(rnaReads, t.getReads());
			System.out.println("Found " + (geneInfo.getOverlappingGene(p.getCenter())==null ? "extra" : "inter") + "genic peak transcript " + t.toString() + " with " + t.getReads().length + " reads for peak " + p.toString() + " with nearest TES gene " + geneInfo.getNearestTESGene(p.getCenter()).toString());
		    }
		}
		//else 
		//  System.out.println(p.toString() + " overlaps K4me3 peak.");
	    }
	}
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	    saveTranscriptInfo(readDensSaveFile, false, peakRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	/*if (minAAs>0) {
	    String type = "coding";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", false, minAAs, codingRegions, type);
	    }*/
	return rnaReads;
    }

    public static ESTCIData analyzeConservedPeaks(ABIRead[] rnaReads, int chr, boolean forward, Vector<Peak> peaks, Vector<Peak> k4me3peaks, RefSeqUCSCEnsemblGenes geneInfo, double normFactor, FeatureInfo featureInfo, int upstream, int downstream, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped, TreeSet<EST> ests, int maxESTDist, int ciSlackDist) {
	ESTCIData estciData = new ESTCIData();
	HashMap<EST, Peak> ests2Peaks = new HashMap<EST, Peak>();
	for (Peak p : peaks) {
	    //Skip peaks that overlap an me3 peak
	    RefSeqUCSCEnsemblGene nearestTSS = geneInfo.getNearestTSSGene(p.getCenter());
	    if (p.getChromosome()==chr && !geneInfo.overlapsGene(p.getStart(), p.getEnd()) && Math.abs(p.getCenter() - nearestTSS.getStart())>2000) {
		TreeSet<ConservationIsland> cis = consIslands.getOverlappingConservationIslands(chr, p.getCenter() - ciSlackDist, p.getCenter() + ciSlackDist);
		if (cis!=null && !cis.isEmpty()) {
		    boolean foundCI = false;
		    for (ConservationIsland ci : cis) {
			if (!consIslandsOverlapped.contains(ci)) {
			    consIslandsOverlapped.add(ci);
			    foundCI = true;
			}
		    }
		    if (foundCI && (k4me3peaks==null || !p.overlaps(k4me3peaks))) {
			//System.out.println("Nearest TSS for " + p.toString() + " is " + Math.abs(p.getCenter() - nearestTSS.getStart()) + " for " + nearestTSS.toString());
			estciData.addCI(cis);
			featureInfo.addLocInfo(chr, p.getCenter(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, p.getCenter(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, true);
			for (EST est : ests) {
			    if (est.getStrand()==forward && Math.abs(est.getStart() - p.getCenter())<maxESTDist) {
				estciData.addCI(cis, est);
			    }
			}
			for (EST est : ests2Peaks.keySet())
			    ests.remove(est);
		    }
		}
	    }
	}
	return estciData;
    }

    public static ESTCIData analyzeConservedRemaining(ABIRead[] rnaReads, int chr, boolean forward, RefSeqUCSCEnsemblGenes geneInfo, double normFactor, ConservationIslands consIslands, HashSet<ConservationIsland> consIslandsOverlapped, TreeSet<EST> ests, int maxESTDist) {
	//Use this method to locate ESTs near CI
	ESTCIData estciData = new ESTCIData();
	HashMap<EST, ConservationIsland> ests2CIs = new HashMap<EST, ConservationIsland>();
	for (ConservationIsland ci : consIslands.getConservationIslands()) {
	    RefSeqUCSCEnsemblGene nearestTSS = geneInfo.getNearestTSSGene(ci.getCenter());
	    if (ci!=null && !consIslandsOverlapped.contains(ci) && !geneInfo.overlapsGene(ci.getStart(), ci.getEnd()) && Math.abs(ci.getCenter() - nearestTSS.getStart())>2000) {
		estciData.addCI(ci);
		//if (k4me3peaks==null || !p.overlaps(k4me3peaks)) {
		for (EST est : ests) {
		    if (est.getStrand()==forward && Math.abs(est.getStart() - ci.getCenter())<maxESTDist) {
			//Check if EST overlaps known TSS
			RefSeqUCSCEnsemblGene g = geneInfo.getOverlappingTSS(forward ? est.getStart() : est.getEnd(), forward ? est.getEnd() : est.getStart());
			if (g!=null)
			    System.out.println(est.toString() + " overlaps TSS " + g.toString());
			//else
			//System.out.println(est.toString() + " starts near " + ci.toString());
			estciData.addCI(ci, est);
		    }
		}
		for (EST e : ests2CIs.keySet())
		    ests.remove(e);
		//}
		//}
	    }
	}
	return estciData;
    }

    public static ABIRead[] analyzeH3K4me3Peaks(ABIRead[] rnaReads, int chr, boolean forward, int windowSize, Vector<Peak> k4me3peaks, int minReads, String readDensSaveFile, double normFactor, BufferedWriter bw, int upstream, int downstream, ConservationIslands consIslands) {
	//Use this function to analyze the expression levels surrounding the H3K4me3 peaks
	int startReads = rnaReads.length;
	TreeSet<TranscribedRegionMerge> peakRegions = new TreeSet<TranscribedRegionMerge>();
	//Go through the peaks and create a region for each peak that has transcripts and is intragenic
	for (Peak p : k4me3peaks) {
	    if (p.getChromosome()==chr) {
		TranscribedRegionMerge t = findERNATranscript(rnaReads, minReads, chr, p.getCenter() + (forward ? 0 : -windowSize), p.getCenter() + (forward ? windowSize : 0), forward, windowSize);
		if (t!=null) {
		    t.assignReads(rnaReads);
		    boolean conserved = consIslands!=null ? consIslands.overlapsWithin(chr, forward ? t.getStart() : t.getEnd(), FindConservationIslands.MAX_ENHANCER_DIST) : null;
		    //featureInfo.addLocInfo(chr, forward ? t.getStart() : t.getEnd(), forward, FindReadsAcrossFeature2.getReadsHistogram(chr, forward ? t.getStart() : t.getEnd(), forward, downstream, upstream, rnaReads, 1.0), normFactor, upstream, conserved);
		    t.setConservedStart(conserved);
		    peakRegions.add(t);
		    rnaReads = subtractReads(rnaReads, t.getReads());
		    System.out.println("Found H3K4me3 peak transcript " + t.toString() + " with " + t.getReads().length + " reads for H3K4me3 peak " + p.toString());
		}
	    }
	}
	//Now we can calculate how many reads that were assigned to
	//coding regions, etc
	saveTranscriptInfo(readDensSaveFile, false, peakRegions, 0, normFactor, startReads - rnaReads.length, bw);
	/*if (minAAs>0) {
	    String type = "coding";
	    saveSequence(chr, forward, readDensSaveFile.substring(0, readDensSaveFile.lastIndexOf("/")) + SEQUENCE_DIR + type + "_transcripts_" + (forward ? "f" : "r") + "_" + chr + ".fa", false, minAAs, codingRegions, type);
	    }*/
	return rnaReads;
    }

    public static TreeMap<Integer, Integer> analyzeRemainingReads(ABIRead[] rnaReads, int windowSize, int chr, boolean forward, int minReads, String readDensSaveFile, String species, double normFactor, BufferedWriter bw, TreeSet<TranscribedRegion2> repeatRNAs) {
	//If consecutive windows have more than minReads, we assign them to a transcript which we save in saveFileName
	int startReads = rnaReads.length, nArtefacts = 0, nArtefactReads = 0, nFoundByHaTriC = 0;
	TreeMap<Integer, Integer> readsHist = new TreeMap<Integer, Integer>();
	TreeSet<TranscribedRegionMerge> remainingRegions = new TreeSet<TranscribedRegionMerge>();
	int windowStart = 0, transcriptStart = -1, nReadsInTranscript = 0;
	while (windowStart + windowSize<RefSeqUCSCEnsemblGenes.getChromosomeLength(chr, species)) {
	    ABIRead rs[] = getReadSubset(rnaReads, chr, forward, windowStart, windowStart + windowSize);
	    if (!isArtefact(rs)) {
		int nReads = rs.length;
		if (nReads>minReads && transcriptStart==-1) {
		    transcriptStart = windowStart;
		    nReadsInTranscript += nReads;
		}
		else if (nReads<=minReads && transcriptStart!=-1 && bw!=null) {
		    TranscribedRegionMerge t = new TranscribedRegionMerge(chr, forward, transcriptStart, windowStart);
		    t.assignReads(rnaReads);
		    rnaReads = subtractReads(rnaReads, t.getReads());
		    remainingRegions.add(t);
		    transcriptStart = -1;
		    nReadsInTranscript = 0;
		    //System.out.println("Found remaining transcript " + t.toString() + " with " + t.getReads().length);
		}
		if (!readsHist.containsKey(nReads))
		    readsHist.put(nReads, 0);
		readsHist.put(nReads, readsHist.get(nReads) + 1);
	    }
	    else {
		nArtefacts++;
		nArtefactReads += rs.length;
	    }
	    windowStart += windowSize;
	}
	System.out.println("Found " + nArtefacts + " loci that were considered artefacts with " + nArtefactReads);
	saveTranscriptInfo(readDensSaveFile, false, remainingRegions, nFoundByHaTriC, normFactor, startReads - rnaReads.length, bw);
	return readsHist;    
    }

}

class ESTCIData {

    //Use this class to store information about ESTs found at a given class of CIs

    protected HashMap<ConservationIsland, HashSet<EST>> cis2est = new HashMap<ConservationIsland, HashSet<EST>>();
    protected String category;
    
    public ESTCIData() { 

    }

    public void addCI(ConservationIsland ci) { cis2est.put(ci, new HashSet<EST>()); }

    public void addCI(ConservationIsland ci, EST est) { 
	if (!cis2est.containsKey(ci))
	    cis2est.put(ci, new HashSet<EST>()); 
	cis2est.get(ci).add(est);
    }

    public void addCI(TreeSet<ConservationIsland> cis) { 
	for (ConservationIsland ci : cis)
	    cis2est.put(ci, new HashSet<EST>()); 
    }

    public void addCI(TreeSet<ConservationIsland> cis, EST est) { 
	for (ConservationIsland ci : cis) {
	    if (!cis2est.containsKey(ci))
		cis2est.put(ci, new HashSet<EST>()); 
	    cis2est.get(ci).add(est); //Assign to all of the CIs
	}
    }

    public void putAll(ESTCIData d) {
	//Merge operation
	cis2est.putAll(d.cis2est);
    }


    public double ciHasESTProb(boolean spliced) {
	//Calculate the probability that a CI will have an EST associated with it
	
	return cis2est.size()==0 ? 0 : ((double)(cis2est.size() - numberOfFeaturesNoESTs(spliced)))/cis2est.size();
    }

    public int numberOfFeaturesNoESTs(boolean spliced) { 
	int nZero = 0;
	for (ConservationIsland ci : cis2est.keySet()) {
	    boolean estFound = false;
	    for (EST est : cis2est.get(ci)) {
		if (est.isSpliced()==spliced) {
		    estFound = true;
		    break;
		}
	    }
	    nZero += !estFound ? 1 : 0;
	}
	return nZero;
    }

    public int numberOfFeatures() { return cis2est.size(); }

    public int numberOfFeaturesESTCrossTSS(RefSeqUCSCEnsemblGenes geneInfo, boolean forward) {
	//Check how many CIs have an EST that cross an annotated TSS
	int n = 0;
	for (ConservationIsland ci : cis2est.keySet()) {
	    for (EST est : cis2est.get(ci)) {
		RefSeqUCSCEnsemblGene g = geneInfo.getOverlappingTSS(forward ? est.getStart() : est.getEnd(), forward ? est.getEnd() : est.getStart());
		if (g!=null && g.getStrand()==forward) {
		    n++;
		    break;
		}
		//System.out.println(est.toString() + " overlaps TSS " + g.toString());
	    }
	}
	return n;
   }

}