//This class represents an exon from JMGs allEXNs.tsv file. We assign
//exons to genes by looking for overlap which is reasonably accurate,
//but not perfect.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class Exon implements Comparable {

    protected int start, end, chr, ind;
    protected String gene;

    public Exon(int c, int s, int e, String n) {
	chr = c;
	start = s;
	end = e;
	gene = n;
    }

    public int getChromosome() { return chr; }

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public boolean overlaps(ConservationIsland c) { return c.getChromosome()==chr && overlaps(c.getStart(), c.getEnd()); }

    public boolean overlaps(int s, int e) {
	return (start<=s && end>=s) || (start<=e && end>=e) || (start>s && end<e || (start<s && end>e));
    }

    public TreeSet<ConservationIsland> getOverlappingCIs(ConservationIslands consIslands) {
	TreeSet<ConservationIsland> overlappingCIs = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : consIslands.getConservationIslands()) {
	    if (overlaps(c))
		overlappingCIs.add(c);
	}
	return overlappingCIs;
    }

    //////////////////////////////////////////////////////////////////

    public boolean equals(Object o) {
	if (!(o instanceof Exon))
	    return false;
	Exon e = (Exon)o;
	return e.chr==chr && e.start==start && e.end==end;
    }

    public int compareTo(Object o) {
	if (!(o instanceof Exon))
	    return 0;
	Exon e = (Exon)o;
	if (chr<e.chr)
	    return -1;
	else if (chr>e.chr)
	    return 1;
	else {
	    if (start<e.start)
		return -1;
	    else if (start>e.start)
		return 1;
	    else {
		if (end<e.end)
		    return -1;
		else if (end>e.end)
		    return 1;
		else
		    return 0;
	    }
	}
    }

    public int hashCode() {
	return gene.hashCode() + start*23 + end*37 + chr*93;
    }

    public String toString() {
	return "chr" + chr + ":" + start + "-" + end;
    }

}