//Use this method to extract the sequences surrounding the peaks. In
//particular, we want to be able to pick a subset defined by the
//proximity of enhancers, TSSs, or some other feature.
//
//Example:
//
//bsub -q shared_12h java ExtractPeakSequences -p MmDataPeakFiles/tf_kcl_names --tf Npas4_Npas4_KCl_B1 --enhancers /groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/enhancers_b2_null_all.txt -w 200 -o
//
//Find the peaks near a specific type of transcripts:
//
////bsub -q shared_12h java ExtractPeakSequences -p MmDataPeakFiles/tf_kcl_names --tf Npas4_Npas4_KCl_B1 --transcripts /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c_b1/hr6c_b1 --type Divergent -W 1500 -w 100 -o
//
//bsub -q shared_12h java ExtractPeakSequences -p MmDataPeakFiles/cbp_names --tf CBP_Millipore_KCl_B1 -w 1500 -o -c 19 --upstream

//bsub -q shared_2h java ExtractPeakSequences --enhancers /groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/enhancers_b2_null_all.txt -w 1500 -o -c 19 --intergenic
//
//bsub -q shared_2h java ExtractPeakSequences -p MmDataPeakFiles/cbp_names --tf CBP_Millipore_KCl_B1 -w 100 -o -c 1
//
//bsub -q shared_2h java ExtractPeakSequences -f Brenda/promoter_Npas4_Npas4_KCl_B1.txt -w 100 
//
//bsub -q shared_12h java ExtractPeakSequences -TSS -w 2500
//
//For the ncor peaks:
//
//bsub -q shared_2h java -Xmx1024m ExtractPeakSequences -w 2000 -f MeCP2DataPeakFileNames/ncor_names_combos_center.txt -o -s Harrison/ncor_seqs_w_2000.fa 
//
//For getting the sequences around the peaks that were classified using various schemes:
//
//bsub -q shared_2h java ExtractPeakSequences -f /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/PeakLocations/Coding_inactive_tfbs_locations.txt -w 100 -o
//
//Get the sequences near a subset of TSSs:
//
//bsub -q shared_12h java ExtractPeakSequences -TSS -w 2500 --select Ivo/inhibitory_specific_enhancers_d_10000_s_3.5_genes.txt
//
//for ((i=1; i<=21; i++)); do bsub -q shared_2h -o /dev/null java -Xmx1024m ExtractPeakSequences -w 100 -f Matlab/Athar/ap1_fos_peaks_distal.txt -o --remove TGANTCA -c $i ; done
//
//for ((i=1; i<=21; i++)); do bsub -q shared_2h -o /dev/null java -Xmx512m ExtractPeakSequences -w 100 -f Athar/AP1Files/ap1_locs_$i.txt -s Athar/AP1Files/Sequences/ap1_locs_w_100_$i.fa -o --remove TGANTCA -c $i ; done
//
//To get the sequences for the DHSs in Wui's project
//
//cell_types=(GM12878_all HepG2_all H1hESC_all K562_all ); for cell_type in ${cell_types[@]}; do for ((i=1; i<=24; i++)); do bsub -q shared_2h -o /dev/null java -Xmx512m ExtractPeakSequences -w 100 -f Wui/PeakLocs/$cell_type\_locs.txt -s Wui/PeakLocs/Sequences/$cell_type\_locs_w_100_$i.fa -o -c $i --species hs --hg19 ; done ; done
//
//For getting control sequences
//
//cell_types=(GM12878_all HepG2_all H1hESC_all K562_all ); for cell_type in ${cell_types[@]}; do for ((i=1; i<=24; i++)); do bsub -q shared_2h -o /dev/null java -Xmx512m ExtractPeakSequences -w 100 -f Wui/PeakLocs/$cell_type\_locs.txt -s Wui/PeakLocs/Sequences/$cell_type\_control_locs_w_100_$i.fa -o -c $i --species hs --hg19 --offset 2000 ; done ; done
//
//For getting the sequences near the DHSs in 33 cell types
//
//for ((i=1; i<=24; i++)); do bsub -q shared_2h -o /dev/null java -Xmx512m ExtractPeakSequences -w 100 -f Transcripts/HeatMaps/dnase1_tier3_names_peaks_heatmap_$i.txt -s Wui/PeakLocs/Sequences/Tier3/dnase1_tier3_names_locs_w_100_$i.fa -o -c $i --species hs --hg19 ; done
//
//For getting the sequences near the DHSs in 79 cell types
//
//for ((i=1; i<=24; i++)); do bsub -q shared_7d -o /dev/null java -Xmx3512m ExtractPeakSequences -w 50 -f Transcripts/HeatMaps/dnase_stam1_names_peaks_heatmap_$i.txt -s /groups/kreiman/hemberg/SeqData/Wui/PeakLocs/Sequences/Tier3/dnase1_stam1_names_locs_w_50_$i.fa -o -c $i --species hs --hg19 ; done
//
//
//for ((i=1; i<=24; i++)); do bsub -q shared_7d -o /dev/null java -Xmx3512m ExtractPeakSequences -w 50 -f Wui/HeatMaps/dnase_stam1_names_dnase_stam2_names_peaks_heatmap_$i.txt -s /groups/kreiman/hemberg/SeqData/Wui/PeakLocs/Sequences/Tier3/dnase1_stam1_names_dnase_stam2_names_locs_w_50_$i.fa -o -c $i --species hs --hg19 ; done
//
//Extract the single validated negative region to use as background for synthetic regions
//
//java ExtractPeakSequences -w 500 -f /home/mh168/Projects/EnhMut/validated_neg_reg.txt -s /home/mh168/Projects/EnhMut/validated_neg_reg_w_500.fa -o -c 10

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class ExtractPeakSequences {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option strongestPeaksFlag = argParser.addIntegerOption("top");	    
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option featureFileNameFlag = argParser.addStringOption('f', "featureFileName");
	    CmdLineParser.Option tfNameFlag = argParser.addStringOption("tf");
	    //Use these flags when we only want peaks in the vicinity of
	    //a certain type of transcripts
	    CmdLineParser.Option transcriptFileNameFlag = argParser.addStringOption("transcripts");
	    CmdLineParser.Option transcriptTypeFlag = argParser.addStringOption("type");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option transcriptDistanceFlag = argParser.addIntegerOption('W', "transcriptDistance");
	    CmdLineParser.Option reverseStrandFlag = argParser.addBooleanOption('r', "reverse");
	    
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption("enhancers");
	    CmdLineParser.Option tssFlag = argParser.addBooleanOption("TSS");
	    CmdLineParser.Option intragenicFlag = argParser.addBooleanOption("intragenic");
	    CmdLineParser.Option intergenicFlag = argParser.addBooleanOption("intergenic");
	    CmdLineParser.Option downstreamFlag = argParser.addBooleanOption("downstream");
	    CmdLineParser.Option upstreamFlag = argParser.addBooleanOption("upstream");
	    CmdLineParser.Option subsetFlag = argParser.addStringOption("subset");
	    CmdLineParser.Option selectFileNameFlag = argParser.addStringOption("select");
	    //Use to remove a certain sequence
	    CmdLineParser.Option removeMotifFlag = argParser.addStringOption("remove");
	    CmdLineParser.Option hg19Flag = argParser.addBooleanOption("hg19");
	    //Use this flag to add an offset to all locations. Useful when creating control sequences
	    CmdLineParser.Option offsetFlag = argParser.addIntegerOption("offset");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean tss = argParser.getOptionValue(tssFlag)!=null;
	    boolean intragenic = argParser.getOptionValue(intragenicFlag)!=null;
	    boolean intergenic = argParser.getOptionValue(intergenicFlag)!=null;
	    boolean downstream = argParser.getOptionValue(downstreamFlag)!=null;
	    boolean upstream = argParser.getOptionValue(upstreamFlag)!=null;
	    boolean reverseStrand = argParser.getOptionValue(reverseStrandFlag)!=null;
	    boolean hg19 = argParser.getOptionValue(hg19Flag)!=null; 
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(strongestPeaksFlag);
	    int nStrongestPeaks = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(tfNameFlag);
	    String tfName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(featureFileNameFlag);
	    String featureFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(subsetFlag);
	    String subset = tmp==null ? "" : tmp.toString();
	    tmp = argParser.getOptionValue(transcriptFileNameFlag);
	    String transcriptFileName = tmp==null ? null : tmp.toString();
	    tmp = argParser.getOptionValue(selectFileNameFlag);
	    String selectFileName = tmp==null ? null : tmp.toString();
	    tmp = argParser.getOptionValue(removeMotifFlag);
	    String removeMotif = tmp==null ? null : tmp.toString();
	    int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	    tmp = argParser.getOptionValue(transcriptTypeFlag);
	    String transcriptType = tmp==null ? "" : tmp.toString();
	    tmp = argParser.getOptionValue(transcriptDistanceFlag);
	    int transcriptDistance = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(offsetFlag);
	    int offset = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = null;
	    if (tmp!=null)
		saveFileName = (String)tmp;
	    else {
		if (peakFileName!=null) {
		    if (enhancerFileName!=null)
			saveFileName = peakFileName.replace("MmDataPeakFiles", "MmDataPeakFiles/PeakSequences") + (tfName==null ? "" : "_" + tfName) + "_" + enhancerFileName.substring(enhancerFileName.lastIndexOf("/") + 1, enhancerFileName.lastIndexOf(".")) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa";
		    else if (tss)
			saveFileName = (peakFileName!=null ? peakFileName.replace("MmDataPeakFiles", "MmDataPeakFiles/PeakSequences") : "/groups/neuroduo/martin/Genomes/Mm/v9/")  + (tfName==null ? "" : "_" + tfName + "_") + (nStrongestPeaks>0 ? "n_" + nStrongestPeaks + "_" : "") + (geneIDsKeep==null ? "" : selectFileName.substring(selectFileName.lastIndexOf("/") + 1) + "_") + "TSS_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa";
		    else if (intragenic)
			saveFileName = peakFileName.replace("MmDataPeakFiles", "MmDataPeakFiles/PeakSequences") + (tfName==null ? "" : "_" + tfName) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_intragenic_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa";
		    else if (intergenic)
			saveFileName = peakFileName.replace("MmDataPeakFiles", "MmDataPeakFiles/PeakSequences") + (tfName==null ? "" : "_" + tfName) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_intergenic_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa";
		    else if (transcriptFileName!=null)
			saveFileName = peakFileName.replace("MmDataPeakFiles", "MmDataPeakFiles/PeakSequences") + (tfName==null ? "" : "_" + tfName) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_" + transcriptFileName.substring(transcriptFileName.lastIndexOf("/") + 1) + "_" + transcriptType + "_d_" + transcriptDistance + "_w_" + windowSize + (reverseStrand ? "_r" : "_f") + (chr==-1 ? "" : "_" + chr) + ".fa";
		    else 
			saveFileName = peakFileName.replace("MmDataPeakFiles", "MmDataPeakFiles/PeakSequences") + (tfName==null ? "" : "_" + tfName) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_w_" + windowSize + (downstream ? "d" : upstream ? "u" : "") + (chr==-1 ? "" : "_" + chr) + ".fa";
		}
		else if (enhancerFileName!=null)
		    saveFileName = enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_w_" + windowSize + (downstream ? "d" : upstream ? "u" : "") + (chr==-1 ? "" : "_" + chr) + ".fa";
		else if (featureFileName!=null) {
		    saveFileName = featureFileName.substring(0, featureFileName.lastIndexOf("/") + 1) + "Sequences/";
		    saveFileName += featureFileName.substring(featureFileName.lastIndexOf("/") + 1, featureFileName.lastIndexOf(".")) + (tfName==null ? "" : "_" + tfName) + (nStrongestPeaks>0 ? "_n_" + nStrongestPeaks : "") + "_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa";
		}
		else if (tss)
		    saveFileName = "MmDataPeakFiles/PeakSequences/" + (geneIDsKeep==null ? "" : selectFileName.substring(selectFileName.lastIndexOf("/") + 1) + "_") + "TSS_w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".fa";
	    }
	    ///////////////////////////////////////////////////////
	    int nChr = species.toLowerCase().equals("mm") ? 21 : 24;
	    //Get the peaks
	    TreeMap<Integer, Peak> peaks = null;
	    if (peakFileName!=null) {
		peaks = FindEnhancerSites.parsePeakFile(peakFileName, chr, false);
		System.err.println("Warning! Inter and extragenic subsets currently not working!");
		/*TreeSet<Peak> removePeaks = new TreeSet<Peak>();
		for (Peak p : peaks.values()) {
		    RefSeqGene overlappingGene = null;//geneInfo[i-1].getOverlappingGene(p.getCenter());
		    boolean subsetOK = subset==null || subset.equals("") || (subset.equals(FindReadsAcrossFeature.INTERGENIC_STR) && overlappingGene==null) || (subset.equals(FindReadsAcrossFeature.INTRONIC_STR) && overlappingGene!=null);
		    if (!subsetOK)
			removePeaks.add(p);
		}
		for (Peak p : removePeaks) 
		peaks.remove(p.getCenter());*/
		
		TreeSet<String> peakNames = new TreeSet<String>(FindEnhancerSites.parsePeakFileHeader(peakFileName, chr).values());
		System.out.println("Found " + peaks.size() + " combo peaks with " + peakNames.size() + " factors from " + peakFileName);
	    }
	    Vector<Enhancer> enhancers = enhancerFileName==null ? null : FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	    TreeSet<TranscribedRegion2> transcripts = transcriptFileName==null ? null : loadTranscripts(transcriptFileName, transcriptType, reverseStrand);
	    if (enhancers!=null)
		Collections.sort(enhancers);
	    Vector<Peak> ps = null;
	    if (nStrongestPeaks>0) 
		ps = findStrongestPeaks(peaks, tfName, nStrongestPeaks, chr, enhancers, tss, windowSize, orchestra);
	    else if (peakFileName!=null)
		ps = new Vector<Peak>(peaks.values());
	    //Save the sequences as well
	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 ||i==chr) {
		    RefSeqGenes geneInfo = new RefSeqGenes(i, orchestra);
		    FastaSequence2 fastaReader = new FastaSequence2(species, hg19, i, orchestra);
		    if (peaks!=null) {
			for (Peak p : ps) {
			    if (p.getChromosome()==i && (tfName==null || p.hasFactor(tfName))) {
				if (enhancers!=null) { //Make sure that there is a nearby enhancer
				    if (nearEnhancer(p, enhancers, windowSize))
					bw.write(">chr" + i + ":" + (p.getCenter() - windowSize) + "-" + (p.getCenter() + windowSize) + "\n" + fastaReader.getSequence(p.getCenter() - windowSize, 1 + 2*windowSize) + "\n");
				}
				else if (transcripts!=null) {
				    if (nearTranscript(p, transcripts, transcriptDistance))
					bw.write(">chr" + i + ":" + (p.getCenter() - windowSize) + "-" + (p.getCenter() + windowSize) + "\n" + fastaReader.getSequence(p.getCenter() - windowSize, 1 + 2*windowSize) + "\n");
				}
				else if (tss) {
				    if (Math.abs(p.getCenter() - geneInfo.getNearestTSS(p.getCenter()))<windowSize)
					bw.write(">chr" + i + ":" + (p.getCenter() - windowSize) + "-" + (p.getCenter() + windowSize) + "\n" + fastaReader.getSequence(p.getCenter() - windowSize, 1 + 2*windowSize) + "\n");
				}
				else if (intragenic) {
				    if (geneInfo.overlapsGene(p.getCenter()) && Math.abs(geneInfo.getNearestTSS(p.getCenter()))>windowSize)
					bw.write(">chr" + i + ":" + (p.getCenter() - windowSize) + "-" + (p.getCenter() + windowSize) + "\n" + fastaReader.getSequence(p.getCenter() - windowSize, 1 + 2*windowSize) + "\n");
				}
				else if (intergenic) {
				    if (!geneInfo.overlapsGene(p.getCenter()) && Math.abs(geneInfo.getNearestTSS(p.getCenter()))>windowSize)
					bw.write(">chr" + i + ":" + (p.getCenter() - windowSize) + "-" + (p.getCenter() + windowSize) + "\n" + fastaReader.getSequence(p.getCenter() - windowSize, 1 + 2*windowSize) + "\n");
				}
				else {
				    if (!upstream && !downstream)
					bw.write(">chr" + i + ":" + (p.getCenter() - windowSize) + "-" + (p.getCenter() + windowSize) + "\n" + fastaReader.getSequence(p.getCenter() - windowSize, 1 + 2*windowSize) + "\n");
				    else {
					bw.write(">chr" + i + ":" + (p.getCenter() - (downstream ? 0 : windowSize)) + "-" + (p.getCenter() + (upstream ? windowSize : 0)) + "\n");

					if (downstream)
					    bw.write(fastaReader.getSequence(Math.max(1, p.getCenter()), windowSize) + "\n");
					else 
					    bw.write(FastaSequence2.complement(FastaSequence2.reverse(fastaReader.getSequence(Math.max(1, p.getCenter() - windowSize - 1), windowSize))) + "\n");
				    }
				}
			    }
			}
		    }
		    else {
			if (tss) {
			    for (RefSeqGene g : geneInfo.getGenes()) {
				if (g.getStrand() && (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0)) {
				    bw.write(">chr" + i + ":" + (g.getStart() - windowSize) + "-" + (g.getStart() + windowSize) + "(+)_ID:" + g.getID() + "\n");
				    bw.write(fastaReader.getSequence(Math.max(1, g.getStart() - windowSize), 1 + 2*windowSize) + "\n");
				}
			    }
			    fastaReader = new FastaSequence2(i, orchestra);
			    for (RefSeqGene g : geneInfo.getGenes()) {
				if (!g.getStrand() && (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0)) {
				    bw.write(">chr" + i + ":" + (g.getStart() - windowSize) + "-" + (g.getStart() + windowSize) + "(-)_ID:" + g.getID() + "\n");
				    bw.write(FastaSequence2.complement(FastaSequence2.reverse(fastaReader.getSequence(Math.max(1, g.getStart() - windowSize), 1 + 2*windowSize))) + "\n");
				}
			    }
			}
			else if (enhancers!=null) {
			    for (Enhancer en : enhancers) {
				if (en.getChromosome()==i) {
				    bw.write(">chr" + i + ":" + (en.getCenter() - (downstream ? 0 : windowSize)) + "-" + (en.getCenter() + (upstream ? windowSize : 0)) + "\n");
				    if (downstream)
					bw.write(fastaReader.getSequence(Math.max(1, en.getCenter()), windowSize) + "\n");
				    else if (upstream)
					bw.write(FastaSequence2.complement(FastaSequence2.reverse(fastaReader.getSequence(Math.max(1, en.getCenter() - windowSize - 1), windowSize))) + "\n");
				    else
					bw.write(fastaReader.getSequence(Math.max(1, en.getCenter() - windowSize), 1 + 2*windowSize) + "\n");
				}
			    }
			}
			else {
			    //Assume files with columns containing chr and peaks. A column for the strand is optional, but not yet implemented.
			    int[] locVec = FindReadsAcrossFeature2.parseFeatureFile(featureFileName, i, true);
			    Arrays.sort(locVec);
			    System.out.println(Arrays.toString(locVec));
			    for (int j=0; j<locVec.length; j++) {
				bw.write(">chr" + i + ":" + (locVec[j] - windowSize + offset) + "-" + (locVec[j] + windowSize + offset) + "\n");
				bw.write(filterMotif(fastaReader.getSequence(locVec[j] - windowSize + offset, 1 + 2*windowSize), removeMotif) + "\n");
			    }
			}
		    }
		}
	    }
	    bw.close();
	    //Launch the weeder job
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static String filterMotif(String seq, String removeMotif) {
	if (removeMotif==null)
	    return seq;
	int len = removeMotif.length();
	String removeMotifRevComp = EnumerateMers.reverseComplement(removeMotif);
	String removeMotifRegExp = removeMotif.replaceAll("N", ".");
	String removeMotifRevCompRegExp = removeMotifRevComp.replaceAll("N", ".");
	String mask = removeMotif.replaceAll(".", "N");
	seq = seq.toUpperCase().replaceAll(removeMotifRegExp, mask);
	seq = seq.replaceAll(removeMotifRevCompRegExp, mask);
	return seq;
    }


    public static Vector<Peak> findStrongestPeaks(TreeMap<Integer, Peak> peaks, String tfName, int nStrongestPeaks, int chr, Vector<Enhancer> enhancers, boolean tss, int windowSize, boolean orchestra) {
	TreeMap<Double, Peak> height2peak = new TreeMap<Double, Peak>();
	for (int i=1; i<=21; i++) {
	    if (chr==-1 || i==chr) {
		RefSeqGenes geneInfo = new RefSeqGenes(i, orchestra);
		for (Peak p : peaks.values()) {
		    if (p.getChromosome()==i && p.hasFactor(tfName)) {
			if (enhancers==null || nearEnhancer(p, enhancers, windowSize)) {
			    int tssDist = tss ? p.getCenter() - geneInfo.getNearestTSS(p.getCenter()) : 0;
			    //System.out.println("Nearest TSS is " + tssDist);
			    if (!tss || Math.abs(tssDist)<windowSize) {
				//Use negative values since we want to get the largest ones later
				double key = -p.getFactor(tfName);
				while (height2peak.containsKey(key)) //avoid collisions
				    key += random.nextGaussian()/100000;
				height2peak.put(key, p); 
			    }
			}
		    }
		}
	    }
	}
	System.out.println("Found a total of " + height2peak.size() + " peaks");
	Vector<Peak> ps = new Vector<Peak>();
	for (Double d : height2peak.keySet()) {
	    ps.addElement(height2peak.get(d));
	    if (ps.size()>=nStrongestPeaks)
		break;
	}
	return ps;
    }

    public static boolean nearEnhancer(Peak p, Vector<Enhancer> enhancers, int windowSize) {
	for (Enhancer en : enhancers) {
	    if (en.getChromosome()==p.getChromosome() && Math.abs(en.getCenter() - p.getCenter())<windowSize) 
		return true;
	}
	return false;
    }

    public static boolean nearTranscript(Peak p, TreeSet<TranscribedRegion2> transcripts, int maxDist) {
	for (TranscribedRegion2 t : transcripts) {
	    if (t.getChromosome()==p.getChromosome()) {
		if (t.getStrand() && Math.abs(t.getStart() - p.getCenter())<maxDist)
		    return true;
		else if (!t.getStrand() && Math.abs(t.getEnd() - p.getCenter())<maxDist)
		    return true;
	    }
	}
	return false;
    }

    public static TreeSet<TranscribedRegion2> loadTranscripts(String transcriptFileName, String type, boolean reverse) {
	//Use this method to load transcripts for a given type
	TreeSet<TranscribedRegion2> segments = new TreeSet<TranscribedRegion2>();
	try {
	    String fileName = transcriptFileName + "_transcript_densities_rna_names_" + (reverse ? "r" : "f") + ".txt";
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //chr\tstrand\tstart\tend\tdens\t...\tType
	    Pattern transcriptPattern = Pattern.compile("^(\\d+)\\s+([0|1])\\s+(\\d+)\\s+(\\d+)\\s+([\\d|\\.]+)\\s+.+\\s+(\\w+)\\s+\\d+$");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = transcriptPattern.matcher(line);
		if (m.find()) {
		    if (m.group(6).equals(type) && ((m.group(2).equals("1") && !reverse) || (m.group(2).equals("0") && reverse)) ) {
			TranscribedRegion2 t = new TranscribedRegion2(Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4)), !reverse, false, Integer.parseInt(m.group(1)), Double.parseDouble(m.group(5)));
			segments.add(t);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opening: " + e.toString());
	}
	System.out.println("Found " + segments.size() + " transcripts of type " + type);
	return segments;
    }

}