//Use this method to read a Fasta file and extract the sequences
//corresponding to probe locations.


import java.util.*;
import java.io.*;

public class FastaSequence2 {

    protected String fileName, species;
    protected int chr = -1, posInd = 0;
    protected BufferedReader fastaReader = null;
    protected Vector seqs = null, starts = null, inds = null;

    //////////////////////////////////////////////////////////////////

    public FastaSequence2(String f){ fileName = f; }

    public FastaSequence2(String species, boolean hg19, int chr, boolean orchestra) {
	fileName = orchestra ? "/groups/neuroduo/martin/Genomes/" : "/media/My Book/Genomes/";
	if (species.equals("mm")) {
	    fileName += "Mm/v9/chr";
	    if (chr==20)
	    fileName += "X";
	else if (chr==21) 
	    fileName += "Y";
	else
	    fileName += chr;
	}
	else if (species.equals("hs")) {
	    fileName += hg19 ? "Hs/v37/chr" : "Hs/v36/chr";
	    if (chr==23)
		fileName += "X";
	    else if (chr==24) 
		fileName += "Y";
	    else
		fileName += chr;
	}
	fileName += ".fa";
	System.out.println("Initiating sequence parser from " + fileName);
    }
    
    public FastaSequence2(int chr, boolean orchestra) {
	fileName = orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/";
	if (chr==20)
	    fileName += "Genomes/Mm/v9/chrX";
	else if (chr==21) 
	    fileName += "Genomes/Mm/v9/chrY";
	else
	    fileName += "Genomes/Mm/v9/chr" + chr;
	fileName += ".fa";
	System.out.println("Initiating sequence parser from " + fileName);
    }

    //////////////////////////////////////////////////////////////////

    public long getChromosomeLength(){
	long len = 0;
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null) {
		len += (long)line.length();
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	return len;
    }

    public static long getNumberOfRepeatElements(String fileName){
	long len = 0;
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null) {
		for (int j=0; j<line.length(); j++) {
		    String s = line.substring(j, j+1);
		    len += s.equals("a") || s.equals("c") || s.equals("g") || s.equals("t") ? 1 : 0;
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	return len;
    }

    public String[] getContiguousSequences(long pos[], int probeLength){
	//Use this method to extract the non-overlapping sequences
	//corresponding to the given tiling array.
	seqs = new Vector();
	starts = new Vector();
	inds = new Vector();
	int posIndex = 0;
	long genomeIndex = 0, rem = 0, genomePartLength = 0;	
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Do not use the fancy BioJava stuff. Instead, we just
	    //read the data since it is more or less just a flat
	    //file. This version assumes that there is only one
	    //sequence and header in the entire file. This will be a
	    //little bit tricky since the pos argument are based on an
	    //index starting at 1 while the strings used in Java are
	    //indexed from 0.
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    int segStartIndex = 0;
	    int segStopIndex = findContiguousRegion(pos, segStartIndex, probeLength);
	    while (segStopIndex<pos.length) {
		line = br.readLine();
		int lineLength = line.length();
		while (line!=null && genomeIndex + lineLength<=pos[segStartIndex]-1) {
		    genomePart = line;
		    genomeIndex += genomePart.length();
		    line = br.readLine();
		}
		while (line!=null && genomeIndex + genomePartLength<pos[segStopIndex] + probeLength - 1 && genomeIndex<=pos[segStartIndex]-1) {
		    lineLength = line.length();
		    genomePart += line;
		    genomePartLength = genomePart.length();
		    line = br.readLine();
		}
		//Read enough to be able to extract the sequence
		//for probe starting at pos[posIndex]
		if ((pos[segStartIndex] - genomeIndex - 1)<0)
		    System.out.println("Index out of bound for start " + pos[segStartIndex] + ", stop " + pos[segStopIndex] + " and " + genomeIndex);
		if (pos[segStopIndex] - genomeIndex + probeLength - 1<0)
		    System.out.println("Index out of bound for start " + pos[segStartIndex] + ", stop " + pos[segStopIndex] + " and " + genomeIndex);
		seqs.addElement(genomePart.substring((int)(pos[segStartIndex] - genomeIndex - 1), (int)(pos[segStopIndex] - genomeIndex + probeLength - 1)));
		starts.addElement(new Long(pos[segStartIndex]));
		inds.addElement(new Integer(segStartIndex));
		segStartIndex = segStopIndex + 1;
		segStopIndex = findContiguousRegion(pos, segStartIndex, probeLength);
		//System.out.println(pos[segStartIndex] + "\t" + pos[segStopIndex]);
		//Remove the parts which are no longer needed from
		//the String genomePart
		if (segStartIndex<pos.length && pos[segStartIndex] - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(pos[segStartIndex] - 1 - genomeIndex));
		    genomeIndex += pos[segStopIndex] - 1 - genomeIndex;
		    genomePartLength = genomePart.length();
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		    genomePartLength = 0;
		}
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	String[] s = new String[seqs.size()];
	for (int i=0; i<seqs.size(); i++)
	    //make sure that all letters are capital which is required by TFMScan
	    s[i] = ((String)seqs.elementAt(i)).toUpperCase();
	return s;
    }

    /*public String[] getContiguousSequences(Vector enrichedRegions){
	//Use this method to extract the non-overlapping sequences
	//corresponding to the given tiling array.
	seqs = new Vector();
	starts = new Vector();
	inds = new Vector();
	int fragStarts[] = new int[enrichedRegions.size()], fragLengths[] = new int[enrichedRegions.size()];
	for (int i=0; i<enrichedRegions.size(); i++) {
	    starts.addElement(new Integer(((EnrichedRegion)enrichedRegions.elementAt(i)).getStart()));
	    fragStarts[i] = ((EnrichedRegion)enrichedRegions.elementAt(i)).getStart();
	    fragLengths[i] = ((EnrichedRegion)enrichedRegions.elementAt(i)).getLength();
	}
	int posIndex = 0;
	long genomeIndex = 0, rem = 0, genomePartLength = 0;	
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    for (int i=0; i<fragStarts.length; i++) {
		line = br.readLine();
		int lineLength = line.length();
		while (line!=null && genomeIndex + lineLength<=fragStarts[i]-1) {
		    genomePart = line;
		    genomeIndex += genomePart.length();
		    line = br.readLine();
		}
		while (line!=null && genomeIndex + genomePartLength<fragStarts[i] + fragLengths[i] - 1 && genomeIndex<=fragStarts[i]-1) {
		    lineLength = line.length();
		    genomePart += line;
		    genomePartLength = genomePart.length();
		    line = br.readLine();
		}
		//Read enough to be able to extract the sequence
		//for probe starting at pos[posIndex]
		seqs.addElement(genomePart.substring((int)(fragStarts[i] - genomeIndex - 1), (int)(fragStarts[i] + fragLengths[i] - genomeIndex - 1)));
		//Remove the parts which are no longer needed from
		//the String genomePart
		if (i<fragStarts.length && fragStarts[i] - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(fragStarts[i] - 1 - genomeIndex));
		    genomeIndex += fragStarts[i] - 1 - genomeIndex;
		    genomePartLength = genomePart.length();
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		    genomePartLength = 0;
		}
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	String[] s = new String[seqs.size()];
	for (int i=0; i<seqs.size(); i++)
	    //make sure that all letters are capital which is required by TFMScan
	    s[i] = ((String)seqs.elementAt(i)).toUpperCase();
	return s;
	}*/

    public long[] getContiguousSequencesStart(){
	//Use this method to extract the starting positions of the
	//contiguous sequences.
	if (starts==null) {
	    System.err.println("Error, must use getContiguousSequences(long[] pos, int probeLength) first.");
	    return null;
	}
	else {
	    long[] s = new long[starts.size()];
	    for (int i=0; i<starts.size(); i++)
		s[i] = ((Long)starts.elementAt(i)).longValue();
	    return s;
	}
    }

    public int[] getContiguousSequencesStartIndexes(){
	//Use this method to extract the starting positions of the
	//contiguous sequences.
	if (inds==null) {
	    System.err.println("Error, must use getContiguousSequences(long[] pos, int probeLength) first.");
	    return null;
	}
	else {
	    int[] s = new int[inds.size()];
	    for (int i=0; i<inds.size(); i++)
		s[i] = ((Integer)inds.elementAt(i)).intValue();
	    return s;
	}
    }

    private int findContiguousRegion(long[] pos, int startInd, int probeLength){
	//Returns the index of the last element of pos which forms a
	//contiguous region with all indices starting from startInd.
	while (startInd+1<pos.length && pos[startInd] + (long)probeLength>pos[startInd+1])
	    startInd++;
	//if ((startInd+1)==pos.length) //Handle the last probe
	//return pos[startInd] + probeLength>pos[pos.length-1] ? pos.length - 1 : pos.length - 2;
	return startInd;
    }


    public void saveSequences(String fileName, long pos[], int probeLength){
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    String seqs[] = getSequences(pos, (long)probeLength, false);
	    for (int i=0; i<pos.length; i++) {
		bw.write(">" + pos[i] + "-" + (pos[i] + probeLength - 1) + "\n");
		bw.write(seqs[i] + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public void saveSequence(String fileName, long start, int length) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    bw.write(">" + start + "-" + (start + length - 1) + "\n");
	    long pos[] = new long[1];
	    pos[0] = start;
	    bw.write(getSequences(pos, (long)length, false)[0] + "\n");
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public static void saveSequence(String fileName, String header, String sequence) {
	//Use this method for saving a sequence in Fasta-format
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    bw.write(">" + header + "\n");
	    bw.write(sequence + "\n");
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public String[] getSequences(int pos[], int probeLength[]){ return getSequences(pos, probeLength, false); }

    public String[] getSequences(int pos[], int probeLength[], boolean toUpper){
	//Use this method when retrieving multiple sequences of
	//different lengths. pos and probeLength should have the same
	//length.
	String seqs[] = new String[pos.length];
	int posIndex = 0;
	int genomeIndex = 0, rem = 0, genomePartLength = 0;
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Do not use the fancy BioJava stuff. Instead, we just
	    //read the data since it is more or less just a flat
	    //file. This version assumes that there is only one
	    //sequence and header in the entire file. This will be a
	    //little bit tricky since the pos argument are based on an
	    //index starting at 1 while the strings used in Java are
	    //indexed from 0.
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null && posIndex<pos.length) {
		int lineLength = line.length();
		genomePart += line;
		genomePartLength = genomePart.length();
		while (posIndex<pos.length && genomeIndex + genomePartLength>pos[posIndex] + probeLength[posIndex] - 1 && genomeIndex<=pos[posIndex]-1) {
		    //Read enough to be able to extract the sequence
		    //for probe starting at pos[posIndex]
		    if (toUpper)
			seqs[posIndex] = genomePart.substring((int)(pos[posIndex] - genomeIndex - 1), (int)(pos[posIndex] - genomeIndex + probeLength[posIndex] - 1)).toUpperCase();
		    else
			seqs[posIndex] = genomePart.substring((int)(pos[posIndex] - genomeIndex - 1), (int)(pos[posIndex] - genomeIndex + probeLength[posIndex] - 1));
		    //Remove the parts which are no longer needed from
		    //the String genomePart
		    posIndex++;
		}
		if (posIndex<pos.length && pos[posIndex] - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(pos[posIndex] - 1 - genomeIndex));
		    genomeIndex += pos[posIndex] - 1 - genomeIndex;
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	    System.err.println(posIndex + " = posIndex, " + genomeIndex + " = genomeIndex, " + genomePartLength + " = genomePartLength");
	    System.err.println(pos[posIndex] - 1 - genomeIndex);
	    char wildcards[] = new char[probeLength[posIndex]];
	    Arrays.fill(wildcards, 'N');
	    seqs[posIndex] = new String(wildcards);
	    //return null;
	}
	return seqs;
    }

    public boolean hasMoreSequences() { 
	try {
	    if (fastaReader!=null) { return fastaReader.ready(); } 
	 }
	catch (IOException e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}   
	return true; 
    }

    public String getNextSequence(int pos, int probeLength){
	//Use this method when stepping forward through the genome
	String seq = new String();
	try {
	    if (fastaReader==null)
		fastaReader = new BufferedReader(new FileReader(fileName));
	    if (pos<posInd) 
		System.err.println("Error, already past index " + pos);
	    while (fastaReader.ready() && posInd<pos+probeLength-1) {
		char c = MerTree.removeComments(fastaReader);
		if (c!='\n') {
		    if (++posInd>=pos) { seq += c; }
		}
	    }
	    //if (!fastaReader.ready())
	    //fastaReader.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	//System.out.println("pos = " + pos + ", probeLength = " + probeLength + ", index = " + posInd + " and seq = " + seq);
	return seq;
    }

    public String getNextBase(){ return getNextBases(1); }

    public String getNextBases(int nBases){ 
	String seq = new String();
	int pos = posInd + nBases;
	try {
	    if (fastaReader==null)
		fastaReader = new BufferedReader(new FileReader(fileName));
	    while (fastaReader.ready() && posInd<pos) {
		char c = MerTree.removeComments(fastaReader);
		if (c!='\n') {
		    if (++posInd<=pos) { seq += c; }
		}
	    }
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	return seq;
    }


    public String[] getSequences(int pos[], int probeLength){
	long[] p = new long[pos.length];
	for (int i=0; i<p.length; i++)
	    p[i] = (long)pos[i];
	return getSequences(p, (long)probeLength, false);
    }

    public String[] getSequences(ABIRead reads[]) {
	long[] p = new long[reads.length];
	for (int i=0; i<p.length; i++)
	    p[i] = (long)reads[i].getPosition();
	return getSequences(p, (long)reads[0].getLength(), false);
    }

    public String[] getSequences(long p[], long probeLength, boolean toUpper){ return getSequences(p, probeLength, toUpper, false); }

    public String[] getSequences(long p[], long probeLength, boolean toUpper, boolean retry){
	//Use this method to extract all the sequences starting at pos
	//and extending probeLength basepairs. Assume that pos is
	//sorted in ascending order.
	int posIndex = 0;
	long genomeIndex = 0, rem = 0, genomePartLength = 0, pos[] = new long[p.length];
	for (int i=0; i<p.length; i++) { pos[i] = p[i]; }
	String seqs[] = new String[pos.length];
	if (pos.length>0) {
	    try {
		//Open the file
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		//Do not use the fancy BioJava stuff. Instead, we just
		//read the data since it is more or less just a flat
		//file. This version assumes that there is only one
		//sequence and header in the entire file. This will be a
		//little bit tricky since the pos argument are based on an
		//index starting at 1 while the strings used in Java are
		//indexed from 0.
		String line = br.readLine(), genomePart = "";
		//Discard the header
		while (line!=null && line.startsWith(">"))
		    line = br.readLine();
		if (pos[posIndex] - 1 - genomeIndex<0) {
		    char wildcards[] = new char[(int)probeLength];
		    Arrays.fill(wildcards, 'N');
		    seqs[posIndex] = new String(wildcards);
		    posIndex++;
		}
		while (line!=null && posIndex<pos.length) {
		    int lineLength = line.length();
		    genomePart += line;
		    genomePartLength = genomePart.length();
		    while (posIndex<pos.length && genomeIndex + genomePartLength>pos[posIndex] + probeLength - 1 && genomeIndex<=pos[posIndex]-1) {
			//Read enough to be able to extract the sequence
			//for probe starting at pos[posIndex]
			seqs[posIndex] = genomePart.substring((int)(pos[posIndex] - genomeIndex - 1), (int)(pos[posIndex] - genomeIndex + probeLength - 1));
			//Remove the parts which are no longer needed from
			//the String genomePart
			if (seqs[posIndex]==null)
			    System.out.println(genomePart);
			posIndex++;
		    }
		    if (posIndex<pos.length && pos[posIndex] - 1<genomeIndex + genomePartLength) {
			genomePart = genomePart.substring((int)(Math.max(0, pos[posIndex] - 1 - genomeIndex)));
			genomeIndex += Math.max(0, pos[posIndex] - 1 - genomeIndex);
		    }
		    else {
			genomeIndex += genomePartLength;
			genomePart = new String("");
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (Exception e) {
		System.err.println("Error reading file " + fileName + ": " + e.toString());
		System.err.println(posIndex + " = posIndex, " + genomeIndex + " = genomeIndex, " + genomePartLength + " = genomePartLength, pos[posIndex] = " + pos[posIndex]);
		System.err.println(pos[posIndex] - 1 - genomeIndex);
		/*if (!retry) {//Try to recover, by getting the remaining stuff from a separate process
		  FastaSequence2 fastaSeq = new FastaSequence2(fileName);
		  long posLeft[] = new long[pos.length - posIndex];
		  for (int i=0; i<posLeft.length; i++) {
		  posLeft[i] = p[posIndex + i];
		  //System.out.println(p[posIndex + i]);
		  }
		  String seqsLeft[] = fastaSeq.getSequences(posLeft, probeLength, toUpper, true);
		for (int i=0; i<seqsLeft.length; i++)
		seqs[posIndex + i] = seqsLeft[i];
		}
		else {*/
		char wildcards[] = new char[(int)probeLength];
		Arrays.fill(wildcards, 'N');
		seqs[posIndex] = new String(wildcards);
		//}
	    }
	    if (toUpper) {
		for (int i=0; i<seqs.length; i++)
		    seqs[i] = seqs[i].toUpperCase();
	    }
	}
	return seqs;
    }

    public String getSequence(int pos, int probeLength){ return getSequences(new int[]{pos}, probeLength)[0]; }

    public String getSequence(int pos, int probeLength, boolean reverseComplement){ return reverseComplement ? reverse(complement(getSequences(new int[]{pos}, probeLength)[0])) : getSequence(pos, probeLength); }

    /*public String getSequence(int pos, int probeLength){
	//Use this method to extract all the sequences starting at pos
	//and extending probeLength basepairs. Assume that pos is
	//sorted in ascending order.
	int posIndex = 0;
	int genomeIndex = 0, rem = 0, genomePartLength = 0;
	String seq = "";

	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Do not use the fancy BioJava stuff. Instead, we just
	    //read the data since it is more or less just a flat
	    //file. This version assumes that there is only one
	    //sequence and header in the entire file. This will be a
	    //little bit tricky since the pos argument are based on an
	    //index starting at 1 while the strings used in Java are
	    //indexed from 0.
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null) {
		int lineLength = line.length();
		genomePart += line;
		genomePartLength = genomePart.length();
		while (genomeIndex + genomePartLength>pos + probeLength - 1 && genomeIndex<=pos-1) {
		    //Read enough to be able to extract the sequence
		    //for probe starting at pos[posIndex]
		    seq = genomePart.substring((int)(pos - genomeIndex - 1), (int)(pos - genomeIndex + probeLength - 1));
		}
		if (seq.length()>0)
		    break;
		if (pos - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(pos - 1 - genomeIndex));
		    genomeIndex += pos - 1 - genomeIndex;
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	return seq;
	}*/

    public int[][] getMATCoefficients(long pos[], int probeLength){
	String seqs[] = getSequences(pos, (long)probeLength, true);
	int[][] coeffs = new int[pos.length][probeLength*3 + 5];
	for (int i=0; i<pos.length; i++){
	    int nA = 0, nC = 0, nG = 0, nT = 0;
	    for (int j=0; j<probeLength; j++) {
		if (seqs[i].substring(j, j+1).equals("A")) {
		    coeffs[i][j] = 1;
		    nA++;
		}
		else if (seqs[i].substring(j, j+1).equals("C")) {
		    coeffs[i][j+probeLength] = 1;
		    nC++;
		}
		else if (seqs[i].substring(j, j+1).equals("G")) {
		    coeffs[i][j+probeLength*2] = 1;
		    nG++;
		}
		else if (seqs[i].substring(j, j+1).equals("T")) 
		    nT++;
	    }
	    coeffs[i][probeLength*3] = nT;
	    coeffs[i][probeLength*3+1] = nA*nA;
	    coeffs[i][probeLength*3+2] = nC*nC;
	    coeffs[i][probeLength*3+3] = nG*nG;
	    coeffs[i][probeLength*3+4] = nT*nT;
	}
	return coeffs;
    }

    public int[] scoreMotif(int[][] pfm, long start, int length) {
	return scoreMotif(pfm, start, length, .7, true);
    }

    public int[] scoreMotif(int[][] pfm, long start, int length, double threshold) {
	return scoreMotif(pfm, start, length, threshold, true);
    }

    public int[] scoreMotif(int[][] pfm, long start, int length, double threshold, boolean single) {
	//Use this method to scan for a motif between start and start
	//+ length given a representation pfm. Return an array with
	//the indices for the motif of interest. These indices are
	//with respect to the true coordinates.
	return null;
    }
    
    public static String complement(String s) {
	s = s.toUpperCase();
	String c = "";
	for (int i=0; i<s.length(); i++) {
	    if (s.charAt(i)=='A')
		c += "T";
	    else if (s.charAt(i)=='C')
		c += "G";
	    else if (s.charAt(i)=='G')
		c += "C";
	    else if (s.charAt(i)=='T')
		c += "A";
	    else if (s.charAt(i)=='N')
		c += "N";
	}
	return c;
    }

    public static String reverse(String s) {
	String r = "";
	char c[] = s.toCharArray(), cr[] = new char[s.length()];
	for (int i=c.length-1; i>=0; i--) 
	    cr[c.length-1-i] = c[i];
	return new String(cr);
    }

    public static String getColorCode(String s) {
	String c = "";
	for (int i=1; i<s.length(); i++) {
	    if (s.charAt(i-1)==s.charAt(i))
		c += "0";
	    else if (s.charAt(i-1)=='A' && s.charAt(i)=='C')
		c += "1";
	    else if (s.charAt(i-1)=='A' && s.charAt(i)=='G')
		c += "2";
	    else if (s.charAt(i-1)=='A' && s.charAt(i)=='T')
		c += "3";
	    else if (s.charAt(i-1)=='C' && s.charAt(i)=='A')
		c += "1";
	    else if (s.charAt(i-1)=='C' && s.charAt(i)=='G')
		c += "3";
	    else if (s.charAt(i-1)=='C' && s.charAt(i)=='T')
		c += "2";
	    else if (s.charAt(i-1)=='G' && s.charAt(i)=='A')
		c += "2";
	    else if (s.charAt(i-1)=='G' && s.charAt(i)=='C')
		c += "3";
	    else if (s.charAt(i-1)=='G' && s.charAt(i)=='T')
		c += "1";
	    else if (s.charAt(i-1)=='T' && s.charAt(i)=='A')
		c += "3";
	    else if (s.charAt(i-1)=='T' && s.charAt(i)=='C')
		c += "2";
	    else if (s.charAt(i-1)=='T' && s.charAt(i)=='G')
		c += "1";
	    else
		c += "N";
	}
	return c;
    }

    public static String getPseudoColorCode(String s) {
	String c = getColorCode(s), p = "";
	for (int i=0; i<c.length(); i++) {
	    if (c.charAt(i)=='0')
		p += "A";
	    else if (c.charAt(i)=='1')
		p += "C";
	    else if (c.charAt(i)=='2')
		p += "G";
	    else if (c.charAt(i)=='3')
		p += "T";
	    else
		p += "N";
	}
	return p;
    }

    public static String convert2AAs(String seq, int frame) {
	String aas = "";
	for (int i=frame % 3; i<seq.length()-3; i+=3) 
	    aas += getAA(seq.substring(i, i+3));
	return aas;
    }

    public static char getAA(String codon) {
	if (codon.length()<3)
	    System.err.println("Error, codon is too short!");
	else if (codon.length()>3)
	    System.err.println("Error, codon is too long!");
	else {
	    if (codon.toUpperCase().charAt(0)=='A') {
		if (codon.toUpperCase().charAt(1)=='A') 
		    return codon.toUpperCase().charAt(2)=='A' || codon.toUpperCase().charAt(2)=='G' ? 'K' : 'N';
		else if (codon.toUpperCase().charAt(1)=='C') 
		    return 'T';
		else if (codon.toUpperCase().charAt(1)=='G') 
		    return codon.toUpperCase().charAt(2)=='A' || codon.toUpperCase().charAt(2)=='G' ? 'R' : 'S';
		else if (codon.toUpperCase().charAt(1)=='T') 
		    return codon.toUpperCase().charAt(2)=='G' ? 'M' : 'I';
	    }
	    else if (codon.toUpperCase().charAt(0)=='C') {
		if (codon.toUpperCase().charAt(1)=='A') 
		    return codon.toUpperCase().charAt(2)=='A' || codon.toUpperCase().charAt(2)=='G' ? 'Q' : 'H';
		else if (codon.toUpperCase().charAt(1)=='C') 
		    return 'P';
		else if (codon.toUpperCase().charAt(1)=='G') 
		    return 'R';
		else if (codon.toUpperCase().charAt(1)=='T') 
		    return 'L';
	    }
	    else if (codon.toUpperCase().charAt(0)=='G') {
		if (codon.toUpperCase().charAt(1)=='A') 
		    return codon.toUpperCase().charAt(2)=='A' || codon.toUpperCase().charAt(2)=='G' ? 'E' : 'D';
		else if (codon.toUpperCase().charAt(1)=='C') 
		    return 'A';
		else if (codon.toUpperCase().charAt(1)=='G') 
		    return 'G';
		else if (codon.toUpperCase().charAt(1)=='T') 
		    return 'V';
	    }
	    else if (codon.toUpperCase().charAt(0)=='T') {
		if (codon.toUpperCase().charAt(1)=='A') 
		    return codon.toUpperCase().charAt(2)=='A' || codon.toUpperCase().charAt(2)=='G' ? '*' : 'Y'; //* is STOP codon
		else if (codon.toUpperCase().charAt(1)=='C') 
		    return 'S';
		else if (codon.toUpperCase().charAt(1)=='G') 
		    return codon.toUpperCase().charAt(2)=='C' || codon.toUpperCase().charAt(2)=='T' ? 'C' : codon.toUpperCase().charAt(2)=='A' ? '*' : 'W'; //* is STOP codon
		else if (codon.toUpperCase().charAt(1)=='T') 
		    return codon.toUpperCase().charAt(2)=='A' || codon.toUpperCase().charAt(2)=='G' ? 'L' : 'F'; //* is STOP codon
	    }
	}
	return 'X';
    }		
}
