import java.util.*;

public class Filter {

    private String[] filters;

    //////////////////////////////////////////////////////////////////

    public Filter(String s[]) { filters = s; Arrays.sort(filters); }

    public Filter(String s) { filters = new String[]{s}; Arrays.sort(filters); }

    public Filter(Filter f) {filters = f.filters; Arrays.sort(filters); }

    public Filter(Filter f, String s) {filters = f.filters; add(s); }

    //////////////////////////////////////////////////////////////////

    public void add(String s) {
	String tmp[] = filters;
	filters = new String[tmp.length + 1];
	filters[0] = s;
	for (int i=0; i<tmp.length; i++)
	    filters[i+1] = tmp[i];
	Arrays.sort(filters);
    }

    public boolean distanceOK() {
	return Arrays.binarySearch(filters, FindEnhancerSites.DISTAL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.PROXIMAL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.INTRONIC_FILTER_STR, null)<0; 
    }

    public boolean rnaOK() {
	return Arrays.binarySearch(filters, FindEnhancerSites.RNA_BI_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.RNA_SS_FILTER_STR, null)<0;
    }

    public boolean cbpOK() {
	return Arrays.binarySearch(filters, FindEnhancerSites.CBP_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.CBP_KCL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.CBP_KCL_B2_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.REPRODUCIBLE_CBP_KCL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.CBP_UN_FILTER_STR, null)<0;
    }

    public boolean pol2OK() {
	return Arrays.binarySearch(filters, FindEnhancerSites.POL2_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.POL2_KCL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.POL2_UN_FILTER_STR, null)<0;
    }

    public boolean me1OK() {
	return Arrays.binarySearch(filters, FindEnhancerSites.H3K4ME1_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.H3K4ME1_KCL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.H3K4ME1_UN_FILTER_STR, null)<0;
    }
    
    public boolean me3OK() {
	return Arrays.binarySearch(filters, FindEnhancerSites.H3K4ME3_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.H3K4ME3_KCL_FILTER_STR, null)<0 && Arrays.binarySearch(filters, FindEnhancerSites.H3K4ME3_UN_FILTER_STR, null)<0;
    }

    public boolean containsKClFilter() {
	for (int i=0; i<filters.length; i++) 
	    if (filters[i].contains("KCl")) { return true; }
	return false;
    }

    public boolean containsUnFilter() {
	for (int i=0; i<filters.length; i++) 
	    if (filters[i].contains("un")) { return true; }
	return false;
    }

    public static boolean stimulusFilter(String f) {
	return !(f.equals(FindEnhancerSites.CONSERVATION_FILTER_STR) || f.equals(FindEnhancerSites.RNA_SS_FILTER_STR) || f.equals(FindEnhancerSites.RNA_BI_FILTER_STR) || f.equals(FindEnhancerSites.DISTAL_FILTER_STR) || f.equals(FindEnhancerSites.INTRONIC_FILTER_STR) || f.equals(FindEnhancerSites.PROXIMAL_FILTER_STR));
    }

    public boolean filterOK(String f) {
	if (contains(f))
	    return false;
	if (stimulusFilter(f)) {
	    if (containsKClFilter() && !f.contains("KCl"))
		return false;
	    if (containsUnFilter() && !f.contains("un"))
		return false;
	    if (!(containsUnFilter() || containsKClFilter()) && (f.contains("un") || f.contains("KCl")))
		return false;
	}
	if (distanceOK() || (!distanceOK() && !(f.equals(FindEnhancerSites.DISTAL_FILTER_STR) || f.equals(FindEnhancerSites.PROXIMAL_FILTER_STR) || f.equals(FindEnhancerSites.INTRONIC_FILTER_STR)))) {
	    if (rnaOK() || (!rnaOK() && !(f.equals(FindEnhancerSites.RNA_BI_FILTER_STR) || f.equals(FindEnhancerSites.RNA_SS_FILTER_STR)))) {
		if (cbpOK() || (!cbpOK() && !(f.equals(FindEnhancerSites.CBP_FILTER_STR) || f.equals(FindEnhancerSites.CBP_KCL_FILTER_STR) || f.equals(FindEnhancerSites.CBP_KCL_B2_FILTER_STR) || f.equals(FindEnhancerSites.REPRODUCIBLE_CBP_KCL_FILTER_STR) || f.equals(FindEnhancerSites.CBP_UN_FILTER_STR)))) {
		    if (pol2OK() || (!pol2OK() && !(f.equals(FindEnhancerSites.POL2_FILTER_STR) || f.equals(FindEnhancerSites.POL2_KCL_FILTER_STR) || f.equals(FindEnhancerSites.POL2_UN_FILTER_STR)))) {
			if (me1OK() || (!me1OK() && !(f.equals(FindEnhancerSites.H3K4ME1_FILTER_STR) || f.equals(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) || f.equals(FindEnhancerSites.H3K4ME1_UN_FILTER_STR)))) {
			    if (me3OK() || (!me3OK() && !(f.equals(FindEnhancerSites.H3K4ME3_FILTER_STR) || f.equals(FindEnhancerSites.H3K4ME3_KCL_FILTER_STR) || f.equals(FindEnhancerSites.H3K4ME3_UN_FILTER_STR))))
				return true;
			}
		    }
		}
	    }
	}
	return false;
    }
	

    public boolean hasRequired() {
	return contains(FindEnhancerSites.CBP_KCL_FILTER_STR) || contains(FindEnhancerSites.CBP_KCL_B2_FILTER_STR) || contains(FindEnhancerSites.REPRODUCIBLE_CBP_KCL_FILTER_STR) || contains(FindEnhancerSites.CBP_UN_FILTER_STR) || contains(FindEnhancerSites.CBP_FILTER_STR);// || contains(FindEnhancerSites.H3K4ME1_FILTER_STR);
    }

    public boolean hasRequired(boolean kcl) {
	if (kcl)
	    return contains(FindEnhancerSites.CBP_KCL_FILTER_STR) || contains(FindEnhancerSites.CBP_KCL_B2_FILTER_STR) || contains(FindEnhancerSites.REPRODUCIBLE_CBP_KCL_FILTER_STR);
	return contains(FindEnhancerSites.CBP_UN_FILTER_STR);
    }

    public boolean highQuality() {
	return contains(FindEnhancerSites.POL2_FILTER_STR) && contains(FindEnhancerSites.CBP_FILTER_STR) && contains(FindEnhancerSites.H3K4ME1_FILTER_STR) && contains(FindEnhancerSites.RNA_BI_FILTER_STR) && !distanceOK();
    }

    public boolean highQuality(boolean kcl) {
	if (kcl)
	    return contains(FindEnhancerSites.POL2_KCL_FILTER_STR) && (contains(FindEnhancerSites.CBP_KCL_FILTER_STR) && contains(FindEnhancerSites.CBP_KCL_B2_FILTER_STR) && contains(FindEnhancerSites.REPRODUCIBLE_CBP_KCL_FILTER_STR)) && contains(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) && contains(FindEnhancerSites.RNA_BI_FILTER_STR) && !distanceOK();
	return contains(FindEnhancerSites.POL2_UN_FILTER_STR) && contains(FindEnhancerSites.CBP_UN_FILTER_STR) && contains(FindEnhancerSites.H3K4ME1_UN_FILTER_STR) && contains(FindEnhancerSites.RNA_BI_FILTER_STR) && !distanceOK();
    }

    public boolean mediumQuality() {
	return hasRequired() && contains(FindEnhancerSites.H3K4ME1_FILTER_STR) && (contains(FindEnhancerSites.RNA_BI_FILTER_STR) || contains(FindEnhancerSites.RNA_SS_FILTER_STR)) && !distanceOK();
    }

    public boolean nullFilter() {
	return contains(FindEnhancerSites.H3K4ME1_FILTER_STR) && contains(FindEnhancerSites.CBP_FILTER_STR) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && !distanceOK();
    }

    public boolean nullFilter(boolean kcl) {
	if (kcl)
	    return contains(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) && (contains(FindEnhancerSites.CBP_KCL_FILTER_STR) || contains(FindEnhancerSites.CBP_KCL_B2_FILTER_STR)) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && distanceOK();
	return contains(FindEnhancerSites.H3K4ME1_UN_FILTER_STR) && contains(FindEnhancerSites.CBP_UN_FILTER_STR) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && distanceOK();
    }

    public boolean nullFilter(boolean kcl, boolean cons) {
	if (cons)
	    return nullFilter(kcl);
	if (kcl)
	    return contains(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) && (contains(FindEnhancerSites.CBP_KCL_FILTER_STR) || contains(FindEnhancerSites.CBP_KCL_B2_FILTER_STR)) && !contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && distanceOK();
	return contains(FindEnhancerSites.H3K4ME1_UN_FILTER_STR) && contains(FindEnhancerSites.CBP_UN_FILTER_STR) && !contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && distanceOK();
    }

    public boolean nullFilterRep(boolean kcl) {
	if (kcl)
	    return contains(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) && contains(FindEnhancerSites.REPRODUCIBLE_CBP_KCL_FILTER_STR) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && !distanceOK();
	return contains(FindEnhancerSites.H3K4ME1_UN_FILTER_STR) && contains(FindEnhancerSites.CBP_UN_FILTER_STR) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && !distanceOK();
    }

    public boolean p300Filter(boolean kcl) {
	if (kcl)
	    return contains(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) && (contains(FindEnhancerSites.P300_KCL_B1_FILTER_STR) || contains(FindEnhancerSites.P300_KCL_B2_FILTER_STR)) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && distanceOK();
	return contains(FindEnhancerSites.H3K4ME1_UN_FILTER_STR) && (contains(FindEnhancerSites.P300_UN_B1_FILTER_STR) || contains(FindEnhancerSites.P300_UN_B2_FILTER_STR)) && contains(FindEnhancerSites.CONSERVATION_FILTER_STR) && distanceOK();
    }

    public boolean p300Filter(boolean kcl, boolean cons) {
	if (cons)
	    return p300Filter(kcl);
	if (kcl)
	    return contains(FindEnhancerSites.H3K4ME1_KCL_FILTER_STR) && (contains(FindEnhancerSites.P300_KCL_B1_FILTER_STR) || contains(FindEnhancerSites.P300_KCL_B2_FILTER_STR)) && distanceOK();
	return contains(FindEnhancerSites.H3K4ME1_UN_FILTER_STR) && (contains(FindEnhancerSites.P300_UN_B1_FILTER_STR) || contains(FindEnhancerSites.P300_UN_B2_FILTER_STR)) && distanceOK();
    }

    ////////////////////////////////////////////

    public boolean contains(String f) { return Arrays.binarySearch(filters, f, null)>=0; }

    public int size() { return filters.length; }

    public String get(int i) { return filters[i]; }

    public String saveString(Vector<String> allFilters) { 
	String str = "";
	for (String f : allFilters)
	    str += (contains(f) ? "Y" : "N") + "\t";
	return str;
    }

    public int numerOfFilters() { return filters.length; }

    ///////////////////////////////////////////////////////

    public int hashCode() { return Arrays.hashCode(filters); }

    public boolean equals(Object o) { return Arrays.equals(filters, ((Filter)o).filters); }

    public String toString() { return Arrays.toString(filters); }

}
