//Use this class to find which windows are significant for a given
//pair of thresholds (one for #ChIP-#input and one restricting the
//maximum #input). This method utilizes the output from
//CalculateCriticalWindowDifference.java and takes the histograms of
//the distributions as well as the list of windows
//generated. Unfortunately, the analysis cannot be done in Matlab
//directly since we run out of memory because of Matlab's crappy
//method for handling it.
//
//Example:
//
//>java -Xmx512m FindCodingRegionsPol2Enrichment -n Pol2/4H8_KCl/4H8_KCl -i Mef2/inputKCl/inputKCl -u -e -p .01 -c 19
//



import jargs.gnu.CmdLineParser;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
//import umontreal.iro.lecuyer.stat.Tally;
import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;
import org.apache.commons.math.distribution.BinomialDistributionImpl;


public class FindCodingRegionsPol2Enrichment {

    public static DecimalFormat format = new DecimalFormat("###.########");

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .fa-files
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    CmdLineParser.Option qualityScoreFlag = argParser.addDoubleOption('q', "qualityScore");
	    //Decide what file to write the results to
	    CmdLineParser.Option saveFileFlag = argParser.addStringOption('f', "saveFile");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('p', "fdr");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    String GENOME_LOC = "./";
	    int nChr = 21;
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileFlag);
	    String saveFileStart = tmp!=null && !tmp.toString().equals("") ? tmp.toString() : fileName;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    boolean extend = argParser.getOptionValue(extendFlag)!=null;
	    tmp = argParser.getOptionValue(qualityScoreFlag);
	    double qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : 10;
	    boolean unique = argParser.getOptionValue(uniqueFlag)!=null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    int nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    //////////////////////////////////////////////////////////////////
	    Vector inputDensities = new Vector(), readsDensities = new Vector(), relativeEnrichments = new Vector();
	    TreeMap binTestPVals = new TreeMap();
	    MRNAInfo mRNAInfo = new MRNAInfo(chr, orchestra);
	    //Parse all the reads for ChIP and input
	    ABIRead reads[] = PeakFinder2.parseFiles(chr, fileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, null); //getSaveFileName("read_stats", chr, "txt"));
	    ABIRead inputReads[] = PeakFinder2.parseFiles(chr, inputFileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, null); //getSaveFileName("input_stats", c, "txt"));
	    //The proportion of ChIP reads
	    double chipProp = (double)reads.length/(reads.length + inputReads.length);
	    System.out.println(chipProp);
	    Iterator it = mRNAInfo.getGenes().iterator();
	    while (it.hasNext()) {
		MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
		int nCodingReads = 0, nCodingInputReads = 0, readInd = 0, inputInd = 0;
		//Find the ones within coding regions
		while (readInd<reads.length && reads[readInd].getPosition()<gene.getEnd()) 
		    nCodingReads += reads[readInd++].getPosition()>=gene.getStart() ? 1 : 0;
		while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<gene.getEnd()) 
		    nCodingInputReads += inputReads[inputInd++].getPosition()>=gene.getStart() ? 1 : 0;
		if (nCodingInputReads>0 && nCodingReads>0) {
		    //For each gene, we compare using a binomial test
		    BinomialDistributionImpl binDist = new BinomialDistributionImpl(nCodingInputReads+nCodingReads, chipProp);
		    if ((1 - binDist.cumulativeProbability(nCodingReads))<.01)
			System.out.println(gene.toString() + "\t" + nCodingReads + "\t" + nCodingInputReads + "\t" + (((double)nCodingReads/nCodingInputReads)/chipProp) + "\t" + (1 - binDist.cumulativeProbability(nCodingReads)));
		    binTestPVals.put(gene, 1 - binDist.cumulativeProbability(nCodingReads));
		    //We also calculate the denisities
		    readsDensities.addElement(new Double((double)nCodingReads/gene.getLength()));
		    inputDensities.addElement(new Double((double)nCodingInputReads/gene.getLength()));
		    relativeEnrichments.addElement(new Double(((double)nCodingReads/nCodingInputReads)/chipProp));
		}
	    }
	    System.out.print("rel_en = [");
	    for (int i=0; i<relativeEnrichments.size(); i++)
		System.out.print(format.format(((Double)relativeEnrichments.elementAt(i)).doubleValue()) + "\t");
	    System.out.print("]\n");
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }
}
