//Use this method to characterize how many of the islands of high
//degree of conservation (as determined by PhastCons) there are.
//
//bsub -q shared_2h java -Xmx4096m -server FindConservationIslands -o -c 19 -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/merged_total/merged_total
//
//bsub -q shared_2h java -Xmx4096m -server FindConservationIslands -o -c 22 --species hs
//
//for ((i=1; i<=24; i++)); do bsub -q shared_12h java -Xmx4096m -server FindConservationIslands -o -c $i --species hs --hg19; done
//
//for ((i=1; i<=21; i++)); do bsub -q all_2h -o /dev/null java -Xmx4096m -server FindConservationIslands -o -c $i; done

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class FindConservationIslands {

    public final static int CODING_PROMOTER_CI_TYPE = 0;
    public final static int NON_CODING_PROMOTER_CI_TYPE = 1;
    public final static int CODING_EXON_CI_TYPE = 2;
    public final static int NON_CODING_EXON_CI_TYPE = 3;
    public final static int ENHANCER_CI_TYPE = 4;
    public final static int INSULATOR_CI_TYPE = 5;
    public final static int PEAK_CI_TYPE = 6;
    public final static int K4ME3_PEAK_CI_TYPE = 7;
    public final static int INTRONIC_CI_TYPE = 8;
    public final static int EXTRAGENIC_CI_TYPE = 9;
    public final static int NOVEL_PROMOTER_CI_TYPE = 10;
    public final static int MAR_CI_TYPE = 11;

    public final static int NUMBER_OF_CI_TYPES = 12;

    public final static int MAX_PROMOTER_DIST = 2000;
    public final static int MAX_ENHANCER_DIST = 100;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    //Ugly, but lazy... Keep track of the length distributions for various categories
    public static TreeMap<Integer, Integer> promoterLengths = new TreeMap<Integer, Integer>(), ncPromoterLengths = new TreeMap<Integer, Integer>(), exonLengths = new TreeMap<Integer, Integer>(), ncExonLengths = new TreeMap<Integer, Integer>(), otherLenghts = new TreeMap<Integer, Integer>(), enhancerLengths = new TreeMap<Integer, Integer>(), peakLengths = new TreeMap<Integer, Integer>(), ctcfLengths = new TreeMap<Integer, Integer>(), intronicLengths = new TreeMap<Integer, Integer>(), marLengths = new TreeMap<Integer, Integer>();


    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option hg19Flag = argParser.addBooleanOption("hg19");
	    CmdLineParser.Option excludeExonsFlag = argParser.addBooleanOption('e', "excludeExons");
	    CmdLineParser.Option excludePromotersFlag = argParser.addBooleanOption('p', "excludePromoters");
	    CmdLineParser.Option promoterDistFlag = argParser.addIntegerOption('P', "PromoterDistance");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");	    
	    CmdLineParser.Option minDistFlag = argParser.addIntegerOption('d', "minDist");	    
	    CmdLineParser.Option maxERNADistFlag = argParser.addIntegerOption("maxERNADist");	    
	    CmdLineParser.Option seqFlag = argParser.addBooleanOption('S', "sequence");
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option estsFlag = argParser.addBooleanOption("EST");
	    CmdLineParser.Option upstreamFlag = argParser.addIntegerOption('U', "upstream");
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('D', "downstream");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option reverseFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option randomizeFlag = argParser.addBooleanOption('R', "random");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean excludeExons = argParser.getOptionValue(excludeExonsFlag)!=null;
	    boolean excludePromoters = argParser.getOptionValue(excludePromotersFlag)!=null;
	    boolean ests = argParser.getOptionValue(estsFlag)!=null;
	    boolean sequence = argParser.getOptionValue(seqFlag)!=null;
	    boolean forward = argParser.getOptionValue(reverseFlag)==null;
	    boolean randomize = argParser.getOptionValue(randomizeFlag)!=null;
	    boolean hg19 = argParser.getOptionValue(hg19Flag)!=null;
	    tmp = argParser.getOptionValue(promoterDistFlag);
	    int promoterDist = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/hela_tf_names.txt" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/tf_kcl_names";
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(minDistFlag);
	    int minIslandDist = tmp!=null ? ((Integer)tmp).intValue() : 50;
	    tmp = argParser.getOptionValue(maxERNADistFlag);
	    int maxERNADist = tmp!=null ? ((Integer)tmp).intValue() : 1500;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : (hg19 ? "Hs/v37" : "Hs/v36")) + "/conservation_islands_" + (species.equals("mm") ? "" : "hs_") + (excludeExons ? "exons_" : "") + (excludePromoters ? "promoters_" : "") + "w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".bed";
	    String saveFileNameSeq = saveFileName.replace(".bed", ".fa");
	    String saveFileNameList = saveFileName.replace(".bed", ".txt");
	    String saveFileNameLengths = saveFileName.replace(".bed", "_lengths.txt");
	    String savePromoterCIs = tmp!=null ? (String)tmp : "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : (hg19 ? "Hs/v37" : "Hs/v36")) + "/conservation_islands_" + (species.equals("mm") ? "" : "hs_") + "promoters_" + "w_" + windowSize + (chr==-1 ? "" : "_" + chr) + ".txt";
	    String saveExonCIs = savePromoterCIs.replace("promoters", "exons");
	    String saveLongCIs = savePromoterCIs.replace("promoters", "ncRNAs");
	    //For saving the lengths of the conserved islands associated with each enhancer
	    String saveERNAConservationLengthsName = saveFileName.replace(".bed", "_eRNA_" + maxERNADist + "_lengths.txt");
	    ///////////////////////////////////////////////////////
	    int nChr = FindTranscripts3.numberOfChromosomes(species), nWindows = 0, maxEnhancerDist = 100, maxPromoterDist = 2000, nBpsBridged = 0, nBpsCovered = 0, readLength = 35;
	    boolean repeatMask = true;
	    String rnaFileName = fileName!=null ? fileName + ".tsv" : null;
	    double normFactor = fileName!=null ? 1e6/FindReadsAcrossFeature2.countRNAReads(fileName, nChr) : -1;
	    String repeatFileName = species.toLowerCase().equals("mm") ? EstimateGenicTranscription3.REPEAT_RNA_FILE_MM : EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_36;
	    String exonFileName = "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : (hg19 ? "Hs/v37" : "Hs/v36")) + "/all_EXNs.tsv";
	    RefSeqUCSCEnsemblGenes geneInfo[] = new RefSeqUCSCEnsemblGenes[nChr];

	    String extragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lRNA_lMe3_real.tsv";
	    TreeSet<Enhancer2> extragenicEnhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
	    String intragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_intragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_in_lMe3_real.tsv";
	    TreeSet<Enhancer2> intragenicEnhancers = FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName);
	    TreeSet<Enhancer2> allEnhancers = new TreeSet<Enhancer2>();
	    allEnhancers.addAll(extragenicEnhancers);
	    allEnhancers.addAll(intragenicEnhancers);

	    String macroRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    String lncRNAFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    String lncRNAExonFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/exonsNPC.gff";

	    //Consider the novel transcripts

	    String estFileName = !species.equals("mm") ? "/groups/neuroduo/jesse/human_annot_for_mh/human_ESTs/" : "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/";
	    LinkedHashMap<String, EST> estTagsMapForward = null, estTagsMapReverse = null;

	    Vector<Peak> peaks = new Vector<Peak>(), ctcfPeaks = new Vector<Peak>();
	    String tfNamesInStudy[] = null;
	    if (species.equals("mm")) {
		if (peakFileName.contains("tf_kcl_names")) {
		    peaks = new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values()); 
		    tfNamesInStudy = new String[]{"CBP_Millipore_KCl_B1","CREB_SC_KCl_B1","Npas4_Npas4_KCl_B1","SRF_H300_KCl_B1"};
		}
	    }
	    else {
		peaks = new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(peakFileName, "Helas3", chr));
		//Also get CTCF peaks if available
		String ctcfPeakFileName = "/home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3Ctcf.narrowPeak";
		ctcfPeaks = new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(ctcfPeakFileName, "Helas3", chr));
		System.out.println("Found " + ctcfPeaks.size() + " CTCF peaks.");
	    }
	    Collections.sort(peaks);
	    System.out.println("Found " + peaks.size() + " peaks.");

	    String k4me3PeakFileName = !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/H3K4Me3_KCl_peaks_names", cellType = "Helas3";
	    Vector<Peak> k4me3peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileCellType(k4me3PeakFileName, cellType, chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(k4me3PeakFileName, chr).values());

	    ////////////////////////////////////////////
	    int nBins = FindReadsAcrossFeature.numberOfBins(upstream, downstream, 1);
	    LinkedHashMap<String, FeatureInfo> features = new LinkedHashMap<String, FeatureInfo>();
	    features.put("TSS", new FeatureInfo("TSS", nChr, nBins));
	    features.put("ncTSS", new FeatureInfo("ncTSS", nChr, nBins));
	    features.put("novel", new FeatureInfo("novel", nChr, nBins));
	    features.put("extra_enhancer", new FeatureInfo("extra_enhancer", nChr, nBins));
	    features.put("intra_enhancer", new FeatureInfo("intra_enhancer", nChr, nBins));
	    features.put("RFBS", new FeatureInfo("RFBS", nChr, nBins));
	    features.put("insulator", new FeatureInfo("insulator", nChr, nBins));
	    features.put("k4me3", new FeatureInfo("k4me3", nChr, nBins));
	    features.put("MAR", new FeatureInfo("MAR", nChr, nBins));

	    //Calculate the average number of CIs as a function of the length and the number of exons of the transcript. We use this to estimate the number of conserved islands accounted for by the unannotated transcripts
	    TreeMap<Integer, Integer> ncExons2CIs = new TreeMap<Integer, Integer>(), ncLength2Exons = new TreeMap<Integer, Integer>();
	    TreeMap<Enhancer2, Integer> enhancerConservedBPs = new TreeMap<Enhancer2, Integer>();

	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    BufferedWriter bwSeq = sequence ? new BufferedWriter(new FileWriter(saveFileNameSeq)) : null;
	    System.out.println("Saving results to " + saveFileName);
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    ABIRead[] rnaReads = rnaFileName!=null ? ABIReads.loadMRNAReadsChr(rnaFileName, forward, i, readLength, false, repeatFileName) : null;
		    //Load the annotation
		    geneInfo[i-1] = new RefSeqUCSCEnsemblGenes(i, orchestra, species);
		    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species, i);
		    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species, i);

		    //Get all the exons as well
		    TreeSet<Exon> exons = parseExons(exonFileName, species, i);
		    TreeSet<Exon> macroRNAExons = parseMacroRNAExons(macroRNAFileName, i);
		    TreeSet<Exon> lincRNAExons = parseLincRNAExons(lncRNAExonFileName, i, lncRNAs);
		    if (ests) { //Get ESTs
			estTagsMapForward = CompareCageEstRNASeqStarts.parseEstTags(estFileName, i, true, species);
			estTagsMapReverse = CompareCageEstRNASeqStarts.parseEstTags(estFileName, i, false, species);
		    }
		    //Load the MARs
		    TreeSet<MAR> mars = loadMARs(i, species, repeatMask, randomize);

		    PhastConsParser2 phastCons = new PhastConsParser2(i, true, orchestra, species, hg19);
		    FastaSequence2 sequenceParser = sequence ? new FastaSequence2(species, hg19, i, orchestra) : null;
		    ConservationIslands consIslands = new ConservationIslands(i, species);

		    //Go through the genome one bin at a time, exclude the ones that are near promoters or overlapping exons
		    int windowStart = 0, islandStart = -1;
		    double meanConsIsland = 0;
		    Iterator<Exon> exonIt = exons.iterator();
		    Exon exon = exonIt.hasNext() ? exonIt.next() : null;
		    Iterator<RefSeqUCSCEnsemblGene> geneIt = geneInfo[i-1]==null ? (new TreeSet<RefSeqUCSCEnsemblGene>()).iterator() : geneInfo[i-1].getGenes().iterator();
		    RefSeqUCSCEnsemblGene gene = geneIt.hasNext() ? geneIt.next() : null;
		    TreeMap<Integer, Integer> islandStartEnd = new TreeMap<Integer, Integer>();
		    while (windowStart + windowSize<RefSeqUCSCEnsemblGenes.getChromosomeLength(i, species) && phastCons.hasMoreIslands()) {
			//if (!overlapsExon(windowStart, windowStart + windowSize, exons) && !overlapsPromoter(windowStart, windowStart + windowSize, geneInfo[i-1], promoterDist)) {
			//overlapsExon = !(windowStart + windowSize<exon.getStart()) || overlapsExon;
			//overlapsPromoter = !(windowStart + windowSize<gene.getStart() - promoterDist) || overlapsPromoter;
			if (true) {
			//if ((exon==null || excludeExons || !overlapsExon) && (gene==null || excludePromoters || !overlapsPromoter)) {
			    double cons[] = phastCons.getNextIsland(windowSize), totCons = 0;
			    for (int j=0; j<cons.length; j++) { totCons += cons[j]; }
			    if (totCons>cons.length*.9) { //Store in .bed format
				if (islandStart==-1)
				    islandStart = windowStart;
				nWindows++;
				meanConsIsland += totCons/windowSize;
			    }
			    else if (islandStart!=-1) {
				islandStartEnd.put(islandStart, windowStart + windowSize);
				islandStart = -1;
				meanConsIsland = 0;
			    }
			}
			else 
			    phastCons.getNextIsland(windowSize);
			windowStart = phastCons.getNextIslandStart();
		    }
		    if (islandStart!=-1)  //Write the last window
			islandStartEnd.put(islandStart, windowStart);
		    //Merge islands that are within minIslandDist
		    int islandEnd = islandStartEnd.get(islandStartEnd.firstKey());
		    islandStart = islandStartEnd.firstKey();
		    for (Integer s : islandStartEnd.keySet()) {
			if (s - islandEnd>minIslandDist) {
			    String chrStr = Integer.toString(i);
			    if (species.equals("mm")) {
				if (i==20)
				    chrStr = "X";
				else if(i==21)
				    chrStr = "Y";
			    }
			    else if (species.equals("hs")) {
				if (i==23)
				    chrStr = "X";
				else if(i==24)
				    chrStr = "Y";
			    }
			    if (sequence)
				bwSeq.write(">chr" + chrStr + ":" + islandStart + "-" + islandEnd + "\n" + sequenceParser.getNextSequence(islandStart, islandEnd - islandStart) + "\n");
			    ConservationIsland ci = new ConservationIsland(i, islandStart, islandEnd);
			    consIslands.add(ci);

			    int len = islandEnd - islandStart;
			    nBpsCovered += len;
			    //Find a list of the most conserved enhancers
			    Enhancer2 enhancer = getOverlappingEnhancer(extragenicEnhancers, i, islandStart, islandEnd, maxERNADist);
			    if (enhancer!=null) {
				if (!enhancerConservedBPs.containsKey(enhancer))
				    enhancerConservedBPs.put(enhancer, 0);
				enhancerConservedBPs.put(enhancer, enhancerConservedBPs.get(enhancer) + len);
			    }
			    rnaReads = assignType(ci, geneInfo[i-1], maxPromoterDist, maxEnhancerDist, macroRNAs, lncRNAs, exons, macroRNAExons, lincRNAExons, allEnhancers, peaks, tfNamesInStudy, ctcfPeaks, k4me3peaks, mars, features, upstream, downstream, forward, rnaReads, normFactor);

			    bw.write("chr" + chrStr + "\t" + islandStart + "\t" + islandEnd + "\t" + ci.getType() + "\n");
			    islandStart = s;
			    islandEnd = islandStartEnd.get(s);
			}
			else {
			    nBpsBridged += s - islandEnd;
			    islandEnd = islandStartEnd.get(s);
			}
		    }
		    //Write the last window
		    String chrStr = Integer.toString(i);
		    if (species.equals("mm")) {
			if (i==20)
			    chrStr = "X";
			else if(i==21)
			    chrStr = "Y";
		    }
		    else if (species.equals("hs")) {
			if (i==23)
			    chrStr = "X";
			else if(i==24)
			    chrStr = "Y";
		    }
		    ConservationIsland ci = new ConservationIsland(i, islandStart, islandEnd);
		    rnaReads = assignType(ci, geneInfo[i-1], maxPromoterDist, maxEnhancerDist, macroRNAs, lncRNAs, exons, macroRNAExons, lincRNAExons, allEnhancers, peaks, tfNamesInStudy, ctcfPeaks, k4me3peaks, mars, features, upstream, downstream, forward, rnaReads, normFactor);
		    bw.write("chr" + chrStr + "\t" + islandStart + "\t" + islandStartEnd.get(islandStartEnd.lastKey()) + "\t" + ci.getType() + "\n");
		    if (sequence)
			bwSeq.write(">chr" + chrStr + ":" + islandStart + "-" + islandStartEnd.get(islandStartEnd.lastKey()) + "\n" + sequenceParser.getNextSequence(islandStart, islandStartEnd.get(islandStartEnd.lastKey()) - islandStart) + "\n");
		    
		    consIslands.add(ci);
		    bw.close();
		    if (sequence)
			bwSeq.close();
		    if (rnaFileName!=null) { //Save the feature info
			//Save the aggregate profiles and the scatter plots
			String scatterSaveFileStart = EstimateGenicTranscription3.SAVE_DIR + "scatter_cons_islands_" + fileName.substring(fileName.lastIndexOf("/") + 1) + "_" + (fileName.toLowerCase().contains("rna") ? (forward ? "f_" : "r_") : "") + "w_" + windowSize + (upstream==0 ? "d" : (downstream==0 ? "u" : ""));
			String profileSaveFile = EstimateGenicTranscription3.SAVE_DIR + "profile_cons_islands_" + fileName.substring(fileName.lastIndexOf("/") + 1) + (fileName.toLowerCase().contains("rna") ? (forward ? "_f" : "_r") : "") + (upstream==0 ? "_d" : (downstream==0 ? "_u" : "")) + (chr!=-1 ? "_" + chr : "") + ".txt";
			EstimateGenicTranscription3.saveScatterAndProfile(features, i, nChr, upstream, scatterSaveFileStart, profileSaveFile);
		    }
		    ///////////////////////////////////////////////////////
		    //Create another list for estimating the average number of exons and CIs per transcript
		    BufferedWriter bwPromoterCIs = new BufferedWriter(new FileWriter(savePromoterCIs)), bwExonCIs = new BufferedWriter(new FileWriter(saveExonCIs)), bwLongCIs = new BufferedWriter(new FileWriter(saveLongCIs));
		    System.out.println("Saving conserved exon info to " + saveExonCIs);
		    bwPromoterCIs.write("#ID\tStart\tEnd\tStrand\tnCIs\tCoding\n");
		    bwExonCIs.write("#ID\tGene_Start\tGene_End\tStrand\tExon_Start\tExon_End\tnCIs\tCoding\n");
		    bwLongCIs.write("#ID\tGene_Start\tGene_End\tStrand\tnCIs\tCoding\n");
		    HashSet<Exon> exonsFound = new HashSet<Exon>();
		    HashSet<ConservationIsland> cisFound = new HashSet<ConservationIsland>();
		    for (RefSeqUCSCEnsemblGene g : geneInfo[i-1].getGenes()) {
			//Find all the CIs overlapping the promoter
			boolean strand = g.getStrand();
			TreeSet<ConservationIsland> overlappingPromoterCIs = g.getOverlappingPromoterCIs(consIslands, maxPromoterDist);
			int nCIs = findNewCIs(overlappingPromoterCIs, cisFound);
			bwPromoterCIs.write(g.getID() + "\t" + (g.getStart()) + "\t" + (g.getStart()) + "\t" + (strand ? 1 : 0) + "\t" + nCIs + "\t" + (g.isNonCoding() ? 0 : 1) + "\n");
			cisFound.addAll(overlappingPromoterCIs);
			//Find all the exons overlapping the gene
			TreeSet<Exon> overlappingExons = g.getOverlappingExons(exons);
			for (Exon ex : overlappingExons) {
			    if (!exonsFound.contains(ex)) {
				//Find all the CIs overlapping the exon
				TreeSet<ConservationIsland> overlappingCIs = ex.getOverlappingCIs(consIslands);
				nCIs = findNewCIs(overlappingCIs, cisFound);
				bwExonCIs.write(g.getID() + "\t" + (strand ? g.getStart() : g.getEnd()) + "\t" + (strand ? g.getEnd() : g.getStart()) + "\t" + (strand ? 1 : 0) + "\t" + ex.getStart() + "\t" + ex.getEnd() + "\t" + overlappingCIs.size() + "\t" + (g.isNonCoding() ? 0 : 1) + "\n");
				cisFound.addAll(overlappingCIs);
			    }
			}
		    }
		    for (RefSeqUCSCEnsemblGene g : geneInfo[i-1].getNonOverlappingGenes()) {
			if (g.isNonCoding()) {
			    TreeSet<ConservationIsland> overlappingLong = g.getOverlappingCIs(consIslands);
			    //nCIs = findNewCIs(overlappingLong, cisFound);
			    bwLongCIs.write(g.getID() + "\t" + (g.getStart()) + "\t" + (g.getEnd()) + "\t" + (g.getStrand() ? 1 : 0) + "\t" + overlappingLong.size() + "\t" + (g.isNonCoding() ? 0 : 1) + "\n");
			}
		    }

		    //Check the macroRNAs and LincRNAs as well
		    for (MacroRNA m : macroRNAs) {
			boolean strand = m.getStrand();
			TreeSet<ConservationIsland> overlappingPromoterCIs = m.getOverlappingPromoterCIs(consIslands, maxPromoterDist);
			int nCIs = findNewCIs(overlappingPromoterCIs, cisFound);
			bwPromoterCIs.write(m.getID() + "\t" + (m.getStart()) + "\t" + (m.getStart()) + "\t" + (strand ? 1 : 0) + "\t" + nCIs + "\t0\n");
			cisFound.addAll(overlappingPromoterCIs);
			//Find all the exons overlapping the gene
			TreeSet<Exon> overlappingExons = m.getOverlappingExons(macroRNAExons);
			TreeSet<ConservationIsland> overlappingLong = m.getOverlappingCIs(consIslands);
			//nCIs = findNewCIs(overlappingLong, cisFound);
			bwLongCIs.write(m.getID() + "\t" + (m.getStart()) + "\t" + (m.getEnd()) + "\t" + (strand ? 1 : 0) + "\t" + overlappingLong.size() + "\t0\n");
			for (Exon ex : overlappingExons) {
			    if (!exonsFound.contains(ex)) {
				//Find all the CIs overlapping the exon
				TreeSet<ConservationIsland> overlappingCIs = ex.getOverlappingCIs(consIslands);
				nCIs = findNewCIs(overlappingCIs, cisFound);
				bwExonCIs.write(m.getID() + "\t" + m.getStart() + "\t" + m.getEnd() + "\t" + (strand ? 1 : 0) + "\t" + ex.getStart() + "\t" + ex.getEnd() + "\t" + overlappingCIs.size() + "\t0\n");
				cisFound.addAll(overlappingCIs);
			    }
			}
		    }
		    for (LncRNA m : lncRNAs) {
			TreeSet<ConservationIsland> overlappingPromoterCIs = m.getOverlappingPromoterCIs(consIslands, maxPromoterDist);
			int nCIs = findNewCIs(overlappingPromoterCIs, cisFound);
			bwPromoterCIs.write(m.getID() + "\t" + (m.getStart()) + "\t" + (m.getStart()) + "\t1\t" + nCIs + "\t0\n");
			cisFound.addAll(overlappingPromoterCIs);
			//Find all the exons overlapping the gene
			TreeSet<ConservationIsland> overlappingLong = m.getOverlappingCIs(consIslands);
			//nCIs = findNewCIs(overlappingLong, cisFound);
			bwLongCIs.write(m.getID() + "\t" + (m.getStart()) + "\t" + (m.getEnd()) + "\t1\t" + overlappingLong.size() + "\t0\n");
			TreeSet<Exon> overlappingExons = m.getOverlappingExons(macroRNAExons);
			for (Exon ex : overlappingExons) {
			    if (!exonsFound.contains(ex)) {
				//Find all the CIs overlapping the exon
				TreeSet<ConservationIsland> overlappingCIs = ex.getOverlappingCIs(consIslands);
				nCIs = findNewCIs(overlappingCIs, cisFound);
				bwExonCIs.write(m.getID() + "\t" + m.getStart() + "\t" + m.getEnd() + "\t1\t" + ex.getStart() + "\t" + ex.getEnd() + "\t" + overlappingCIs.size() + "\t0\n");
				cisFound.addAll(overlappingCIs);
			    }
			}
		    }
		    bwPromoterCIs.close();
		    bwLongCIs.close();
		    bwExonCIs.close();
		    //Save the list with conservation at eRNAs
		    bw = new BufferedWriter(new FileWriter(saveERNAConservationLengthsName));
		    for (Enhancer2 e : extragenicEnhancers) {
			if (e.getChromosome()==i)
			    bw.write(e.getChromosome() + "\t" + e.getCenter() + "\t" + (enhancerConservedBPs.containsKey(e) ? enhancerConservedBPs.get(e) : 0) + "\n");
		    }
		    bw.close();
		}
	    }
	    System.out.println("Bridged " + nBpsBridged + " and covered " + nBpsCovered);
	    //Convert to .bw format and create a load-line
	    String trackInfoStr = "track type=bigWig name=\"conservation_islands" + (excludeExons ? "_exons" : "") + (excludePromoters ? "_promoters" : "") + "\" description=\"conservation_islands" + (excludeExons ? "_exons" : "") + (excludePromoters ? "_promoters" : "") + "\" priority=" + (1567 + (excludeExons ? 1 : 0) + (excludePromoters ? 2 : 0)) + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:1 color=\"255,127,0\" smoothingWindow=off bigDataUrl=http://neuroduo.hms.harvard.edu/bigWig/mh/" + (saveFileName.substring(saveFileName.lastIndexOf("/") + 1)).replace("bed", "bw");
	    //CreateWigFiles.addTrackInfo(trackInfoStr, CreateWigFiles.TRACK_INFO_FILE_NAME);
	    //CreateWigFiles.createWWWLink(saveFileName, "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/", species, false);
	    System.out.println("Found " + nWindows + " that were conserved.");
	    //Save the lengths of the conserved islands
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "promoter_lengths")));
	    for (Integer k : promoterLengths.keySet())
		bw.write(k + "\t" + promoterLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "ncpromoter_lengths")));
	    for (Integer k : ncPromoterLengths.keySet())
		bw.write(k + "\t" + ncPromoterLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "exon_lengths")));
	    for (Integer k : exonLengths.keySet())
		bw.write(k + "\t" + exonLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "ncexon_lengths")));
	    for (Integer k : ncExonLengths.keySet())
		bw.write(k + "\t" + ncExonLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "enhancer_lengths")));
	    for (Integer k : enhancerLengths.keySet())
		bw.write(k + "\t" + enhancerLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "peak_lengths")));
	    for (Integer k : peakLengths.keySet())
		bw.write(k + "\t" + peakLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "ctcf_lengths")));
	    for (Integer k : ctcfLengths.keySet())
		bw.write(k + "\t" + ctcfLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "intronic_lengths")));
	    for (Integer k : intronicLengths.keySet())
		bw.write(k + "\t" + intronicLengths.get(k) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", "other_lengths")));
	    for (Integer k : otherLenghts.keySet())
		bw.write(k + "\t" + otherLenghts.get(k) + "\n");
	    bw.close();
	    System.out.println("Saving MARs to " + saveFileNameLengths.replace("lengths", randomize ? "random_mar_lengths" : "mar_lengths"));
	    bw = new BufferedWriter(new FileWriter(saveFileNameLengths.replace("lengths", randomize ? "random_mar_lengths" : "mar_lengths")));
	    for (Integer k : marLengths.keySet())
		bw.write(k + "\t" + marLengths.get(k) + "\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static ABIRead[] assignType(ConservationIsland ci, RefSeqUCSCEnsemblGenes geneInfo, int maxPromoterDist, int maxEnhancerDist, TreeSet<MacroRNA> macroRNAs, TreeSet<LncRNA> lncRNAs, TreeSet<Exon> exons, TreeSet<Exon> macroRNAExons, TreeSet<Exon> lincRNAExons, TreeSet<Enhancer2> allEnhancers, Vector<Peak> peaks, String tfNamesInStudy[], Vector<Peak> ctcfPeaks, Vector<Peak> k4me3peaks, TreeSet<MAR> mars, LinkedHashMap<String, FeatureInfo> features, int upstream, int downstream, boolean forward, ABIRead[] rnaReads, double normFactor) {
	FeatureInfo featureInfo = null;
	int islandStart = ci.getStart(), islandEnd = ci.getEnd(), chr = ci.getChromosome();
	RefSeqUCSCEnsemblGene g = geneInfo.getOverlappingTSS(islandStart - maxPromoterDist, islandEnd + maxPromoterDist);
	if (g!=null) {
	    if (g.isNonCoding()) {
		ci.setType(NON_CODING_PROMOTER_CI_TYPE);
		if (ci.overlaps(g.getStart()))
		    featureInfo = features.get("TSS");
		//Not a pretty solution...
		if (!ncPromoterLengths.containsKey(ci.getLength()))
		    ncPromoterLengths.put(ci.getLength(), 0);
		ncPromoterLengths.put(ci.getLength(), ncPromoterLengths.get(ci.getLength()) + 1);
	    }
	    else {
		ci.setType(CODING_PROMOTER_CI_TYPE);
		if (ci.overlaps(g.getStart()))
		    featureInfo = features.get("ncTSS");
		if (!promoterLengths.containsKey(ci.getLength()))
		    promoterLengths.put(ci.getLength(), 0);
		promoterLengths.put(ci.getLength(), promoterLengths.get(ci.getLength()) + 1);
	    }
	}
	else if (overlapsMacroRNAPromoter(chr, islandStart - maxPromoterDist, islandEnd + maxPromoterDist, macroRNAs) || overlapsLncRNAPromoter(chr, islandStart - maxPromoterDist, islandEnd + maxPromoterDist, lncRNAs)) {
	    //Check lincRNA and macroRNA promoters as well
	    ci.setType(NON_CODING_PROMOTER_CI_TYPE);
	    featureInfo = features.get("ncTSS");
	    if (!ncPromoterLengths.containsKey(ci.getLength()))
		ncPromoterLengths.put(ci.getLength(), 0);
	    ncPromoterLengths.put(ci.getLength(), ncPromoterLengths.get(ci.getLength()) + 1);
	}
	else {
	    //Check exons
	    Exon ex = getOverlappingExon(islandStart, islandEnd, exons);
	    if (ex!=null) {
		g = geneInfo.getOverlappingGene(islandStart, islandEnd);
		if (g!=null) {
		    if (g.isNonCoding()) {
			ci.setType(NON_CODING_EXON_CI_TYPE);
			if (!ncExonLengths.containsKey(ci.getLength()))
			    ncExonLengths.put(ci.getLength(), 0);
			ncExonLengths.put(ci.getLength(), ncExonLengths.get(ci.getLength()) + 1);
		    }
		    else {
			ci.setType(CODING_EXON_CI_TYPE);
			if (!exonLengths.containsKey(ci.getLength()))
			    exonLengths.put(ci.getLength(), 0);
			exonLengths.put(ci.getLength(), exonLengths.get(ci.getLength()) + 1);
		    }
		}
		//else 
		//	System.err.println("Could not find overlapping gene for exon " + ex.toString());
	    }
	    else if (overlapsExon(islandStart, islandEnd, macroRNAExons) || overlapsExon(islandStart, islandEnd, lincRNAExons)) {
		ci.setType(NON_CODING_EXON_CI_TYPE);
		if (!ncExonLengths.containsKey(ci.getLength()))
		    ncExonLengths.put(ci.getLength(), 0);
		ncExonLengths.put(ci.getLength(), ncExonLengths.get(ci.getLength()) + 1);
	    }
	    else if (overlapsEnhancer(allEnhancers, chr, islandStart, islandEnd, maxEnhancerDist)) {
		ci.setType(ENHANCER_CI_TYPE);
		if (g==null)
		    featureInfo = features.get("extra_enhancer");
		if (!enhancerLengths.containsKey(ci.getLength()))
		    enhancerLengths.put(ci.getLength(), 0);
		enhancerLengths.put(ci.getLength(), enhancerLengths.get(ci.getLength()) + 1);
	    }
	    else if (overlapsPeak(k4me3peaks, chr, islandStart, islandEnd, maxEnhancerDist, tfNamesInStudy)) {
		ci.setType(K4ME3_PEAK_CI_TYPE);
		featureInfo = features.get("k4me3");
		if (!peakLengths.containsKey(ci.getLength()))
		    peakLengths.put(ci.getLength(), 0);
		peakLengths.put(ci.getLength(), peakLengths.get(ci.getLength()) + 1);
	    }
	    else if (overlapsPeak(peaks, chr, islandStart, islandEnd, maxEnhancerDist, tfNamesInStudy)) {
		ci.setType(PEAK_CI_TYPE);
		featureInfo = features.get("RFBS");
		if (!peakLengths.containsKey(ci.getLength()))
		    peakLengths.put(ci.getLength(), 0);
		peakLengths.put(ci.getLength(), peakLengths.get(ci.getLength()) + 1);
	    }
	    else if (overlapsPeak(ctcfPeaks, chr, islandStart, islandEnd, maxEnhancerDist, null)) {
		ci.setType(INSULATOR_CI_TYPE);
		if (!ctcfLengths.containsKey(ci.getLength()))
		    ctcfLengths.put(ci.getLength(), 0);
		ctcfLengths.put(ci.getLength(), ctcfLengths.get(ci.getLength()) + 1);
	    }
	    else if (overlapsMAR(islandStart, islandEnd, mars)) {
		ci.setType(MAR_CI_TYPE);
		featureInfo = features.get("MAR");
		if (!marLengths.containsKey(ci.getLength()))
		    marLengths.put(ci.getLength(), 0);
		marLengths.put(ci.getLength(), marLengths.get(ci.getLength()) + 1);
	    }
	    /*else if (ests && (findOverlappingESTs(estTagsMapForward, i, true, islandStart - maxPromoterDist, islandEnd + maxPromoterDist)>0 || findOverlappingESTs(estTagsMapReverse, i, false, islandStart - maxPromoterDist, islandEnd + maxPromoterDist)>0)) {
	      if (!estLengths.containsKey(len))
	      estLengths.put(len, 0);
	      estLengths.put(len, estLengths.get(len) + 1);
	      }*/
	    else {
		//Check if overlaps a gene -> intronic
		if (geneInfo.getOverlappingGene(islandStart, islandEnd)!=null) {
		    ci.setType(INTRONIC_CI_TYPE);
		    if (!intronicLengths.containsKey(ci.getLength()))
			intronicLengths.put(ci.getLength(), 0);
		    intronicLengths.put(ci.getLength(), intronicLengths.get(ci.getLength()) + 1);
		}
		else {
		    ci.setType(EXTRAGENIC_CI_TYPE);
		    if (!otherLenghts.containsKey(ci.getLength()))
			otherLenghts.put(ci.getLength(), 0);
		    otherLenghts.put(ci.getLength(), otherLenghts.get(ci.getLength()) + 1);
		}
	    }
	}
	if (featureInfo!=null) {
	    //int[] rs = FindReadsAcrossFeature2.getReadsHistogram(chr, ci.getCenter(), forward, downstream, upstream, rnaReads, 1.0);
	    //featureInfo.addLocInfo(chr, ci.getCenter(), forward, rs, normFactor, upstream, true);
	    //rnaReads = EstimateGenicTranscription3.subtractReads(rnaReads, rs);
	}
	return rnaReads;
    }
	
    public static int findNewCIs(TreeSet<ConservationIsland> newCIs, HashSet<ConservationIsland> foundCIs) {
	int n = 0;
	for (ConservationIsland c : newCIs) 
	    n += foundCIs.contains(c) ? 0 : 1;
	return n;
    }

    public static boolean overlapsEnhancer(TreeSet<Enhancer2> enhancers, Peak p, int maxDist) { return overlapsEnhancer(enhancers, p.getChromosome(), p.getStart(), p.getEnd(), maxDist); }

    public static boolean overlapsEnhancer(TreeSet<Enhancer2> enhancers, int chr, int start, int end, int maxDist) { return getOverlappingEnhancer(enhancers, chr, start, end, maxDist)!=null; }

    public static Enhancer2 getOverlappingEnhancer(TreeSet<Enhancer2> enhancers, int chr, int start, int end, int maxDist) {
	for (Enhancer2 e : enhancers) {
	    if (chr==e.getChromosome() && e.overlaps(start, end, maxDist)) {
		//System.out.println(e.toString() + " overlaps chr" + chr + ":" + start + "-" + end + " with length " + (end-start));
		return e;
	    }
	}
	return null;
    }

    public static boolean overlapsPeak(Vector<Peak> peaks, int chr, int start, int end, int maxDist, String tfNamesInStudy[]) { return getOverlappingPeak(peaks, chr, start, end, maxDist, tfNamesInStudy)!=null; }

    public static Peak getOverlappingPeak(Vector<Peak> peaks, int chr, int start, int end, int maxDist, String tfNamesInStudy[]) {
	for (Peak p : peaks) {
	    if (chr==p.getChromosome() && p.overlaps(start, end, maxDist) && (tfNamesInStudy==null || p.hasFactor(tfNamesInStudy)))
		return p;
	}
	return null;
    }

    public static Exon getOverlappingExon(int start, int end, TreeSet<Exon> exons) {
	for (Exon e : exons) {
	    if (e.overlaps(start, end))
		return e;
	}
	return null;
    }

    public static boolean overlapsExon(int start, int end, TreeSet<Exon> exons) { return getOverlappingExon(start, end, exons)!=null; }

    public static boolean overlapsExon(int center, TreeSet<Exon> exons) { return getOverlappingExon(center, center, exons)!=null; }
	
    public static MAR getOverlappingMAR(int start, int end, TreeSet<MAR> mars) {
	for (MAR m : mars) {
	    if (m.overlaps(start, end, true))
		return m;
	}
	return null;
    }

    public static boolean overlapsMAR(int start, int end, TreeSet<MAR> mars) { return getOverlappingMAR(start, end, mars)!=null; }


    public static boolean overlapsPromoter(int start, int end, RefSeqUCSCEnsemblGenes geneInfo, int promoterDist) {
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes()) {
	    if ((g.getStart() - promoterDist<=start && g.getStart() + promoterDist>=start) || (g.getStart() - promoterDist<=end && g.getStart() + promoterDist>=end) || (g.getStart() - promoterDist>start && g.getStart() + promoterDist<end))
		return true;
	}
	return false;
    }

    public static boolean isConservationIsland(double consScores[]) {
	double consSum = 0, maxCons = 0;
	for (int i=0; i<consScores.length; i++) {
	    consSum += consScores[i];
	    maxCons = consScores[i]>maxCons ? consScores[i] : maxCons;
	}
	return /*maxCons>.9 || */consSum>consScores.length/2;
    }

    public static TreeSet<Exon> parseExons(String exonFileName, String species, int chr) {
	TreeSet<Exon> exons = new TreeSet<Exon>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(exonFileName));
	    String line = br.readLine();
	    //name chr start end
	    Pattern exonPattern = Pattern.compile("([\\w|-|_]+)\\s+chr(\\w+)\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = exonPattern.matcher(line);
		if (m.find()) {
		    int ch = 0;
		    if (species.equals("mm")) {
			if (m.group(2).equals("X"))
			    ch = 20;
			else if (m.group(2).equals("Y"))
			    ch = 21;
			else if (m.group(2).equals("M") || m.group(2).contains("random") || m.group(2).contains("hap")) {
			    line = br.readLine();
			    continue;
			}
			else
			    ch = Integer.parseInt(m.group(2));
		    }
		    else if (species.equals("hs")) {
			if (m.group(2).equals("X"))
			    ch = 23;
			else if (m.group(2).equals("Y"))
			    ch = 24;
			else if (m.group(2).equals("M") || m.group(2).contains("random") || m.group(2).contains("hap") || m.group(2).contains("n")) {
			    line = br.readLine();
			    continue;
			}
			else
			    ch = Integer.parseInt(m.group(2));
		    }
		    if (ch==chr || chr==-1) {
			Exon ex = new Exon(chr, Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4)), m.group(1));
			if (!exons.contains(ex))
			    exons.add(ex);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing " + exonFileName + " (exon file) " + e.toString());
	}
	System.out.println("Found " + exons.size() + " exons for chr " + chr);
	return exons;
    }

    public static TreeSet<Exon> parseMacroRNAExons(String macroRNAFileName, int chr) {
	TreeSet<Exon> macroRNAExons = new TreeSet<Exon>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(macroRNAFileName));
	    String line = br.readLine();
	    //bin    matches misMatches      repMatches      nCount  qNumInsert      qBaseInsert     tNumInsert      tBaseInsert     strand  qName   qSize   qStart qEnd     tName   tSize   tStart  tEnd    blockCount      blockSizes      qStarts tStarts
	    Pattern macrornaPattern = Pattern.compile("([+|-])\\s+(\\w+)\\s+\\d+\\s+\\d+\\s+\\d+\\s+chr([X|Y|\\d]+)\\s+\\d+\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = macrornaPattern.matcher(line);
		if (m.find()) {
		    int c = 0;
		    if (m.group(3).equals("X"))
			c = 20;
		    else if (m.group(3).equals("Y"))
			c = 21;
		    else
			c = Integer.parseInt(m.group(3));
		    if (c==chr) {
			String tokens[] = line.split("\\s+");
			int nExons = Integer.parseInt(tokens[tokens.length - 4]);
			String exonStarts[] = tokens[tokens.length - 1].split(",");
			String exonLengths[] = tokens[tokens.length - 3].split(",");
			for (int i=0; i<nExons; i++) {
			    Exon ex = new Exon(chr, Integer.parseInt(exonStarts[i]), Integer.parseInt(exonStarts[i]) + Integer.parseInt(exonLengths[i]), m.group(2));
			    //System.out.println("Found macroRNA exon " + ex.toString());
			    macroRNAExons.add(ex);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing exon file " + e.toString());
	}
	catch (NullPointerException e) {
	    System.err.println("Error parsing exon file " + e.toString());
	}
	System.out.println("Found " + macroRNAExons.size() + " macroRNA exons for chr " + chr);
	return macroRNAExons;
    }

    public static TreeSet<Exon> parseLincRNAExons(String lncRNAExonFileName, int chr, TreeSet<LncRNA> lncRNAs) {
	TreeSet<Exon> lincRNAExons = new TreeSet<Exon>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(lncRNAExonFileName));
	    String line = br.readLine();
	    Pattern lincrnaPattern = Pattern.compile("chr([X|Y|\\d]+):(\\d+)-(\\d+)");
	    while (line!=null) {
		Matcher m = lincrnaPattern.matcher(line);
		if (m.find()) {
		    int c = 0;
		    if (m.group(1).equals("X"))
			c = 20;
		    else if (m.group(1).equals("Y"))
			c = 21;
		    else
			c = Integer.parseInt(m.group(1));
		    Exon ex = new Exon(chr, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), "ScriptureExon");
		    if (c==chr && overlapsLncRNA(chr, ex.getStart(), ex.getEnd(), lncRNAs))
			//Make sure that the exon overlaps a lincRNA
			lincRNAExons.add(ex);
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing exon file " + e.toString());
	}
	catch (NullPointerException e) {
	    System.err.println("Error parsing exon file " + e.toString());
	}
	System.out.println("Found " + lincRNAExons.size() + " lincRNA exons for chr " + chr);
	return lincRNAExons;
    }

    public static boolean overlapsLncRNA(int chr, int start, int end, TreeSet<LncRNA> lncRNAs) {
	for (LncRNA l : lncRNAs) {
	    if (l.getChromosome()==chr && l.overlaps(start, end))
		return true;
	}
	return false;
    }

    public static boolean overlapsLncRNA(int chr, int pos, TreeSet<LncRNA> lncRNAs) { return overlapsLncRNA(chr, pos, pos, lncRNAs); }

    public static boolean overlapsMacroRNA(int chr, int start, int end, TreeSet<MacroRNA> macroRNAs) {
	for (MacroRNA l : macroRNAs) {
	    if (l.getChromosome()==chr && l.overlaps(start, end))
		return true;
	}
	return false;
    }

    public static boolean overlapsMacroRNA(int chr, int pos, TreeSet<MacroRNA> macroRNAs) { return overlapsMacroRNA(chr, pos, pos, macroRNAs); }

    public static int findOverlappingESTs(LinkedHashMap<String, EST> estTagsMap, int chr, boolean forward, int start, int end) {
	int nNear = 0;
	//System.out.println("Checking for ESTs near " + chr + ":" + center);
	//System.out.println(estTagsMap.get(estTagsMap.keySet().iterator().next()).toString());
	for (String k : estTagsMap.keySet()) { //go through all ESTs
	    EST est = estTagsMap.get(k);
	    if (start<est.getStart() && end>est.getStart()) 
		nNear++;
	}
	return nNear;
    }

    public static boolean overlapsMacroRNAPromoter(int chr, int start, int end, TreeSet<MacroRNA> macroRNAs) {
	for (MacroRNA m : macroRNAs) {
	    if (chr==m.getChromosome() && start<m.getStart() && end>m.getStart())
		return true;
	}
	return false;
    }

    public static boolean overlapsLncRNAPromoter(int chr, int start, int end, TreeSet<LncRNA> lncRNAs) {
	for (LncRNA l : lncRNAs) {
	    if (chr==l.getChromosome() && ((start<l.getStart() && end>l.getStart()) || (start<l.getEnd() && end>l.getEnd())))
		return true;
	}
	return false;
    }

    public static TreeSet<MAR> loadMARs(int chr, String species, boolean repeatMask) { return loadMARs(chr, species, repeatMask, false); }

    public static TreeSet<MAR> loadMARs(int chr, String species, boolean repeatMask, boolean randomize) {
	String loadFileName = "MARs/" + species + (repeatMask ? "_repeatMask" : "") + "_MARs_Hrule.txt";
	TreeSet<MAR> mars = new TreeSet<MAR>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(loadFileName));
	    String line = br.readLine();
	    while (line!=null) {
		if (!line.startsWith("#")) {
		    String tokens[] = line.split("\\s+");
		    int c = Integer.parseInt(tokens[0]);
		    if (chr==c) {
			if (randomize) {
			    //Choose a random location instead
			    int randLoc = random.nextInt(species.equals("hs") ? RefSeqGenes.CHROMOSOME_LENGTH_HS[c-1] : RefSeqGenes.CHROMOSOME_LENGTH[c-1]);
			mars.add(new MAR(c, randLoc, randLoc + Integer.parseInt(tokens[2]) - Integer.parseInt(tokens[1])));
			}
			else
			    mars.add(new MAR(c, Integer.parseInt(tokens[1]), Integer.parseInt(tokens[2])));
		    }
		    line = br.readLine();
		}
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.out.println("Error opening MAR file " + e.toString());
	}
	System.out.println("Found " + mars.size() + " MARs.");
	return mars;
    }


}


class MAR implements Comparable {

    protected int start, end, chr;

    public MAR(int c, int s, int e) {
	chr = c;
	start = s;
	end = e;
    }

    public int getChromosome() { return chr; }

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public boolean overlaps(ConservationIsland c) { return c.getChromosome()==chr && overlaps(c.getStart(), c.getEnd(), true); }

    public boolean overlaps(int s, int e, boolean strict) {
	if (strict && e-s>0) {
	    double overlap = Math.min(e, end) - Math.max(s, start);
	    return overlap>0 ? /*overlap/(end - start)>.5 &&*/ overlap>(e - s)/3 : false;
	}
	return (start<=s && end>=s) || (start<=e && end>=e) || (start>s && end<e || (start<s && end>e));
    }

    public TreeSet<ConservationIsland> getOverlappingCIs(ConservationIslands consIslands) {
	TreeSet<ConservationIsland> overlappingCIs = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : consIslands.getConservationIslands()) {
	    if (overlaps(c))
		overlappingCIs.add(c);
	}
	return overlappingCIs;
    }

    //////////////////////////////////////////////////////////////////

    public boolean equals(Object o) {
	if (!(o instanceof MAR))
	    return false;
	MAR m = (MAR)o;
	return m.chr==chr && m.start==start && m.end==end;
    }

    public int compareTo(Object o) {
	if (!(o instanceof MAR))
	    return 0;
	MAR m = (MAR)o;
	if (chr<m.chr)
	    return -1;
	else if (chr>m.chr)
	    return 1;
	else {
	    if (start<m.start)
		return -1;
	    else if (start>m.start)
		return 1;
	    else {
		if (end<m.end)
		    return -1;
		else if (end>m.end)
		    return 1;
		else
		    return 0;
	    }
	}
    }

    public int hashCode() {
	return start*23 + end*37 + chr*93;
    }

    public String toString() {
	return "chr" + chr + ":" + start + "-" + end;
    }

}