//Use this method to calculate the overlap btwn the peaks for a given
//TF from different cell-types. v2 does the same thing as the
//original, but is better structured.
//
//
//For looking at the overlap btwn h3k4me3 in HeLa and DHS sites in other cell types
//
//bsub -q shared_2h java -Xmx4096m -server FindEncodePeaksOverlap2 -o -c 22 -f dnase -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak
//
//Looking for overlap between H3k4me3 in HeLa and other cell types
//
//java -Xmx4096m -server FindEncodePeaksOverlap2 -o -c 22 -f H3k4me3 -p ../../Encode/PeakFiles/h3k4me3_names.txt -F H3k4me3 -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak
//
//Looking for the total number of conserved H3k4me3 that are not explained by annotation:
//
//java -Xmx4096m -server FindEncodePeaksOverlap2 -o --conservation -c 22 -f H3k4me3 -p ../../Encode/PeakFiles/h3k4me3_names.txt
//
//When looking at only a specific category of HeLa DHS and h3k4me3 in other cell types
//
//java -Xmx4096m -server FindEncodePeaksOverlap2 -o -c 22 -f H3k4me3 -p ../../Encode/PeakFiles/h3k4me3_names.txt -P hela_dnase -C Extragenic_other
//
//If we are interested in only two factors from one cell type
//
//bsub -q shared_2h java -Xmx4096m -server FindEncodePeaksOverlap2 -o -c 22 -f dnase -p /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwDnaseSeqPeaksRep1Helas3.narrowPeak -P /home/mh168/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak
//
//Looking for the overlap of Cfos
//
//java -Xmx4096m -server FindEncodePeaksOverlap2 -o -c 22 -f Cfos -p ../../Encode/PeakFiles/cfos_names.txt 
//



import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.Tally;


public class FindEncodePeaksOverlap2 {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption("conservation");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption("nMC");	
	    //This the TF for the set of peaks that we want to use as our reference
	    CmdLineParser.Option factorNameFlag = argParser.addStringOption('f', "factorName");
	    //The TF for the other peaks that we are interested in
	    CmdLineParser.Option overlapFactorNameFlag = argParser.addStringOption('F', "overlapFactorName");
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option overlapPeakFileNameFlag = argParser.addStringOption('P', "overlapPeakFileName");
	    //Only consider peaks that fall into a given category
	    CmdLineParser.Option peakCategoryFlag = argParser.addStringOption('C', "category");
	    //Look at ESTs as well as a way to assess transcriptional activity
	    CmdLineParser.Option estsFlag = argParser.addIntegerOption("EST");
	    CmdLineParser.Option splicedESTFlag = argParser.addBooleanOption("spliced");
	    CmdLineParser.Option unsplicedESTFlag = argParser.addBooleanOption("unspliced");
	    //Investigate the GC content at different loci
	    CmdLineParser.Option cpgFlag = argParser.addIntegerOption("CpG");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    tmp = argParser.getOptionValue(factorNameFlag);
	    String factorName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(overlapFactorNameFlag);
	    String overlapFactorName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : (!orchestra ? "/home/mhemberg" : "/home/mh168") + "/Projects/ChIP-Seq/Encode/PeakFiles/dnase_names.txt";
	    tmp = argParser.getOptionValue(overlapPeakFileNameFlag);
	    String overlapPeakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(estsFlag);
	    int maxESTDist = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean ests = maxESTDist>0;
	    int spliced = argParser.getOptionValue(splicedESTFlag)!=null ? 1 : (argParser.getOptionValue(unsplicedESTFlag)!=null ? -1 : 0);
	    tmp = argParser.getOptionValue(cpgFlag);
	    int sequenceWidth = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean cpg = sequenceWidth>0;
	    tmp = argParser.getOptionValue(peakCategoryFlag);
	    String category = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "HeatMaps/" + (peakFileName.contains("/") ? peakFileName.substring(peakFileName.lastIndexOf("/") + 1, peakFileName.lastIndexOf(".")) : peakFileName.substring(0, Math.max(peakFileName.length(), peakFileName.lastIndexOf("."))));
	    if (tmp==null) {
		if (overlapPeakFileName!=null)
		    saveFileName += "_" + (overlapPeakFileName.contains("/") && overlapPeakFileName.contains(".") ? overlapPeakFileName.substring(overlapPeakFileName.lastIndexOf("/") + 1, overlapPeakFileName.lastIndexOf(".")) : overlapPeakFileName);
		saveFileName += "_peaks_overlap" + (category!=null ? "_" + category : "") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    }
	    
	    ///////////////////////////////////////////////////////
	    boolean removePromoters = false, removeIntergenic = false, removeEnhancerOverlappingPeaks = false;
	    int nChr = 24, maxTSSDist = 1000, maxEnhancerDist = 1000, maxPeakDist = 100, maxDist = 100;

	    String estFileName = "/groups/neuroduo/jesse/human_annot_for_mh/human_ESTs/", species = "hs";
	    
	    LinkedHashMap<String, EST> estTagsMapForward = null, estTagsMapReverse = null;

	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bwOverlapEnrichment = new BufferedWriter(new FileWriter(saveFileName==null ? "/dev/null" : saveFileName));

	    String cellTypes[] = EncodePeaks.getEncodeCellTypes(peakFileName, factorName);
	    System.out.println(Arrays.toString(cellTypes));
	    String overlapCellTypes[] = overlapPeakFileName!=null ? EncodePeaks.getEncodeCellTypes(overlapPeakFileName, overlapFactorName) : cellTypes;
	    System.out.println(Arrays.toString(overlapCellTypes));
	    if ((chr==-1 || chr==1)) {
		bwOverlapEnrichment.write("#chr\tStart\tEnd\tCategory");
		if (conservation)
		    bwOverlapEnrichment.write("\tCons");
		if (ests)
		    bwOverlapEnrichment.write("\tESTs_spliced\tESTs_unspliced");
		if (cpg)
		    bwOverlapEnrichment.write("\tCpG");
		for (int i=0; i<overlapCellTypes.length; i++) 
		    bwOverlapEnrichment.write("\t" + overlapCellTypes[i]);
		bwOverlapEnrichment.write("\n");
	    }
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    EncodePeaks overlapPeaks = null;
		    //RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(i, orchestra, "hs");
		    if (overlapPeakFileName!=null) { //Get a set of peaks and consider how the peaks in peakFileName overlaps this set
			overlapPeaks = new EncodePeaks(overlapPeakFileName, overlapFactorName, i);
			System.out.println("Found " + overlapPeaks.numberOfPeaks() + " peaks from " + overlapPeakFileName);
		    }
		    EncodePeaks encodePeaks = category==null ? new EncodePeaks(peakFileName, factorName, i) : new EncodePeaks(peakFileName, category, i, false);
		    System.out.println("Found " + encodePeaks.numberOfPeaks() + " peaks from " + peakFileName);
		    if (category==null) {//Need to do all the work of categorizing peaks again. However, this time we are mainly concerned about promoters vs non-promoters
			
		    }
		    //TreeSet<Peak> overlappingPeaks = overlapPeaks.findOverlappingEnrichment(encodePeaks);
		    TreeSet<Peak> overlappingPeaks = overlapPeakFileName!=null ? encodePeaks.findOverlappingEnrichment(overlapPeaks, maxDist) : encodePeaks.getPeaks();

		    RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(i, orchestra, species);
		    ConservationIslands conservationIslands = conservation ? new ConservationIslands(i, orchestra, "hs") : null;
		    ESTSplicingInfo estSplicingInfo = new ESTSplicingInfo(i, geneInfo, conservationIslands);
		    if (ests) { //Get ESTs
			estTagsMapForward = CompareCageEstRNASeqStarts.parseEstTags(estFileName, i, true, 0, species);
			estTagsMapReverse = CompareCageEstRNASeqStarts.parseEstTags(estFileName, i, false, 0, species);
		    }
		    FastaSequence2 fastaSeq = cpg ? new FastaSequence2(species, i, orchestra) : null;
		    for (Peak p : overlappingPeaks) {
			HashSet<EST> forwardESTsSpliced = ests ? FindEnhancerSites.getNearbyESTs(estTagsMapForward, p.getChromosome(), true, p.getCenter(), true, maxESTDist) : new HashSet<EST>();
			HashSet<EST> forwardESTsUnspliced = ests ? FindEnhancerSites.getNearbyESTs(estTagsMapForward, p.getChromosome(), true, p.getCenter(), false, maxESTDist) : new HashSet<EST>();
			HashSet<EST> reverseESTsSpliced = ests ? FindEnhancerSites.getNearbyESTs(estTagsMapForward, p.getChromosome(), false, p.getCenter(), true, maxESTDist) : new HashSet<EST>();
			HashSet<EST> reverseESTsUnspliced = ests ? FindEnhancerSites.getNearbyESTs(estTagsMapForward, p.getChromosome(), false, p.getCenter(), false, maxESTDist) : new HashSet<EST>();
			HashSet<EST> allESTsSpliced = new HashSet<EST>(reverseESTsSpliced), allESTsUnspliced = new HashSet<EST>(reverseESTsUnspliced);
			allESTsSpliced.addAll(forwardESTsSpliced);
			allESTsUnspliced.addAll(forwardESTsUnspliced);
			//Determine the category of the peak
			p.setCategory(geneInfo, maxTSSDist, allESTsSpliced);
			bwOverlapEnrichment.write(p.getChromosome() + "\t" + p.getStart() + "\t" + p.getEnd() + "\t" + p.categoryNumber());
			if (conservation)
			    bwOverlapEnrichment.write("\t" + (conservationIslands.overlaps(p) ? 1 : 0));
			if (ests) {
			    //Find spliced ESTs that start from a locus that appears to be enhancers in at least one cell type. We hypothesize that these loci may have a different splicing pattern than ESTs that correspond to mRNAs
			    for (EST est : allESTsSpliced) 
				estSplicingInfo.storeESTSplicingInformation(p, est, maxTSSDist);
			    bwOverlapEnrichment.write("\t" + allESTsSpliced.size() + "\t" + allESTsUnspliced.size());
			    //if (nForward + nReverse>0)
			    //System.out.println(p.toString() + " has " + (nForward + nReverse) + " ESTs.");
			    //Find loci that are conserved with spliced ESTs, where the exons of the EST are not conserved
			}
			if (cpg) {
			    char seq[] = fastaSeq.getSequence(p.getCenter() - sequenceWidth, 2*sequenceWidth + 1).toCharArray();
			    double nCpG = 0, nGpC = 0, nGC = 0;
			    for (int j=0; j<seq.length; j++) {
				nGC += (seq[j]=='c' || seq[j]=='C' || seq[j]=='g' || seq[j]=='G') ? 1 : 0;
				nCpG += (j>0 && ((seq[j-1]=='c' || seq[j-1]=='C') && (seq[j]=='g' || seq[j]=='G'))) ? 1 : 0;
				nGpC += (j>0 && ((seq[j-1]=='g' || seq[j-1]=='G') && (seq[j]=='c' || seq[j]=='C'))) ? 1 : 0;
			    }
			    nGC /= seq.length;
			    nCpG /= seq.length;
			    nGpC /= seq.length;
			    //For non-genic sites, we don't know the strandedness for calculating CpG islands and hence we are generous and consider the one that has the most CG dinucleotides
			    bwOverlapEnrichment.write("\t" + format.format(Math.max(nCpG, nGpC)/((nGC/2)*(nGC/2))));
			    //System.out.println("found cpg " + format.format(Math.max(nCpG, nGpC)/((nGC/2)*(nGC/2))) + " for " + p.toString());
			}
			int nFactors = 0;
			for (int j=0; j<overlapCellTypes.length; j++) {
			    boolean hasFactor = Arrays.binarySearch(p.getFactorNames(false), overlapCellTypes[j])>=0;
			    nFactors += hasFactor ? 1 : 0;
			    bwOverlapEnrichment.write("\t" + (hasFactor ? p.getFactor(overlapCellTypes[j]) : "NaN"));
			}
			//if (nFactors==0)
			//    System.out.println("Could not find any factors for " + p.toString() + " with " + Arrays.toString(p.getFactorNames(false)) + " and " + p.getFactor(p.getFactorNames(false)[0]));
			bwOverlapEnrichment.write("\n");
		    }
		    if (ests)
			estSplicingInfo.getESTSplicingStatistics();
		}
	    }
	    bwOverlapEnrichment.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////


}

//////////////////////////////////////////////////////////////////
//Use this class to keep track of the length of introns and exons in
//various categories of ESTs. We would like to show that some ESTs
//that are found at enhancer-only sites are different from the ones that correspond to 

class ESTSplicingInfo {

    protected int chr;
    protected RefSeqUCSCEnsemblGenes geneInfo;
    protected ConservationIslands conservationIslands;
    protected Tally codingExonLengths = new Tally(), codingIntronLengths = new Tally(), ncExonLengths = new Tally(), ncIntronLengths = new Tally();
    protected Tally intragenicExonLength = new Tally(), intragenicIntronLength = new Tally(), intragenicNcExonLength = new Tally(), intragenicNcIntronLength = new Tally();
    protected Tally unannotatedExonLengths = new Tally(), unannotatedNcExonLengths = new Tally(), otherExonLengths = new Tally(), unannotatedIntronLengths = new Tally(), unannotatedNcIntronLengths = new Tally(), otherIntronLengths = new Tally();
    protected int nCodingESTs = 0, nNCESTs = 0, nIntragenicESTs = 0, nNcIntragenicESTs = 0, nUnannotatedESTs = 0, nUnannotatedNcESTs = 0, nOtherESTs = 0;
    protected int conservedUnannotated = 0, conservedUnannotatedNc = 0, conservedOther = 0;

    public static DecimalFormat format = new DecimalFormat("###.####");

    public ESTSplicingInfo(int c, RefSeqUCSCEnsemblGenes gInfo, ConservationIslands cis) {
	chr = c;
	geneInfo = gInfo;
	conservationIslands = cis;
    }

    //////////////////////////////////////////////////////////////////

    public void getESTSplicingStatistics() {
	System.out.println("Category\tnESTs\tnExons\texon_len\tintron_len\n");
	System.out.println("Coding\t\t" + nCodingESTs + "\t" + codingExonLengths.numberObs() + "\t" + format.format(codingExonLengths.average()) + "\t" + format.format(codingIntronLengths.average()) + "\n");
	System.out.println("Non-coding\t" + nNCESTs + "\t" + ncExonLengths.numberObs() + "\t" + format.format(ncExonLengths.average()) + "\t" + format.format(ncIntronLengths.average()) + "\n");
	System.out.println("Intragenic, coding\t" + nIntragenicESTs + "\t" + intragenicExonLength.numberObs() + "\t" + format.format(intragenicExonLength.average()) + "\t" + format.format(intragenicIntronLength.average()) + "\n");
	System.out.println("Intragenic, non-coding\t" + nNcIntragenicESTs + "\t" + intragenicNcExonLength.numberObs() + "\t" + format.format(intragenicNcExonLength.average()) + "\t" + format.format(intragenicNcIntronLength.average()) + "\n");
	System.out.println("Unannotated, coding\t" + nUnannotatedESTs + "\t" + unannotatedExonLengths.numberObs() + "\t" + format.format(unannotatedExonLengths.average()) + "\t" + format.format(unannotatedIntronLengths.average()) + "\t" + conservedUnannotated + "\n");
	System.out.println("Unannotated, non-coding\t" + nUnannotatedNcESTs + "\t" + unannotatedNcExonLengths.numberObs() + "\t" + format.format(unannotatedNcExonLengths.average()) + "\t" + format.format(unannotatedNcIntronLengths.average()) + "\t" + conservedUnannotatedNc + "\n");
	System.out.println("Other\t" + nOtherESTs + "\t" + otherExonLengths.numberObs() + "\t" + format.format(otherExonLengths.average()) + "\t" + format.format(otherIntronLengths.average()) + "\t" + conservedOther + "\n");
    }

    public void storeESTSplicingInformation(Peak p, EST est, int maxDist) {
	//Find out if the peak is close to a TSS of a coding or ncRNA, intronic or extragenic
	Tally intronLengthTally, exonLengthTally;
	boolean studyConservation = false;
	if (p.getCategory().equals(Peak.CODING)) { 
		exonLengthTally = codingExonLengths;
		intronLengthTally = codingIntronLengths;
		nCodingESTs++;
	}
	else if (p.getCategory().equals(Peak.NON_CODING)) {
	    exonLengthTally = ncExonLengths;
	    intronLengthTally = ncIntronLengths;
	    nNCESTs++;
	}
	else if (p.getCategory().equals(Peak.INTRAGENIC_CODING)) {
	    exonLengthTally = intragenicExonLength;
	    intronLengthTally = intragenicIntronLength;
	    nIntragenicESTs++;
	}
	else if (p.getCategory().equals(Peak.INTRAGENIC_NON_CODING)) {
	    exonLengthTally = intragenicNcExonLength;
	    intronLengthTally = intragenicNcIntronLength;
	    nNcIntragenicESTs++;
	}
	else if (p.getCategory().equals(Peak.UNANNOTATED_CODING)) {   
	    exonLengthTally = unannotatedExonLengths;
	    intronLengthTally = unannotatedIntronLengths;
	    nUnannotatedESTs++;
	    System.out.println("Found unannotated promoter EST " + est.toString() + " near " + p.toString() + " with " + est.numberOfIntrons() + " and first exon length " + est.getExonLength(0) + " starting at " + est.getExonStart(0));
	    //for (int i=0; i<est.numberOfIntrons(); i++) 
	    //	System.out.println("Intron " + i + " has length " + est.getIntronLength(i+1) + " starting at " + est.getIntronStart(i) + ", and exon " + est.getExonLength(i+1) + " starting at " + est.getExonStart(i+1));
	    studyConservation = true;
	}
	else if (p.getCategory().equals(Peak.UNANNOTATED_NON_CODING)) {   
	    exonLengthTally = unannotatedNcExonLengths;
	    intronLengthTally = unannotatedNcIntronLengths;
	    nUnannotatedNcESTs++;
	    System.out.println("Found unannotated non-coding promoter EST " + est.toString() + " near " + p.toString() + " with " + est.numberOfIntrons() + " and first exon length " + est.getExonLength(0) + " starting at " + est.getExonStart(0));
	    //for (int i=0; i<est.numberOfIntrons(); i++) 
	    //	System.out.println("Intron " + i + " has length " + est.getIntronLength(i+1) + " starting at " + est.getIntronStart(i) + ", and exon " + est.getExonLength(i+1) + " starting at " + est.getExonStart(i+1));
	    studyConservation = true;
	}
	else {
	    exonLengthTally = otherExonLengths;
	    intronLengthTally = otherIntronLengths;
	    nOtherESTs++;
	    System.out.println("Found other EST " + est.toString() + " near " + p.toString() + " with " + est.numberOfIntrons());
	    //for (int i=0; i<est.numberOfIntrons(); i++) 
	    //System.out.println("Intron " + i + " has length " + est.getIntronLength(i+1) + " starting at " + est.getIntronStart(i) + ", and exon " + est.getExonLength(i+1) + " starting at " + est.getExonStart(i+1));
	    studyConservation = true;
	}
	if (est.numberOfIntrons()>0) {
	    exonLengthTally.add(Math.log10(est.getExonLength(0)));
	    int nConservedExons = 0;
	    for (int i=0; i<est.numberOfIntrons(); i++) {
		intronLengthTally.add(Math.log10(est.getIntronLength(i+1)));
		if (i<est.numberOfIntrons()-1) //Don't bother with the last exon
		    exonLengthTally.add(Math.log10(est.getExonLength(i+1)));
		if (studyConservation) {//check for conservation
		    int exonStart = est.getStrand() ? est.getExonStart(i+1) : est.getExonStart(i+1) - est.getExonLength(i+1);
		    int exonEnd = est.getStrand() ? est.getExonStart(i+1) + est.getExonLength(i+1) : est.getExonStart(i+1);
		    if (conservationIslands!=null && conservationIslands.overlaps(chr, exonStart, exonEnd) && !geneInfo.overlapsGene(exonStart, exonEnd)) {
			System.out.println("Exon " + i + " at " + exonStart + "-" + exonEnd + " overlaps a conservation island.");
			if (p.getCategory().equals(Peak.UNANNOTATED_CODING))
			    conservedUnannotated++;
			else if (p.getCategory().equals(Peak.UNANNOTATED_NON_CODING))
			    conservedUnannotatedNc++;
			else if (p.getCategory().equals(Peak.OTHER))
			    conservedOther++;
		    }
		}
	    }
	}
    }

}
