//Use this method to determine which peaks are located at
//transcription initiation sites (based on the mRNA data). We use the
//loc FDR framework to determine which peaks should be included. The
//null set which we compare agains is generated using
//GeneratePeakmRNANullDist.java.

//Example:

//java -Xmx512m FindInitiationPeaks -p PeakFiles/8WG16_names_combos.txt -m mrna_names.txt -w 1000 -f .01 -n mRNA/NullDistributions/8WG16_names_combos.txt

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class FindInitiationPeaks{

    public static String GENOME_LOC;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    //Keep track of all the peaks in this datastructure
    protected static TreeSet[] peaks = new TreeSet[21];


    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option nullPeakFileNameFlag = argParser.addStringOption('n', "nullPeakFileName");
	    //File containing a list of all the mRNA data sets that we
	    //want to investigate. This needs to correspond to the
	    //file which was used for generating the null peaks
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    //This should also be the same as for the null peaks
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('f', "fdr");
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('i', "isolated");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(nullPeakFileNameFlag);
	    String nullPeakFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/NullDistributions/" + peakFileName.substring(peakFileName.lastIndexOf("/") + 1, peakFileName.lastIndexOf("."));
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    //////////////////////////////////////////////////////////////////
	    int nullPeaksData[][] = loadPeakData(nullPeakFileName);
	    int peakPeaksData[][] = loadPeakData(peakFileName);
	    Vector mRNAFileNames = GeneratePeakmRNANullDist.parseMRNAFileList(mRNAFileName);
	    Iterator it = mRNAFileNames.iterator();
	    //Save the mRNAs to a file so that we can analyze separately in matlab
	    BufferedWriter bwForwardPeak = new BufferedWriter(new FileWriter("forward_peak_mrna_tmp.txt"));
	    BufferedWriter bwForwardNull = new BufferedWriter(new FileWriter("forward_null_mrna_tmp.txt"));
	    double background = .083/35;
	    while (it.hasNext()) {
		String mRNAKey = (String)it.next();
		/*double intergenicPoissonRate = *///intergenicMRNADistribution(windowSize, mRNAKey, true, orchestra);
		//intergenicMRNADistribution(windowSize, mRNAKey, false, orchestra);
		mRNADistanceDistribution("nongenic", mRNAKey, true, orchestra);
		mRNADistanceDistribution("nongenic", mRNAKey, false, orchestra);
		TreeMap diffHistsPeakForward = new TreeMap(), diffHistsPeakReverse = new TreeMap();
		TreeMap diffHistsNullForward = new TreeMap(), diffHistsNullReverse = new TreeMap();
		for (int i=1; i<=21; i++) {
		    TreeMap rnaReadsMapForward = GeneratePeakmRNANullDist.readMRNAInfo(mRNAKey, true, i);
		    //diffHistsPeakForward = getDiffHist(peakPeaksData, rnaReadsMapForward, i, windowSize, diffHistsPeakForward, bwForwardPeak);
		    //diffHistsNullForward = getDiffHist(nullPeaksData, rnaReadsMapForward, i, windowSize, diffHistsNullForward, bwForwardNull);
		    TreeMap rnaReadsMapReverse = GeneratePeakmRNANullDist.readMRNAInfo(mRNAKey, false, i);
		    //diffHistsPeakReverse = getDiffHist(peakPeaksData, rnaReadsMapReverse, i, windowSize, diffHistsPeakReverse, null);
		    //diffHistsNullReverse = getDiffHist(nullPeaksData, rnaReadsMapReverse, i, windowSize, diffHistsNullReverse, null);
		    System.out.println("Found histograms for " + mRNAKey + " on chr " + i);
		}
		//int cutOffsForward[] = getFDRCutOffs(diffHistsNullForward, diffHistsPeakForward, fdr);
		//int cutOffsForward[] = getFDRCutOffs(diffHistsPeakForward, background*windowSize, fdr);
		//int cutOffsReverse[] = getFDRCutOffs(diffHistsNullReverse, diffHistsPeakReverse, fdr);
		//int cutOffsReverse[] = getFDRCutOffs(diffHistsPeakReverse, background*windowSize, fdr);
		//System.out.println(mRNAKey + " has cut-offs " + cutOffsForward[0] + " and " + cutOffsForward[1] + " forward and " + cutOffsReverse[0] + " and " + cutOffsReverse[1] + " reverse.");
		//break;
	    }
	    bwForwardPeak.close();
	    bwForwardNull.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static TreeMap getDiffHist(int peaksData[][], TreeMap rnaReadsMap, int chr, int windowSize, TreeMap diffHist, BufferedWriter bw) throws IOException {
	//Build a histogram of the differences. Make the bin sizes in
	//log2 scale.
	RefSeqGenes geneInfo = new RefSeqGenes(chr, false);
	for (int i=0; i<peaksData.length; i++) {
	    if (peaksData[i][0]==chr) {
		//boolean overlapsTSS = geneInfo.overlapsTSS(peaksData[i][1], peaksData[i][2]);
		//Use an int, where -1 represents negative strand, +1
		//positive strand and 0 not overlapping coding
		//region. If a TSS is overlapped, this counts as
		//0. Note that in some cases, a peak may overlap a TSS
		//on one strand and a coding region on the other.
		//int overlapsCodingRegion = geneInfo.overlapsCodingRegion(peaksData[i][1], peaksData[i][2]);
		//if (!(forward && overlapsCodingRegion==1)) {
		    int nLeft = GeneratePeakmRNANullDist.numberOfReads(rnaReadsMap, peaksData[i][1] - windowSize, peaksData[i][1]);
		    int nRight = GeneratePeakmRNANullDist.numberOfReads(rnaReadsMap, peaksData[i][2], peaksData[i][2] + windowSize);
		    int diff = nLeft - nRight;
		    Integer bin = new Integer((int)Math.floor(diff>=0 ? log2(1 + diff) : -log2(-diff + 1)));
		    int n = diffHist.containsKey(bin) ? ((Integer)diffHist.get(bin)).intValue() : 0;
		    diffHist.put(bin, new Integer(n + 1));
		    if (bw!=null)
			bw.write(chr + "\t" + peaksData[i][1] + "\t" + peaksData[i][2] + "\t" + nLeft + "\t" + nRight + "\n");
		    /*}
		else {
		    System.out.println("Invalid peak!");
		    nInvalid++;
		    }*/
	    }
	}
	return diffHist;
    }

    public static double log2(int n) { return log2((double)n); }

    public static double log2(double d) { return Math.log(d)/Math.log(2); }

    public static int[] getFDRCutOffs(TreeMap peakDiffHist, double poissonRate, double locFDR) {
	//Calculate the ratio which is used for the FDR
	//cut-off. Assume that the background is Poisson => difference
	//is Skellam.
	Iterator it = peakDiffHist.keySet().iterator(); //First count the total number of peaks
	int lowCutOff = Integer.MIN_VALUE, highCutOff = Integer.MAX_VALUE, nPeaks = 0;	
	while (it.hasNext()) 
	    nPeaks += ((Integer)peakDiffHist.get((Integer)it.next())).intValue();
	TreeMap localRatio = getLocalRatios(peakDiffHist, poissonRate, nPeaks);
	it = peakDiffHist.keySet().iterator(); 
	while (it.hasNext()) {
	    int cutOff = ((Integer)it.next()).intValue();
	    double locRatio = ((Double)localRatio.get(new Integer(cutOff))).doubleValue();
	    //System.out.println("Ratio is " + format.format(locRatio) + " for " + cutOff);
	    lowCutOff = cutOff<0 && locRatio<locFDR ? cutOff : lowCutOff;
	    highCutOff = cutOff>0 && locRatio<locFDR ? cutOff : highCutOff;
	}
	return new int[]{lowCutOff, highCutOff};
    }

    public static TreeMap getLocalRatios(TreeMap peakDiffHist, double poissonRate, int nPeaks) {
	TreeMap locRatios = new TreeMap();
	Iterator itPeak = peakDiffHist.keySet().iterator();
	System.out.println("Poisson rate = " + format.format(poissonRate));
	while (itPeak.hasNext()) {
	    int key = ((Integer)itPeak.next()).intValue();
	    double nullProb = 0; //Since we are using a log-scale, we need to sum over all the values that go into a given bin
	    if (Math.pow(2, Math.abs(key))/poissonRate<1025) {
		int start = 0, end = 1;
		if (key<0) {
		    start = -(int)Math.pow(2, -key + 1) + 1;
		    end = -(int)Math.pow(2, -key) + 1;
		}
		else if (key>0) {
		    start = (int)Math.pow(2, key - 1) + 1;
		    end = (int)Math.pow(2, key) + 1;
		}
		//System.out.println("key = " + key + ", start = " + start + " and end = " + end);
		for (int i=start; i<end; i++)
		    nullProb += PeakFinder2.skellamPDF(i, poissonRate, poissonRate);
	    }
	    double empiricalProb = (double)((Integer)peakDiffHist.get(new Integer(key))).intValue()/nPeaks;
	    locRatios.put(new Integer(key), new Double(nullProb/empiricalProb));
	    System.out.println(format.format(nullProb/empiricalProb) + ", found " + format.format(nullProb) + " null prob and " + format.format(empiricalProb) + " empirical prob for " + key + " with " + ((Integer)peakDiffHist.get(new Integer(key))).intValue() + " peaks in the bin");
	}
	return locRatios;
    }

    //////////////////////////////////////////////////////////////////

    public static int[] getFDRCutOffs(TreeMap nullDiffHist, TreeMap peakDiffHist, double locFDR) {
	//Calculate the ratio which is used for the FDR cut-off
	TreeMap localRatio = getLocalRatios(nullDiffHist, peakDiffHist);
	int lowCutOff = Integer.MIN_VALUE, highCutOff = Integer.MAX_VALUE;
	Iterator it = localRatio.keySet().iterator();
	while (it.hasNext()) {
	    int cutOff = ((Integer)it.next()).intValue();
	    double locRatio = ((Double)localRatio.get(new Integer(cutOff))).doubleValue();
	    //System.out.println("Ratio is " + format.format(locRatio) + " for " + cutOff);
	    lowCutOff = cutOff<0 && locRatio<locFDR ? cutOff : lowCutOff;
	    highCutOff = cutOff>0 && locRatio<locFDR ? cutOff : highCutOff;
	}
	return new int[]{lowCutOff, highCutOff};
    }

    public static TreeMap getLocalRatios(TreeMap nullDiffHist, TreeMap peakDiffHist) {
	TreeMap locRatios = new TreeMap();
	Iterator itNull = nullDiffHist.keySet().iterator();
	int nNullPeaks = 0, nPeaks = 0;
	while (itNull.hasNext()) {
	    Integer key = (Integer)itNull.next();
	    if (peakDiffHist.containsKey(key)) {
		locRatios.put(key, new Double((double)((Integer)nullDiffHist.get(key)).intValue()/((Integer)peakDiffHist.get(key)).intValue()));
		System.out.println("Found " + ((Integer)nullDiffHist.get(key)) + " null peaks and " + ((Integer)peakDiffHist.get(key)) + " true peaks for " + key);
		nNullPeaks += ((Integer)nullDiffHist.get(key)).intValue();
		nPeaks += ((Integer)peakDiffHist.get(key)).intValue();
	    }
	    else
		System.err.println("Error, no peakDiffs found for " + key.toString());
	}
	Iterator itPeak = peakDiffHist.keySet().iterator();
	while (itPeak.hasNext()) {
	    Integer key = (Integer)itPeak.next();
	    if (!locRatios.containsKey(key)) {
		locRatios.put(key, new Double(0));
		nPeaks += ((Integer)peakDiffHist.get(key)).intValue();
	    }
	}
	System.out.println("Found " + nNullPeaks + " null peaks and " + nPeaks + " real ones.");
	return locRatios;
    }

    public static int[][] loadPeakData(String fileName) {
	//Returns an array containing the chromosome, start and end for each peak
	int peaks[][] = null;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern peakPattern = Pattern.compile("^(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    int i =0, nPeaks = 0;
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		if (m.find()) { nPeaks++; }
		line = br.readLine();
	    }
	    br.close();
	    peaks = new int[nPeaks][3];
	    br = new BufferedReader(new FileReader(fileName));
	    line = br.readLine();
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		if (m.find()) { 
		    peaks[i][0] = Integer.parseInt(m.group(1));
		    peaks[i][1] = Integer.parseInt(m.group(2));
		    peaks[i][2] = Integer.parseInt(m.group(3));
		    i++;
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading peaks from file " + fileName);
	}
	return peaks;
    }

    //////////////////////////////////////////////////////////////////

    public static TreeMap intergenicMRNADistribution(int binSize, String mRNAKey, boolean forward, boolean orchestra) {
	//Use this method to find the distribution of intergenic
	//mRNA. Consider non-overlapping bins which are not
	//overlapping an annotated gene.
	String fileStart = mRNAKey.substring(mRNAKey.lastIndexOf("/") + 1);
	String saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/NullDistributions/" + fileStart + "_nongenic_" + (forward ? "f" : "r") + "_" + binSize + ".txt"; //Try to open file with results
	TreeMap hist = new TreeMap();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(saveFileName));
	    String line = br.readLine();
	    Pattern pattern = Pattern.compile("(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = pattern.matcher(line);
		if (m.find()) 
		    hist.put(new Integer(m.group(1)), new Integer(m.group(2)));
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Could not find file with results, recomputing....");
	    for (int j=1; j<=21; j++) {
		RefSeqGenes geneInfo = new RefSeqGenes(j, orchestra);

		TreeMap mRNA = GeneratePeakmRNANullDist.readMRNAInfo(mRNAKey, forward, j);
		int nReads = 0, nBins = 0;
		//Find all the intergenic regions
		for (int i=0; i<PeakFinder2.CHROMOSOME_LENGTH[j - 1]/binSize; i++) {
		    int start = i*binSize, end = (i+1)*binSize;
		    if (!geneInfo.overlapsGene(start, end)) { //intergenic, this is a very slow and stupid method for checking that we are not overlapping a gene.
			Integer nRNA = new Integer(GeneratePeakmRNANullDist.numberOfReads(mRNA, start, end));
			int n = hist.containsKey(nRNA) ? ((Integer)hist.get(nRNA)).intValue() : 0;
			hist.put(nRNA, new Integer(n + 1));
			nBins++;
		    }
		}
	    }
	    System.err.println("Saving results to " + saveFileName);
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		Iterator it = hist.keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    bw.write(key.toString() + "\t" + ((Integer)hist.get(key)).toString() + "\n");
		}
		bw.close();
	    }
	    catch (IOException ex) {
		System.err.println("Error saving results to " + saveFileName);
	    }
	}
	return hist;
    }

    public static TreeMap<Integer, Integer> mRNADistanceDistribution(String type, String mRNAKey, boolean forward, boolean orchestra) {
	//Use this method to calculate the
	String fileStart = mRNAKey.substring(mRNAKey.lastIndexOf("/") + 1);
	String saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/NullDistributions/" + fileStart + "_" + type + "_distances_" + (forward ? "f" : "r") + ".txt"; //Try to open file with results
	TreeMap<Integer, Integer> minDistHist = new TreeMap<Integer, Integer>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(saveFileName));
	    String line = br.readLine();
	    Pattern pattern = Pattern.compile("(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = pattern.matcher(line);
		if (m.find()) 
		    minDistHist.put(new Integer(m.group(1)), new Integer(m.group(2)));
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Could not find file with results for distances, recomputing....");
	    for (int j=1; j<=21; j++) {
		RefSeqGenes geneInfo = new RefSeqGenes(j, orchestra);

		TreeMap mRNA = GeneratePeakmRNANullDist.readMRNAInfo(mRNAKey, forward, j);
		System.out.println("Found results for chr " + j);
		int nReads = 0, nBins = 0;
		if (type.equals("nongenic")) {
		    //Find all reads in non-genic regions. We find the
		    //regions where there are no genes and study the
		    //subset of of reads which are in these regions.
		    int nonGenicStart = 0, nonGenicEnd = 0;
		    Iterator it = geneInfo.getGenes().iterator();
		    Iterator rnaIt = mRNA.keySet().iterator();
		    RefSeqGene g = null;
		    while (it.hasNext()) {
			g = (RefSeqGene)it.next();
			if (g.getStrand()==forward) {
			    nonGenicEnd = forward ? g.getStart() : g.getEnd();
			    break;
			}
		    }
		    //Get all the reads in the non-genic region
		    //and find their distances
		    int prevMRNAStart = ((Integer)rnaIt.next()).intValue();
		    System.out.println("First gene starts at " + nonGenicEnd + " and first read at " + prevMRNAStart);
		    int mRNAStart = ((Integer)rnaIt.next()).intValue();
		    while (rnaIt.hasNext()) {
			int nextMRNAStart = ((Integer)rnaIt.next()).intValue();
			if (prevMRNAStart>nonGenicStart && nextMRNAStart<nonGenicEnd) {
			    int minDist = Math.min(mRNAStart - prevMRNAStart, nextMRNAStart - mRNAStart);
			    System.out.println("Found min dist " + minDist + " for " + mRNAStart);
			    minDistHist.put(minDist, 1 + (minDistHist.containsKey(minDist) ? minDistHist.get(minDist) : 0));
			}
			else if (g!=null) {
			    nonGenicStart = forward ? g.getEnd() : g.getStart();
			    while (it.hasNext()) {
				g = (RefSeqGene)it.next();
				if (g.getStrand()==forward)
				    nonGenicEnd = g.getStrand() ? g.getStart() : g.getEnd();
				else
				    g = null;
			    }
			    if (g==null) //reached last gene
				nonGenicEnd = PeakFinder2.CHROMOSOME_LENGTH[j - 1];
			    break;
			}
			prevMRNAStart = mRNAStart;
			mRNAStart = nextMRNAStart;
		    }
		}
	    }
	    System.err.println("Saving results to " + saveFileName);
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		Iterator it = minDistHist.keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    bw.write(key.toString() + "\t" + ((Integer)minDistHist.get(key)).toString() + "\n");
		}
		bw.close();
	    }
	    catch (IOException ex) {
		System.err.println("Error saving results to " + saveFileName);
	    }
	}
	return minDistHist;
    }

    public static int[][] findNonGenicRegions(int chr, boolean strands[], int tsss[], int tses[], boolean forward) {
	int nGenes = 0;
	for (int i=0; i<strands.length; i++) {
	    if (strands[i]==forward) { nGenes++; }
	}
	int genic[][] = new int[nGenes][2], j = 0;
	for (int i=0; i<strands.length; i++) {
	    if (strands[i]==forward) { 
		genic[j][0] = tsss[i];
		genic[j++][1] = tses[i];
	    }
	}
	int nonGenic[][] = new int[nGenes + 2][2];
	nonGenic[0][0] = 0;
	nonGenic[0][1] = genic[0][0];
	for (int i=0; i<genic.length-1; i++) {
	    nonGenic[i+1][0] = genic[i][1] + 1;
	    nonGenic[i+1][1] = genic[i+1][0];
	}
	nonGenic[nonGenic.length-1][0] = genic[genic.length-1][1] + 1;
	nonGenic[nonGenic.length-1][1] = PeakFinder2.CHROMOSOME_LENGTH[chr-1];
	return nonGenic;
    }

}
