//Use this class to find which windows are significant for a given
//pair of thresholds (one for #ChIP-#input and one restricting the
//maximum #input). This method utilizes the output from
//CalculateCriticalWindowDifference.java and takes the histograms of
//the distributions as well as the list of windows
//generated. Unfortunately, the analysis cannot be done in Matlab
//directly since we run out of memory because of Matlab's crappy
//method for handling it.
//
//Example:
//
//bsub -q shared_2h java -Xmx512m FindOverlappingPeaks -n MmDataPeakFiles/fos_un_names.txt -o -p .001 -l 150 -g 4 -m 4 -w 240 -W 10 -c 19 -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Malik/un/B1/Input_Malik_un_B1


import jargs.gnu.CmdLineParser;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
//import umontreal.iro.lecuyer.stat.Tally;
//import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;
import org.apache.commons.math.distribution.BinomialDistributionImpl;
import org.apache.commons.math.stat.regression.SimpleRegression;

public class FindOverlappingPeaks {

    public static final String OVERLAP_SUFFIX = "_combos_";
    public static final String OVERLAP_SUFFIX_CENTER = "_combos_center_";
    public static final String CBP_DISTANCE_SUFFIX = "_cbp_distance_";
    public static final String OVERLAP_TSS_SUFFIX = "_combos_TSS_";
    public static final String OVERLAP_PROMOTER_SUFFIX = "_combos_promoter_";
    public static final String PEAKS_SUFFIX = "_peaks_";
    public static final String CODING_REGION_PEAKS_SUFFIX = "_peaks_coding_";
    public static final String SEQUENCE_SUFFIX = "_peak_sequences_";
    public static final String SATURATION_SUFFIX = "_peak_list_";
    public static final int MIN_TOTAL_READS = 5;
    public static final int READ_LENGTH = 35;

    public static final String INPUT_KCL_FILE_NAME = "/groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1";
    public static final String INPUT_UN_FILE_NAME = "/groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/un/B1/Input_Input_un_B1";
    public static final String INPUT_MNASE_FILE_NAME = "/groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/MNase/B1/Input_Input_MNase_B1";

    public static DecimalFormat format = new DecimalFormat("###.##");

    protected static int chr, windowStep;
    protected static Vector<String> tfNames = new Vector<String>(), longTFNames = new Vector(), mRNANames = new Vector();


    protected static double inputFDR = .001;

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the file containg
	    //the list of directories to be investigated
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    //Location of the directories for factors which are not
	    //assumed to be peaked
	    CmdLineParser.Option longFileNameFlag = argParser.addStringOption('N', "longFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Decide what file to write the results to
	    CmdLineParser.Option saveFileFlag = argParser.addStringOption('f', "saveFile");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('p', "fdr");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option windowStepFlag = argParser.addIntegerOption('W', "windowStep");
	    CmdLineParser.Option mappabilityBinsFlag = argParser.addIntegerOption('m', "mappabilityBins");
	    CmdLineParser.Option gcBinsFlag = argParser.addIntegerOption('g', "GCBins");
	    CmdLineParser.Option nUpstreamFlag = argParser.addIntegerOption('U', "nUpstream");
	    CmdLineParser.Option nDownstreamFlag = argParser.addIntegerOption('d', "nDownstream");
	    CmdLineParser.Option qualityScoreFlag = argParser.addDoubleOption('q', "qualityScore");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('M', "mismatches");
	    //Use this flag for the bins of the empirical bayes dist
	    CmdLineParser.Option binWidthFlag = argParser.addDoubleOption('b', "binWidth");
	    //Use this flag to include the mRNA data
	    CmdLineParser.Option mRNAFlag = argParser.addStringOption("mRNA");
	    //Use when we want to include conservation info
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption('C', "conservation");
	    //Use this flag to extract the DNA-sequences for each peak
	    CmdLineParser.Option sequenceFlag = argParser.addBooleanOption('s', "sequence");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //Use this flag if we want to use the other condition (KCl/un) as input instead of the normal input
	    CmdLineParser.Option conditionInputFlag = argParser.addBooleanOption('I', "conditionInput");
	    //Use this to provide a list of crappy regions that one wants to exclude from the analysis
	    CmdLineParser.Option excludeLociFlag = argParser.addStringOption("exclude");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    String GENOME_LOC = "./";
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? tmp.toString() : "Mm";
	    int nChr = species.equals("Mm") ? 21 : 24;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(longFileNameFlag);
	    String longFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileFlag);
	    String saveFileStart = tmp!=null && !tmp.toString().equals("") ? tmp.toString() : fileName;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 240;
	    tmp = argParser.getOptionValue(windowStepFlag);
	    windowStep = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(gcBinsFlag);
	    int nGCBins = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(mappabilityBinsFlag);
	    int nMappabilityBins = tmp!=null ? ((Integer)tmp).intValue() : 6;
	    tmp = argParser.getOptionValue(nUpstreamFlag);
	    int nUpstream = tmp!=null ? ((Integer)tmp).intValue() : 8000;
	    tmp = argParser.getOptionValue(nDownstreamFlag);
	    int nDownstream = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    tmp = argParser.getOptionValue(qualityScoreFlag);
	    double qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : 10;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    int nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(binWidthFlag);
	    double binWidth = tmp!=null ? ((Integer)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(mRNAFlag);
	    String mRNAFileName = (tmp!=null && !tmp.equals("")) ? (String)tmp : null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean sequence = argParser.getOptionValue(sequenceFlag)!=null;
	    boolean conditionInput = argParser.getOptionValue(conditionInputFlag)!=null;
	    tmp = argParser.getOptionValue(excludeLociFlag);
	    String excludeLociFileName = (tmp!=null && !tmp.equals("")) ? (String)tmp : null;
	    //////////////////////////////////////////////////////////////////
	    Vector<String> fileNames = getFileNames(fileName, tfNames);
	    Collections.sort(tfNames);
	    Vector longFileNames = getFileNames(longFileName, longTFNames);
	    Collections.sort(longTFNames);
	    Vector mRNAFileNames = getFileNames(mRNAFileName, mRNANames);
	    Collections.sort(mRNANames);
	    Vector<int[]> excludeLoci = excludeLociFileName==null ? null : parseExcludeLociFile(excludeLociFileName);
	    //use this map to keep track of sites with key number of
	    //TFBS, regardless of which combination is present.
	    TreeMap<String, Vector<Window>> wins = new TreeMap<String, Vector<Window>>();
	    //Use this for a list of WindowFinders when we want to go
	    //through again
	    TreeMap<String, WindowFinder> wfs = new TreeMap<String, WindowFinder>();
	    TreeMap bws = new TreeMap(), longWFs = new TreeMap(), longBWs = new TreeMap(), longCutOffs = new TreeMap();
	    int nTFs = fileNames.size();
	    Iterator it = fileNames.iterator();
	    TreeMap<String, Double> normFactors = new TreeMap<String, Double>();
	    boolean kclFound = false, unFound = false, histFound = false;
	    //Super-duper ugly hack!!!
	    TreeMap<String, String> tf2InputMap = new TreeMap<String, String>();
	    if (inputFileName==null)
		inputFDR = -1;
	    ///////////////////////////////////////////////////////
	    while (it.hasNext()) {
		//Read the windows and pull out the ones which are
		//significant at the current FDR for each TF
		String fName = (String)it.next();
		String tfName = fName.endsWith(".bed") ? fName.substring(fName.lastIndexOf("/") + 1, fName.lastIndexOf(".")) : fName.substring(fName.lastIndexOf("/") + 1);
		kclFound = kclFound || tfName.contains("KCl");
		unFound = unFound || tfName.contains("un");
		histFound = histFound || tfName.startsWith("H\\d+K\\d+");
		if (conditionInput && fileName.contains("mecp2")) {
		    tf2InputMap.put(tfName, fName.replaceAll("2hr", "0hr"));
		    System.out.println("Adding " + tfName + " -> " + tf2InputMap.get(tfName));
		    normFactors.put(tf2InputMap.get(tfName), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/FindReadsAcrossFeature2.calculateTotalReads(tf2InputMap.get(tfName), null, nChr, fragmentLength, qualityScore)); //countTotalReads(tf2InputMap.get(tfName), fragmentLength));
		}
		else if (fileName.contains("Mikkelsen")) {
		    tf2InputMap.put(tfName, "/groups/neuroduo/martin/ChIP-Seq/MikkelsenData/ES/Input/ES_Input");
		    System.out.println("Adding " + tfName + " -> " + tf2InputMap.get(tfName));
		    normFactors.put(tf2InputMap.get(tfName), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/countTotalReads(tf2InputMap.get(tfName), fragmentLength, nChr));
		}
		else if (fileName.endsWith(".bed")) {
		    tf2InputMap.put(tfName, inputFileName);
		    System.out.println("Adding " + tfName + " -> " + tf2InputMap.get(tfName));
		    normFactors.put(tf2InputMap.get(tfName), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/CreateWigFiles.getTotalNumberOfReads(tf2InputMap.get(tfName), nChr, fragmentLength)); //countTotalReads(tf2InputMap.get(tfName), fragmentLength));
		}
		else {
		    tf2InputMap.put(tfName, inputFileName);
		    System.out.println("Adding " + tfName + " -> " + tf2InputMap.get(tfName));
		}
		//Calculate the total number of reads to use for normalization later
		normFactors.put(tfName, (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/CreateWigFiles.getTotalNumberOfReads(fName, nChr, fragmentLength)); //FindReadsAcrossFeature2.calculateTotalReads(fName, null, nChr, fragmentLength, qualityScore)); //countTotalReads(fName, fragmentLength));
		WindowFinder windowFinder = new WindowFinder(fName, true, windowSize, windowStep, nMappabilityBins, nGCBins);
		Vector merged = windowFinder.findSignificantWindows(chr, fragmentLength, new double[]{fdr}, inputFDR, !species.equals("Mm"), true)[0];
		System.out.println("Found " + merged.size() + " peaks for " + tfName);
		//Save the list of peaks so that we can calculate the saturation
		saveSaturationPeakList(fName, chr, merged);
		int nReads = windowFinder.numberOfReads();
		//System.out.println("Found " + merged.size() + " peaks for " + fName);
		wins.put(tfName, merged);
		wfs.put(tfName, new WindowFinder(fName, chr, fragmentLength, true, windowSize, windowStep, nMappabilityBins, nGCBins, !species.equals("Mm")));
		BufferedWriter bw = new BufferedWriter(new FileWriter(fName + WindowFinder.PEAKS_SUFFIX_NON_SIGNIFICANT + chr + ".bed"));
		if (chr==1) { WindowFinder.writeBedFileHeader(bw, fName, false);}
		bws.put(tfName, bw);
	    }
	    if (inputFileName!=null) //Add the input reads
		normFactors.put(inputFileName, (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/CreateWigFiles.getTotalNumberOfReads(inputFileName, nChr, fragmentLength)); //FindReadsAcrossFeature2.calculateTotalReads(inputFileName, null, nChr, fragmentLength, qualityScore)); //countTotalReads(inputFileName, fragmentLength));
	    /*if (kclFound)
		normFactors.put(INPUT_KCL_FILE_NAME.substring(INPUT_KCL_FILE_NAME.lastIndexOf("/") + 1), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/countTotalReads(INPUT_KCL_FILE_NAME, fragmentLength));
	    if (unFound)
		normFactors.put(INPUT_UN_FILE_NAME.substring(INPUT_UN_FILE_NAME.lastIndexOf("/") + 1), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/countTotalReads(INPUT_UN_FILE_NAME, fragmentLength));
	    if (histFound)
		normFactors.put(INPUT_MNASE_FILE_NAME.substring(INPUT_MNASE_FILE_NAME.lastIndexOf("/") + 1), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/countTotalReads(INPUT_MNASE_FILE_NAME, fragmentLength));*/
	    for (String k : normFactors.keySet())
		System.out.println(k + "\t" + normFactors.get(k));
	    //Find the overlapping peaks
	    TreeSet<OverlapRegion> overlapRegions = findOverlappingRegions(wins);
	    //Now we need to go through the data again and find the
	    //significance level of the TFs which were not enriched at
	    //the sites of interest.
	    /*it = overlapRegions.iterator();
	    for (OverlapRegion reg : overlapRegions) {
		for (String tfName : wfs.keySet()) {
		    if (reg.readsEnrichment(tfName)==0) {
			Window win = ((WindowFinder)wfs.get(tfName)).findReadsInRegion(reg.getStart(), reg.getEnd(), (BufferedWriter)bws.get(tfName), false);
			if (win!=null)
			    reg.addTF(win, tfName, false);
		    }			    
		}
		}*/
	    Iterator tfNameIt = wfs.keySet().iterator();
	    while (tfNameIt.hasNext()) {
		String tfName = (String)tfNameIt.next();
		((BufferedWriter)bws.get(tfName)).close();
		((WindowFinder)wfs.get(tfName)).closeWindowReader();
	    }
	    //////////////////////////////////////////////////////////////////
	    //Find the number of mRNA reads in each locus
	    Iterator mRNANameIt = mRNAFileNames.iterator();
	    while (mRNANameIt.hasNext()) {
		String mName = (String)mRNANameIt.next();
		String mRNAName = mName.substring(mName.lastIndexOf("/") + 1);
		findMRNAReads(overlapRegions, mName, mRNAName);
		System.out.println("Found mRNA reads for " + mName);
	    }
	    //////////////////////////////////////////////////////////////////
	    if (conservation) { //Find the conservation scores for each peak
		findConservationScores(overlapRegions, orchestra);
		System.out.println("Found conservation scores.");
	    }
	    //////////////////////////////////////////////////////////////////
	    //Now we want to study the behavior of the spread out
	    //binding factors in the coding regions. Here we assume
	    //that the results have already been pre-computed. Also,
	    //we are currently not concerned with finding which are
	    //significantly enriched
	    RefSeqGenes refSeqGenes = new RefSeqGenes(chr, orchestra);
	    it = longFileNames.iterator();
	    //Use this data structure to keep track of the z-values
	    //associated with each gene for each TF.
	    TreeMap longTF2gene2z = new TreeMap();
	    while (it.hasNext()) {
		String fName = (String)it.next();
		String tfName = fName.substring(fName.lastIndexOf("/") + 1) + "_coding";
		BufferedWriter bw = new BufferedWriter(new FileWriter(fName + CODING_REGION_PEAKS_SUFFIX + chr + ".bed"));
		if (chr==1) { WindowFinder.writeCodingRegionsBedFileHeader(bw, tfName);}
		//use this object to determine which genes are significantly enriched
		//longCutOffs.put(tfName, new WiderPeakLocFDR(fName, .1, windowSize, windowStep, nMappabilityBins, nGCBins, orchestra));
		//load the z-values
		TreeMap gene2z = WiderPeakLocFDR.loadZValues(fName, refSeqGenes);
		//Save the z-values in the .bed file
		saveCodingRegionBedFile(bw, chr, gene2z);
		//Store so that we can obtain the value later when we want to write to file
		longTF2gene2z.put(fName, gene2z);
	    }
	    //////////////////////////////////////////////////////////////////
	    //Find the subset of peaks which are sitting right on top
	    //of an annotated promoter
	    /*TreeSet tssOverlapingRegions = new TreeSet();
	    BufferedWriter bwTSS  = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + OVERLAP_TSS_SUFFIX + chr + ".txt"));
	    BufferedWriter bwPromoter  = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + OVERLAP_PROMOTER_SUFFIX + chr + ".txt"));
	    int k = 0;
	    if (chr==1) { 
		writeHeaders(bwTSS, conservation, tfNames, longTFNames, mRNAFileNames, false); 
		writeHeaders(bwPromoter, conservation, tfNames, longTFNames, mRNAFileNames, true); 
	    }
	    it = refSeqGenes.getGenes().iterator();
	    while (it.hasNext()) {
		RefSeqGene gene = (RefSeqGene)it.next();
		boolean pol2PeakFound = false;
		//Find out if there is a window overlapping the TSS
		Vector tssOverlaps = findTSSOverlap(gene, overlapRegions, true);
		//Check the TSE
		Vector tseOverlaps = findTSSOverlap(gene, overlapRegions, false);
		boolean tfbsFound = false;
		//Find the subset of peaks which are in the promoter
		//region of an annotated promoter
		OverlapRegion promoterRegion = gene.getStrand() ? new OverlapRegion(chr, gene.getStart() - nUpstream, gene.getStart() + nDownstream, 0, tfNames, mRNANames) : new OverlapRegion(chr, gene.getStart() - nDownstream, 0, gene.getStart() + nUpstream, tfNames, mRNANames);
		//Count the total number of peaks for each TF in the region
		TreeMap nTFsMap = new TreeMap();
		if (overlapRegions.tailSet(promoterRegion)!=null && overlapRegions.tailSet(promoterRegion).size()>0) {
		    TreeSet promoterPeaks = new TreeSet(overlapRegions.tailSet(promoterRegion));
		    promoterPeaks.addAll(overlapRegions.headSet(promoterRegion));
		    Iterator it3 = tfNames.iterator();
		    while (it3.hasNext())  //Initialize
			nTFsMap.put((String)it3.next(), new Integer(0));
		    Iterator it2 = promoterPeaks.iterator();
		    while (it2.hasNext()) {
			OverlapRegion r = (OverlapRegion)it2.next();
			if (r.overlaps(promoterRegion)) {
			    it3 = tfNames.iterator();
			    //System.out.println("Found " + r.toString() + " for " + gene.getName());
			    while (it3.hasNext()) {
				String tfName = (String)it3.next();
				if (r.readsEnrichment(tfName)>0) {
				    nTFsMap.put(tfName, new Integer(((Integer)nTFsMap.get(tfName)).intValue() + 1));
				    tfbsFound = true;
				}
			    }
			}
		    }
		}
		//Save if there is Pol2 enrichment or binding
		if (pol2PeakFound || tfbsFound) {
		    bwPromoter.write(chr + "\t" + promoterRegion.getStart() + "\t" + promoterRegion.getEnd());
		    Iterator it3 = tfNames.iterator();
		    while (it3.hasNext()) 
			bwPromoter.write("\t" + ((Integer)nTFsMap.get((String)it3.next())).toString());
		    bwPromoter.write("\t" + gene.getName() + "\t" + gene.getID());
		    Iterator longIt = longTF2gene2z.keySet().iterator();
		    while (longIt.hasNext()) {
			String tfName = (String)longIt.next();
			TreeMap map = (TreeMap)longTF2gene2z.get(tfName);
			double z = map.containsKey(gene) ? ((Double)map.get(gene)).doubleValue() : 0;
			bwPromoter.write("\t" + format.format(z));
		    }
		    bwPromoter.write("\n");
		}
		//Save the TSS stuff
		Iterator tssIt = tssOverlaps.iterator();
		while (tssIt.hasNext()) {
		    OverlapRegion tssOverlap = (OverlapRegion)tssIt.next();
		    bwTSS.write(tssOverlap.toString() + "\t" + gene.getName() + "\t" + gene.getID());
		    Iterator longIt = longTF2gene2z.keySet().iterator();
		    while (longIt.hasNext()) {
			String tfName = (String)longIt.next();
			TreeMap map = (TreeMap)longTF2gene2z.get(tfName);
			double z = map.containsKey(gene) ? ((Double)map.get(gene)).doubleValue() : 0;
			bwTSS.write("\t" + format.format(z));
		    }
		    bwTSS.write("\t1\n");
		}
		Iterator tseIt = tseOverlaps.iterator();
		while (tseIt.hasNext()) {
		    OverlapRegion tseOverlap = (OverlapRegion)tseIt.next();
		    bwTSS.write(tseOverlap.toString() + "\t" + gene.getName() + "\t" + gene.getID());
		    Iterator longIt = longTF2gene2z.keySet().iterator();
		    while (longIt.hasNext()) {
			String tfName = (String)longIt.next();
			TreeMap map = (TreeMap)longTF2gene2z.get(tfName);
			double z = map.containsKey(gene) ? ((Double)map.get(gene)).doubleValue() : 0;
			bwTSS.write("\t" + format.format(z));
		    }
		    bwTSS.write("\t0\n");
		}
	    }
	    bwPromoter.close();
	    bwTSS.close();*/
	    //Find the peaks which overlap with starting mRNA reads as
	    //defined by the ChIP-Seq


	    //save a list of all locations with mulitple TFs, use one
	    //file for each combination which was present
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + OVERLAP_SUFFIX + chr + ".txt"));
	    if (chr==1) 
		writeHeaders(bw, conservation, tfNames, null, mRNANames, true);
	    it = overlapRegions.iterator();
	    while (it.hasNext()) {
		OverlapRegion r = (OverlapRegion)it.next();
		if (excludeLoci==null || !r.overlapsLocus(excludeLoci))
		    bw.write(r.toString(normFactors, tf2InputMap) + "\n");
	    }
	    bw.close();
	    ////////////////////////////////////////////
	    //Save with peak center info as well
	    bw = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + OVERLAP_SUFFIX_CENTER + chr + ".txt"));
	    if (chr==1) 
		writeHeaders(bw, conservation, tfNames, null, mRNANames, true, true);
	    it = overlapRegions.iterator();
	    while (it.hasNext()) {
		OverlapRegion r = (OverlapRegion)it.next();
		if (excludeLoci==null || !r.overlapsLocus(excludeLoci))
		    bw.write(r.toString(normFactors, tf2InputMap) + "\n");
	    }
	    bw.close();
	    //Also store a more succinct format
	    int[] nTFBS = new int[tfNames.size()+1];
	    Arrays.fill(nTFBS, 0);
	    it = overlapRegions.iterator();
	    while (it.hasNext()) {
		OverlapRegion r = (OverlapRegion)it.next();
		nTFBS[r.numberOfTFs()]++;
	    }
	    System.out.println("Distribution of the number of overlapping TFBS:\n#TFBS\t#sites");
	    for (int i=0; i<=tfNames.size(); i++) 
		System.out.println((i) + "\t" + nTFBS[i]);
	    if (sequence) { //Do the sequences last since it takes a long time
		savePeakSequences(overlapRegions, fileName.substring(0, fileName.lastIndexOf(".")), orchestra);
		System.out.println("Found peak sequences.");
	    }
	    ////////////////////////////////////////////
	    //Check the distance for each TF to the CBP peak, do this
	    //only for the KCl case.
	    /*LinkedHashMap<String, Vector<Integer>> peakDists = new LinkedHashMap<String, Vector<Integer>>();
	    peakDists.put("pos", new Vector<Integer>());
	    String cbpName = null;
	    for (String name : tfNames) {
		if (name.contains("KCl") && name.contains("Millipore") && name.contains("CBP")) 
		    cbpName = name;
		else if (!name.contains("Millipore") && !name.contains("CBP")) 
		    peakDists.put(name, new Vector<Integer>());
	    }
	    for (OverlapRegion r : overlapRegions) {
		for (String name : tfNames) {
		    if (peakDists.containsKey(name)) {
			if (r.readsEnrichment(cbpName)>0 && r.readsEnrichment(name)>0) 
			    peakDists.get(name).addElement(r.getTFBSDistance(cbpName, name));
			else
			    peakDists.get(name).addElement(-1);
		    }
		}
		peakDists.get("pos").addElement(r.getCenter());
	    }
	    bw = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + CBP_DISTANCE_SUFFIX + chr + ".txt"));
	    if (chr==1) {
		bw.write("#chr\t");
		for (String name : peakDists.keySet()) {
		    if (peakDists.containsKey(name)) 
			bw.write(name + "\t");
		}
		bw.write("\n");
	    }
	    for (int i=0; i<overlapRegions.size(); i++) {
		bw.write(chr + "\t");
		for (String name : peakDists.keySet())
		    bw.write(peakDists.get(name).elementAt(i) + "\t");
		bw.write("\n");
	    }
	    bw.close();*/
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    //////////////////////////////////////////////////////////////////

    protected static void saveSaturationPeakList(String fName, int chr, Vector merged) {
	//Use this method to save the peaks in a file format which makes it easy to apply the saturation analysis algorithm
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fName + SATURATION_SUFFIX + chr + ".txt"));
	    Iterator it = merged.iterator();
	    while(it.hasNext()) {
		Window w = (Window)it.next();
		bw.write(w.toString() + "\t" + format.format(w.getProportionAboveLimit()) + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving saturation file " + fName);
	}
    }
    
    protected static void savePeakSequences(TreeSet overlapRegions, String fileStart, boolean orchestra) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileStart + SEQUENCE_SUFFIX + chr + ".fa"));
	    FastaSequence2 fs = new FastaSequence2(chr, orchestra);
	    Iterator it = overlapRegions.iterator();
	    while (it.hasNext()) {
		OverlapRegion r = (OverlapRegion)it.next();
		r.setSequence(fs.getSequence(r.getStart(), r.getLength()));
		bw.write(">chr" + chr + ":" + r.getStart() + "-" + r.getEnd() + "\n" + r.getSequence() + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving " + fileStart + ", " + e.toString());
	}
	
    }

    protected static void findConservationScores(TreeSet overlapRegions, boolean orchestra) {
	int longestRegion = 0;
	Iterator it = overlapRegions.iterator();
	while (it.hasNext()) {
	    OverlapRegion r = (OverlapRegion)it.next();
	    longestRegion = Math.max(longestRegion, r.getLength());
	}
	PhastConsParser pspv = new PhastConsParser(chr, true, orchestra, longestRegion);
	PhastConsParser pspe = new PhastConsParser(chr, false, orchestra, longestRegion);;
	it = overlapRegions.iterator();
	while (it.hasNext()) {
	    OverlapRegion r = (OverlapRegion)it.next();
	    r.setConservationScores(pspv.findScores(r.getStart(), r.getEnd()), true);
	    //System.out.println("Found conservation " + r.meanConservationScore(true) + " for " + r.getStart() + "-" + r.getEnd());
	    r.setConservationScores(pspe.findScores(r.getStart(), r.getEnd()), false);
	    //System.out.println("Found conservation " + r.meanConservationScore(false) + " for " + r.getStart() + "-" + r.getEnd());
	}
    }

    public static void saveCodingRegionBedFile(BufferedWriter bw, int chr, TreeMap gene2z) throws IOException {
	//Problems may occur where there are overlapping genes. In
	//that case, we will simply start the one which comes last
	//where the overlap stops.
	String c = "chr" + chr;
	if (chr==20) { c = "chrX"; }
	else if (chr==21) { c = "chrY"; }
	int prevEnd = 0;
	Iterator it = gene2z.keySet().iterator();
	while (it.hasNext()) {
	    RefSeqGene gene = (RefSeqGene)it.next();
	    int start = gene.getStrand() ? Math.max(gene.getStart(), prevEnd) : Math.max(gene.getEnd(), prevEnd);
	    int end = gene.getStrand() ? gene.getEnd() : gene.getStart();
	    if (end>prevEnd) {
		bw.write(c + "\t" + start + "\t" + end + "\t" + format.format(((Double)gene2z.get(gene)).doubleValue()) + "\n");
		prevEnd = end;
	    }
	}
	bw.close();
    }

    public static TreeMap<String, String> getFileNames(String fileName) throws IOException {
	Vector shortNames = new Vector(), fileNames = getFileNames(fileName, shortNames);
	TreeMap<String, String> map = new TreeMap<String, String>();
	for (int i=0; i<fileNames.size(); i++)
	    map.put((String)fileNames.elementAt(i), (String)shortNames.elementAt(i));
	return map;
    }	

    public static Vector getFileNames(String fileName, Vector shortNames) throws IOException {
	//Use this to parse the list containing the directories which
	//should be compared.
	Vector fileNames = new Vector();
	if (fileName!=null) {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    while (line!=null) {
		if (line!=null && line.length()>0 && !line.startsWith("#")) {
		    fileNames.addElement(line);
		    if (shortNames!=null) {
			String sName = line.contains(".") ? line.substring(line.lastIndexOf("/") + 1, line.lastIndexOf(".")) : line.substring(line.lastIndexOf("/") + 1);
			shortNames.add(sName);
		    }
		    //System.out.println(shortNames.lastElement().toString());
		}
		line = br.readLine();
	    }
	    br.close();
	}
	return fileNames;
    }

    protected static void writeHeaders(BufferedWriter bw, boolean conservation, Vector tfNames, Vector longTFNames, Vector mRNANames, boolean tss) throws IOException { writeHeaders(bw, conservation, tfNames, longTFNames, mRNANames, tss, false); }

    protected static void writeHeaders(BufferedWriter bw, boolean conservation, Vector tfNames, Vector longTFNames, Vector mRNANames, boolean tss, boolean centers) throws IOException {
	bw.write("#chr\tStart\tEnd\tGC");
	if (conservation)
	    bw.write("\tMean cons 11\tStd dev cons 11\tMax cons 11\tMean cons 30\tStd dev cons 30\tMax cons 30");
	Iterator it = tfNames.iterator();
	while (it.hasNext()) {
	    String tf = (String)it.next();
	    bw.write("\t" + tf + (centers ? "\t" + tf + "_center" : ""));
	}
	if (mRNANames!=null) {
	    it = mRNANames.iterator();
	    while (it.hasNext()) {
		String tf = (String)it.next();
		bw.write("\t" + tf + OverlapRegion.FORWARD_SUFFIX);
		bw.write("\t" + tf + OverlapRegion.REVERSE_SUFFIX);
	    }
	}
	if (longTFNames!=null) {
	    bw.write("\tGene\tGeneID");
	    it = longTFNames.iterator();
	    while (it.hasNext()) {
		String tf = (String)it.next();
		bw.write("\t" + tf);
	    }
	    if (tss)
		bw.write("\tTSS/TSE\n");
	    else
		bw.write("\n");
	}
	else 
	    bw.write("\n");
    }


    public static TreeSet<OverlapRegion> findOverlappingRegions(TreeMap<String, Vector<Window>> wins) {
	//Use this method to loop through all the windows and find the
	//ones which are phyiscally overlapping. These are merged.
	TreeSet<OverlapRegion> overlapRegions = new TreeSet<OverlapRegion>();
	Iterator it = tfNames.iterator();
	while (it.hasNext()) {
	    String fName = (String)it.next();
	    System.out.println(wins.keySet().toString() + "\t" + fName);
	    for (Window win : wins.get(fName)) {
		//store them in the map
		OverlapRegion r = new OverlapRegion(win, fName, tfNames, mRNANames);
		boolean merged = false;
		TreeSet<OverlapRegion> overlapping = win.findOverlappingRegions(overlapRegions);
		for (OverlapRegion o : overlapping) {
		    r.merge(o);
		    overlapRegions.remove(o);
		}
		overlapRegions.add(r);
	    }
	}
	it = overlapRegions.iterator();
	TreeMap<String, Integer> nPeaksWithTF = new TreeMap<String, Integer>();
	while (it.hasNext()) {
	    OverlapRegion r = (OverlapRegion)it.next();
	    Iterator it2 = tfNames.iterator();
	    while (it2.hasNext()) {
		String tfName = (String)it2.next();
		if (r.readsEnrichment(tfName)>0)
		    nPeaksWithTF.put(tfName, nPeaksWithTF.containsKey(tfName) ? nPeaksWithTF.get(tfName) + 1 : 1);
	    }
	}
	for (String tfName : nPeaksWithTF.keySet())
	    System.out.println(tfName + " has " + nPeaksWithTF.get(tfName));
	return overlapRegions;
    }

    public static Vector findTSSOverlap(RefSeqGene gene, TreeSet overlapRegions, boolean tss) {
	Vector tssOverlaps = new Vector();
	for (int i=0; i<(tss ? gene.numberOfTSS() : gene.numberOfTSE()); i++) {
	    OverlapRegion tmp2 = tss ? new OverlapRegion(chr, gene.getStart(i), gene.getStart(i), 0, tfNames, mRNANames) : new OverlapRegion(chr, gene.getEnd(i), gene.getEnd(i), 0, tfNames, mRNANames);
	    if (overlapRegions.tailSet(tmp2)!=null && overlapRegions.tailSet(tmp2).size()>0) {
		OverlapRegion r = (OverlapRegion)overlapRegions.tailSet(tmp2).first();
		if (r!=null && r.overlaps(tmp2)) 
		    tssOverlaps.addElement(r);
	    }
	}
	return tssOverlaps;
    }

    protected static void findMRNAReads(TreeSet overlapRegions, String mRNAFileName, String mRNAName) {
	try {
	    //Go through the list of all transcripts and list of regions
	    Iterator it = overlapRegions.iterator();
	    if (it.hasNext()) {
		OverlapRegion r = (OverlapRegion)it.next();
		BufferedReader br = new BufferedReader(new FileReader(mRNAFileName + ".tsv"));
		Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([R|F])\\s+(\\d+)");
		String line = br.readLine();
		while (line!=null) {
		    Matcher m = mRNAPattern.matcher(line);
		    if (m.find()) {
			int c = Integer.parseInt(m.group(1));
			if (c>chr) { break; }
			else if (c==chr) {
			    int start = Integer.parseInt(m.group(2));
			    boolean strand = m.group(3).equals("F"), bre = false;
			    int count = Integer.parseInt(m.group(4));
			    //Check if the read falls under the region r
			    while (r.getEnd()<start) {
				if (it.hasNext()) { r = (OverlapRegion)it.next(); }
				else { bre = true; break; }
			    }
			    if (bre) { break; }
			    if (r.overlaps(start))
				r.addMRNAReads(mRNAName, count, strand);
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	}
	catch (IOException e) {
	    
	}
    }

    public static int countTotalReads(String fileName, int fragmentLength, int nChr) {
	//Get all the reads
	int n = 0;
	double qualityScore = 10;
	for (int i=1; i<=nChr; i++) {
	    ABIRead[] reads = null;
	    if (fileName.contains("MikkelsenData")) 
		reads = PeakFinder2.parseBroadFiles(i, fileName, fragmentLength, true);
	    else if (fileName.endsWith(".bed")) 
		reads = CreateWigFiles.parseBedFiles(i, fileName, true, fragmentLength);
	    else
		reads = PeakFinder2.parseFiles(i, fileName, fragmentLength, true, 0, qualityScore);
	    if (reads!=null)
		n += reads.length;
	}
	System.out.println("Found " + n + " reads for " + fileName);
	return n;
    }

    public static Vector<int[]> parseExcludeLociFile(String excludeLociFileName) {
	//Assume three columns with chr start end
	Vector<int[]> excludeLoci = new Vector<int[]>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(excludeLociFileName));
	    String line = br.readLine();
	    while (line!=null) {
		if (!line.startsWith("#")) {
		    String tokens[] = line.split("\\s+");
		    int locus[] = new int[3];
		    for (int i=0; i<locus.length; i++)
			locus[i] = Integer.parseInt(tokens[i]);
		    excludeLoci.addElement(locus);
		}
		line = br.readLine();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error parsing exlude loci file " + e.toString());
	}
	return excludeLoci;
    }


}
    

