import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class GenerateRandomReads{

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    //c, o, r, t
	    //For the mef2 input file r = 8922641
	    CmdLineParser argParser = new CmdLineParser();
	    //How many random reads to generate
	    CmdLineParser.Option randomReadsFlag = argParser.addIntegerOption('r', "random");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag when generating multiple files which will
	    //later be merged
	    CmdLineParser.Option countFlag = argParser.addIntegerOption('c', "count");
	    //Total number of reads, for naming the directory and files
	    CmdLineParser.Option totalFlag = argParser.addIntegerOption('t', "totalNumber");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(randomReadsFlag);
	    int nRandomReads = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(countFlag);
	    int count = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(totalFlag);
	    int total = tmp!=null ? ((Integer)tmp).intValue() : nRandomReads;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    //////////////////////////////////////////////////////////////////
	    RandomReadsGenerator rrg = new RandomReadsGenerator(orchestra);
	    rrg.generateRandomReads(nRandomReads, count, total);
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()
	
}
    
class RandomReadsGenerator {
    
    protected Random random = new Random();
    protected boolean orchestra;
    protected long totChrLen = 0;

    public RandomReadsGenerator(boolean o) { 
	orchestra = o;
	for (int i=0; i<PeakFinder.CHROMOSOME_LENGTH.length; i++) { totChrLen += PeakFinder.CHROMOSOME_LENGTH[i]; }
    }

    protected void generateRandomReads(int nRandomReads, int count, int total) {
	//Generate randomized reads. Save using the color code so that
	//they can be analyzed using the software for mapping reads.
	try {
	    String dirName = "/groups/neuroduo/martin/Randomized" + total;
	    File dir = new File(dirName);
	    if (orchestra && !dir.exists()) { //Create a new directory
		Runtime runTime = Runtime.getRuntime();
		runTime.exec("mkdir /groups/neuroduo/martin/Randomized" + total);
	    }
	    String indStr = count>0 ? String.valueOf(count) : "";
	    String fileName = orchestra ? "/groups/neuroduo/martin/Randomized" + total + "/rand_" + total + "_reads" + indStr + ".csfasta" : "Randomized/rand_" + total + "_reads" + indStr + ".csfasta";
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    int readsLeft = nRandomReads;
	    while (readsLeft>0) {
		int nSaved = generateRandomReads(bw, readsLeft, totChrLen);
		readsLeft = readsLeft - nSaved;
		System.out.println("There are " + readsLeft + " sequences left.");
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }

    protected int generateRandomReads(BufferedWriter bw, int readsLeft, long totChrLen) {
	int nRepeated = 0, nSaved = 0;
	Random random = new Random();
	long r[] = new long[readsLeft];
	for (int i=0; i<readsLeft; i++) { r[i] = (long)Math.abs(random.nextLong()) % totChrLen; }
	Arrays.sort(r);
	try {
	    String GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	    for (int j=0; j<PeakFinder.CHROMOSOME_LENGTH.length; j++) {
		FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(j + 1));
		int nChr = 0, k = 0;
		for (int i=0; i<readsLeft; i++) {
		    if (r[i]<PeakFinder.CHROMOSOME_LENGTH[j] && r[i]>0) { nChr++; }
		}
		if (nChr>0) {
		    long rChr[] = new long[nChr];
		    for (int i=0; i<readsLeft; i++) {
			if (r[i]<PeakFinder.CHROMOSOME_LENGTH[j] && r[i]>0) { rChr[k++] = r[i]; }
		    }
		    String seqsChr[] = fastaSeq.getSequences(rChr, AnalyzeControls.READ_LENGTH, false);
		    for (int i=0; i<seqsChr.length; i++) {
			if (seqsChr[i]!=null && !containsRepeatElement(seqsChr[i])) { //check if there were wildcards
			    boolean strand = random.nextBoolean();
			    if (!strand) { //Make it a reverse read
				seqsChr[i] = reverseComplement(seqsChr[i]);
				rChr[i] = rChr[i] - 1 + AnalyzeControls.READ_LENGTH;
			    }
			    bw.write(">" + (j+1) + "_" + (strand ? "" : "-") + (rChr[i] - 1) + "\nT" + FastaSequence2.getColorCode("T" + seqsChr[i].toUpperCase()) + "\n");
			    nSaved++;
			}
			else
			    nRepeated++;
		    }
		}
		for (int i=0; i<readsLeft; i++) 
		    r[i] -= PeakFinder.CHROMOSOME_LENGTH[j];
	    }
	}
	catch (IOException e) { }
	return nSaved;
    }

    public static String reverseComplement(String seq) {
	char chars[] = seq.toCharArray(), rc[] = new char[chars.length];
	for (int i=chars.length; i>0; i--) {
	    if (chars[i-1]=='A')
		rc[rc.length-i] = 'T';
	    else if (chars[i-1]=='T')
		rc[rc.length-i] = 'A';
	    else if (chars[i-1]=='C')
		rc[rc.length-i] = 'G';
	    else if (chars[i-1]=='G')
		rc[rc.length-i] = 'C';
	    else if (chars[i-1]=='a')
		rc[rc.length-i] = 't';
	    else if (chars[i-1]=='t')
		rc[rc.length-i] = 'a';
	    else if (chars[i-1]=='c')
		rc[rc.length-i] = 'g';
	    else if (chars[i-1]=='g')
		rc[rc.length-i] = 'c';
	}
	return new String(rc);
    }

    public static boolean containsRepeatElement(String s) {
	return s.indexOf("N")!=-1 || s.indexOf("a")!=-1 || s.indexOf("c")!=-1 || s.indexOf("g")!=-1 || s.indexOf("t")!=-1;
    }

}
