
import java.util.*;
import java.util.regex.*;
import java.io.*;

public class KnownGene {
    
    public LinkedHashSet[] knownGenes = new LinkedHashSet[21];
    
    public KnownGene() {
	for (int i=0; i<knownGenes.length; i++)
	    knownGenes[i] = new LinkedHashSet();
	try { //Read from the file with the data
	    BufferedReader reader = new BufferedReader(new FileReader("/media/exthd/Genomes/Mm/v9/knownGene.txt"));
	    String line = reader.readLine();
	    //Use a regular expression to find all the info in the file
	    Pattern annotationPattern1 = Pattern.compile("([\\d|\\w|\\_]+)\\s+chr([\\d|X|Y]{1,3})\\s+([\\+|\\-])\\s+([\\d]+)\\s+([\\d]+)\\s+([\\d]+)\\s+([\\d]+)\\s+([\\d]+)");
	    while (line!=null) {
		Matcher m = annotationPattern1.matcher(line);
		if (m.find()) {
		    //Store the data as annotation objects
		    String n = m.group(1);
		    int chromosome = 0;
		    if (m.group(2).equals("X"))
			chromosome = 20;
		    else if (m.group(2).equals("Y"))
			chromosome = 21;
		    else
			chromosome = Integer.parseInt(m.group(2));
		    boolean strand = m.group(3).equals("+");
		    int tss = Integer.parseInt(m.group(4));
		    int tse = Integer.parseInt(m.group(5));
		    int codingStart = Integer.parseInt(m.group(6));
		    int codingEnd  = Integer.parseInt(m.group(7));
		    //Variable number of exons
		    int exonCount = Integer.parseInt(m.group(8));
		    //Use ugly solution with two patterns since we have a
		    //variable component
		    Pattern exonStartPattern = Pattern.compile("\\s?([\\d]+),");
		    String rest = line.substring(m.end());
		    int exonStarts[] = new int[exonCount], exonEnds[] = new int[exonCount];
		    for (int i=0; i<exonCount; i++) {
			m = exonStartPattern.matcher(rest);
			if (m.find()) {
			    exonStarts[i] = Integer.parseInt(m.group(1));
			    rest = rest.substring(m.end());
			}
		    }
		    for (int i=0; i<exonCount; i++) {
			m = exonStartPattern.matcher(rest);
			if (m.find()) {
			    exonEnds[i] = Integer.parseInt(m.group(1));
			    rest = rest.substring(m.end());
			}
		    }
		    Pattern geneIDPattern = Pattern.compile("\\s+([\\d|\\w|\\_]+)\\s+([\\d|\\w|\\_]+)");
		    m = geneIDPattern.matcher(rest);
		    if (m.find()) {
			String refSeq = m.group(1);
			String alignName = m.group(2);
			//Add to the storage
			knownGenes[chromosome-1].add(new KnownGeneEntry(n, chromosome, strand, tss, tse, codingStart, codingEnd, exonStarts, exonEnds, refSeq, alignName));
		    }
		}
		line = reader.readLine();
	    }
	    reader.close();
	}
	catch (IOException ex) {
	    System.err.println("Error: " + ex.toString());
	}
    }

    public String findNearestGene(int chr, int pos) {
	int nearestUpstream = -pos-1, nearestDownstream = Integer.MAX_VALUE;
	Iterator it = knownGenes[chr-1].iterator();
	KnownGeneEntry prevGene = null;
	while (it.hasNext()) {
	    KnownGeneEntry k = (KnownGeneEntry)it.next();
	    int dist = k.tss - pos;
	    if (dist<0 && dist>nearestUpstream) //Negative values
		nearestUpstream = dist;
	    else if (dist>0) {
		nearestDownstream = dist;
		if (Math.abs(nearestDownstream)>Math.abs(nearestUpstream))
		    return prevGene.refSeq;
		return k.refSeq;
	    }
	    prevGene = k;
	}
	return prevGene.refSeq;
    }

    public int findNearestGeneTSS(int chr, int pos) {
	int nearestUpstream = -pos-1, nearestDownstream = Integer.MAX_VALUE;
	Iterator it = knownGenes[chr-1].iterator();
	KnownGeneEntry prevGene = null;
	while (it.hasNext()) {
	    KnownGeneEntry k = (KnownGeneEntry)it.next();
	    int dist = k.tss - pos;
	    if (dist<0 && dist>nearestUpstream) //Negative values
		nearestUpstream = dist;
	    else if (dist>0) {
		nearestDownstream = dist;
		if (Math.abs(nearestDownstream)>Math.abs(nearestUpstream))
		    return prevGene.tss;
		return k.tss;
	    }
	    prevGene = k;
	}
	return prevGene.tss;
    }
    
}

class KnownGeneEntry {

    public String name, refSeq, alignID;
    public int chr, tss, tse, codingStart, codingEnd, nExons;
    public int[] exonStart, exonEnd;
    public boolean strand;

    public KnownGeneEntry(String n, int c, boolean st, int ts, int te, int cs, int ce, int exs[], int exe[], String ref, String al) {
	name = n;
	chr = c;
	strand = st;
	tss = ts;
	tse = te;
	codingStart = cs;
	codingEnd = ce;
	nExons = exe.length;
	exonStart = exs;
	exonEnd = exe;
	refSeq = ref;
	alignID = al;
    }

    public int hashCode() { return name.hashCode() + 3*tss + 7*nExons + 13*refSeq.hashCode(); }

}
