//Use this class to represent one lincRNA

import java.util.*;
import edu.uci.ics.jung.graph.impl.*;

public class LincRNA extends UndirectedSparseVertex implements Comparable {

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    protected int chr, id, start, end;
    protected boolean strand;
    //Keep track of the number of reads, can either be across coding
    //region or just at TSS
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();

    //////////////////////////////////////////////////////////////////

    public LincRNA(int c, int s, int e, boolean st) {
	chr = c;
	start = s;
	end = e;
	strand = st;
    }

    public LincRNA(int c, int s, int e) {
	chr = c;
	start = s;
	end = e;
    }


    ///////////////////////////////////////////////////////

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public int getLength() { return Math.abs(start - end); }

    //////////////////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : null; }

    public void addFactorReads(String f, ABIRead[] reads) { 
	ABIRead rs[] = getFactorReads(f);
	if (rs!=null) {
	    ABIRead[] newRs = new ABIRead[rs.length + reads.length];
	    for (int i=0; i<rs.length; i++) { newRs[i] = rs[i]; }
	    for (int i=0; i<reads.length; i++) { newRs[i + rs.length] = reads[i]; }
	    factorReads.put(f, newRs); 
	}
	else
	    factorReads.put(f, reads);
    }

    public double getReadDensity(String tfName) { return (double)getFactorReads(tfName).length/getLength(); }

    //////////////////////////////////////////////////////////////////

    public boolean getStrand() { return strand; }

    public int getChromosome() { return chr; }


    /////////////////////////////////

    public boolean overlaps(int s, int e) {
	//if (strand)
	//  return end>maxStart && start<maxEnd;
	//return start<maxStart && end>maxEnd;
	if ((s>start && s<end) || (e>start && e<end)) //Start or end of region inside gene
	    return true;
	else if ((start>s && start<e) || (end>s && end<e)) //Start or end of exon inside region
	    return true;
	return false;
    }


    //////////////////////////////////////////////////////////////////
    //Stuff for the JUNG package
    

    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof LincRNA))
	    System.err.println("ERROR comparing lincRNA!");
	LincRNA li = (LincRNA)o;
	if (chr<li.chr)
	    return -1;
	else if (chr>li.chr)
	    return 1;
	else {
	    if (getStart()<li.getStart())
		return -1;
	    else if (getStart()>li.getStart())
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return chr + ":" + start + "-" + end;
    }

    public boolean equals(Object o) {
	if (!(o instanceof LincRNA))
	    System.err.println("ERROR comparing lincRNA!");
	LincRNA mi = (LincRNA)o;
	return mi.chr==chr && mi.start==start && mi.end==end;
    }

    public int hashCode() { return chr*29 + 5*end + 13*start; }

}
