//Use this class to keep track of the mappability index. It does so by
//reading from the file. It can be queried efficiently if all queries
//are in terms of ascending coordinate values.
//
//The Mappability-index is calculated using a c++ program:
//
//First we need to calculate the .bbb index for the full genome
//
//
//
///home/mh168/Softwares/bbwt2/src/BBBWTgen -F -i /groups/neuroduo/martin/Genomes/Hs/v37/hs_chr_all.fa -o /groups/neuroduo/martin/Genomes/Hs/v37/hs_chr_all.bbb
//
//Finally, calculate the index showing sites which are uniquely mappable vs those that are not
//
///home/mh168/Softwares/bbwt2/src/run_mappability_one_hs.sh 35

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class Mappability {
    
    protected int nBasepairs, nUniqueBasePairs, nWildcards;
    protected int readLength = 35, fragmentLength = 120, windowSize, windowStart = 0;
    protected boolean window[], hg19 = false;
    protected String bedFileName, statsFileName, species;
    //Pattern for parsing the file
    protected Pattern mappabilityPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)");
    protected int minDist, segmentEnd;
    //Use these to keep track of where in the genome the current
    //mappability index corresponds to
    protected BufferedReader mappabilityReader;

    /*public Mappability(String genomeLoc, int chr, int w, int l, String sp) {
	windowSize = w;
	window = new boolean[windowSize];
	fragmentLength = l;
	species = sp;
	bedFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_U_" + AnalyzeControls.READ_LENGTH + ".bed";
	statsFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	//String bedFileName = genomeLoc + "chrY_U_" + readLength + ".bed";
	//String statsFileName = genomeLoc + "chrY_stats_" + readLength + ".txt";
	initialize();
    }

    public Mappability(String genomeLoc, int chr, int w, String sp) {
	windowSize = w;
	species = sp;
	window = new boolean[windowSize];
	bedFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_U_" + AnalyzeControls.READ_LENGTH + ".bed";
	statsFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	//String bedFileName = genomeLoc + "chrY_U_" + readLength + ".bed";
	//String statsFileName = genomeLoc + "chrY_stats_" + readLength + ".txt";
	initialize();
	}*/

    public Mappability(boolean orchestra, int chr, int w, int l, String sp) {
	windowSize = w;
	fragmentLength = l;
	species = sp;
	window = new boolean[windowSize];
	String genomeLoc = (orchestra ? "/groups/neuroduo/martin" : "/media/My Book") + "/Genomes/" + (species.equals("mm") ? "Mm/v9/" : "Hs/v" + (hg19 ? "37" : "36") + "/");
	bedFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_U_" + AnalyzeControls.READ_LENGTH + ".bed";
	statsFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	initialize();
    }

    public Mappability(boolean orchestra, int chr, int w, int l, String sp, boolean hg19) {
	windowSize = w;
	fragmentLength = l;
	species = sp;
	window = new boolean[windowSize];
	String genomeLoc = (orchestra ? "/groups/neuroduo/martin" : "/media/My Book") + "/Genomes/" + (species.equals("mm") ? "Mm/v9/" : "Hs/v" + (hg19 ? "37" : "36") + "/");
	bedFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_U_" + AnalyzeControls.READ_LENGTH + ".bed";
	statsFileName = genomeLoc + "chr" + AnalyzeControls.getChromosomeNumber(chr, species) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	initialize();
    }

    /*public Mappability(String genomeLoc, int chr) {
	bedFileName = genomeLoc + "chr" + ABIReads.getChromosomeNumber(chr) + "_U_" + AnalyzeControls.READ_LENGTH + ".bed";
	statsFileName = genomeLoc + "chr" + ABIReads.getChromosomeNumber(chr) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	//String bedFileName = genomeLoc + "chrY_U_" + readLength + ".bed";
	//String statsFileName = genomeLoc + "chrY_stats_" + readLength + ".txt";
	initialize();
	}*/

    protected void initialize() {
	try { //Read the data for the summary statistics
	    BufferedReader br = new BufferedReader(new FileReader(statsFileName));
	    Pattern statsPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    String line = br.readLine();
	    Matcher m = statsPattern.matcher(line);
	    if (m.find()) {
		nBasepairs = Integer.parseInt(m.group(1));
		nUniqueBasePairs = Integer.parseInt(m.group(2));
		nWildcards = Integer.parseInt(m.group(3));
	    }
	    br.close();
	    mappabilityReader = new BufferedReader(new FileReader(bedFileName));
	    //Remove the first line since it contains comments
	    mappabilityReader.readLine();
	    //Second line may contain an error
	    line = mappabilityReader.readLine();
	    m = mappabilityPattern.matcher(line);
	    if (m.find()) {
		if (Integer.parseInt(m.group(1))==0 && Integer.parseInt(m.group(2))==0) {
		    line = mappabilityReader.readLine();
		    m = mappabilityPattern.matcher(line);
		    if (m.find()) {
			segmentEnd = Integer.parseInt(m.group(2));
			minDist = Integer.parseInt(m.group(3));
		    }
		}
		else { //No error
		    segmentEnd = Integer.parseInt(m.group(2));
		    minDist = Integer.parseInt(m.group(3));
		}
		//System.out.println("windowStart = " + windowStart + ", segmentEnd = " + segmentEnd + " and minDist = " + minDist);
		while (windowStart<window.length-1 ) { //initialize the window
		    updateWindow();
		    window[windowStart++] = minDist>0;
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Error reading " + statsFileName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    //////////////////////////////////////////////////////////////////

    public int getNumberOfUniqueBases() { return nUniqueBasePairs; }
    
    public int getNumberOfBases() { return nBasepairs; }

    //////////////////////////////////////////////////////////////////

    public int readCanStartNextBasePair() {
	if (++windowStart==segmentEnd) {
	    //System.out.println("windowStart = " + windowStart + ", segmentEnd = " + segmentEnd + " and minDist = " + minDist);
	    String line;
	    try {
		line = mappabilityReader.readLine();
		if (line==null) { 
		    mappabilityReader.close();
		    return -1; 
		}
	    }
	    catch (IOException e) {
		System.err.println("Error reading from mappabilityReader");
		return -1;
	    }
	    Matcher m = mappabilityPattern.matcher(line);
	    if (m.find()) {
		segmentEnd = Integer.parseInt(m.group(2));
		minDist = Integer.parseInt(m.group(3));
		//System.out.println("windowStart = " + windowStart + ", segmentEnd = " + segmentEnd + " and minDist = " + minDist);
	    }
	}
	return minDist;
    }

    public int numberOfPotentialReadsNextWindow() { return numberOfPotentialReadsNextWindow(1); }

    public int numberOfPotentialReadsNextWindow(int n) {
	//Slide the window n bp forward and count how many fragments
	//we can theoretically fit there.
	for (int j=0; j<n; j++) {
	    for (int i=0; i<window.length - 1; i++) { window[i] = window[i+1]; }
	    //for (int i=windowSize-readLength-1; i<windowSize-1; i++)
	    int u = readCanStartNextBasePair();
	    if (u==-1)
		return -1;
	    window[windowSize-1] = u>0;
	}
	int nForwardReads = 0;
	for (int i=0; i<window.length-fragmentLength; i++) { nForwardReads += window[i] ? 1 : 0; }
	//Count the number of reads on the reverse strand which can be
	//made to fit into this window. This is slightly more
	//complicated and we must take into account that the start of
	//the read is not the same location as the coordinate given by
	//ChromosomeWindow which transforms it to the positive strand.
	int nReverseReads = 0;
	for (int i=fragmentLength-readLength; i<window.length-readLength; i++) { nReverseReads += window[i] ? 1 : 0; }
	//System.out.println("Window has " + nForwardReads + " potential forward and " + nReverseReads + " potential reverse.");
	return nForwardReads + nReverseReads;
    }

    protected void updateWindow() {
	//use this method to figure out if we need to read a new line
	//from the Mappability file
	try {
	    if (windowStart==segmentEnd) {
		String line = mappabilityReader.readLine();
		Matcher m = mappabilityPattern.matcher(line);
		if (m.find()) {
		    segmentEnd = Integer.parseInt(m.group(2));
		    minDist = Integer.parseInt(m.group(3));
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Error reading from mappabilityReader");
	}
    }

    public int numberOfPotentialReadsRegion(int start, int end) {
	//Slide the window up to start and move through to end,
	//counting the number of reads which may occur in the region.
	while (windowStart<start) { numberOfPotentialReadsNextWindow(); } //slide window forward
	int prevPotential = numberOfPotentialReadsNextWindow();
	int totPotential = prevPotential;
	while (windowStart<end) {
	    int nPotential = numberOfPotentialReadsNextWindow();
	    totPotential += nPotential>prevPotential ? nPotential - prevPotential : 0;
	    prevPotential = nPotential;
	}
	return totPotential;
    }

    public int getWindowStart() { return windowStart; }

    public void moveWindowStartTo(int s) { numberOfPotentialReadsNextWindow(s - windowStart); }

}

