//Use this class to represent a non-genic transcript


class NonGenicTranscript implements Comparable {

    protected int nReads = 0;
    protected boolean strand;
    //start and end are defined with respect to the 5' and 3'
    //ends. Thus, start>end for transcripts on the negative strand.
    protected int start, end, chr;

    public NonGenicTranscript(int c, boolean st, int s, int e, int n) {
	chr = c;
	strand = st;
	start = s;
	end = e;
	nReads = n;
    }

    public int getChromosome() { return chr; }
    
    public int getStart() { return start; }

    public int getEnd() { return end; }

    public int numberOfReads() { return nReads; }

    public int getLength() { return Math.abs(end - start); }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	return AnalyzeControls.getChromosomeNumber(chr) + ":" + start + "-" + end + "\t" + "(" + (strand ? "+" : "-") + ")\t" + nReads;
    }

    public int compareTo(Object o) {
	if (!(o instanceof NonGenicTranscript)) {
	    System.out.println("Error comparing with " + o.toString());
	    return 0;
	}
	NonGenicTranscript n = (NonGenicTranscript)o;
	if (this.chr<n.chr)
	    return -1;
	else if (this.chr>n.chr)
	    return 1;
	else {
	    if (this.start<n.start)
		return -1;
	    else if (this.start>n.start)
		return 1;
	    else
		return 0;
	}
    }
    
    public boolean equals(Object o) {
	if (!(o instanceof NonGenicTranscript)) {
	    System.out.println("Error comparing with " + o.toString());
	    return false;
	}
	NonGenicTranscript n = (NonGenicTranscript)o;
	return n.chr==chr && n.start==start && n.end==end && n.strand==strand && n.nReads==nReads;
    }

}
