import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class RepeatInfo {

    protected int chr;
    protected int strand;
    protected String species = "mm";
    //Keep track of all the repeat regions
    protected TreeSet<RepeatRegion> repeatRegions = new TreeSet<RepeatRegion>();
    protected TreeMap<String, String> name2class = new TreeMap<String, String> ();

    //////////////////////////////////////////////////////////////////

    public RepeatInfo(int c, boolean f, boolean orchestra, String sp) {
	chr = c;
	strand = f ? 1 : -1;
	species = sp;
	String fileName = orchestra && species.equals("mm") ? "/groups/neuroduo/martin/Genomes/Mm/v9/repeat_mask_mm9.tsv" : null;
	if (fileName!=null)
	    parseFile(fileName, strand);
    }

    public RepeatInfo(int c, boolean orchestra, String sp) {
	chr = c;
	strand = 0;
	species = sp;
	String fileName = orchestra && species.equals("mm") ? "/groups/neuroduo/martin/Genomes/Mm/v9/repeat_mask_mm9.tsv" : null;
	if (fileName!=null)
	    parseFile(fileName, strand);
    }

    public RepeatInfo(boolean orchestra, String sp) {
	chr = -1;
	strand = 0;
	species = sp;
	String fileName = orchestra && species.equals("mm") ? "/groups/neuroduo/martin/Genomes/Mm/v9/repeat_mask_mm9.tsv" : null;
	if (fileName!=null)
	    parseFile(fileName, strand);
    }


    public void parseFile(String fileName, int strand) {
	//returns a list of all repName and their corresponding
	//repClass. This is useful when we want to save so that we can
	//list all of them every time.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //#bin    swScore milliDiv        milliDel        milliIns        genoName        genoStart       genoEnd genoLeft        strand  repName repClass        repFamily   repStart        repEnd  repLeft id
	    
	    //Pattern repeatPattern = Pattern.compile("chr(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+[-|\\d]+\\s+([\\+|-])\\s+([\\w|_|-|\\(|\\)]+)\\s+([\\w|_|-]+)\\s+([\\w|_|-]+)");
	    while (line!=null) {
		//Matcher m = repeatPattern.matcher(line);
		if (!line.startsWith("#")) {
		    //Get all the 
		    String tokens[] = line.split("\\s"), chrStr = tokens[5].substring(3);
		    int ch;
		    if (chrStr.equals("X"))
			ch = 20;
		    else if (chrStr.equals("Y"))
			ch = 21;
		    else if (chrStr.equals("M") || chrStr.contains("random")) {
			line = br.readLine();
			continue;
		    }
		    else
			ch = Integer.parseInt(chrStr);
		    if ((chr==-1 || ch==chr) && (strand==0 || (tokens[9].equals("+") && strand>0) || (!tokens[9].equals("+") && strand<0))) {
			repeatRegions.add(new RepeatRegion(ch, tokens[9].equals("+"), Integer.parseInt(tokens[6]), Integer.parseInt(tokens[7]), tokens[10], tokens[11]));
			name2class.put(tokens[10], tokens[11]);
		    }
		    //}
		    //else if (!line.startsWith("#"))
		    //  System.err.println("Error, could not find repeat region for: " + line);
		}
		line = br.readLine();
	    }
	    System.out.println("Found " + repeatRegions.size() + " repeat regions for chromosome " + chr);
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing repeat region file " + fileName + ", " + e.toString());
	}
    }

    public TreeSet<RepeatRegion> getRepeatRegions() { return repeatRegions; }

    public TreeMap<String, String> getName2classMap() { return name2class; }

    public boolean overlaps(int pos) {
	for (RepeatRegion r : repeatRegions) {
	    if (r.overlaps(pos))
		return true;
	}
	return false;
    }

    public RepeatRegion getOverlappingRegion(int chr, int pos) {
	for (RepeatRegion r : repeatRegions) {
	    if (chr==r.getChromosome() && r.overlaps(pos))
		return r;
	}
	return null;
    }

}
