//Use this class to extract the contiguous sequence from a set of
//tiling arrays and then scan for motifs using the TFMscan software.

import java.util.*;
import java.util.regex.*;
import java.io.*;
import Synergism.FastaSequence;

public class SidowParser {

    //Store the reverse strand by using a negative sign for the positions.
    protected int nReads = 0, chr, maxPos;
    //Stores the first point of each read for the chromosome
    protected TreeMap reads = new TreeMap(), randReads = null;
    //Store the start and end points for each contig. This is useful
    //when we want to retrieve the full mapping
    protected Random random = new Random();
    protected int readLength = 25;
    protected String fileName;

    //////////////////////////////////////////////////////////////////

    public SidowParser(String fileName, int chr) {
	this.chr = chr;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern sidowPattern = Pattern.compile("([A|C|G|T|\\.]) [0-9] [0-9] chr([0-9|X|Y]+):([0-9]+) ([F|R])");
	    while (line!=null) {
		Matcher m = sidowPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(2).equals("X"))
			c = 23;
		    else if (m.group(2).equals("Y"))
			c = 24;
		    else
			c = Integer.parseInt(m.group(2));
		    if (c==chr) {
			int pos = Integer.parseInt(m.group(3));
			for (int i=0; i<readLength; i++) {
			    Integer key = new Integer(pos + i);
			    if (reads.containsKey(key))
				reads.put(key, new Integer(((Integer)reads.get(key)).intValue() + 1));
			    else
				reads.put(key, new Integer(1));
			}
			//Update the list of contigs
			//if (!contigs.contains(pos, pos + readLength))
			//  contigs.add(pos, pos + readLength);
			nReads++;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    maxPos = ((Integer)reads.lastKey()).intValue() + readLength;
	}
	catch(IOException e) {

	}
    }

    public int getNumberOfReads(){
	return nReads;
    }

    public int[] getReads() {
	//Return the number of reads in a compact array
	Iterator it = reads.keySet().iterator();
	int i = 0, r[] = new int[reads.size()];
	while (it.hasNext())
	    r[i++] = ((Integer)reads.get((Integer)it.next())).intValue();
	return r;
    }

    public int[] getPositions() {
	//Return the number of reads in a compact array
	Object[] a = reads.keySet().toArray();
	int[] p = new int[a.length];
	for (int i=0; i<a.length; i++)
	    p[i] = ((Integer)a[i]).intValue();
	return p;
    }

    public void generateRandomReads() {
	//Spread out the same number of reads as in the real
	//experiment randomly.
	//Find out how long the genome really is
	/*String genomeLoc = "/media/exthd/Genomes/Hs/v36/chr";
	if (chr==23)
	    genomeLoc += "X.fa";
	else if (chr==24)
	    genomeLoc += "Y.fa";
	else
	    genomeLoc += (new Integer(chr)).toString() + ".fa";
	    FastaSequence fastaSequence = new FastaSequence(genomeLoc);*/
	//int chrLength = (int)fastaSequence.getChromosomeLength();
	int chrLength = maxPos;
	randReads = new TreeMap();
	//Remove the original reads to save memory
	reads = null;
	for (int i=0; i<nReads; i++){
	    //Get a randomized starting position
	    int randStart = random.nextInt(chrLength-readLength);
	    for (int j=0; j<readLength; j++) {
		Integer key = new Integer(randStart + j);
		if (randReads.containsKey(key))
		    randReads.put(key, new Integer(((Integer)randReads.get(key)).intValue() + 1));
		else
		    randReads.put(key, new Integer(1));
	    }
	}
    }

    public int[] getRandomReads() {
	if (randReads!=null) {
	    //Return the number of reads in a compact array
	    Iterator it = randReads.keySet().iterator();
	    int i = 0, r[] = new int[randReads.size()];
	    while (it.hasNext())
		r[i++] = ((Integer)randReads.get((Integer)it.next())).intValue();
	    return r;
	}
	System.err.println("Must generate the random reads first!");
	return null;
    }

    public int[] getRandomPositions() {
	//Return the number of reads in a compact array
	if (randReads!=null) {
	    Object[] a = randReads.keySet().toArray();
	    int[] p = new int[a.length];
	    for (int i=0; i<a.length; i++)
		p[i] = ((Integer)a[i]).intValue();
	    return p;
	}
	System.err.println("Must generate the random reads first!");
	return null;
    }

}
