//Use this method to split the reads into one file for each chromosome.
//
//Example:
//
//java SplitReadsFiles -n /groups/neuroduo/martin/ChIP-Seq/GrimmondData/Hs/mRNA/heart.tsv -o -c 24


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class SplitReadsFiles{
    
    
    public static void main(String[] args) {
	String line = null;
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    //c, n
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "nChromosomes");
	    //use this flag to extract and save the sequence
	    CmdLineParser.Option sequenceFlag = argParser.addBooleanOption('s', "sequence");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //Use this flag when we want to treat RNA-files as though the data was in the ChIP format (useful when not pre-processed by David)
	    CmdLineParser.Option chipFlag = argParser.addBooleanOption('C', "ChIP");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int nChromosomes = tmp!=null ? ((Integer)tmp).intValue() : 21;
	    boolean sequence = argParser.getOptionValue(sequenceFlag)!=null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean chip = argParser.getOptionValue(chipFlag)!=null;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    boolean extend = fragmentLength>AnalyzeControls.READ_LENGTH;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    //////////////////////////////////////////////////////////////////
	    boolean rna = fileName.contains("RNA") && !chip;
	    //chr loc strand nReads [total_len]
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)\\s*(\\d*)");
	    BufferedWriter bws[] = new BufferedWriter[nChromosomes];
	    for (int i=0; i<nChromosomes; i++) 
		bws[i] = new BufferedWriter(new FileWriter(rna ? fileName.replace(".tsv", "_" + (i+1) + ".tsv") : fileName + "_" + (i + 1)));
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    line = br.readLine();
	    while (line!=null) {
		Matcher m = rna ? mRNAPattern.matcher(line) : PeakFinder2.rawPattern1.matcher(line);
		if (m.find()) {
		    int c = Integer.parseInt(m.group((rna ? 1 : 2)));
		    if (c<=nChromosomes) {
			bws[c-1].write(line + "\n");
			if (!rna)
			    bws[c-1].write(br.readLine() + "\n"); //write the code
		    }
		}
		line = br.readLine();
	    }
	    for (int i=0; i<nChromosomes; i++) { bws[i].close(); }
	    br.close();
	    if (sequence) {
		for (int i=0; i<nChromosomes; i++) { //Now we want to extract the sequence for each read
		    //PeakFinder2 pf = new PeakFinder2(orchestra, i+1, fileName, null, fragmentLength, extend, 0);
		    //pf.saveReadsAndSequenceData(PeakFinder2.getProcessedReadsFileName(fileName, i+1, extend, fragmentLength));
		}
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	    System.err.println("For line: " + line);
	} // end of try-catch
    } // end of main()
    
}
