//Use this class to compare transcripts from two experiments. This
//class is called from CompareTranscripts.java and does all the heavy
//lifting.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.*;

public class TranscriptComparer {

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    protected ReducedModel r1, r2;
    protected int chr, maxReadLength = 35;
    protected boolean forward, orchestra;
    protected String species, rnaFileName1, rnaFileName2;
    protected ABIRead[] rnaReads1, rnaReads2;

    protected TreeMap<TranscribedRegion, TranscribedRegion> matchingTranscripts = new TreeMap<TranscribedRegion, TranscribedRegion>();
    protected TreeMap<TranscribedRegion, TreeSet<TranscribedRegion>> matchingTranscriptsAll1 = new TreeMap<TranscribedRegion, TreeSet<TranscribedRegion>>(), matchingTranscriptsAll2 = new TreeMap<TranscribedRegion, TreeSet<TranscribedRegion>>();
    protected TreeSet<TranscribedRegion> nonMatching1 = new TreeSet<TranscribedRegion>(), nonMatching2 = new TreeSet<TranscribedRegion>(), notIn1 = new TreeSet<TranscribedRegion>(), notIn2 = new TreeSet<TranscribedRegion>();
    protected TreeMap<String, TreeMap<String, Integer>> categoryCountMap1 = new TreeMap<String, TreeMap<String, Integer>>(), categoryCountMap2 = new TreeMap<String, TreeMap<String, Integer>>();

    //////////////////////////////////////////////////////////////////

    public TranscriptComparer(int c, boolean f, String s, boolean or, ReducedModel r1, ReducedModel r2, ABIRead[] reads1, String rnaName1, ABIRead[] reads2, String rnaName2) {
	chr = c;
	forward = f;
	species = s;
	orchestra = or;
	this.r1 = r1;
	this.r2 = r2;
	rnaReads1 = reads1;
	rnaReads2 = reads2;
	rnaFileName1 = rnaName1;
	rnaFileName2 = rnaName2;
    }


    ///////////////////////////////////////////////////////

    public TreeMap<TranscribedRegion, TranscribedRegion> compareTranscripts() {
	//Create a transition matrix showing what category (if any)
	//each transcript maps to in the other sample
	for (TranscribedRegion t1 : r1.getSegments()) {
	    for (TranscribedRegion t2 : r2.getSegments()) {
		if (t1.matchesUniquely(t2)) {
		    matchingTranscripts.put(t1, t2);
		    break;
		}
	    }
	}
	for (TranscribedRegion t : r1.getSegments()) {
	    if (!matchingTranscripts.containsKey(t))
		nonMatching1.add(t);
	}
	for (TranscribedRegion t : r2.getSegments()) {
	    if (!matchingTranscripts.containsValue(t))
		nonMatching2.add(t);
	}
	return matchingTranscripts;
    }

    public TreeMap<TranscribedRegion, TreeSet<TranscribedRegion>> compareTranscriptsAll(boolean one2two) {
	//Create a transition matrix showing what category (if any)
	//each transcript maps to in the other sample
	ReducedModel rOne = one2two ? r1 : r2;
	ReducedModel rTwo = one2two ? r2 : r1;
	TreeMap<TranscribedRegion, TreeSet<TranscribedRegion>> matchingTranscriptsAll = one2two ? matchingTranscriptsAll1 : matchingTranscriptsAll2;
	for (TranscribedRegion t1 : rOne.getSegments()) {
	    matchingTranscriptsAll.put(t1, new TreeSet<TranscribedRegion>());
	    for (TranscribedRegion t2 : rTwo.getSegments()) {
		if (t1.overlaps(t2) || t2.overlaps(t1))
		    matchingTranscriptsAll.get(t1).add(t2);
	    }
	}
	return matchingTranscriptsAll;
    }

    public TreeMap<String, TreeMap<String, Integer>> countCategoryTransitions(boolean one2two, String saveFileName) {
	//Use this method to find out how the transcripts map between
	//the different samples. Also calculate the average density of
	//reads from rnaReads2 for the regions covered by the segments
	//in r1 and save to a file
	TreeMap<String, TreeMap<String, Integer>> categoryCountMap = one2two ? categoryCountMap1 : categoryCountMap2;
	try { //Save the densities for both experiments for each transcript
	    System.out.println("Saving density info to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (chr==1) {
		String rna1 = one2two ? rnaFileName1.substring(rnaFileName1.lastIndexOf("/") + 1) : rnaFileName2.substring(rnaFileName2.lastIndexOf("/") + 1);
		String rna2 = one2two ? rnaFileName2.substring(rnaFileName2.lastIndexOf("/") + 1) : rnaFileName1.substring(rnaFileName1.lastIndexOf("/") + 1);
		bw.write("#chr\tstrand\tstart\tend\t" + rna1 + "_density\t" + rna2 + "_density\t" + rna1 + "_category\t" + rna2 + "_category\n");
	    }
	    TreeSet<TranscribedRegion> notInOther = new TreeSet<TranscribedRegion>();
	    TreeMap<TranscribedRegion, TreeSet<TranscribedRegion>> matchingTranscriptsAll = one2two ? matchingTranscriptsAll1 : matchingTranscriptsAll2;
	    TreeSet<TranscribedRegion> uniqueMatches = new TreeSet<TranscribedRegion>(one2two ? matchingTranscripts.keySet() : matchingTranscripts.values());
	    for (TranscribedRegion t : matchingTranscriptsAll.keySet()) {
		if (!categoryCountMap.containsKey(t.getType()))
		    categoryCountMap.put(t.getType(), new TreeMap<String, Integer>());
		Vector<String> cat2s = new Vector<String>();
		if (uniqueMatches.contains(t)) { //Check if there is a unique match for this transcript
		    if (one2two)
			cat2s.addElement(matchingTranscripts.get(t).getType());
		    else {
			for (TranscribedRegion t2 : matchingTranscripts.keySet()) {
			    if (matchingTranscripts.get(t2).equals(t)) {
				cat2s.addElement(t2.getType());
				break;
			    }
			}
		    }
		}
		else { //one-to-many or many-to-one
		    if (matchingTranscriptsAll.get(t).size()==0) {
			notInOther.add(t);
			//cat2s.addElement(CompareTranscripts.NULL_STR);
		    }
		    else {
			for (TranscribedRegion t2 : matchingTranscriptsAll.get(t))
			    cat2s.addElement(t2.getType());
		    }
		}
		TranscribedRegion tOther = new TranscribedRegion(t.getStart(), t.getEnd(), forward, false, chr, one2two ? rnaReads2 : rnaReads1, maxReadLength, true);
		bw.write(chr + "\t" + (forward ? "1" : "0") + "\t" + t.getStart() + "\t" + t.getEnd() + "\t" + format.format(t.getDensity()) + "\t" + format.format(tOther.getDensity()) + "\t" + t.getType() + "\t");
		for (String cat2 : cat2s) {
		    categoryCountMap.get(t.getType()).put(cat2, categoryCountMap.get(t.getType()).containsKey(cat2) ? categoryCountMap.get(t.getType()).get(cat2) + 1 : 1);
		    bw.write(cat2 + ",");
		}
		bw.write("\n");
	    }
	    bw.close();
	    if (one2two)
		notIn2 = notInOther;
	    else
		notIn1 = notInOther;
	}
	catch (IOException e) {
	    System.err.println("Error saving to " + e.toString());
	}
	return categoryCountMap;
    }

    public TreeSet<TranscribedRegion> getNotInOther(boolean one2two) { return one2two ? notIn2 : notIn1; }

    //////////////////////////////////////////////////////////////////


    public void saveCategoryTransitions(boolean one2two, String saveFileName) {
	try {
	    TreeMap<String, TreeMap<String, Integer>> categoryCountMap = one2two ? categoryCountMap1 : categoryCountMap2;
	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    for (String cat1 : categoryCountMap.keySet()) {
		for (String cat2 : categoryCountMap.get(cat1).keySet()) 
		    bw.write(cat1 + "\t" + cat2 + "\t" + categoryCountMap.get(cat1).get(cat2) + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving to " + e.toString());
	}
    }

    public void saveTranscripts(TreeSet<TranscribedRegion> regions, String saveName, int binSize, Vector<Peak> peaks, boolean conservation) {
	try {
	    System.out.println("Saving " + regions.size() + " transcripts to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    for (TranscribedRegion t : regions) 
		bw.write(t.getBedFileString(!forward, species) + "\n");
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving to " + e.toString());
	}
	ReducedModel r = new ReducedModel(regions, chr, forward, species);

	FindTranscripts.saveTranscriptProperties(r, binSize, peaks!=null ? peaks.firstElement().getFactorNames(false) : null, orchestra, saveName.substring(0, saveName.lastIndexOf(FindTranscripts.BED_FILE_SUFFIX)), !forward, chr, conservation);			
	/*r.saveGeneInfo(saveFileName + FindTranscripts.BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", saveFileName, RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR, chr==1 || chr==-1);
	r.saveGeneInfo(saveFileName + FindTranscripts.BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.MERGED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", saveFileName, RefSeqGene.MERGED_TRANSCRIPT_GENE_STR, chr==1 || chr==-1);
	r.saveGeneInfo(saveFileName + FindTranscripts.BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", saveFileName, RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR, chr==1 || chr==-1);
	r.saveGeneInfo(saveFileName + FindTranscripts.BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.MISSED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", saveFileName, RefSeqGene.MISSED_TRANSCRIPT_GENE_STR, chr==1 || chr==-1);*/
    }



}