//Use this class to represent the table that we use for summarizing
//the transcript information. In particular, it allows us to save the
//information from multiple MC simulations. It should also feature
//methods for merging the results from the different chromosomes.

import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;


public class TranscriptLatexTable {

    public static final String[] fieldNames = new String[]{"Reads"};
    
    protected static LinkedHashMap<String, Vector<Double>> header2DataMap = new LinkedHashMap<String, Vector<Double>>();
    


    public void add(ReducedModel r) {
	
    }
    
public static void saveLatexTable(String latexTableFileName, String[] types, int[] typesCount, LinkedHashMap header2DataMap, boolean includeFractions[], boolean weightedAverage[]) {
	//Assume that everything has the same order as TYPES_ARRAY_STR
	DecimalFormat format = new DecimalFormat("###.###");
	try {
	    System.out.println("Saving summary statistics for all chromosomes to " + latexTableFileName);
	    //Find out how many columns we need
	    int nHeaders = 1 + header2DataMap.size(), nFractions = 0;
	    for (int i=0; i<includeFractions.length; i++) {
		nFractions += includeFractions[i] ? 1 : 0;
		nHeaders += includeFractions[i] ? 1 : 0;
	    }
	    //Keep track of the totals
	    LinkedHashMap<String, Double> header2SumMap = new LinkedHashMap<String, Double>();
	    int totTranscripts = 0;
	    BufferedWriter bw = new BufferedWriter(new FileWriter(latexTableFileName));
	    //write the header
	    bw.write("\\begin{tabular}{lr");
	    for (int i=0; i<nHeaders + nFractions; i++)
		bw.write("r");
	    //Always include the fractions for no transcripts
	    bw.write("}\n\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# Transcripts}}");
	    for (int i=0; i<typesCount.length; i++)
		totTranscripts += typesCount[i];
	    int i = 0;
	    Iterator it = header2DataMap.keySet().iterator();
	    while (it.hasNext()) {
		String h = (String)it.next();
		double d = 0;
		bw.write(" & ");
		if (includeFractions[i])
		    bw.write("\\multicolumn{2}{c}{"); 
		bw.write("\\textbf{" + h + "}");
		if (includeFractions[i])
		    bw.write("}");
		i++;
		//Find the sum for the category
		for (int j=0; j<((double[])header2DataMap.get(h)).length; j++) {
		    if (weightedAverage[i])
			d += ((double[])header2DataMap.get(h))[j]*((double)typesCount[j])/totTranscripts;
		    else
			d += ((double[])header2DataMap.get(h))[j];
		}
		header2SumMap.put(h, d);
	    }
	    bw.write("\\\\\n\\hline\\\\\n");
	    //write the data for each category
	    for (int j=0; j<types.length; j++) {
		bw.write(types[j] + " & " + typesCount[j] + " & (" + format.format((double)typesCount[j]/totTranscripts) + ")");
		i = 0;
		it = header2DataMap.keySet().iterator();
		while (it.hasNext()) {
		    String h = (String)it.next();
		    bw.write(" & " + format.format(((double[])header2DataMap.get(h))[j]));
		    if (includeFractions[i++])
			bw.write(" & (" + format.format(((double[])header2DataMap.get(h))[j]/header2SumMap.get(h)) + ")");
		}
		bw.write("\\\\\n");
	    }
	    //Write the sum
	    bw.write("\\hline\n");
	    bw.write("Total & " + totTranscripts + " & ");
	    i = 0;
	    for (String h : header2SumMap.keySet()) {
		bw.write(" & " + format.format(header2SumMap.get(h))); 
		if (includeFractions[i++])
		    bw.write(" & ");
	    }
	    bw.write("\n");
	    bw.write("\\end{tabular}\n"); //footer
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }

}
